// anova1e.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-24 09:53:14 tac>
//
// 対応の無い多群間で一元配置分散分析を行う。
// 2群の場合は2標本t検定も出力する。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W anova1e anova1e.C
//               -I/usr/local/include L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "anova1e  Version 7.42e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:anova1e [<OPTION> ...] [<DATA FILE> ...]",
  "  The one-way ANOVA(including two sample t-test) with the first item",
  "  as the group(factor A) and the next item as the analysis data.",
  "<OPTION>",
  "  -al=<ALPHA>:specify the significace level for power analysis",
  "    (0<ALPHA<1, default:0.05)",
  "  -dm=<DIFFERENCE>:specify the scientific significant difference",
  "    for power analysis(default:difference in sample mean)",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={bar|box|ci|data|msd|mse|sd[<COEF>]|sr[<COEF>]}[,...]",
  "      :select the data for Ngraph",
  "    bar:bar chart  box:box-and-whisker plot  ci:mean and CI  data:raw data",
  "    msd:mean and D  mse:mean and SE",
  "    sd[<COEF>]:mean and coefficient x SD",
  "    sr[<COEF>]:mean and coefficient x residuals  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=anova1e_ngp.dat",
  "  -po=<1-BETA>:specify the power for power analysis(0<1-BETA<1,default:0.8)",
  "  -res=[-][{def|all|at|cc[<C-COEF)>]|ft|m[{a|d}][{b|d|f|s|t]|mci|pa|r|stat",
  "       |tt|vt}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    at:ANOVA  cc[<C-COEF>]:confidence intervals(0<=C-COEF<1)",
  "    ft:2-tailed F-test to compare 2 variances",
  "    m[{a|d}][{b|d|f|s|t}]:multiple comparisons(ma:all 2 group comparisons",
  "    md:comparisons of group-1 vs other groups)",
  "    mci:confidence interval of means",
  "    pa:power analysis  r:correlation ratio  stat:basic statistics",
  "    tt:2 sample t-test  vt:test for equal variances",
  "    default:auto judgment according to data"
};

const string Title = "=== Comparison of Means of Multiple Groups ==="; // 表題
const string NGPNAMEdef = "anova1e_ngp.dat"; // 省略時Ngraphファイル名

using listdbl = list<double>;   // 倍精度実数リストの別名定義

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double val;                   // 有意水準
  double vcc;                   // 信頼係数
  double vdm;                   // 検出差
  mctype vma;                   // 全ての2群ごとの多重比較法
  mctype vmd;                   // 対照群と他の群の多重比較法
  double vpo;                   // 検出力
  double vtv;                   // -ngp=s[dr]<係数>
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 結果出力リスト
  map<string, bool> vresmap;    // 結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double al() const             // 有意水準を返す
  { return val; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double dm() const             // 検出差を返す
  { return vdm; }
  mctype ma() const             // 全ての2群ごとの多重比較法を返す
  { return vma; }
  mctype md() const             // 対照群と他の群の多重比較法を返す
  { return vmd; }
  double po() const             // 検出力を返す
  { return vpo; }
  double tv() const             // SDの係数を返す
  { return vtv; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 解析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), val(ALdef), vcc(CCdef), vdm(BLANKf), vma(mctuk), vmd(mcdun), vpo(POdef), vtv(BLANKf), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "al")           // 有意水準の指定
  {
    val = opt[1].empty() ? ALdef : strtodf(opt[1]);
    if (val <= 0.0 || 1.0 <= val) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "dm")      // 検出差の指定
  {
    vdm = opt[1].empty() ? BLANKf : strtodf(opt[1]);
    if (vdm < 0.0) err = badopt(vpname, optstr, USAGE, 0.0, DBL_MAX);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "po")      // 検出力の指定
  {
    vpo = opt[1].empty() ? POdef : strtodf(opt[1]);
    if (vpo <= 0.0 || 1.0 <= vpo) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni != 2) return disperr(vpname, "Number of items must be 2 !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["bar"] = vngpmap["box"] = vngpmap["ci"] = vngpmap["data"] = vngpmap["msd"] = vngpmap["mse"] = vngpmap["sd"] = vngpmap["sr"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["bar"] = vngpmap["box"] = vngpmap["ci"] = vngpmap["data"] = vngpmap["msd"] = vngpmap["mse"] = vngpmap["sd"] = vngpmap["sr"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngp[0] == 's')
      {
        switch (ngp[1])
        {
        case 'd' : vngpmap["sd"] = true;
          break;
        case 'r' : vngpmap["sr"] = true;
          break;
        default : return disperr(vpname, "Wrong -ngp=" + ngp + " option !");
        }
        vtv = (ngp.size() > 2) ? strtodf(ngp.substr(2)) : BLANKf;
        if (vtv < 0.0) return disperr(vpname, "Wrong -ngp" + ngp + " option !");
      }
      else if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["stat"] = vresmap["tt"] = true;
  vresmap["at"] = vresmap["ma"] = vresmap["md"] = vresmap["mci"] = vresmap["pa"] = vresmap["r"] = vresmap["vt"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["stat"] = vresmap["tt"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["stat"] = vresmap["tt"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["ft"] = vresmap["ma"] = vresmap["md"] = vresmap["mci"] = vresmap["pa"] = vresmap["r"] = vresmap["stat"] = vresmap["tt"] = vresmap["vt"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "m")      // -res=mと指定された時
      { vresmap["ma"] = vresmap["md"] = true; }
      else if (res[0] == 'm' && res[1] == 'a' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vma = mcbon;
          break;
        case 'f' : vma = mcfis;
          break;
        case 's' : vma = mcsch;
          break;
        case 't' : vma = mctuk;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["ma"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'd' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmd = mcbon;
          break;
        case 'd' : vmd = mcdun;
          break;
        case 'f' : vmd = mcfis;
          break;
        case 's' : vmd = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["md"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  map<string, listdbl> gfdmap;    // 群データマップ
  listdbl dumfdlst;               // ダミーの群データリスト
  vector<string> gcodevec;      // 群コードベクトル
  vector<statv<double>> gstatvec; // 基礎統計量ベクトル
  double vsa;                   // 要因A平方和
  int vdfa;                     // 要因A自由度
  double vsr;                   // 残差平方和
  int vdfr;                     // 残差自由度
  double vst;                   // 全体平方和
  int vdft;                     // 全体自由度
public:
  varclass() : gfdmap(), dumfdlst(), gcodevec(), gstatvec(), vsa(0.0), vdfa(0), vsr(0.0), vdfr(0), vst(0.0), vdft(0)
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int cal(ldata& datalst);      // データを集計して変数をセットする
  size_t ng() const             // 群数を返す
  { return gfdmap.size(); }
  const string& gcode(size_t idx) const // 群コードを返す
  { return (idx < gcodevec.size()) ? gcodevec[idx] : BLANKs; }
  listdbl& fdlst(size_t idx);   // 群データリストを返す
  int n(size_t idx) const       // 群の例数を返す
  { return (idx <= gfdmap.size()) ? gstatvec[idx].n() : gstatvec[gfdmap.size()].n(); }
  double mean(size_t idx) const // 群の平均値を返す
  { return (idx <= gfdmap.size()) ? gstatvec[idx].mean() : gstatvec[gfdmap.size()].mean(); }
  double ss(size_t idx) const   // 群の平方和を返す
  { return (idx <= gfdmap.size()) ? gstatvec[idx].ss() : gstatvec[gfdmap.size()].ss(); }
  double vr(size_t idx) const   // 群の不偏分散を返す
  { return (idx <= gfdmap.size()) ? gstatvec[idx].vr() : gstatvec[gfdmap.size()].vr(); }
  double sd(size_t idx) const   // 群の標準偏差を返す
  { return (idx <= gfdmap.size()) ? gstatvec[idx].sd() : gstatvec[gfdmap.size()].sd(); }
  double se(size_t idx) const   // 群の標準誤差を返す
  { return (idx <= gfdmap.size()) ? gstatvec[idx].se() : gstatvec[gfdmap.size()].se(); }
  int n() const                 // 全体の例数を返す
  { return gstatvec[gfdmap.size()].n(); }
  double mean() const           // 全体の平均値を返す
  { return gstatvec[gfdmap.size()].mean(); }
  double ss() const             // 全体の平方和を返す
  { return gstatvec[gfdmap.size()].ss(); }
  double vr() const             // 全体の不偏分散を返す
  { return gstatvec[gfdmap.size()].vr(); }
  double sd() const             // 全体の標準偏差を返す
  { return gstatvec[gfdmap.size()].sd(); }
  double se() const             // 全体の標準誤差を返す
  { return gstatvec[gfdmap.size()].se(); }
  double sa() const             // 要因A平方和を返す
  { return vsa; }
  int dfa() const               // 要因A自由度を返す
  { return vdfa; }
  double va() const             // 要因A分散を返す
  { return (gfdmap.size() > 1) ? vsa / (double)(gfdmap.size() - 1) : 0.0; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  double vrr() const            // 残差分散を返す
  { return (vdfr > 0) ? vsr / (double)vdfr : 0.0; }
  double st() const             // 全体平方和を返す
  { return vst; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
};

int varclass::cal(ldata& datalst) // データを集計して変数をセットする
{
  for (auto& cdata : datalst)
  {
    if (cdata.sdata(0) != BLANKs && cdata.fdata(1) != BLANKf)
    {
      map<string, listdbl>::iterator gitr = gfdmap.find(cdata.sdata(0));
      if (gitr == gfdmap.end()) // 新規群コード
      {
        listdbl tfdlst(1, cdata.fdata(1));
        gfdmap.insert(map<string, listdbl>::value_type(cdata.sdata(0), tfdlst));
      }
      else
      { gitr->second.push_back(cdata.fdata(1)); }
    }
  }
  size_t ng = gfdmap.size();    // 群数
  gcodevec.resize(ng);  gstatvec.resize(ng + 1);
  size_t i = 0;
  for (const auto& gfd : gfdmap)
  {
    gcodevec[i] = gfd.first;  // 群コードをべクルトに格納する
    for (double fdata : gfd.second) (gstatvec[i]).sigma(fdata); // データ累積
    ++i;
  }
  vsa = 0.0;
  for (size_t i = 0; i < ng; ++i)
  {
    double m = gstatvec[i].mean();
    vsa += (double)(gstatvec[i].n()) * m * m;
    gstatvec[ng] += gstatvec[i];
  }
  vdft = gstatvec[ng].n() - 1;  vdfa = ng - 1;
  vdfr = vdft - vdfa;
  vst = gstatvec[ng].ss();
  if (vdft < 1 || sgn(vst) <= 0) return 0; // 分散分析不可能
  double m = gstatvec[ng].mean();
  vsa -= (double)(gstatvec[ng].n()) * m * m;
  const double Seps = DBL_EPSILON * fmax(5.0, vst);
  if (vdfa < 1 || vsa < Seps) vsa = 0.0;
  vsr = vst - vsa;
  if (vdfr < 1 || vsr < Seps) vsr = 0.0;
  return (int)ng;               // 群数を返す
}

listdbl& varclass::fdlst(size_t idx) // 群のデータリストを返す
{
  if (idx >= gcodevec.size()) return dumfdlst; // ダミー群データを返す
  map<string, listdbl>::iterator gpos = gfdmap.find(gcodevec[idx]);
  if (gpos == gfdmap.end()) return dumfdlst; // ダミー群データを返す
  return gfdmap[gcodevec[idx]];              // 群データリストを返す
}

class anova1class               // anova1メインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 分析結果のタイトルとデータ識別情報を出力する
  void put_res(ostream& fout, option& opt, idd& id, const varclass& var) const;
  // 解析結果を出力する
  void put_res2g(ostream& fout, option& opt, const varclass& var) const;
  // 2群比較の解析結果を出力する
  void put_resmg(ostream& fout, option& opt, const varclass& var) const;
  // 3群以上の時の解析結果を出力する
  void put_stat(ostream& fout, const varclass& var) const;
  // 基礎統計量を出力する
  void put_mci(ostream& fout, const option& opt, const varclass& var) const;
  // 群別平均値の信頼区間出力
  void put_rratio(ostream& fout, const option& opt, const varclass& var) const;
  // 相関比の出力
  void put_ttest(ostream& fout, option& opt, const varclass& var) const;
  // 2標本t検定とWelchの検定
  void put_ftest(ostream& fout, option& opt, double f, int df1, int df2, ldbl pr) const;
  // 分散比Fの両側検定と推定結果を出力して両側p値を返す
  void put_pa(ostream& fout, const option& opt, const varclass& var) const;
  // 検出力分析の検定
  void put_vtest(ostream& fout, const varclass& var) const;
  // 分散の均一性の検定結果出力
  void put_mcop(ostream& fout, const varclass& var, mctype mc, bool ma) const;
  // 2群間の多重比較出力
  void put_mdci(ostream& fout, const option& opt, const varclass& var, mctype mc, bool ma) const;
  // 2群間の平均値の差の信頼区間出力
  void put_mpa(ostream& fout, const option& opt, const varclass& var, mctype mc, bool ma) const;
  // 2群間の平均値の差の検出力分析結果出力
  void put_dns(ostream& fout, option& opt, const varclass& var) const;
  // 分散分析の結果をデータファイル形式で出力する
  void plot_ngp(option& opt, idd& id, const ldata& datalst, varclass& var) const;
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_bar(ofstream& fout, const varclass& var) const;
  // Ngraph用にバーチャートデータを出力する
  void plot_box(ofstream& fout, varclass& var) const;
  // Ngraph用に箱ヒゲプロットデータを出力する
  void plot_ci(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に平均±信頼区間データを出力する
  void plot_data(ofstream& fout, varclass& var) const;
  // Ngraph用に個々のデータを出力する
  void plot_msd(ofstream& fout, const varclass& var) const;
  // Ngraph用に平均±SDを出力する
  void plot_mse(ofstream& fout, const varclass& var) const;
  // Ngraph用に平均±SEを出力する
  void plot_sdt(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に平均±係数×SDを出力する
  void plot_srt(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に平均±係数×残差SDを出力する
public:
  anova1class() : fname()       // コンストラクタ
  {}
  ~anova1class()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込む
  void put(option& opt, idd& id, ldata& datalst) const; // 解析結果を出力する
};

void anova1class::put_resid(ostream& fout, idd& id) const
// 分析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM(A):" << id.item(0) << "\n";
  fout << "ANALYSIS ITEM:" << id.item(1) << "\n";
}

void anova1class::put_res(ostream& fout, option& opt, idd& id, const varclass& var) const
// 解析結果を出力する
{
  put_resid(fout, id);
  if (opt.res("stat"))          // 基礎統計量出力
  {
    fout << "\n";
    put_stat(fout, var);
  }
  if (var.dft() < 1) return;    // 分散分析不可能
  if (opt.res("mci"))           // 平均値の信頼区間出力
  {
    fout << "\n";
    put_mci(fout, opt, var);
  }
  if (var.ng() == 2) put_res2g(fout, opt, var); // 2群の時
  else put_resmg(fout, opt, var); // 3群以上の時
}

void anova1class::put_res2g(ostream& fout, option& opt, const varclass& var) const
// 2群比較の解析結果を出力する
{
  if (opt.res("r"))             // 相関比出力
  {
    fout << "\n";
    put_rratio(fout, opt, var);
  }
  if (opt.res("at"))           // 2群の時は-res=atの時だけ分散分析出力
  {
    if (opt.res("vt"))          // 等分散性の検定出力
    {
      fout << "\n";
      put_vtest(fout, var);
    }
    fout << "\n<ONE-WAY LAYOUT ANALYSIS OF VARIANCE(ANOVA1)>\n";
    putatab(fout, "GROUP(A)", var.sa(), var.dfa(), var.sr(), var.dfr(), var.st(), var.dft());
  }
  if (opt.res("tt"))            // 2標本t検定出力
  {
    fout << "\n";
    put_ttest(fout, opt, var);
  }
  if (opt.res("pa"))            // 検出力分析出力
  {
    fout << "\n";
    put_pa(fout, opt, var);
  }
}

void anova1class::put_resmg(ostream& fout, option& opt, const varclass& var) const
// 3群以上の時の解析結果を出力する
{
  if (opt.res("def") || opt.res("r")) // 省略時または-res=rの時に相関比出力
  {
    fout << "\n";
    put_rratio(fout, opt, var);
  }
  if (opt.res("vt"))            // 等分散性の検定出力
  {
    fout << "\n";
    put_vtest(fout, var);
  }
  if (opt.res("def") || opt.res("at")) // 省略時または-res=atの時に分散分析出力
  {
    fout << "\n";
    fout << "<ONE-WAY LAYOUT ANALYSIS OF VARIANCE(ANOVA1)>\n";
    putatab(fout, "GROUP(A)", var.sa(), var.dfa(), var.sr(), var.dfr(), var.st(), var.dft());
  }
  if (opt.res("ma"))            // 全ての2群の多重比較出力
  {
    fout << "\n";
    put_mcop(fout, var, opt.ma(), true);
    if (opt.res("cc"))          // 信頼区間出力
    {
      fout << "\n";
      put_mdci(fout, opt, var, opt.ma(), true);
    }
    if (opt.res("pa"))          // 検出力分析出力
    {
      fout << "\n";
      put_mpa(fout, opt, var, opt.ma(), true);
    }
  }
  if (opt.res("md"))            // 群1と他群の多重比較出力
  {
    fout << "\n";
    put_mcop(fout, var, opt.md(), false);
    if (opt.res("cc"))          // 信頼区間出力
    {
      fout << "\n";
      put_mdci(fout, opt, var, opt.md(), false);
    }
    if (opt.res("pa"))          // 検出力分析出力
    {
      fout << "\n";
      put_mpa(fout, opt, var, opt.md(), false);
    }
  }
}

void anova1class::put_stat(ostream& fout, const varclass& var) const
// 基礎統計量を出力する
{
  constexpr int LENline = 76;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  fout << "GROUP     :BASIC STATISTICS\n";
  putline(fout, LENline);
  for (size_t i = 0; i < var.ng(); ++i)
  {
    string gcodei(var.gcode(i));
    int npad = gcodei.size() - strclen(gcodei, CHARwide);
    fout << left << setw(LENhed + npad) << gcodei << right << ":";
    putmsd(fout, var.n(i), var.mean(i), var.vr(i));
  }
  putline(fout, LENline);
  string tstr("TOTAL");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(LENhed + npad) << tstr << right << ":";
  putmsd(fout, var.n(), var.mean(), var.vr());
}

void anova1class::put_mci(ostream& fout, const option& opt, const varclass& var) const
// 群別平均値の信頼区間を出力する
{
  constexpr int LENline = 58;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  fout << "<" << opt.cc() * 100.0 << "% CONFIDENCE INTERVALS OF MEANS>\n";
  fout << "GROUP             Mean   Precision   Lower-Lim   Upper-Lim\n";
  putline(fout, LENline);
  for (size_t i = 0; i < var.ng(); ++i)
  {
    string gcodei(var.gcode(i));
    int npad = gcodei.size() - strclen(gcodei, CHARwide);
    int df = var.n(i) - 1;
    double me = var.mean(i);
    double t = (df > 0) ? (double)(ptot((ldbl)(1.0 - opt.cc()), df)) : 0.0;
    double ci = t * var.se(i);
    fout << left << setw(LENhed + npad) << gcodei << right;
    fout  << setw(12) << me << setw(12) << ci << setw(12) << me - ci << setw(12) << me + ci << "\n";
  }
  putline(fout, LENline);
  string tstr("TOTAL");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  int df = var.n() - 1;
  double me = var.mean();
  double t = (df > 0) ? (double)(ptot((ldbl)(1.0 - opt.cc()), df)) : 0.0;
  double ci = t * var.se();
  fout << left << setw(LENhed + npad) << tstr << right;
  fout  << setw(12) << me << setw(12) << ci << setw(12) << me - ci << setw(12) << me + ci << "\n";
}

void anova1class::put_rratio(ostream& fout, const option& opt, const varclass& var) const
// 相関比を出力する
{
  double r2 = (var.st() > 0.0) ? var.sa() / var.st() : 0.0;
  fout << "Correlation Ratio:r=" << sqrt(r2) << "  r-squared=" << r2 << "\n";
  double f = (var.vrr() > 0.0) ? var.va() / var.vrr() : 0.0;
  double crl = 0.0, cru = 0.0;
  calcrci(crl, cru, opt.cc(), var.dfa(), var.dfr(), f);
  fout << opt.cc() * 100.0 << "% CI of r:Lower=" << crl << "  Upper=" << cru << "\n";
}

void anova1class::put_ttest(ostream& fout, option& opt, const varclass& var) const
// 2標本t検定とWelchの検定結果を出力する
{
  int n1 = var.n(0), n2 = var.n(1);
  double ss1 = var.ss(0), ss2 = var.ss(1);
  if (n1 < 2 || n2 < 2 || (ss1 <= 0.0 && ss2 <= 0.0)) return; // 出力なし
  double v1 = var.vr(0), v2 = var.vr(1);
  double f = (v2 > 0.0) ? v1 / v2 : 0.0;
  int df1 = n1 - 1, df2 = n2 - 1;
  if (v1 < v2)
  {
    f = (v1 > 0.0) ? v2 / v1 : 0.0;
    df1 = n2 - 1;  df2 = n1 - 1;
  }
  ldbl pr = ftop((ldbl)f, df1, df2);
  pr = (pr < 0.5L) ? 2.0L * pr : 1.0L; // 両側有意確率にする
  if (opt.res("ft"))                   // 等分散性の両側F検定出力
  {
    put_ftest(fout, opt, f, df1, df2, pr);
    fout << "\n";
  }
  fout << "<TWO SAMPLE T-TEST(->:Adopt This Result)>\n";
  string sel1 = "->", sel2 = "  ";
  if (pr <= 0.05L) { sel1 = "  ";  sel2 = "->"; }
  double d = var.mean(1) - var.mean(0);
  int df = n1 + n2 - 2;
  double vd = (ss1 + ss2) / df * (double)(n1 + n2) / (double)n1 / (double)n2;
  double t = (vd > 0.0) ? d / sqrt(vd) : 0.0;
  pr = ttop((ldbl)t, df);
  fout << sel1 << "t-test of Equal Variances(=ANOVA1):t=" << t;
  fout << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
  double ci = 0.0;
  if (opt.res("cc"))            // 平均値の差の信頼区間出力
  {
    ci = calmcci(mcfis, opt.cc(), 2, 1, df, 1, vd);
    fout << sel1 << opt.cc() * 100.0 << "% CI of Difference in Means:";
    fout << "Lower=" << d - ci << "  Upper=" << d + ci << "\n";
  }
  v1 /= n1;  v2 /= n2;
  vd = v1 + v2;
  df = (int)(vd * vd / (v1 * v1 / (n1 - 1) + v2 * v2 / (n2 -1)));
  t = (vd > 0.0) ? d / sqrt(vd) : 0.0;
  pr = ttop((ldbl)t, df);
  fout << sel2 << "Aspin-Welch Test:t=" << t;
  fout << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
  if (opt.res("cc"))            // 平均値の差の信頼区間出力
  {
    ci = calmcci(mcfis, opt.cc(), 2, 1, df, 1, vd);
    fout << sel2 << opt.cc() * 100.0 << "% CI of Difference in Means:";
    fout << "Lower=" << d - ci << "  Upper=" << d + ci << "\n";
  }
}

void anova1class::put_ftest(ostream& fout, option& opt, double f, int df1, int df2, ldbl pr) const
// 分散比Fの両側検定と推定結果を出力する
{
  fout << "<TWO-TAILED F-TEST TO COMPARE TWO VARIANCES>\n";
  fout << "F=" << f << "  DF1=" << df1 << "  DF2=" << df2 << "  two-tailed p=" << pr << pind(pr) << "\n";
  if (opt.res("cc"))            // 分散比の信頼区間出力
  {
    double fl = 0.0, fu = 0.0;
    calfci(fl, fu, opt.cc(), df1, df2, f);
    fout << opt.cc() * 100.0 << "% CI of F:Lower=" << fl << "  Upper=" << fu << "\n";
  }
}

void anova1class::put_pa(ostream& fout, const option& opt, const varclass& var) const
// 検出力分析結果を出力する
{
  int n1 = var.n(0), n2 = var.n(1);
  double ss1 = var.ss(0), ss2 = var.ss(1);
  if (n1 < 2 || n2 < 2 || (ss1 <= 0.0 && ss2 <= 0.0)) return; // 出力なし
  int nt = n1 + n2;
  int df = nt - 2;
  double sd = sqrt((ss1 + ss2) / (double)df);
  double dm = (opt.dm() != BLANKf) ? opt.dm() : fabs(var.mean(1) - var.mean(0));
  const double SDeps = DBL_EPSILON * fmax(5.0, sd);
  if (dm < SDeps) dm = sd;      // 平均値の差が0の時はSDにする
  double d = dm / sd;
  double lamda = d / sqrt(1.0 / (double)n1 + 1.0 / (double)n2);
  double ta = (double)(ptot((ldbl)(opt.al()), df));
  double t2b = lamda - ta;
  double b = (t2b > 0.0) ? (double)(ttop((ldbl)t2b, df)) * 0.5 : 1.0 - (double)(ttop((ldbl)(-t2b), df)) * 0.5;
  double b0 = (double)(ncttolp((ldbl)ta, df, (ldbl)lamda));
  fout << "<POWER ANALYSIS>\n";
  fout << "Significance Level=" << opt.al() << "  SD of Diff of Means=" << sd << "\n";
  fout << "power=" << 1.0 - b0 << " at Specified Diff=" << dm << "(Effect Size=" << d << ")\n";
  t2b = (double)(ptot((ldbl)(2.0 * (1.0 - opt.po())), df));
  double s = (double)n2 / (double)n1;
  double nc = ta * ta / 4.0 / (1.0 - 1.0 / (double)nt);
  if (n1 <= nc) nc = 0.0;
  d = (ta + t2b) * sqrt((1.0 + s) / s / ((double)n1 - nc));
  dm = d * sd;
  fout << "Diff of Means=" << dm << "(Effect Size=" << d << ") at Specified Power=" << opt.po() << "\n";
  ta = (double)(ptoz((ldbl)(opt.al())));
  t2b = (double)(ptoz((ldbl)(2.0 * (1.0 - opt.po()))));
  if (opt.po() < 0.5) t2b = (double)(-ptoz((ldbl)(2.0 * opt.po())));
  dm = (opt.dm() != BLANKf) ? opt.dm() : fabs(var.mean(1) - var.mean(0));
  if (dm < SDeps) dm = sd;      // 平均値の差が0の時はSDにする
  d = dm / sd;
  double nsqrt = (ta + t2b) / d;
  double n0 = 2.0 * nsqrt * nsqrt;
  n1 = (int)(n0 + 0.99);
  n1 = (n1 < 2) ? n1 + 1 : (int)(n0 + ta * ta / 4.0 / (1.0 - 1.0 / (double)n1) + 0.99);
  fout << "n/group=" << n1 << "(Total n=" << n1 * 2 << ")";
  fout << " at Specified Power=" << opt.po() << " & Diff=" << dm << endl;
}

void anova1class::put_vtest(ostream& fout, const varclass& var) const
// 分散の均一性の検定結果を出力する
{
  int ng = var.ng();
  int d = -ng;
  double w = 0.0, ss = 0.0, lnss = 0.0;
  for (int i = 0; i < ng; ++i)
  {
    int df = var.n(i) - 1;
    d += var.n(i);
    w += 1.0 / (double)df;
    ss += (double)df * var.vr(i);
    lnss += (double)df * log(var.vr(i));
  }
  double m = d * log(ss / d) - lnss;
  double c = 1.0 + 1.0 / 3.0 / (double)(ng - 1) * (w - 1.0 / (double)d);
  double csqr = m / c;
  int df = ng - 1;
  ldbl p = ctop((ldbl)csqr, df);
  fout << "Bartlett Test:chi-squared=" << csqr;
  fout << "  DF=" << df << "  p=" << p << pind(p) << "\n";
}

void anova1class::put_mcop(ostream& fout, const varclass& var, mctype mc, bool ma) const
// 2群間の多重比較結果を出力する
{
  const int LENline = 76;       // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  double vr = var.vrr();        // 残差分散
  if (vr <= 0.0) return;        // 出力なし
  int ng = var.ng();
  int df1 = 1, df2 = var.dfr(), dc = 1;
  string method = "TUKEY";
  switch (mc)
  {
  case mcbon : method = "BONFERRONI";
    dc = ma ? ng * (ng - 1) / 2 : ng - 1;
    break;
  case mcdun : method = "DUNNETT";
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";
    df1 = ng - 1;
  }
  fout << "<" << method << " TYPE MULTIPLE COMPRISONS OF MEANS>\n";
  if (mc == mctuk)
  {
    fout << "GROUP     - GROUP          q-value      Groups          DF       p-value\n";
  }
  else if (mc == mcdun)
  {
    fout << "GROUP     - GROUP          d-value      Groups          DF       p-value\n";
  }
  else
  {
    fout << "GROUP     - GROUP          F-value         DF1         DF2       p-value\n";
  }
  putline(fout, LENline);
  int iend = ma ? ng - 1 : 1;
  for (int i = 0; i < iend; ++i)
  {
    string gcodei(var.gcode(i));
    int npadi = gcodei.size() - strclen(gcodei, CHARwide);
    double ni = (double)(var.n(i));
    double mi = var.mean(i);
    for (int j = i + 1; j < ng; ++j)
    {
      double nj = (double)(var.n(j));
      double dm = var.mean(j) - mi;
      double f = dm * dm / df1 / (1.0 / ni + 1.0 / nj) / vr;
      double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
      ldbl pr = calmcpr(mc, f, ng, df1, df2, dc, r);
      string gcodej(var.gcode(j));
      int npad = gcodej.size() - strclen(gcodej, CHARwide);
      fout << left << setw(LENhed + npad) << gcodej;
      fout << "- ";
      fout << setw(LENhed + npadi) << gcodei << right << setw(12) << f;
      if (mc == mctuk || mc == mcdun) fout << setw(12) << ng;
      else fout << setw(12) << df1;
      fout << setw(12) << df2 << setw(14) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, LENline);
}

void anova1class::put_mdci(ostream& fout, const option& opt, const varclass& var, mctype mc, bool ma) const
// 2群間の平均値の差の信頼区間を出力する
{
  constexpr int LENline = 70;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  double vr = var.vrr();        // 残差分散
  if (vr <= 0.0) return;        // 出力なし
  int ng = var.ng();
  int df1 = 1, df2 = var.dfr(), dc = 1;
  string method = "TUKEY";
  switch (mc)
  {
  case mcbon : method = "BONFERRONI";
    dc = ma ? ng * (ng - 1) / 2 : ng - 1;
    break;
  case mcdun : method = "DUNNETT";
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";
    df1 = ng - 1;
  }
  fout << "<" << method << " TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CONFIDENCE INTERVALS>\n";
  fout << "GROUP     - GROUP    Diff of Means   Precision   Lower-Lim   Upper-Lim\n";
  putline(fout, LENline);
  int iend = ma ? ng - 1 : 1;
  for (int i = 0; i < iend; ++i)
  {
    string gcodei(var.gcode(i));
    int npadi = gcodei.size() - strclen(gcodei, CHARwide);
    int ni = var.n(i);
    double mi = var.mean(i);
    for (int j = i + 1; j < ng; ++j)
    {
      int nj = var.n(j);
      double dm = var.mean(j) - mi;
      double vd = (1.0 / ni + 1.0 / nj) * vr;
      double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
      double ci = calmcci(mc, opt.cc(), ng, df1, df2, dc, vd, r);
      string gcodej(var.gcode(j));
      int npad = gcodej.size() - strclen(gcodej, CHARwide);
      fout << left << setw(LENhed + npad) << gcodej;
      fout << "- ";
      fout << setw(LENhed + npadi) << gcodei << right << setw(12) << dm;
      fout << setw(12) << ci << setw(12) << dm - ci << setw(12) << dm + ci;
      fout << "\n";
    }
  }
  putline(fout, LENline);
}

void anova1class::put_mpa(ostream& fout, const option& opt, const varclass& var, mctype mc, bool ma) const
// 2群間の平均値の差の検出力分析結果出力
{
  constexpr int LENline = 70;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  double vr = var.vrr();        // 残差分散
  if (vr <= 0.0) return;        // 出力なし
  double sd = sqrt(vr);
  int ng = var.ng();
  int df1 = 1, df2 = var.dfr(), dc = 1;
  double poc = opt.po(), ac = opt.al();
  double bc = 1.0 - poc, bcc = 1.0;
  double tac = (double)(ptot((ldbl)ac, df2)), zac = (double)(ptoz((ldbl)ac));
  double t2bc = (poc > 0.5) ? (double)(ptot((ldbl)(2.0 * bc), df2)) : (ldbl)(-ptot((ldbl)(2.0 * poc), df2));
double z2bc = (poc > 0.5) ? (double)(ptoz((ldbl)(2.0 * bc))) : (double)(-ptoz((ldbl)(2.0 * poc)));
  string method = "FISHER";
  switch (mc)
  {
  case mcdun :
  case mctuk :
  case mcbon : method = "BONFERRONI";
    dc = ma ? ng * (ng - 1) / 2 : ng - 1;
    ac = opt.al() / dc;
    tac = (double)(ptot((ldbl)ac, df2));
    zac = (double)(ptoz((ldbl)ac));
    break;
  case mcsch : method = "SCHEFFE";
    df1 = ng - 1;
    tac = (double)(sqrt(ptof((ldbl)ac, df1, df2) * (ldbl)df1));
    zac = (double)(sqrt(ptoc((ldbl)ac, df1)));
  }
  fout << "<POWER ANALYSIS:" << method << " TYPE MULTIPLE COMPARISONS>\n";
  fout << "Significance Level=" << opt.al() << "  power=" << opt.po() << "\n";
  fout << "GROUP     - GROUP    Diff of Means   Adj-Power    Adj-Diff     n/group\n";
  putline(fout, LENline);
  const double SDeps = DBL_EPSILON * fmax(5.0, sd);
  int iend = ma ? ng - 1 : 1;
  for (int i = 0; i < iend; ++i)
  {
    string gcodei(var.gcode(i));
    int npadi = gcodei.size() - strclen(gcodei, CHARwide);
    double ni = (double)(var.n(i));
    double mi = var.mean(i);
    for (int j = i + 1; j < ng; ++j)
    {
      double nj = (double)(var.n(j));
      double dm = (opt.dm() != BLANKf) ? opt.dm() : fabs(var.mean(j) - mi);
      if (dm < SDeps) dm = sd; // 平均値の差が0の時はSDにする
      double lamda = dm / sqrt((1.0 / ni + 1.0 / nj) * vr);
      double b = ncttolp(tac, df2, lamda);
      double po = 1.0 - b;
      bcc *= b;
      double s = nj / ni;
      double nc = tac * tac / 4.0 / (1.0 - 1.0 / (ni + nj));
      if (ni <= nc) nc = 0.0;
      double d = (tac + t2bc) * sqrt((1.0 + s) / s / (ni - nc));
      double dmc = d * sd;
      double nsqrt = (zac + z2bc) * sd / dm;
      double n0 = 2.0 * nsqrt * nsqrt;
      int n = (int)(n0 + 0.99);
      n = (n < 2) ? n + 1 : (int)(n0 + zac * zac / 4.0 / (1.0 - 1.0 / (double)n) + 0.99);
      string gcodej(var.gcode(j));
      int npad = gcodej.size() - strclen(gcodej, CHARwide);
      fout << left << setw(LENhed + npad) << gcodej;
      fout << "- ";
      fout << setw(LENhed + npadi) << gcodei << right << setw(12) << dm;
      fout << setw(12) << po << setw(12) << dmc << setw(12) << n;
      fout << "\n";
    }
  }
  putline(fout, LENline);
  if (mc != mcfis) fout << "Adjusted Total Power=" << 1.0 - bcc << "(beta=" << bcc << ")" << "\n";
}

void anova1class::plot_ngp(option& opt, idd& id, const ldata& datalst, varclass& var) const
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("bar"))           // バーチャート
  {
    fout << "-s1 -zx0," << var.ng() + 1 << ",1 -drectangle_solid_fill -x1 -y3\n";
    plot_ngpid(fout, id);
    plot_bar(fout, var);
    delml = true;
  }
  if (opt.ngp("box"))           // 箱ひげプロット
  {
    if (delml)                   // 区切り行を出力する
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << var.ng() + 1 << ",1 -drectangle -x1 -y3 -w80\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_box(fout, var);
  }
  if (opt.ngp("ci"))            // 平均値±CI
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << var.ng() + 1 << ",1 -dmark,78 -w80 -o400\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_ci(fout, opt, var);
    delml = true;
  }
  if (opt.ngp("data"))          // 個々のデータ
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << var.ng() + 1 << ",1 -dmark,00 -o300\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_data(fout, var);
  }
  if (opt.ngp("msd"))           // 平均値±SD
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << var.ng() + 1 << ",1 -dline -w80\n";
      fout << "%When drawing error bars:-derrorbar_y -w80 -o500\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_msd(fout, var);
  }
  if (opt.ngp("mse"))           // 平均値±SE
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << var.ng() + 1 << ",1 -dline -w80\n";
      fout << "%When drawing error bars:-derrorbar_y -w80 -o500\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_mse(fout, var);
  }
  if (opt.ngp("sd"))            // 平均値±係数×SDの折れ線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << var.ng() + 1 << ",1 -dline -w80\n";
      fout << "%When drawing error bars:-derrorbar_y -w80 -o500\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_sdt(fout, opt, var);
  }
  if (opt.ngp("sr"))            // 平均値±係数×残差SD
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << var.ng() + 1 << ",1 -dline -w80\n";
      fout << "%When drawing error bars:-derrorbar_y -w80 -o500\n";
    }
    plot_ngpid(fout, id);
    plot_srt(fout, opt, var);
  }
}

void anova1class::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%GROUP(FACTOR-A):" << id.item(0) << "\n";
  fout << "%ANALYSIS ITEM  :" << id.item(1) << "\n";
}

void anova1class::plot_bar(ofstream& fout, const varclass& var) const
// Ngraph用にバーチャートデータを出力する
{
  fout << "%<BAR CHART>\n";
  for (size_t i = 0; i < var.ng(); ++i)
  {
    double x = (double)(i + 1);
    fout << x - 0.2 << " 0 " << x + 0.2 << " " << var.mean(i) << "\n";
  }
}

struct pro                      // 最小・最大値と髭の下限・上限と四分位値
{
  double min, wlow, q1, q2, q3, wup, max;
};

void anova1class::plot_box(ofstream& fout, varclass& var) const
// Ngraph用に箱ヒゲプロットデータを出力する
{
  size_t ng = var.ng();
  vector<struct pro> provec(ng);
  fout << "%<BOX-WHISKER PLOT:BOX>\n";
  for (size_t i = 0; i < ng; ++i)
  {
    int n = var.n(i);
    vecdbl fdvec(n);
    (var.fdlst(i)).sort();      // 群のデータリストをソートする
    int j = 0;
    for (listdbl::const_iterator fditr = (var.fdlst(i)).begin(); fditr != (var.fdlst(i)).end(); ++fditr, ++j) fdvec[j] = *fditr;
    provec[i].min = provec[i].wlow = fdvec[0];
    provec[i].max = provec[i].wup = fdvec[n - 1];
    provec[i].q1 = fdvec[0];  provec[i].q2 = fdvec[0];
    provec[i].q3 = fdvec[n - 1];
    if (n > 1)
    {
      double q1fidx = (double)(n - 1) * 0.25;
      double q2fidx = (double)(n - 1) * 0.5;
      double q3fidx = (double)(n - 1) * 0.75;
      int q1idx = (int)q1fidx, q2idx = (int)q2fidx, q3idx = (int)q3fidx;
      provec[i].q1 = fdvec[q1idx] + (fdvec[q1idx + 1] - fdvec[q1idx]) * (q1fidx - (double)q1idx);
      provec[i].q2 = fdvec[q2idx] + (fdvec[q2idx + 1] - fdvec[q2idx]) * (q2fidx - (double)q2idx);
      provec[i].q3 = fdvec[q3idx] + (fdvec[q3idx + 1] - fdvec[q3idx]) * (q3fidx - (double)q3idx);
    }
    double h = provec[i].q3 - provec[i].q1;
    double hlow = provec[i].q1 - h * 1.5, hup = provec[i].q3 + h * 1.5;
    bool schwlow = true, schwup = true;
    for (j = 0; (schwlow || schwup) && j < n; ++j)
    {
      if (schwlow && hlow <= fdvec[j])
      { schwlow = false;  provec[i].wlow = fdvec[j]; }
      if (fdvec[j] <= hup) provec[i].wup = fdvec[j];
      else schwup = false;
    }
    double x = (double)i + 0.9;
    fout << x << " " << provec[i].q1 << " ";
    fout << x + 0.2 << " " << provec[i].q2 << "\n";
    fout << x << " " << provec[i].q2 << " ";
    fout << x + 0.2 << " " << provec[i].q3 << "\n";
  }
  fout << "= =\n";
  fout << "%<BOX-WHISKER PLOT:WHISKER(-zx0," << ng + 1 << ",1 -derrorbar_y -x1 -y2 -w40 -o500)>\n";
  for (size_t i = 0; i < ng; ++i)
  {
    fout << i + 1 << " " << provec[i].q1;
    fout << " " << provec[i].wlow - provec[i].q1 << " 0\n";
    fout << i + 1 << " " << provec[i].q3;
    fout << " 0 " << provec[i].wup - provec[i].q3 << "\n";
  }
  fout << "= =\n";
  fout << "%<BOX-WHISKER PLOT:MEAN(-s1 -zx0," << ng + 1 << ",1 -dmark,72 -w80 -o400)>\n";
  for (size_t i = 0; i < ng; ++i) fout << i + 1 << " " << var.mean(i) << "\n";
  fout << "= =\n";
  fout << "%<BOX-WHISKER PLOT:OUTLIER(-s1 -zx0," << ng + 1 << ",1 -dmark,2 -w80 -o400)>\n";
  for (size_t i = 0; i < ng; ++i)
  {
    for (double fdata : var.fdlst(i))
    {
      if (fdata < provec[i].wlow) fout << i + 1 << " " << fdata << "\n";
      else if (provec[i].wup < fdata) fout << i + 1 << " " << fdata << "\n";
    }
  }
}

void anova1class::plot_ci(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に平均±信頼区間データを出力する
{
  size_t ng = var.ng();
  fout << "%<" << opt.cc() * 100.0 << "% confident intervals(-zx0," << ng + 1 << ",1 -derrorbar_y -w0 -o500)>\n";
  for (size_t i = 0; i < ng; ++i)
  {
    double t = (double)(ptot((ldbl)(1.0 - opt.cc()), var.n(i) - 1));
    double ci = t * var.se(i);
    fout << i + 1 << " " << var.mean(i) << " " << ci << " " << -ci << "\n";
  }
}

void anova1class::plot_data(ofstream& fout, varclass& var) const
// Ngraph用に個々のデータを出力する
{
  fout << "%<RAW DATA>\n";
  for (size_t i = 0; i < var.ng(); ++i)
  {
    for (double fdata : var.fdlst(i)) fout << i + 1 << " " << fdata << "\n";
  }
}

void anova1class::plot_msd(ofstream& fout, const varclass& var) const
// Ngraph用に平均±SDを出力する
{
  size_t ng = var.ng();
  fout << "%<MEAN AND SD(-zx0," << ng + 1 << ",1 -derrorbar_y -w40 -o500)>\n";
  for (size_t i = 0; i < ng; ++i)
  {
    double sd = var.sd(i);
    fout << i + 1 << " " << var.mean(i) << " " << sd << " " << -sd << "\n";
  }
}

void anova1class::plot_mse(ofstream& fout, const varclass& var) const
// Ngraph用に平均±SEを出力する
{
  size_t ng = var.ng();
  fout << "%<MEAN AND SE(-zx0," << ng + 1 << ",1 -derrorbar_y -w40 -o500)>\n";
  for (size_t i = 0; i < ng; ++i)
  {
    double se = var.se(i);
    fout << i + 1 << " " << var.mean(i) << " " << se << " " << -se << "\n";
  }
}

void anova1class::plot_sdt(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に平均±係数×SDを出力する
{
  int dfr = var.dfr();          // 残差自由度
  double t = (opt.tv() != BLANKf) ? opt.tv() : (double)(ptot((ldbl)(1.0 - opt.cc()), dfr));
  fout << "%<MEAN AND " << t << "×SD>\n";
  for (size_t i = 0; i < var.ng(); ++i) fout << i + 1 << " " << var.mean(i) << "\n";
  fout << "= =\n";
  fout << "%Lower Limit\n";
  for (size_t i = 0; i < var.ng(); ++i)
  { fout << i + 1 << " " << var.mean(i) - t * var.sd(i) << "\n"; }
  fout << "= =\n";
  fout << "%Upper Limit\n";
  for (size_t i = 0; i < var.ng(); ++i)
  { fout << i + 1 << " " << var.mean(i) + t * var.sd(i) << "\n"; }
}

void anova1class::plot_srt(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に平均±係数×残差SDを出力する
{
  int dfr = var.dfr();          // 残差自由度
  double t = (opt.tv() != BLANKf) ? opt.tv() : (double)(ptot((ldbl)(1.0 - opt.cc()), dfr));
  double tsd = t * sqrt(var.vrr());
  fout << "%<MEAN AND " << t << "×RESIDUAL SD=MEAN AND " << tsd << ">\n";
  for (size_t i = 0; i < var.ng(); ++i) fout << i + 1 << " " << var.mean(i) << "\n";
  fout << "= =\n";
  fout << "%Lower Limit\n";
  for (size_t i = 0; i < var.ng(); ++i)
  { fout << i + 1 << " " << var.mean(i) - tsd << "\n"; }
  fout << "= =\n";
  fout << "%Upper Limit\n";
  for (size_t i = 0; i < var.ng(); ++i)
  { fout << i + 1 << " " << var.mean(i) + tsd << "\n"; }
}

int anova1class::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;                    // 読み込んだケース数を返す
}

void anova1class::put(option& opt, idd& id, ldata& datalst) const
// 解析結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var;                 // 計算用変数クラス
  if (var.cal(datalst) < 1) return; // 群なし
  if (opt.res("res"))
  {
    put_res(cout, opt, id, var);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(opt, id, datalst, var); // Ngraph用データ出力
}

int main(int argc, char* argv[]) // メインルーチン
{
  anova1class anova1;           // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) anova1.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { anova1.get(argv[i], opt, id, datalst);  fok = false; }
    }
    anova1.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
