// anova2e.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-11 14:42:49 tac>
//
// 対応のある多時期間で二元配置分散分析を行う。
// 繰り返し測定型二元配置分散分析も含む。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -w anova2e anova2e.C -I/usr/local/include
//              -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "anova2e  Version 7.42e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:anova2e [<OPTION> ...] [<DATA FILE> ...]",
  "  The two-way ANOVA with the first item as the group(factor A)",
  "  and the subsequent items as the period(factor B).",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -res=[-][{def|all|a2|cc[<C-COEF)>]|icc|m[{a|b|c|d|e|f|m|x}][{b|d|f|s|t}]",
  "       |rm|rp[<NUMBER>]|sd|stat}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    a2:standard ANOVA2  cc[<C-COEF>]:confidence intervals",
  "    icc:intraclass correlation coefficients",
  "    m[{a|b|c|d|e|f|m|x}][{b|d|f|s|t}]:multiple comparisons",
  "      ma:all 2 group  mb:period-1 vs other periods  mc:all 2 periods",
  "      md:group-1 vs other groups  me:group-1 vs other groups by period",
  "      mf:group-1 vs other groups of interactions",
  "      mm:all groups comparisons by period",
  "      mx:all groups comparisons of interactions",
  "      b:Bonferroni type  d:Dunnett type  f:Fisher type  s:Scheffe type",
  "      t:Tukey type",
  "    rm:two-way RMANOVA  rp[<NUMBER>]:pared ANOVA2",
  "    sd:SD of each factor  stat:basic statistics",
  "    default:auto judgment according to data"
};

const string Title = "=== Comparison of Means of Multiple Groups & Periods ==="; // 表題

using vecstatv = vector<statv<double>>; // 統計量ベクトル配列の別名定義

enum atabtype                   // 分散分析表の種類
{
  atabnon,                      // 出力なし
  at2,                          // 繰り返しのない二元配置分散分析表
  at2r,                         // 繰り返しのある二元配置分散分析表
  at2rm,                        // 繰り返し測定型二元配置分散分析表
  at2rp,                        // 対応のある二元配置分散分析表
  atabetc                       // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  mctype vma;                   // 群の多重比較法(全群比較)
  mctype vmb;                   // 時期の多重比較法(時期1との比較)
  mctype vmc;                   // 時期の多重比較法(全時期比較)
  mctype vmd;                   // 群の多重比較法(群1との比較)
  mctype vme;                   // 時期別平均値の多重比較法(群1との比較)
  mctype vmf;                   // 交互作用の多重比較法(群1との比較)
  mctype vmm;                   // 時期別平均値の多重比法較(全群比較)
  mctype vmx;                   // 交互作用の多重比較法(全群比較)
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  int vrp;                      // 対応のある二元配置分散分析の比較回数
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  mctype ma() const             // 群の多重比較法を返す(全群比較)
  { return vma; }
  mctype mb() const           // 時期の多重比較法を返す(時期1との比較)
  { return vmb; }
  mctype mc() const             // 時期の多重比較法を返す(全時期比較)
  { return vmc; }
  mctype md() const             // 群の多重比較法を返す(群1との比較)
  { return vmd; }
  mctype me() const     // 時期別平均値の多重比較法を返す(群1との比較)
  { return vme; }
  mctype mf() const         // 交互作用の多重比較法を返す(群1との比較)
  { return vmf; }
  mctype mm() const        // 時期別平均値の多重比較法を返す(全群比較)
  { return vmm; }
  mctype mx() const            // 交互作用の多重比較法を返す(全群比較)
  { return vmx; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rp() const         // 対応のある二元配置分散分析の比較回数を返す
  { return vrp; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vma(mctuk), vmb(mcdun), vmc(mctuk), vmd(mcdun), vme(mcbon), vmf(mcbon), vmm(mcbon), vmx(mcbon), vreslst(), vresmap(), vrp(1)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error ->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 3) return disperr(vpname, "At least 3 items are required !");
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["stat"] = true;
  vresmap["a2"] = vresmap["icc"] = vresmap["ma"] = vresmap["mb"] = vresmap["mc"] = vresmap["md"] = vresmap["me"] = vresmap["mf"] = vresmap["mm"] = vresmap["mx"] = vresmap["rm"] = vresmap["rp"] = vresmap["sd"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      { vresmap["def"] = vresmap["cc"] = vresmap["stat"] = true; }
      else if (res == "all")    // 全て出力
      {
        vresmap["a2"] = vresmap["cc"] = vresmap["icc"] = vresmap["ma"] = vresmap["mb"] = vresmap["mc"] = vresmap["md"] = vresmap["me"] = vresmap["mf"] = vresmap["mm"] = vresmap["mx"] = vresmap["rm"] = vresmap["sd"] = vresmap["stat"] = true;
        // rp以外は全てtrue
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "m")      // mと指定された時
      { vresmap["ma"] = vresmap["mx"] = true; }
      else if (res[0] == 'm' && res[1] == 'a' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vma = mcbon;
          break;
        case 'f' : vma = mcfis;
          break;
        case 's' : vma = mcsch;
          break;
        case 't' : vma = mctuk;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["ma"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'b' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmb = mcbon;
          break;
        case 'd' : vmb = mcdun;
          break;
        case 'f' : vmb = mcfis;
          break;
        case 's' : vmb = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["mb"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'c' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmc = mcbon;
          break;
        case 'f' : vmc = mcfis;
          break;
        case 's' : vmc = mcsch;
          break;
        case 't' : vmc = mctuk;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["mc"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'd' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmd = mcbon;
          break;
        case 'd' : vmd = mcdun;
          break;
        case 'f' : vmd = mcfis;
          break;
        case 's' : vmd = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["md"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'e' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vme = mcbon;
          break;
        case 'f' : vme = mcfis;
          break;
        case 's' : vme = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["me"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'f' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmf = mcbon;
          break;
        case 'f' : vmf = mcfis;
          break;
        case 's' : vmf = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["mf"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'm' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmm = mcbon;
          break;
        case 'f' : vmm = mcfis;
          break;
        case 's' : vmm = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["mm"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'x' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmx = mcbon;
          break;
        case 'f' : vmx = mcfis;
          break;
        case 's' : vmx = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res+ " option !");
        }
        vresmap["mx"] = true;
      }
      else if (res[0] == 'r' && res[1] == 'p' && res.size() > 2)
      {
        vrp = int(strtodf(res.substr(2)));
        if (vrp < 1) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["rp"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
    if ((vresmap["def"] || vresmap["a2"]) && vresmap["rp"]) return disperr(vpname, "-res=a2 and rp cannot be specified together !");
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  map<string, vecstatv> vgsmap; // 群統計量マップ
  vector<string> vgcodevec;     // 群コードベクトル
  vecstatv vtstatvec;           // 群統計量ベクトル
  double vsa;                   // 要因A平方和
  int vdfa;                     // 要因A自由度
  double vssr;                  // 個体残差
  int vdfsr;                    // 個体残差自由度
  double vss;                   // 個体間平方和
  int vdfs;                     // 個体自由度
  double vsb;                   // 要因B平方和
  int vdfb;                     // 要因B自由度
  double vsab;                  // 要因AxB平方和
  int vdfab;                    // 要因AxB自由度
  double vsr;                   // 残差平方和
  int vdfr;                     // 残差自由度
  double vst;                   // 全体平方和
  int vdft;                     // 全体自由度
  double statv_sigma(vecstatv& statvec, const vecdbl& fdatavec);
  // データを累積する
  int set_gsmap(option& opt);  // 群統計量マップを整理して群数を返す
public:
  varclass(size_t ni) : vgsmap(), vgcodevec(), vtstatvec(ni), vsa(0.0), vdfa(0), vssr(0.0), vdfsr(0), vss(0.0), vdfs(0), vsb(0.0), vdfb(0), vsab(0.0), vdfab(0), vsr(0.0), vdfr(0), vst(0.0), vdft(0)
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int cal(option& opt, ldata& datalst); // データを集計して変数をセットする
  int g() const                 // 要因A(群)の数を返す
  { return (int)(vgsmap.size()); }
  int b() const                 // 要因B(時期)の数を返す
  { return (int)(vtstatvec.size()) - 1; }
  int ns() const                // 個体数を返す
  { return vtstatvec[0].n(); }
  const string& gcode(size_t idx) const // 群コードを返す
  { return (idx < vgcodevec.size()) ? vgcodevec[idx] : BLANKs; }
  int n(size_t iidx, size_t jidx); // 群の例数を返す
  double mean(size_t iidx, size_t jidx); // 群の平均値を返す
  double vr(size_t iidx, size_t jidx);   // 群の不偏分散を返す
  double ss(size_t iidx, size_t jidx);   // 群の平方和を返す
  double se(size_t iidx, size_t jidx);   // 群の標準誤差を返す
  double sa() const             // 要因A平方和を返す
  { return vsa; }
  int dfa() const               // 要因A自由度を返す
  { return vdfa; }
  double ssr() const            // 個体残差平方和を返す
  { return vssr; }
  double ssr(double ssr0)       // 個体残差平方和を代入する
  { return (vssr = ssr0); }
  int dfsr() const              // 個体残差自由度を返す
  { return vdfsr; }
  int dfsr(int dfsr0)           // 個体残差由度を代入する
  { return (vdfsr = dfsr0); }
  double ss() const             // 個体平方和を返す
  { return vss; }
  int dfs() const               // 個体自由度を返す
  { return vdfs; }
  double sb() const             // 要因B平方和を返す
  { return vsb; }
  int dfb() const               // 要因B由度を返す
  { return vdfb; }
  double sab() const            // 要因AxB平方和を返す
  { return vsab; }
  int dfab() const              // 要因AxB由度を返す
  { return vdfab; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  double sr(double sr0)         // 残差平方和を代入する
  { return (vsr = sr0); }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  int dfr(int dfr0)             // 残差自由度を代入する
  { return (vdfr = dfr0); }
  double st() const             // 全体平方和を返す
  { return vst; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
};

double varclass::statv_sigma(vecstatv& statvec, const vecdbl& fdatavec)
// データを累積する
{
  double sum = 0.0;
  for (size_t j = 0; j < fdatavec.size(); ++j)
  {
    statvec[j].sigma(fdatavec[j]);
    sum += fdatavec[j];
  }
  return sum * sum / fdatavec.size(); // 個体平方和用の値を返す
}

int varclass::set_gsmap(option& opt)  // 群統計量マップを整理して群数を返す
{
  int g = (int)vgsmap.size(), b = (int)vtstatvec.size() - 1;
  vgcodevec.resize(g);
  vsa = 0.0;  vsab = 0.0;
  int i = 0;
  for (map<string, vecstatv>::iterator gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr, ++i)
  {
    vgcodevec[i] = gitr->first; // 群コードをべクルトに格納する
    for (int j = 0; j < b; ++j)
    {
      gitr->second[b] += gitr->second[j];
      vtstatvec[j] += gitr->second[j];
      double m = gitr->second[j].mean();
      vsab += (double)(gitr->second[j].n()) * m * m;
    }
    double m = gitr->second[b].mean();
    vsa += (double)(gitr->second[b].n()) * m * m;
  }
  for (int j = 0; j < b; ++j)
  {
    vtstatvec[b] += vtstatvec[j];
    double m = vtstatvec[j].mean();
    vsb += (double)(vtstatvec[j].n()) * m * m;
  }
  double m = vtstatvec[b].mean();
  double cf = (double)(vtstatvec[b].n()) * m * m;
  int ns = vtstatvec[0].n();
  vdfs = ns - 1;  vdfb = b - 1;  vss -= cf;
  if (ns == g && (opt.res("def") == false && opt.res("a2") == false && opt.res("rp")))
  {                   // 対応のある二元配置分散分析の場合(-res=a2優先)
    vst = vtstatvec[b].sum2();  vdft = vtstatvec[b].n();
    vsab = vsb - cf;  vdfab = b - 1;
    vssr = vsa - cf;  vdfsr = g - 1;
    vsa = cf;  vdfa = 1;  vsb = vsab;
    vsr = vst - vsa - vssr - vsab;  vdfr = vdft - vdfa - vdfsr - vdfab;
  }
  else
  {
    vst = vtstatvec[b].ss();  vdft = vtstatvec[b].n() - 1;
    vsa -= cf;  vdfa = g - 1;  vsb -= cf;
    vssr = vss - vsa;  vdfsr = vdfs - vdfa;
    vsab -= cf + vsa + vsb;  vdfab = vdfa * vdfb;
    vsr = vst - vss - vsb - vsab;  vdfr = vdft - vdfs - vdfb - vdfab;
  }
  const double Seps = DBL_EPSILON * fmax(5.0, vst);
  if (vdfa < 1 || vsa < Seps) vsa = 0.0;
  if (vdfb < 1 || vsb < Seps) vsb = 0.0;
  if (vdfs < 1 || vss < Seps) vss = 0.0;
  if (vdfsr < 1 || vssr < Seps) vssr = 0.0;
  if (vdfr < 1 || vsr < Seps) vsr = 0.0;
  return g;                     // 群数を返す
}

int varclass::cal(option& opt, ldata& datalst)
// データを集計して変数をセットする
{
  size_t ni = vtstatvec.size();
  vecdbl fdatavec(ni - 1);
  vss = 0.0;
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    bool blank = (ditr->sdata(0) == BLANKs) ? true : false;
    size_t j = 1;
    while (blank == false && j < ni)
    {
      if (ditr->fdata(j) == BLANKf)
      { blank = true; }
      else
      {
        fdatavec[j - 1] = ditr->fdata(j);
        ++j;
      }
    }
    if (blank == false)
    {
      map<string, vecstatv >::iterator gitr = vgsmap.find(ditr->sdata(0));
      if (gitr == vgsmap.end())     // 新規群コード
      {
        vecstatv tstatvec(ni);
        vss += statv_sigma(tstatvec, fdatavec);
        vgsmap.insert(map<string, vecstatv>::value_type(ditr->sdata(0), tstatvec));
      }
      else                      // すでにある群
      { vss += statv_sigma(gitr->second, fdatavec); }
    }
  }
  return set_gsmap(opt);        // 群統計量マップを整理して群数を返す
}

int varclass::n(size_t iidx, size_t jidx) // 群の例数を返す
{
  int n = 0, b = (int)vtstatvec.size() - 1;
  if (iidx < vgsmap.size())     // 群別
  {
    if (jidx < vtstatvec.size()) n = (vgsmap[vgcodevec[iidx]])[jidx].n();
    else n = (vgsmap[vgcodevec[iidx]])[b].n();
  }
  else                          // 群合計
  {
    if (jidx < vtstatvec.size()) n = vtstatvec[jidx].n();
    else n = vtstatvec[b].n();
  }
  return n;
}

double varclass::mean(size_t iidx, size_t jidx) // 群の平均値を返す
{
  int b = (int)vtstatvec.size() - 1;
  double m = 0.0;
  if (iidx < vgsmap.size())     // 群別
  {
    if (jidx < vtstatvec.size()) m = (vgsmap[vgcodevec[iidx]])[jidx].mean();
    else m = (vgsmap[vgcodevec[iidx]])[b].mean();
  }
  else                          // 群合計
  {
    if (jidx < vtstatvec.size()) m = vtstatvec[jidx].mean();
    else m = vtstatvec[b].mean();
  }
  return m;
}

double varclass::vr(size_t iidx, size_t jidx) // 群の不偏分散を返す
{
  int b = (int)vtstatvec.size() - 1;
  double vr = 0.0;
  if (iidx < vgsmap.size())     // 群別
  {
    if (jidx < vtstatvec.size()) vr = (vgsmap[vgcodevec[iidx]])[jidx].vr();
    else vr = (vgsmap[vgcodevec[iidx]])[b].vr();
  }
  else                          // 群合計
  {
    if (jidx < vtstatvec.size()) vr = vtstatvec[jidx].vr();
    else vr = vtstatvec[b].vr();
  }
  return vr;
}

double varclass::ss(size_t iidx, size_t jidx) // 群の平方和を返す
{
  int b = (int)vtstatvec.size() - 1;
  double ss = 0.0;
  if (iidx < vgsmap.size())     // 群別
  {
    if (jidx < vtstatvec.size()) ss = (vgsmap[vgcodevec[iidx]])[jidx].ss();
    else ss = (vgsmap[vgcodevec[iidx]])[b].ss();
  }
  else                          // 群合計
  {
    if (jidx < vtstatvec.size()) ss = vtstatvec[jidx].ss();
    else ss = vtstatvec[b].ss();
  }
  return ss;
}

double varclass::se(size_t iidx, size_t jidx) // 群の標準誤差を返す
{
  int b = (int)vtstatvec.size() - 1;
  double se = 0.0;
  if (iidx < vgsmap.size())     // 群別
  {
    if (jidx < vtstatvec.size()) se = (vgsmap[vgcodevec[iidx]])[jidx].se();
    else se = (vgsmap[vgcodevec[iidx]])[b].se();
  }
  else                          // 群合計
  {
    if (jidx < vtstatvec.size()) se = vtstatvec[jidx].se();
    else se = vtstatvec[b].se();
  }
  return se;
}

class anova2class               // anova2メインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_id(ostream& fout);   // タイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, option& opt, varclass& var);
  // 群別基礎統計量を出力する
  void put_icc(ostream& fout, option& opt, varclass& var);
  // 級内相関係数(信頼性係数)を出力する
  void put_iccci(ostream& fout, const string& kicc, double cc, double v1, int df1, double v2, int df2, double b) const;
  // 級内相関係数の信頼区間を出力する
  void put_icc2ci(ostream& fout, double cc, double icc21, double va, double vb, double vr, int dfa, int dfb) const;
  // ICC(2,1)とICC(2,b)の信頼区間を出力する
  void put_rmanova2(ostream& fout, option& opt, varclass& var);
  // 繰り返し測定型二元配置分散分析結果を出力する
  void put_ranova2(ostream& fout, option& opt, varclass& var);
  // 繰り返しのある二元配置分散分析結果を出力する
  void put_nranova2(ostream& fout, option& opt, varclass& var);
  // 繰り返しのない二元配置分散分析結果を出力する
  void put_panova2(ostream& fout, option& opt, varclass& var);
  // 対応のある二元配置分散分析結果を出力する
  void put_anova2(ostream& fout, atabtype attype, varclass& var) const;
  // 二元配置分散分析表を出力する
  void put_sd(ostream& fout, atabtype attype, varclass& var) const;
  // 各要因のSDを出力する
  void put_mcmpad(ostream& fout, mctype mc, bool ma, double sr, int dfr, varclass& var) const;
  // 群の多重比較を出力する(全群比較／群1との比較)
  void put_mdciad(ostream& fout, const option& opt, mctype mc, bool ma, double sr, int dfr, varclass& var) const;
  // 2群間の平均値の差の信頼区間を出力する(全群比較／群1との比較)
  void put_mcmpbc(ostream& fout, mctype mc, bool mb, double sr, int dfr, varclass& var) const;
  // 時期の多重比較を出力する(時期1との比較／全時期比較)
  void put_mdcibc(ostream& fout, const option& opt, mctype mc, bool mb, double sr, int dfr, varclass& var) const;
  // 2時期間の平均値の差の信頼区間を出力する(時期1との比較／全時期比較)
  void put_mcmpxf(ostream& fout, mctype mc, bool mcx, double sr, int dfr, varclass& var) const;
  // 交互作用の多重比較を出力する(全群比較／群1との比較)
  void put_mcmpme(ostream& fout, mctype mc, bool mm, varclass& var) const;
  // 時期別群別平均値の多重検定を出力する(全群比較／群1との比較)
  void put_mdcime(ostream& fout, const option& opt, mctype mc, bool mm, varclass& var) const;
  // 時期別群別平均値の差の信頼区間を出力する(全群比較／群1との比較)
  void put_ttest1(ostream& fout, option& opt, double ssr, int dfsr, varclass& var);
  // 総平均値の1標本t検定(対応のある二元配置の時)
  void put_mcmppm(ostream& fout, const option& opt, varclass& var) const;
  // 時期別平均値の多重検定を出力する(対応のある二元配置の時、全群比較)
  void put_mdcipm(ostream& fout, const option& opt, varclass& var) const;
  // 時期別平均値の信頼区間を出力する(対応のある二元配置の時、時期1との比較)
public:
  anova2class() : fname(), id(), datalst() // コンストラクタ
  {}
  ~anova2class()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み保存・累積する
  void put(option& opt);                     // 集計結果を出力する
};

void anova2class::put_id(ostream& fout) // タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "FACTOR-A  :" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "FACTOR-B" << setw(2) << i;
    fout << ":" << id.item(i) << "\n";
  }
}

void anova2class::put_stat(ostream& fout, option& opt, varclass& var)
// 群別基礎統計量を出力する
{
  constexpr int LENline = 76;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  int g = var.g(), b = var.b(), ns = var.ns();
  if (ns > g || opt.res("a2"))  // 繰り返しあり/一般型二元配置分散分析
  {
    if (ns == g)                // 繰り返しのない二元配置分散分析
    { fout << "SUBJECT   :BASIC STATISTICS OF ALL PERIOD(B) BY SUBJECT(A)\n"; }
    else
    { fout << "GROUP     :BASIC STATISTICS OF ALL PERIOD(B) BY GROUP(A)\n"; }
    putline(fout, LENline);
    for (int i = 0; i < g; ++i)
    {
      string gcodei(var.gcode(i));
      int npad = gcodei.size() - strclen(gcodei, CHARwide);
      fout << left << setw(LENhed + npad) << gcodei << right << ":";
      putmsd(fout, var.n(i, b), var.mean(i, b), var.vr(i, b));
    }
    putline(fout, LENline);
    fout << "\n";
  }
  if (ns == g)                  // 繰り返しのない二元配置分散分析
  { fout << "PERIOD    :BASIC STATISTICS OF ALL SUBJECT(A) BY PERIOD(B)\n"; }
  else
  { fout << "PERIOD    :BASIC STATISTICS OF ALL GROUP(A) BY PERIOD(B)\n"; }
  putline(fout, LENline);
  for (int j = 0; j < b; ++j)
  {
    fout << setw(LENhed - 4) << j + 1 << "    :";
    putmsd(fout, var.n(g, j), var.mean(g, j), var.vr(g, j));
  }
  putline(fout, LENline);
  if (ns > g)                   // 繰り返しあり
  {
    fout << "\nGROUP-PERI:BASIC STATISTICS BY GROUP(A) AND PERIOD(B)\n";
    putline(fout, LENline);
    for (int i = 0; i < g; ++i)
    {
      for (int j = 0; j < b; ++j)
      {
        string gcodei(var.gcode(i));
        int npad = gcodei.size() - strclen(gcodei, CHARwide);
        fout << left << setw(LENhed - 5 + npad) << gcodei << right;
        fout << "-" << setw(4) << j + 1 << ":";
        putmsd(fout, var.n(i, j), var.mean(i, j), var.vr(i, j));
      }
    }
    putline(fout, LENline);
  }
  fout << left << setw(LENhed) << "TOTAL" << ":";
  putmsd(fout, var.n(g, b), var.mean(g, b), var.vr(g, b));
}

void anova2class::put_icc(ostream& fout, option& opt, varclass& var)
// 級内相関係数(信頼性係数)を出力する
{
  double a = (double)var.g(), b = (double)var.b(), sw = var.st() - var.sa();
  int dfw = var.dft() - var.dfa();
  double sr = var.sab();
  int dfr = var.dfab();
  double va = (var.dfa() > 0) ? var.sa() / (double)var.dfa() : 0.0;
  double vb = (var.dfb() > 0) ? var.sb() / (double)var.dfb() : 0.0;
  double vw = (dfw > 0) ? sw / (double)dfw : 0.0;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double vtmp = va + (double)var.dfb() * vw;
  double icc11 = (vtmp > 0.0) ? (va - vw) / vtmp : 0.0;
  double icc1b = (va > 0.0) ? (va - vw) / va : 0.0;
  vtmp = va + (double)var.dfb() * vr + b * (vb - vr) / a;
  double icc21 = (vtmp > 0.0) ? (va - vr) / vtmp : 0.0;
  vtmp = va + (vb - vr) / a;
  double icc2b = (vtmp > 0.0) ? (va - vr) / vtmp : 0.0;
  vtmp = va + (double)var.dfb() * vr;
  double icc31 = (vtmp > 0.0) ? (va - vr) / vtmp : 0.0;
  double icc3b = (va > 0.0) ? (va - vr) / va : 0.0;
  fout << "<INTRACLASS CORRELATION COEFFICIENT>\n";
  double sda = (va - vw) / b, sdr = sqrt(vw);
  sda = (sda > 0.0) ? sqrt(sda) : 0.0;
  fout << "ICC(1,1)=" << icc11 << " ICC(1," << b << ")=" << icc1b;
  fout << "  SA=" << sda << " SR=" << sdr << "\n";
  if (opt.res("cc")) put_iccci(fout, "ICC(1", opt.cc(), va, var.dfa(), vw, dfw, b);
  sda = (va - vr) / b;  sdr = sqrt(vr);
  sda = (sda > 0.0) ? sqrt(sda) : 0.0;
  double sdb = (vb - vr) / a;
  sdb = (sdb > 0.0) ? sqrt(sdb) : 0.0;
  fout << "ICC(2,1)=" << icc21 << " ICC(2," << b << ")=" << icc2b;
  fout << "  SA=" << sda << " SB=" << sdb << " SR=" << sdr << "\n";
  if (opt.res("cc")) put_icc2ci(fout, opt.cc(), icc21, va, vb, vr, var.dfa(), var.dfb());
  fout << "ICC(3,1)=r11=" << icc31 << " ICC(3," << b << ")=alpha=" << icc3b;
  fout << "  SA=" << sda << " SR=" << sdr << "\n";
  if (opt.res("cc")) put_iccci(fout, "ICC(3", opt.cc(), va, var.dfa(), vr, dfr, b);
}

void anova2class::put_iccci(ostream& fout, const string& kicc, double cc, double v1, int df1, double v2, int df2, double b) const
// 級内相関係数の信頼区間を出力する
{
  double f = (v2 > 0.0) ? v1 / v2 : 0.0;
  double ah = (1.0 - cc) * 0.5;
  double fl = f / ptof(ah, df1, df2), fu = f / ptof(1.0 - ah, df1, df2);
  double iccl = (fl - 1.0) / (fl + (b - 1.0));
  double iccu = (fu - 1.0) / (fu + (b - 1.0));
  fout << cc * 100.0 << "%CI of " << kicc << ",1):Lower=" << iccl << "  Upper=" << iccu << "\n";
  iccl = (fl > 0.0) ? (fl - 1.0) / fl : 0.0;
  iccu = (fu > 0.0) ? (fu - 1.0) / fu : 0.0;
  fout << cc * 100.0 << "%CI of " << kicc << "," << b << "):Lower=" << iccl << "  Upper=" << iccu << endl;
}

void anova2class::put_icc2ci(ostream& fout, double cc, double icc21, double va, double vb, double vr, int dfa, int dfb) const
// ICC(2,1)とICC(2,b)の信頼区間を出力する
{
  double f = (vr > 0.0) ? vb / vr : 0.0;
  double a = (double)(dfa + 1), b = (double)(dfb + 1);
  double bicc = b * icc21;
  double t1 = a * (1.0 + (double)dfb * icc21) - bicc;
  double t2 = (double)dfa * (double)dfb * (bicc * f + t1) * (bicc * f + t1);
  int df2 = int(t2 / ((double)dfa * bicc * bicc * f * f + t1 * t1) + 0.5);
  double ah = (1.0 - cc) * 0.5;
  double fl = ptof(ah, dfa, df2), fu = ptof(ah, df2, dfa);
  double iccl = (a * (va - fl * vr)) / (fl * (b * vb + (a * b - a - b) * vr) + a * va);
  double iccu = (a * (fu * va - vr)) / (b * vb + (a * b - a - b) * vr + a * fu * va);
  fout << cc * 100.0 << "%CI of ICC(2,1):Lower=" << iccl << "  Upper=" << iccu << "\n";
  double iccbl = b * iccl / (1.0 + (double)dfb * iccl);
  double iccbu = b * iccu / (1.0 + (double)dfb * iccu);
  fout << cc * 100.0 << "%CI of ICC(2," << b << "):Lower=" << iccbl << "  Upper=" << iccbu << endl;
}

void anova2class::put_rmanova2(ostream& fout, option& opt, varclass& var)
// 繰り返し測定型二元配置分散分析結果を出力する
{
  put_anova2(fout, at2rm, var);
  if (opt.res("sd"))
  {
    fout << "\n";
    put_sd(fout, at2rm, var);
  }
  if (opt.res("ma"))            // 全群の多重比較
  {
    fout << "\n";
    put_mcmpad(fout, opt.ma(), true, var.ssr(), var.dfsr(), var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdciad(fout, opt, opt.ma(), true, var.ssr(), var.dfsr(), var);
    }
  }
  if (opt.res("md"))            // 群1と他群の多重比較
  {
    fout << "\n";
    put_mcmpad(fout, opt.md(), false, var.ssr(), var.dfsr(), var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdciad(fout, opt, opt.md(), false, var.ssr(), var.dfsr(), var);
    }
  }
  if (opt.res("mb"))            // 時期1と他時期の多重比較
  {
    fout << "\n";
    put_mcmpbc(fout, opt.mb(), true, var.sr(), var.dfr(), var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcibc(fout, opt, opt.mb(), true, var.sr(), var.dfr(), var);
    }
  }
  if (opt.res("mc"))            // 全時期の多重比較
  {
    fout << "\n";
    put_mcmpbc(fout, opt.mc(), false, var.sr(), var.dfr(), var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcibc(fout, opt, opt.mc(), false, var.sr(), var.dfr(), var);
    }
  }
  if (opt.res("mx"))            // 全群の交互作用の多重比較
  {
    fout << "\n";
    put_mcmpxf(fout, opt.mx(), true, var.sr(), var.dfr(), var);
  }
  if (opt.res("mf"))            // 群1と他群の交互作用の多重比較
  {
    fout << "\n";
    put_mcmpxf(fout, opt.mf(), false, var.sr(), var.dfr(), var);
  }
  if (opt.res("mm"))            // 全群の時期別多重比較
  {
    fout << "\n";
    put_mcmpme(fout, opt.mm(), true, var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcime(fout, opt, opt.mm(), true, var);
    }
  }
  if (opt.res("me"))            // 群1と他群の時期別多重比較
  {
    fout << "\n";
    put_mcmpme(fout, opt.me(), false, var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcime(fout, opt, opt.me(), false, var);
    }
  }
}

void anova2class::put_ranova2(ostream& fout, option& opt, varclass& var)
// 繰り返しのある二元配置分散分析結果を出力する
{
  int g = var.g(), ns = var.ns();
  atabtype attype = at2r;
  var.sr(var.ssr() + var.sr());  var.dfr(var.dfsr() + var.dfr());
  var.ssr(0.0);  var.dfsr(0);
  put_anova2(fout, attype, var);
  if (opt.res("sd"))
  {
    fout << "\n";
    put_sd(fout, attype, var);
  }
  if (opt.res("ma"))            // 全群の多重比較
  {
    fout << "\n";
    put_mcmpad(fout, opt.ma(), true, var.sr(), var.dfr(), var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdciad(fout, opt, opt.ma(), true, var.sr(), var.dfr(), var);
    }
  }
  if (opt.res("md"))            // 群1と他群の多重比較
  {
    fout << "\n";
    put_mcmpad(fout, opt.md(), false, var.sr(), var.dfr(), var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdciad(fout, opt, opt.md(), false, var.sr(), var.dfr(), var);
    }
  }
  if (opt.res("mb"))            // 時期1と他時期の多重比較
  {
    fout << "\n";
    put_mcmpbc(fout, opt.mb(), true, var.sr(), var.dfr(), var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcibc(fout, opt, opt.mb(), true, var.sr(), var.dfr(), var);
    }
  }
  if (opt.res("mc"))            // 全時期の多重比較
  {
    fout << "\n";
    put_mcmpbc(fout, opt.mc(), false, var.sr(), var.dfr(), var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcibc(fout, opt, opt.mc(), false, var.sr(), var.dfr(), var);
    }
  }
  if (opt.res("mx"))            // 全群の交互作用の多重比較
  {
    fout << "\n";
    put_mcmpxf(fout, opt.mx(), true, var.sr(), var.dfr(), var);
  }
  if (opt.res("mf"))            // 群1と他群の交互作用の多重比較
  {
    fout << "\n";
    put_mcmpxf(fout, opt.mf(), false, var.sr(), var.dfr(), var);
  }
  if (opt.res("mm"))            // 全群の時期別多重比較
  {
    fout << "\n";
    put_mcmpme(fout, opt.mm(), true, var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcime(fout, opt, opt.mm(), true, var);
    }
  }
  if (opt.res("me"))            // 群1と他群の時期別多重比較
  {
    fout << "\n";
    put_mcmpme(fout, opt.me(), false, var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcime(fout, opt, opt.me(), false, var);
    }
  }
}

void anova2class::put_nranova2(ostream& fout, option& opt, varclass& var)
// 繰り返しのない二元配置分散分析結果を出力する
{
  int g = var.g(), ns = var.ns();
  atabtype attype = at2;
  var.sr(var.sab());  var.dfr(var.dfab());
  put_anova2(fout, attype, var);
  if (opt.res("sd")) // 群の多重比較はせず、交互作用と時期別多重比較は不可能
  {
    fout << "\n";
    put_sd(fout, attype, var);
  }
  if (opt.res("mb"))            // 時期1と他時期の多重比較
  {
    fout << "\n";
    put_mcmpbc(fout, opt.mb(), true, var.sr(), var.dfr(), var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcibc(fout, opt, opt.mb(), true, var.sr(), var.dfr(), var);
    }
  }
  if (opt.res("mc"))            // 全時期の多重比較
  {
    fout << "\n";
    put_mcmpbc(fout, opt.mc(), false, var.sr(), var.dfr(), var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcibc(fout, opt, opt.mc(), false, var.sr(), var.dfr(), var);
    }
  }
}

void anova2class::put_panova2(ostream& fout, option& opt, varclass& var)
// 対応のある二元配置分散分析結果を出力する
{
  int g = var.g(), ns = var.ns();
  atabtype attype = at2rp;
  put_anova2(fout, attype, var);
  if (opt.res("sd"))
  {
    fout << "\n";
    put_sd(fout, attype, var);
  }
  fout << "\n";
  put_ttest1(fout, opt, var.ssr(), var.dfsr(), var);
  if (opt.res("mm"))            // 時期別多重比較
  {
    fout << "\n";
    put_mcmppm(fout, opt, var);
    if (opt.res("cc"))
    {
      fout << "\n";
      put_mdcipm(fout, opt, var);
    }
  }
}

void anova2class::put_anova2(ostream& fout, atabtype attype, varclass& var) const
// 二元配置分散分析表を出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  fout << "<";
  if (attype == at2rm) fout << "REPEATED MEASURES ";
  else if (attype == at2rp) fout << "PAIRED ";
  fout << "TWO-WAY LAYOUT ANALYSIS OF VARIANCE(ANOVA2)>\n";
  string tstr("ANOVA Table");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << setw(45 + npad) << tstr << "\n";
  fout << "FACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0, vr = 0.0;
  if (attype == at2rm)          // 繰り返し測定型
  {
    calvfp(v, f, pr, var.sa(), var.dfa(), var.ssr(), var.dfsr());
    tstr = "A:GROUP";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(LENhed + npad) << tstr << right << setw(14) << var.sa();
    fout << setw(10) << var.dfa() << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
    calvfp(v, f, pr, var.ssr(), var.dfsr(), var.sr(), var.dfr());
    tstr = "RESIDUALS";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(LENhed + npad) << tstr << right << setw(14) << var.ssr();
    fout << setw(10) << var.dfsr() << setw(14) << v << "\n";
    putline(fout, LENline);
    calvfp(v, f, pr, var.ss(), var.dfs(), var.sr(), var.dfr());
    tstr = "SUBJECT";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(LENhed + npad) << tstr << right << setw(14) << var.ss();
    fout << setw(10) << var.dfs() << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
  }
  else                          // 繰り返し測定型以外
  {
    calvfp(v, f, pr, var.sa(), var.dfa(), var.sr(), var.dfr());
    if (attype == at2) tstr = "A:SUBJECT";
    else if (attype == at2rp) tstr = "C:TOTAL ME";
    else tstr = "A:GROUP";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(LENhed + npad) << tstr << right << setw(14) << var.sa();
    fout << setw(10) << var.dfa() << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
    if (attype == at2rp)        // 対応がある時
    {
      calvfp(v, f, pr, var.ssr(), var.dfsr(), var.sr(), var.dfr());
      tstr = "AxC:SUBJEC";
      npad = tstr.size() - strclen(tstr, CHARwide);
      fout << left << setw(LENhed + npad) << tstr << right << setw(14) << var.ssr();
      fout << setw(10) << var.dfsr() << setw(14) << v << setw(12) << f;
      fout << setw(14) << pr << "\n";
    }
  }
  if (attype != at2rp)         // 対応のある二元配置分散分析ではない時
  {
    vr = calvfp(v, f, pr, var.sb(), var.dfb(), var.sr(), var.dfr());
    tstr = "B:PERIOD";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(LENhed + npad) << tstr << right << setw(14) << var.sb();
    fout << setw(10) << var.dfb() << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
  }
  if (attype == at2r || attype == at2rm || attype == at2rp) // 繰り返しのある/繰り返し測定型/対応のある時
  {
    vr = calvfp(v, f, pr, var.sab(), var.dfab(), var.sr(), var.dfr());
    if (attype == at2rp) tstr = "BxC:PERIOD";
    else tstr = "AxB";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(LENhed + npad) << tstr << right << setw(14) << var.sab();
    fout << setw(10) << var.dfab() << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
  }
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(LENhed + npad) << tstr << right << setw(14) << var.sr();
  fout << setw(10) << var.dfr() << setw(14) << vr << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(LENhed + npad) << tstr << right << setw(14) << var.st();
  fout << setw(10) << var.dft() << endl;
}

void anova2class::put_sd(ostream& fout, atabtype attype, varclass& var) const
// 各要因のSDを出力する
{
  double va = (var.dfa() > 0) ? var.sa() / (double)var.dfa() : 0.0;
  double vs = (var.dfs() > 0) ? var.ss() / (double)var.dfs() : 0.0;
  double vb = (var.dfb() > 0) ? var.sb() / (double)var.dfb() : 0.0;
  double vab = (var.dfab() > 0) ? var.sab() / (double)var.dfab() : 0.0;
  double vr = (var.dfr() > 0) ? var.sr() / (double)var.dfr() : 0.0;
  double vt = (var.dft() > 0) ? var.st() / (double)var.dft() : 0.0;
  double a = (double)(var.dfa() + 1), b = (double)(var.dfb() + 1);
  double ar = (double)(var.dfs() + 1);
  double r = ar / a;
  double s2a = (va > vr) ? va - vr : 0.0;
  if (attype == at2) s2a /= b;
  else s2a /= b * r;
  double s2s = (vs > vr) ? (vs - vr) / b : 0.0;
  double s2b = (vb > vr) ? vb - vr : 0.0;
  if (attype == at2) s2b /= a;
  else s2b /= ar;
  double s2ab = (vab > vr) ? (vab - vr) / r : 0.0;
  fout << "<VARIACES AND STANDARD DEVIATIONS OF EACH FACTORS>\n";
  fout << "FACTOR-A:" << "V=" << s2a << " SD=" << sqrt(s2a);
  if (attype == at2rm)          // 繰り返し測定型
  {
    fout << "SUBJECT:" << "V=" << s2s << " SD=" << sqrt(s2s);
  }
  fout << "\nFACTOR-B:" << "V=" << s2b << " SD=" << sqrt(s2b);
  if (attype != at2)            // 繰り返しあり
  { fout << "INTERACTION:" << "V=" << s2ab << " SD=" << sqrt(s2ab); }
  fout << "\nRESIDUALS:" << "V=" << vr << " SD=" << sqrt(vr) << endl;
}

void anova2class::put_mcmpad(ostream& fout, mctype mc, bool ma, double sr, int dfr, varclass& var) const
// 群の多重比較を出力する(全群比較／群1との比較)
{
  constexpr int LENline = 72;   // 区切り線の長さ
  int g = var.g(), b = var.b();
  if (g < 2 || sr <= 0.0) return; // 出力なし
  int df1 = 1, dc = 1;
  double vr = sr / (double)dfr;
  string method = "TUKEY";
  switch (mc)
  {
  case mcbon : method = "BONFERRONI";
    dc = ma ? g * (g - 1) / 2 : g - 1;
    break;
  case mcdun : method = "DUNNETT";
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;
  }
  fout << "<" << method << " TYPE MULTIPLE COMPARISONS OF GROUP MEANS>\n";
  if (mc == mctuk)
  {
    fout << "GROUP   - GROUP        q-value      Groups          DF       p-value\n";
  }
  else if (mc == mcdun)
  {
    fout << "GROUP   - GROUP        d-value      Groups          DF       p-value\n";
  }
  else
  {
    fout << "GROUP   - GROUP        F-value         DF1         DF2       p-value\n";
  }
  putline(fout, LENline);
  int iend = ma ? g - 1 : 1;
  for (int i = 0; i < iend; ++i)
  {
    string gcodei(var.gcode(i));
    int npadi = gcodei.size() - strclen(gcodei, CHARwide);
    double ni = (double)var.n(i, b);
    double mi = var.mean(i, b);
    for (int j = i + 1; j < g; ++j)
    {
      double nj = (double)var.n(j, b);
      double dm = var.mean(j, b) - mi;
      double f = dm * dm / df1 / (1.0 / ni + 1.0 / nj) / vr;
      double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
      double pr = calmcpr(mc, f, g, df1, dfr, dc, r);
      string gcodej(var.gcode(j));
      int npad = gcodej.size() - strclen(gcodej, CHARwide);
      fout << left << setw(8 + npad) << gcodej;
      fout << "- ";
      fout << setw(8 + npadi) << gcodei << right << setw(12) << f;
      if (mc == mctuk || mc == mcdun) fout << setw(12) << g;
      else fout << setw(12) << df1;
      fout << setw(12) << dfr << setw(14) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, LENline);
}

void anova2class::put_mdciad(ostream& fout, const option& opt, mctype mc, bool ma, double sr, int dfr, varclass& var) const
// 2群間の平均値の差の信頼区間を出力する(全群比較／群1との比較)
{
  constexpr int LENline = 66;   // 区切り線の長さ
  int g = var.g(), b = var.b();
  if (g < 2 || sr <= 0.0) return; // 出力なし
  int df1 = 1, dc = 1;
  double vr = sr / (double)dfr;
  string method = "TUKEY";
  switch (mc)
  {
  case mcbon : method = "BONFERRONI";
    dc = ma ? g * (g - 1) / 2 : g - 1;
    break;
  case mcdun : method = "DUNNETT";
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;
  }
  fout << "<" << method << " TYPE " << opt.cc() * 100.0 <<  "% SIMULTANEOUS CONFIDENCE INTERVALS>\n";
  fout << "GROUP   - GROUP  Diff of Means  Precisicon   Lower-Lim   Upper-Lim\n";
  putline(fout, LENline);
  int iend = ma ? g - 1 : 1;
  for (int i = 0; i < iend; ++i)
  {
    string gcodei(var.gcode(i));
    int npadi = gcodei.size() - strclen(gcodei, CHARwide);
    int ni = var.n(i, b);
    double mi = var.mean(i, b);
    for (int j = i + 1; j < g; ++j)
    {
      int nj = var.n(j, b);
      double dm = var.mean(j, b) - mi;
      double vd = (1.0 / ni + 1.0 / nj) * vr;
      double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
      double ci = calmcci(mc, opt.cc(), g, df1, dfr, dc, vd, r);
      string gcodej(var.gcode(j));
      int npad = gcodej.size() - strclen(gcodej, CHARwide);
      fout << left << setw(8 + npad) << gcodej << "- ";
      fout << setw(8 + npadi) << gcodei << right << setw(12) << dm;
      fout << setw(12) << ci << setw(12) << dm - ci << setw(12) << dm + ci << "\n";
    }
  }
  putline(fout, LENline);
}

void anova2class::put_mcmpbc(ostream& fout, mctype mc, bool mb, double sr, int dfr, varclass& var) const
// 時期の多重比較を出力する(時期1との比較／全時期比較)
{
  constexpr int LENline = 72;   // 区切り線の長さ
  int g = var.g(), b = var.b();
  if (b < 2 || sr <= 0.0) return; // 出力なし
  int df1 = 1, dc = 1;
  double vr = sr / (double)dfr;
  string method = "DUNNETT";
  switch (mc)
  {
  case mcbon : method = "BONFERRONI";
    dc = mb ? b - 1 : b * (b - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = b - 1;
    break;
  case mctuk : method = "TUKEY";
  }
  fout << "<" << method << " TYPE MULTIPLE COMPARISONS OF PERIOD MEANS>\n";
  if (mc == mcdun)
  {
    fout << "PERIOD  - PERIOD       d-value     Periods          DF       p-value\n";
  }
  else if (mc == mctuk)
  {
    fout << "PERIOD  - PERIOD       q-value     Periods          DF       p-value\n";
  }
  else
  {
    fout << "PERIOD  - PERIOD       F-value         DF1         DF2       p-value\n";
  }
  putline(fout, LENline);
  int iend = mb ? 1 : b - 1;
  for (int i = 0; i < iend; ++i)
  {
    double ni = (double)var.n(g, i);
    double mi = var.mean(g, i);
    for (int j = i + 1; j < b; ++j)
    {
      double nj = (double)var.n(g, j);
      double dm = var.mean(g, j) - mi;
      double f = dm * dm / df1 / (1.0 / ni + 1.0 / nj) / vr;
      double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
      double pr = calmcpr(mc, f, b, df1, dfr, dc, r);
      fout << left << setw(8) << j + 1 << "- " << setw(8) << i + 1 << right;
      fout << setw(12) << f;
      if (mc == mcdun || mc == mctuk) fout << setw(12) << b;
      else fout << setw(12) << df1;
      fout << setw(12) << dfr << setw(14) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, LENline);
}

void anova2class::put_mdcibc(ostream& fout, const option& opt, mctype mc, bool mb, double sr, int dfr, varclass& var) const
// 2時期間の平均値の差の信頼区間を出力する(時期1との比較／全時期比較)
{
  constexpr int LENline = 66;   // 区切り線の長さ
  int g = var.g(), b = var.b();
  if (b < 2 || sr <= 0.0) return; // 出力なし
  int df1 = 1, dc = 1;
  double vr = sr / (double)dfr;
  string method = "DUNNETT";
  switch (mc)
  {
  case mcbon : method = "BONFERRONI";
    dc = mb ? b - 1 : b * (b - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";
    df1 = b - 1;
    break;
  case mctuk : method = "TUKEY";
  }
  fout << "<" << method << " TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CONFIDENCE INTERVALS>\n";
  fout << "PERIOD  - PERIOD Diff of Means  Precisicon   Lower-Lim   Upper-Lim\n";
  putline(fout, LENline);
  int iend = mb ? 1 : b - 1;
  for (int i = 0; i < iend; ++i)
  {
    double ni = (double)var.n(g, i);
    double mi = var.mean(g, i);
    for (int j = i + 1; j < b; ++j)
    {
      double nj = (double)var.n(g, j);
      double dm = var.mean(g, j) - mi;
      double vd = (1.0 / ni + 1.0 / nj) * vr;
      double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
      double ci = calmcci(mc, opt.cc(), b, df1, dfr, dc, vd, r);
      fout << left << setw(8) << j + 1 << "- " << setw(8) << i + 1 << right;
      fout << setw(12) << dm << setw(12) << ci;
      fout << setw(12) << dm - ci << setw(12) << dm + ci << "\n";
    }
  }
  putline(fout, LENline);
}

void anova2class::put_mcmpxf(ostream& fout, mctype mc, bool mcx, double sr, int dfr, varclass& var) const
// 交互作用の多重比較を出力する(全群比較／群1との比較)
{
  constexpr int LENline = 72;   // 区切り線の長さ
  int g = var.g(), b = var.b();
  int df1 = b - 1;
  if (g < 2 || sr <= 0.0) return; // 出力せず
  int dc = mcx ? g * (g - 1) / 2 : g - 1;
  double vr = sr / (double)dfr;
  string method = "BONFERRONI";
  switch (mc)
  {
  case mcfis : method = "FISHER";  dc = 1;
    break;
  case mcsch : method = "SCHEFFE";  df1 = (g - 1) * (b - 1);  dc = 1;
  }
  fout << "<" << method << " TYPE MULTIPLE COMPARISONS OF INTERACTIONS>\n";
  fout << "GROUP   - GROUP        F-value         DF1         DF2       p-value\n";
  putline(fout, LENline);
  int iend = mcx ? g - 1 : 1;
  for (int i = 0; i < iend; ++i)
  {
    string gcodei(var.gcode(i));
    int npadi = gcodei.size() - strclen(gcodei, CHARwide);
    double ni = var.n(i, 0);
    double mi = var.mean(i, b);
    for (int j = i + 1; j < g; ++j)
    {
      double nj = (double)var.n(j, 0);
      double mj = var.mean(j, b);
      double m = -b * (mj - mi) * (mj - mi);
      for (int k = 0; k < b; ++k)
      {
        double dm = var.mean(j, k) - var.mean(i, k);
        m += dm * dm;
      }
      double f = m / df1 / (1.0 / ni + 1.0 / nj) / vr;
      double pr = calmcpr(mc, f, g, df1, dfr, dc);
      string gcodej(var.gcode(j));
      int npad = gcodej.size() - strclen(gcodej, CHARwide);
      fout << left << setw(8 + npad) << gcodej << "- " << setw(8 + npadi) << gcodei << right;
      fout << setw(12) << f << setw(12) << df1 << setw(12) << dfr << setw(14) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, LENline);
}

void anova2class::put_mcmpme(ostream& fout, mctype mc, bool mm, varclass& var) const
// 時期別群別平均値の多重検定を出力する(全群比較／群1との比較)
{
  constexpr int LENline = 70;   // 区切り線の長さ
  int g = var.g(), b = var.b(), df1 = 1;
  if (g < 2 || b < 1) return;   // 出力せず
  int dc = mm ? b * g * (g - 1) / 2 : b * (g - 1);
  if (mc == mcfis)
  {
    dc = 1;
    fout << "<TWO SAMPLE T TEST OF PERIOD-GROUP MEANS>\n";
  }
  else
  { fout << "<BONFERRONI TYPE MULTIPLE TEST OF PERIOD-GROUP MEANS>\n"; }
  fout << "PERIOD  GROUP     - GROUP          t-value          DF     p-value\n";
  putline(fout, LENline);
  for (int k = 0; k < b; ++k)
  {
    int iend = mm ? g - 1 : 1;
    for (int i = 0; i < iend; ++i)
    {
      string gcodei(var.gcode(i));
      int npadi = gcodei.size() - strclen(gcodei, CHARwide);
      double ni = (double)var.n(i, k);
      double mi = var.mean(i, k);
      double ssi = var.ss(i, k);
      for (int j = i + 1; j < g; ++j)
      {
        double nj = (double)var.n(j, k);
        double dm = var.mean(j, k) - mi;
        double ssj = var.ss(j, k);
        int dfr = ni + nj - 2;
        double vr = (ssi + ssj) / (double)dfr * (ni + nj) / ni / nj;
        double t = (vr > 0.0) ? dm / sqrt(vr) : 0.0;
        double f = t * t;
        double pr = calmcpr(mc, f, 2, 1, dfr, dc);
        string gcodej(var.gcode(j));
        int npad = gcodej.size() - strclen(gcodej, CHARwide);
        fout << left << setw(8) << k + 1 << setw(10 + npad) << gcodej << "- ";
        fout << setw(10 + npadi) << gcodei << right << setw(12) << t;
        fout << setw(12) << dfr << setw(12) << pr << pind(pr) << "\n";
      }
    }
    putline(fout, LENline);
  }
}

void anova2class::put_mdcime(ostream& fout, const option& opt, mctype mc, bool mm, varclass& var) const
// 時期別群別平均値の差の信頼区間を出力する(全群比較／群1との比較)
{
  constexpr int LENline = 78;   // 区切り線の長さ
  int g = var.g(), b = var.b(), df1 = 1;
  if (g < 2 || b < 1) return;   // 出力せず
  int dc = mm ? b * g * (g - 1) / 2 : b * (g - 1);
  if (mc == mcfis)
  {
    dc = 1;
    fout << "<" << opt.cc() * 100.0 << "% CONFIDENCE INTERVALS>\n";
  }
  else
  {
    fout << "<BONFERRONI TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CONFIDENCE INTERVALS>\n";
  }
  fout << "PERIOD  GROUP     - GROUP    Diff of Means  Precisicon   Lower-Lim   Upper-Lim\n";
  putline(fout, LENline);
  for (int k = 0; k < b; ++k)
  {
    int iend = mm ? g - 1 : 1;
    for (int i = 0; i < iend; ++i)
    {
      string gcodei(var.gcode(i));
      int npadi = gcodei.size() - strclen(gcodei, CHARwide);
      double ni = (double)var.n(i, k);
      double mi = var.mean(i, k);
      double ssi = var.ss(i, k);
      for (int j = i + 1; j < g; ++j)
      {
        double nj = (double)var.n(j, k);
        double dm = var.mean(j, k) - mi;
        double ssj = var.ss(j, k);
        int dfr = ni + nj - 2;
        double vr = (ssi + ssj) / (double)dfr * (ni + nj) / ni / nj;
        double ci = calmcci(mc, opt.cc(), 2, 1, dfr, dc, vr);
        string gcodej(var.gcode(j));
        int npad = gcodej.size() - strclen(gcodej, CHARwide);
        fout << left << setw(8) << k + 1 << setw(10 + npad) << gcodej;
        fout << "- " << setw(10 + npadi) << gcodei << right << setw(12) << dm;
        fout << setw(12) << ci << setw(12) << dm - ci << setw(12) << dm + ci << "\n";
      }
    }
    putline(fout, LENline);
  }
}

void anova2class::put_ttest1(ostream& fout, option& opt, double ssr, int dfsr, varclass& var)
// 総平均値の1標本t検定(対応のある二元配置の時)
{
  int g = var.g(), b = var.b();
  double me = var.mean(g, b), vr = ssr / (double)dfsr / (double)var.n(g, b);
  double t = (vr > 0.0) ? me / sqrt(vr) : 0.0;
  double f = t * t;
  ldbl pr = calmcpr(mcbon, f, 1, 1, dfsr, opt.rp());
  if (opt.rp() > 1)
  { fout << "<BONFERRONI TYPE MULTIPLE TEST OF TOTAL MEAN>\n"; }
  else
  { fout << "<ONE SAMPLE T TEST OF TOTAL MEAN>\n"; }
  fout << "t=" << t << "  DF=" << dfsr << "  p=" << pr << pind(pr) << "\n";
  if (opt.res("cc"))
  {
    double ci = calmcci(mcbon, opt.cc(), 1, 1, dfsr, opt.rp(), vr);
    if (opt.rp() > 1)
    { fout << "Bonferroni Type " << opt.cc() * 100.0 << "% Simultaneous CI"; }
    else
    { fout << opt.cc() * 100.0 << "% Confidence Interval"; }
    fout << ":Lower=" << me - ci << "  Upper=" << me + ci << "\n";
  }
}

void anova2class::put_mcmppm(ostream& fout, const option& opt, varclass& var) const
// 時期別平均値の多重検定を出力する(対応のある二元配置の時、時期1との比較)
{
  constexpr int LENline = 62;   // 区切り線の長さ
  int g = var.g(), b = var.b();
  if (b < 2) return;            // 出力せず
  int dc = b * opt.rp();
  if (opt.mm() == mcfis)
  {
    dc = 1;
    fout << "<ONE SAMPLE T TEST OF PERIOD MEANS>\n";
  }
  else
  { fout << "<BONFERRONI TYPE MULTIPLE TEST OF PERIOD MEANS>\n"; }
  fout << "PERIOD         Means     t-value          DF       p-value\n";
  putline(fout, LENline);
  for (int j = 0; j < b; ++j)
  {
    int dfj = var.n(g, j) - 1;
    double mj = var.mean(g, j);
    double sej = var.se(g, j);
    double tj = (sej > 0.0) ? mj / sej : 0.0;
    double f = tj * tj;
    double pr = calmcpr(opt.mm(), f, 1, 1, dfj, dc);
    fout << left << setw(4) << j + 1 << right << setw(16) << mj;
    fout << setw(12) << tj << setw(12) << dfj << setw(14) << pr << pind(pr) << "\n";
  }
  putline(fout, LENline);
}

void anova2class::put_mdcipm(ostream& fout, const option& opt, varclass& var) const
// 時期平均値の信頼区間を出力する(対応のある二元配置の時、時期1との比較)
{
  constexpr int LENline = 54;   // 区切り線の長さ
  int g = var.g(), b = var.b();
  if (b < 2) return;            // 出力せず
  int dc = b * opt.rp();
  if (opt.mm() == mcfis)
  {
    dc = 1;
    fout << "<" << opt.cc() * 100.0 << "% CONFIDENCE INTERVALS>\n";
  }
  else
  {
    fout << "<BONFERRONI TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CONFIDENCE INTERVALS>\n";
  }
  fout << "PERIOD       Means   Precision   Lower-Lim   Upper-Lim\n";
  putline(fout, LENline);
  for (int j = 0; j < b; j++)
  {
    int dfj = var.n(g, j) - 1;
    double mj = var.mean(g, j);
    double vrj = var.vr(g, j) / (double)var.n(g, j);
    double ci = calmcci(opt.mm(), opt.cc(), 1, 1, dfj, dc, vrj);
    fout << left << setw(6) << j + 1 << right << setw(12) << mj;
    fout << setw(12) << ci << setw(12) << mj - ci << setw(12) << mj + ci << "\n";
  }
  putline(fout, LENline);
}

int anova2class::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void anova2class::put(option& opt) // 集計結果を出力する
{
  if (opt.res("res") == false) return; // 出力なし
  put_id(cout);
  varclass var(id.ni());        // 計算用変数クラス
  if (var.cal(opt, datalst) < 1) // 群なし
  {
    cout << "\n" << endl;
    return;
  }
  if (opt.res("stat"))
  {
    cout << "\n";
    put_stat(cout, opt, var);
  }
  if (var.g() < 1 || var.ns() < 2 || var.st() <= 0.0) // 分散分析不可能
  {
    cout << "\n" << endl;
    return;
  }
  if (var.ns() > var.g())       // 繰り返しあり
  {
    if (opt.res("def") || opt.res("rm")) // 繰り返し測定型二元配置分散分析
    {
      cout << "\n";
      put_rmanova2(cout, opt, var);
    }
    if (opt.res("a2"))          // 繰り返しのある二元配置分散分析
    {
      cout << "\n";
      put_ranova2(cout, opt, var);
    }
  }
  else                          // 繰り返しなし
  {
    if (opt.res("icc"))
    {
      cout << "\n";
      put_icc(cout, opt, var);
    }
    if (opt.res("def") || opt.res("a2"))
    {
      cout << "\n";
      put_nranova2(cout, opt, var);
    }
    else if (opt.res("rp"))
    {
      cout << "\n";
      put_panova2(cout, opt, var);
    }
  }
  cout << "\n" << endl;
}

int main(int argc, char *argv[]) // メインルーチン
{
  anova2class anova2;           // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) anova2.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { anova2.get(argv[i], opt);  fok = false; }
    }
    anova2.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
