// anovaie.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-21 09:37:55 tac>
//
// 対応の無い多群間で一元配置分散分析を行い、結果をDANSデータとして出力する。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W anovaie anovaie.C -I/usr/local/include
//              -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "anovaie  Version 7.42e       Copyleft(Vo\\oV) 2023 Sugimoto Norio",
  "USAGE:anovaie [<OPTION> ...] [<DATA FILE> ...]",
  "  The one-way ANOVA with the first item as the group(factor A)",
  "  and the subsequent items as the analysis data.",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=anovaie_dns.d",
  "  -dns={fv|me|pv|r|vv}[,...]:select the DANS data",
  "    default:output all",
  "    fv:F-value  me:n and mean  pv:p-value  r:correlation ratio  vv:variance",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -res=[-][{def|all|at|cc[<C-COEF>]|ft|m|r|stat|tt}][,...]",
  "       :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    at:ANOVA  cc[<C-COEF>]:confidence intervals(0=<C-COEF<1)",
  "    ft:2-tailed F-test to compare 2 variances",
  "    m:multiple comparisons of t test  r:correlation ratio",
  "    stat:basic statistics  tt:2 sample t test"
};

const string Title = "=== Comparison of Means of Multiple Groups ==="; // 表題
const string DNSNAMEdef = "anovaie_dns.d"; // 省略時DANSファイル名

using vecstatv = vector<statv<double>>; // 統計量ベクトル配列の別名定義

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  string vdnsname;              // DANSファイル名
  lstr vdnslst;                 // DANSデータ出力リスト
  map<string, bool> vdnsmap;    // DANSデータ出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  bool dns(const string& code); // DANSデータ出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_dns();                // -dnsオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdnsname(), vdnslst(), vdnsmap(), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dns")          // DANSデータの出力指定
  {
    if (strsplit(vdnslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "dnsname") // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

bool option::dns(const string& code) // DANSデータ出力フラッグを返す
{
  map<string, bool>::iterator dnspos = vdnsmap.find(code);
  if (dnspos == vdnsmap.end()) return false; // code無し！
  return vdnsmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "At least 2 items are required !");
  if (set_dns() < 0) return -2; // -dnsオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_dns()           // -dnsオプションのセット
{
  vdnsmap["fv"] = vdnsmap["me"] = vdnsmap["pv"] = vdnsmap["r"] = vdnsmap["vv"] = true;
  if (vdnslst.size() > 0)
  {
    vdnsmap["fv"] = vdnsmap["me"] = vdnsmap["pv"] = vdnsmap["r"] = vdnsmap["vv"] = false;
    for (string dns : vdnslst)
    {
      map<string, bool>::iterator dnspos = vdnsmap.find(dns);
      if (dnspos == vdnsmap.end())
      { return disperr(vpname, "Wrong -dns=" + dns + " option !"); }
      else
      { vdnsmap[dns] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["stat"] = vresmap["tt"] = true;
  vresmap["at"] = vresmap["m"] = vresmap["r"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["stat"] = vresmap["tt"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["stat"] = vresmap["tt"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["at"] = vresmap["cc"] = vresmap["ft"] = vresmap["m"] = vresmap["r"] = vresmap["stat"] = vresmap["tt"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  set<string> vgset;            // 群コードセット
  vector<string> vgcodevec;     // 群コードベクトル
  vecstatv vgstatvec;           // 群別・全体基礎統計量ベクトル
  double vsa;                   // 要因A平方和
  int vdfa;                     // 要因A自由度
  double vsr;                   // 残差平方和
  int vdfr;                     // 残差自由度
  double vst;                   // 全体平方和
  int vdft;                     // 全体自由度
  double vv1;                   // 群1の分散
  int vdf1;                     // 群1の自由度
  double vv2;                   // 群2の分散
  int vdf2;                     // 群2の自由度
  double vdm;                   // 2群の平均値の差
  double vvdmt;                 // 平均値の差の不偏分散(t検定)
  double vvdmw;                 // 平均値の差の不偏分散(Welch検定)
  int vdfw;                     // Welch検定の自由度
public:
  varclass() : vgset(), vgcodevec(), vgstatvec(), vsa(0.0), vdfa(0), vsr(0.0), vdfr(0), vst(0.0), vdft(0), vv1(0.0), vdf1(0), vv2(0.0), vdf2(0), vdm(0.0), vvdmt(0.0), vvdmw(0.0), vdfw(0)
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int setg(ldata& datalst);        // 群コードを分類して群数を返す
  int cala1(ldata& datalst, int jidx); // 分散分析用計算をして有効な群数を返す
  double caltt();               // 2標本t検定とWelchの検定用計算を行う
  size_t g() const              // 群数を返す
  { return vgset.size(); }
  const string& gcode(size_t idx) const // 群コードを返す
  { return (idx < vgcodevec.size()) ? vgcodevec[idx] : BLANKs; }
  int n(size_t idx) const       // 群の例数を返す
  { return (idx <= vgset.size()) ? vgstatvec[idx].n() : 0; }
  double mean(size_t idx) const // 群の平均値を返す
  { return (idx <= vgset.size()) ? vgstatvec[idx].mean() : 0.0; }
  double ss(size_t idx) const   // 群の平方和を返す
  { return (idx <= vgset.size()) ? vgstatvec[idx].ss() : 0.0; }
  double vr(size_t idx) const   // 群の不偏分散を返す
  { return (idx <= vgset.size()) ? vgstatvec[idx].vr() : 0.0; }
  double sa() const             // 要因A平方和を返す
  { return vsa; }
  double va() const             // 要因A分散を返す
  { return (vdfa > 0) ? vsa / (double)vdfa : 0.0; }
  int dfa() const               // 要因A自由度を返す
  { return vdfa; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  double vr() const             // 残差分散を返す
  { return (vdfr > 0) ? vsr / (double)vdfr : 0.0; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  double st() const             // 全体平方和を返す
  { return vst; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
  double v1() const             // 群1の分散を返す
  { return vv1; }
  int df1() const               // 群1の自由度を返す
  { return vdf1; }
  double v2() const             // 群2の分散を返す
  { return vv2; }
  int df2() const               // 群2の自由度を返す
  { return vdf2; }
  double dm() const             // 2群の平均値の差を返す
  { return vdm; }
  double vdmt() const           // 平均値の差の不偏分散を返す(t検定)
  { return vvdmt; }
  double vdmw() const           // 平均値の差の不偏分散を返す(Welch検定)
  { return vvdmw; }
  int dfw() const               // Welch検定の自由度を返す
  { return vdfw; }
};

int varclass::setg(ldata& datalst) // 群コードを分類して群数を返す
{
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    if (ditr->sdata(0) != BLANKs) vgset.insert(ditr->sdata(0));
  }
  size_t g = vgset.size();
  vgcodevec.resize(g);  vgstatvec.resize(g + 1);
  size_t i = 0;
  for (set<string>::iterator gitr = vgset.begin(); gitr != vgset.end(); ++gitr, ++i) vgcodevec[i] = *gitr;
  return (int)g;                // 群数を返す
}

int varclass::cala1(ldata& datalst, int jidx)
// 分散分析用計算をして有効な群数を返す
{
  size_t g = vgset.size();
  for (size_t i = 0; i <= g; ++i) vgstatvec[i].clear();
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    if (ditr->sdata(0) != BLANKs && ditr->fdata(jidx) != BLANKf)
    {
      int i = search_data(vgset, ditr->sdata(0));
      vgstatvec[i].sigma(ditr->fdata(jidx));
    }
  }
  vsa = 0.0;
  int ng = 0;
  for (size_t i = 0; i < g; ++i)
  {
    int n = vgstatvec[i].n();
    if (n > 0)
    {
      double m = vgstatvec[i].mean();
      vsa += (double)n * m * m;
      vgstatvec[g] += vgstatvec[i];
      ++ng;
    }
  }
  vdfa = ng - 1;  vdft = vgstatvec[g].n() - 1;
  vdfr = vdft - vdfa;
  vsa -= (double)vgstatvec[g].n() * vgstatvec[g].mean() * vgstatvec[g].mean();
  vst = vgstatvec[g].ss();
  const double Seps = DBL_EPSILON * fmax(5.0, vst);
  if (vdfa < 1 || vsa < Seps) vsa = 0.0;
  vsr = vst - vsa;
  if (vdft < 1 || vsr < Seps) vsr = 0.0;
  return ng;                    // 有効な群数を返す
}

double varclass::caltt()        // 2標本t検定とWelchの検定用計算を行う
{
  size_t g = vgset.size(), idx1 = 0, idx2 = 0;
  bool sch = true;
  for (size_t i = 0; sch && i < g; ++i)
  {
    if (vgstatvec[i].n() > 0)
    {
      if (idx1 == 0) idx1 = i;      // 群1のインデックス
      else if (idx2 == 0) idx2 = i; // 群2のインデックス
      else sch = false;
    }
  }
  int n1 = vgstatvec[idx1].n(), n2 = vgstatvec[idx2].n();
  double me1 = vgstatvec[idx1].mean(), me2 = vgstatvec[idx2].mean();
  double ss1 = vgstatvec[idx1].ss(), ss2 = vgstatvec[idx2].ss();
  vv1 = vgstatvec[idx1].vr();  vv2 = vgstatvec[idx2].vr();
  double f = (sgn(vv1) > 0) ? vv2 / vv1 : 0.0;
  vdf1 = n1 - 1;  vdf2 = n2 - 1;
  if (vv2 < vv1)
  {
    f = (sgn(vv2) > 0) ? vv1 / vv2 : 0.0;
    vdf1 = n2 - 1;  vdf2 = n1 - 1;
  }
  vdm = me1 - me2;
  vdft = n1 + n2 - 2;
  vvdmt = (ss1 + ss2) / (double)vdft * (double)(n1 + n2) / (double)n1 / (double)n2;
  double vn1 = vv1 / (double)n1,  vn2 = vv2 / (double)n2;
  vvdmw = vn1 + vn2;
  vdfw = (int)(vvdmw * vvdmw / (vn1 * vn1 / (double)vdf1 + vn2 * vn2 / (double)vdf2));
  return f;                     // 分散比Fを返す
}

class anovaiclass               // anovaiメインクラス
{
private:
  option opt;                   // オプション
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout, int iidx);
  // 結果のタイトルとデータ識別情報を出力する
  void put_anova1(ostream& fout, varclass& var);
  // 分散分析の結果を分散分析表形式で出力する
  void put_stat(ostream& fout, const varclass& var) const;
  // 群別基礎統計量を出力する
  void put_rratio(ostream& fout, const varclass& var);
  // 相関比を出力する
  void put_ttest(ostream& fout, ldbl prf, double dm, double vdmt, int dft, double vdmw, int dfw);
  // 2標本t検定結果とWelchの検定結果を出力する
  void put_ftest(ostream& fout, double f, int df1, int df2, ldbl prf);
  // 分散比Fの両側検定と推定結果を出力する
  void put_dnsid(ofstream& fout, const varclass& var);
  // DANSデータのタイトルとデータ識別情報を出力する
  void put_dns(ofstream& fout, const varclass& var, int iidx);
  // 分散分析の結果をケースデータ形式で出力する
public:
  anovaiclass(const char* c_pname = nullptr) : opt(c_pname), fname(), id(), datalst()
  // コンストラクタ
  {}
  ~anovaiclass()                // デストラクタ
  {}
  const string& pname() const   // プログラム名を返す
  { return opt.pname(); }
  void init()      // 初期設定を行う(コンストラクタ中ではできない処理)
  { opt.setdef(); }
  void setopt(const string& optstr, char devc = OPTdev) // オプションを設定する
  { opt.set(optstr, devc); }
  int get(const char* c_fname); // データを読み込み群コードを集計する
  void put(int nc);             // 集計結果を出力する
};

void anovaiclass::put_resid(ostream& fout, int iidx)
// 結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM(A):" << id.item(0) << "\n";
  fout << "ANALYSIS ITEM:" << id.item(iidx) << "\n";
}

void anovaiclass::put_anova1(ostream& fout, varclass& var)
// 分散分析の結果を分散分析表形式で出力する
{
  if (opt.res("stat"))
  {
    fout << "\n";
    put_stat(fout, var);
  }
  if (var.dfa() == 1)           // 2群の場合
  {
    if (opt.res("r"))
    {
      fout << "\n";
      put_rratio(fout, var);
    }
    if (opt.res("at"))
    {
      fout << "\n<ONE-WAY LAYOUT ANALYSIS OF VARIANCE(ANOVA1)>\n";
      putatab(fout, "GROUP(A)", var.sa(), var.dfa(), var.sr(), var.dfr(), var.st(), var.dft());
    }
    if (opt.res("tt"))
    {
      double f = var.caltt();
      ldbl prf = ftop(f, var.df1(), var.df2());
      prf = (prf < 0.5L) ? 2.0L * prf : 1.0L; // 両側p値にする
      if (opt.res("ft"))
      {
        fout << "\n";
        put_ftest(fout, f, var.df1(), var.df2(), prf);
      }
      fout << "\n";
      put_ttest(fout, prf, var.dm(), var.vdmt(), var.dft(), var.vdmw(), var.dfw());
    }
  }
  else                          // 3群以上の場合
  {
    if (opt.res("def") || opt.res("r"))
    {
      fout << "\n";
      put_rratio(fout, var);
    }
    if (opt.res("def") || opt.res("at"))
    {
      fout << "\n<ONE-WAY LAYOUT ANALYSIS OF VARIANCE(ANOVA1)>\n";
      putatab(fout, "GROUP(A)", var.sa(), var.dfa(), var.sr(), var.dfr(), var.st(), var.dft());
    }
  }
  fout << "\n" << endl;
}

void anovaiclass::put_stat(ostream& fout, const varclass& var) const
// 群別基礎統計量を出力する
{
  constexpr int LENline = 76;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  size_t g = var.g();
  fout << "GROUP    :BASIC STATISTICS\n";
  putline(fout, LENline);
  for (size_t i = 0; i < g; ++i)
  {
    int n = var.n(i);
    if (n > 0)
    {
      string gcodei(var.gcode(i));
      int npad = gcodei.size() - strclen(gcodei, CHARwide);
      fout << left << setw(LENhed + npad) << gcodei << right << ":";
      putmsd(fout, n, var.mean(i), var.vr(i));
    }
  }
  putline(fout, LENline);
  string tstr("TOTAL");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(LENhed + npad) << tstr << right << ":";
  putmsd(fout, var.n(g), var.mean(g), var.vr(g));
}

void anovaiclass::put_rratio(ostream& fout, const varclass& var)
// 相関比を出力する
{
  double r2 = var.sa() / var.st();
  fout << "Correlation Ratio:r=" << sqrt(r2) << "  r-squared=" << r2 << "\n";
  double f = (sgn(var.vr()) > 0) ? var.va() / var.vr() : 0.0;
  if (opt.res("cc"))
  {
    double crl = 0.0, cru = 0.0;
    calcrci(crl, cru, opt.cc(), var.dfa(), var.dfr(), f);
    fout << opt.cc() * 100.0 << "% CI:Lower=" << crl << "  Upper=" << cru << "\n";
  }
}

void anovaiclass::put_ttest(ostream& fout, ldbl prf, double dm, double vdmt, int dft, double vdmw, int dfw)
// 2標本t検定結果とWelchの検定結果を出力する
{
  fout << "<TWO SAMPLE T-TEST(->:Adopt This Result)>\n";
  string sel1 = "->", sel2 = "  ";
  if (prf <= 0.05L) { sel1 = "  ";  sel2 = "->"; }
  double t = dm / sqrt(vdmt);
  ldbl pr = ttop(t, dft);
  fout << sel1 << "t-test of Equal Variances(=ANOVA):t=" << t;
  fout << "  DF=" << dft << "  p=" << pr << pind(pr) << "\n";
  if (id.ni() > 2 && opt.res("m"))
  {
    pr = pr * (ldbl)(id.ni() - 1);
    if (pr > 1.0L) pr = 1.0L;
    fout << sel1 << "Bonferroni Type Multiple Comparisons:p=" << pr << pind(pr) << "\n";
  }
  double ci = calmcci(mcfis, opt.cc(), 2, 1, dft, 1, vdmt);
  if (opt.res("cc"))
  {
    fout << sel1 << opt.cc() * 100.0 << "% CI of Difference in Means:";
    fout << "Lower=" << dm - ci << "  Upper=" << dm + ci << "\n";
  }
  if (id.ni() > 2 && opt.res("m") && opt.res("cc"))
  {
    ci = calmcci(mcbon, opt.cc(), 2, 1, dft, id.ni() - 1, vdmt);
    fout << sel1 << opt.cc() * 100.0 << "% Simultaneous CI of Difference in Means:";
    fout << "Lower=" << dm - ci << "  Upper=" << dm + ci << "\n";
  }
  t = dm / sqrt(vdmw);
  pr = ttop(t, dfw);
  fout << sel2 << "Aspin-Welch Test:t=" << t;
  fout << "  DF=" << dfw << "  p=" << pr << pind(pr) << "\n";
  if (id.ni() > 2 && opt.res("m"))
  {
    pr = pr * (ldbl)(id.ni() - 1);
    if (pr > 1.0L) pr = 1.0L;
    fout << sel2 << "Bonferroni Type Multiple Comparisons:p=" << pr << pind(pr) << "\n";
  }
  if (opt.res("cc"))
  {
    ci = calmcci(mcfis, opt.cc(), 2, 1, dfw, 1, vdmw);
    fout << sel2 << opt.cc() * 100.0 << "% CI of Difference in Means:";
    fout << "Lower=" << dm - ci << "  Upper=" << dm + ci << "\n";
  }
  if (id.ni() > 2 && opt.res("m") && opt.res("cc"))
  {
    ci = calmcci(mcbon, opt.cc(), 2, 1, dfw, id.ni() - 1, vdmw);
    fout << sel2 << opt.cc() * 100.0 << "% Simultaneous CI of Difference in Means:";
    fout << "Lower=" << dm - ci << "  Upper=" << dm + ci << "\n";
  }
}

void anovaiclass::put_ftest(ostream& fout, double f, int df1, int df2, ldbl prf)
// 分散比Fの両側検定と推定結果を出力する
{
  fout << "<TWO-TAILED F-TEST TO COMPARE TWO VARIANCES>\n";
  fout << "  F=" << f << " DF1=" << df1 << "  DF2=" << df2 << "  Two-tailed p=" << prf << pind(prf) << "\n";
  if (opt.res("cc"))
  {
    double fl = 0.0, fu = 0.0;
    calfci(fl, fu, opt.cc(), df1, df2, f);
    fout << "  " << opt.cc() * 100.0 << "% CI of F:Lower=" << fl << "  Upper=" << fu << "\n";
  }
}

void anovaiclass::put_dnsid(ofstream& fout, const varclass& var)
// DANSデータのタイトルとデータ識別情報を出力する
{
  size_t g = var.g();
  int tni = 1;
  if (opt.dns("me")) tni += 2 * (g + 1);
  if (opt.dns("vv")) tni += 3;
  if (opt.dns("fv")) ++tni;
  if (opt.dns("pv")) ++tni;
  if (opt.dns("r")) ++tni;
  idd tid(id.title(), tni);
  tid.assign_item(0, "ITEM NAME");
  int j = 1;
  if (opt.dns("me"))
  {
    for (size_t i = 0; i < g; ++i)
    {
      tid.assign_item(j++, "Number of " + id.item(0) + ":" + var.gcode(i));
      tid.assign_item(j++, "Mean of " + id.item(0) + ":" + var.gcode(i));
    }
    tid.assign_item(j++, "Total Number");
    tid.assign_item(j++, "Total Mean");
  }
  if (opt.dns("vv"))
  {
    tid.assign_item(j++, "Variance of Factor-A");
    tid.assign_item(j++, "Variance of Residuals");
    tid.assign_item(j++, "Total Variance");
  }
  if (opt.dns("fv")) tid.assign_item(j++, "F-value");
  if (opt.dns("pv")) tid.assign_item(j++, "p-value");
  if (opt.dns("r")) tid.assign_item(j++, "Corralation Ratio");
  tid.put(fout, formtag);
}

void anovaiclass::put_dns(ofstream& fout, const varclass& var, int iidx)
// 分散分析の結果をケースデータ形式で出力する
{
  size_t g = var.g();
  int tni = 1;
  if (opt.dns("me")) tni += 2 * (g + 1);
  if (opt.dns("vv")) tni += 3;
  if (opt.dns("fv")) ++tni;
  if (opt.dns("pv")) ++tni;
  if (opt.dns("r")) ++tni;
  datac cdata(tni);
  cdata.assign(0, id.item(iidx));
  double fa = (sgn(var.vr()) > 0) ? var.va() / var.vr() : 0.0;
  double r2 = (sgn(var.st()) > 0) ? var.sa() / var.st() : 0.0;
  double pr = ftop(fa, var.dfa(), var.dfr());
  int j = 1;
  if (opt.dns("me"))
  {
    for (size_t i = 0; i < g; ++i)
    {
      int n = var.n(i);
      cdata.assign(j++, (double)n);
      if (n > 0) cdata.assign(j++, var.mean(i));
      else cdata.assign(j++, BLANKs);
    }
    cdata.assign(j++, (double)var.n(g));
    cdata.assign(j++, var.mean(g));
  }
  if (opt.dns("vv"))
  {
    cdata.assign(j++, var.va());
    cdata.assign(j++, var.vr());
    cdata.assign(j++, var.vr(g));
  }
  if (opt.dns("fv")) cdata.assign(j++, fa);
  if (opt.dns("pv")) cdata.assign(j++, pr);
  if (opt.dns("r")) cdata.assign(j++, sqrt(r2));
  cdata.put(fout);
}

int anovaiclass::get(const char* c_fname) // データを読み込み群コードを集計する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void anovaiclass::put(int nc)   // 集計結果を出力する
{
  if (nc < 1)                   // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, 0);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var;                 // 計算用変数クラス
  if (var.setg(datalst) < 1) return; // 群がない
  ofstream foutdns;
  if (opt.dnsname() != "")      // DANSデータ出力
  {
    if (ofopen(foutdns, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
    put_dnsid(foutdns, var);
  }
  for (size_t j = 1; j < id.ni(); ++j)
  {
    int ng = var.cala1(datalst, j);
    if (ng > 0 && sgn(var.st()) > 0)
    {
      if (opt.res("res"))       // 集計結果を分散分析表形式で出力する
      {
        put_resid(cout, j);
        put_anova1(cout, var);
      }
      if (opt.dnsname() != "") put_dns(foutdns, var, j); // 集計結果をケースデータ形式で出力する
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  anovaiclass anovai(argv[0]);  // メインクラス変数
  try
  {
    anovai.init();              // anovaiの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int nc = 0;                 // 読み込んだケース数
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) nc += anovai.get(nullptr); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { anovai.setopt(argv[i] + 1);  fok = true; }
      else
      { nc += anovai.get(argv[i]);  fok = false; }
    }
    anovai.put(nc);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(anovai.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(anovai.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
