// anovame.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-24 09:05:15 tac>
//
// 例数、平均値、SD、SEデータから対応の無い多群間で一元配置分散分析を行う。
// 2群の場合は2標本t検定も出力する。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W anovame anovame.C -I/usr/local/include
//               -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "anovame  Version 7.42e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:anova1e [<OPTION> ...] [<DATA FILE> ...]",
  "  The one-way ANOVA(including two sample t-test) with number and mean",
  "  and SD or SE or Variance.",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=anovame_ngp.dat",
  "  -res=[-][{def|all|at|cc[<C-COEF>]|ft|m[{b|d|f|s|t}]|r|stat|tt|vt}][,...]",
  "      :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    at:ANOVA  cc[<C-COEF>]:confidence intervals(0=<COEF<1)",
  "    ft:2-tailed F-test to compare 2 variances",
  "    m[{b|d|f|s|t}]:multiple comparisons",
  "      b:Bonferroni type  d:Dunnett type  f:Fisher type  s:Scheffe type",
  "      t:Tukey type",
  "    r:corralation ratio  stat:basic statistics  tt:2 sample t-test",
  "    vt:test for equal variances",
  "    default:auto judgment according to data"
};

const string Title = "=== Comparison of Means of Multiple Groups ==="; // 表題
const string NGPNAMEdef = "anovame_ngp.dat"; // 省略時Ngraphファイル名

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  mctype vm;                    // 多重比較法
  string vngpname;              // Ngraphファイル名
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  mctype m() const              // 多重比較法を返す
  { return vm; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check();                  // オプションのチェックとセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vm(mctuk), vngpname(), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check()             // オプションのセットとチェック
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["stat"] = vresmap["tt"] = true;
  vresmap["at"] = vresmap["m"] = vresmap["r"] = vresmap["vt"] = false;
  if (vreslst.size() > 0)
  {
     vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["stat"] = vresmap["tt"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["stat"] = vresmap["tt"] = true;
      }
      else if (res == "all")
      {
        vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["ft"] = vresmap["m"] = vresmap["r"] = vresmap["stat"] = vresmap["tt"] = vresmap["vt"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res[0] == 'm' && res.size() > 1)
      {
        switch (res[1])
        {
        case 'b' : vm = mcbon;
          break;
        case 'd' : vm = mcdun;
          break;
        case 'f' : vm = mcfis;
          break;
        case 's' : vm = mcsch;
          break;
        case 't' : vm = mctuk;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["m"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class gstat                     // 群統計量クラス
{
private:
  string vcode;                 // 群コード
  statv<double> vystat;         // 基礎統計量
public:
  gstat() : vcode(), vystat()   // デフォールトコンストラクタ
  {}
  gstat(const string& code0, int n0, double mean0, double sd0);
  // コンストラクタ
  gstat(const gstat& grp) : vcode(grp.vcode), vystat(grp.vystat)
  // コピーコンストラクタ
  {}
  virtual ~gstat()              // デストラクタ
  {}
  const gstat& operator=(const gstat& gstat0); // コピー代入演算子
  int set(const string& code0, int n0, double mean0, double sd0);
  // データをセットする
  const string& code() const    // 群コードを返す
  { return vcode; }
  const statv<double>& ystat() const // 基礎統計量を返す
  { return vystat; }
};

gstat::gstat(const string& code0, int n0, double mean0, double sd0) : vcode(code0), vystat()
// コンストラクタ
{
  vystat.set(n0, mean0, sd0);
}

const gstat& gstat::operator=(const gstat& gstat0) // コピー代入演算子
{
  if (&gstat0 == this) return *this; // 自分自身への代入を防ぐ
  vcode = gstat0.vcode;  vystat = gstat0.vystat;
  return *this;
}

int gstat::set(const string& code0, int n0, double mean0, double sd0)
// データをセットする
{
  vcode = code0;
  return vystat.set(n0, mean0, sd0);
}

class anovamclass               // anovamメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  list<gstat> gstatlst;         // 群統計量リスト
  void get_stat(gstat& tgstat, const datac& cdata); // 統計量を取得する
  void put(ostream& fout, ofstream& foutngp, option& opt, int nd);
  // 集計結果を出力する
  void put_resid(ostream& fout, int nd);
  // 分析結果のタイトルとデータ識別情報を出力する
  double put_stat(ostream& fout, option& opt, statv<double>& tstat);
  // 群別基礎統計量の出力と累積集計
  void put_rratio(ostream& fout, option& opt, double sa, double sr, double st, int dfa, int dfr);
  // 相関比の出力
  void put_ttest(ostream& fout, option& opt, const statv<double>& tstat);
  // 2標本t検定とWelchの検定
  ldbl put_ftest(ostream& fout, option& opt, double f, int df1, int df2);
  // 分散比Fの検定と推定結果を出力し両側p値を返す
  void put_vtest(ostream& fout, const statv<double>& tstat) const;
  // 分散の均一性の検定結果出力
  void put_mcop(ostream& fout, const option& opt, const statv<double>& tstat, double vr) const;
  // 2群間の多重比較出力
  void put_mcopd(ostream& fout, const option& opt, const statv<double>& tstat, double vr) const;
  // Dunnett型多重比較出力
  void put_mdci(ostream& fout, const option& opt, const statv<double>& tstat, double vr) const;
  // 2群間の平均値の差の信頼区間出力
  void put_mdcid(ostream& fout, const option& opt, const statv<double>& tstat, double vr) const;
  // 2群間の平均値の差のDunnett型信頼区間出力
  void plot_ngpid(ofstream& fout, int nd);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngp(ofstream& fout, option& opt); // Ngraph用データ出力
public:
  anovamclass() : fname(), id(), gstatlst() // コンストラクタ
  {}
  ~anovamclass()                // デストラクタ
  {}
  void getput(const char* c_fname, option& opt);
  // データを読み込んで集計結果を出力する
};

void anovamclass::get_stat(gstat& tgstat, const datac& cdata)
// 統計量を取得する
{
  if (cdata.sdata(0) == BLANKs) throw runtime_error("Wrong group code !");
  int n = 0;
  double mean = BLANKf, sd = BLANKf, se = BLANKf, v = BLANKf;
  string str[2];                // 統計量指定str[0]:項目,str[1]:データ
  for (size_t j = 1; j < cdata.size(); ++j)
  {
    if (getopt(str, cdata.sdata(j)) > 0)
    {
      strlower(str[0]);
      char* ptr = nullptr;
      if (str[0] == "n") n = strtol((str[1]).c_str(), &ptr, 0);
      else if (str[0] == "m") mean = strtodf(str[1]);
      else if (str[0] == "sd") sd = strtodf(str[1]);
      else if (str[0] == "se") se = strtodf(str[1]);
      else if (str[0] == "v") v = strtodf(str[1]);
    }
  }
  if (n < 1 || mean == BLANKf) throw runtime_error("Wrong number or mean !");
  if (sd == BLANKf)             // sdなし
  {
    if (v != BLANKf) sd = sqrt(v);
    else if (se != BLANKf) sd = se * sqrt((double)n);
  }
  if (sd == BLANKf) throw runtime_error("Wrong SD or SE or variance !");
  tgstat.set(cdata.sdata(0), n, mean, sd);
}

void anovamclass::put(ostream& fout, ofstream& foutngp, option& opt, int nd)
// 集計結果を出力する
{
  int g = gstatlst.size();
  statv<double> tstat;
  if (opt.res("res"))           // 計算結果出力
  {
    put_resid(fout, nd);
    double sa = put_stat(fout, opt, tstat);
    double st = tstat.ss();
    if (g < 2 || sgn(st) <= 0)  // 出力中止
    {
      fout << "\n" << endl;
      return;
    }
    double sr = st - sa;
    int dft = tstat.n() - 1, dfr = tstat.n() - g;
    const double Seps = DBL_EPSILON * fmax(5.0, st);
    if (sr < Seps) sr = 0.0;
    if (g == 2)
    {
      if (opt.res("r")) put_rratio(fout, opt, sa, sr, st, g - 1, dfr);
      if (opt.res("at"))        // 2群の時はatを指定された時だけ出力
      {
        if (opt.res("vt")) put_vtest(fout, tstat);
        fout << "\n<ONE-WAY LAYOUT ANALYSIS OF VARIANCE(ANOVA1)>\n";
        putatab(fout, "GROUP(A)", sa, g - 1, sr, dfr, st, dft);
      }
      if (opt.res("tt")) put_ttest(fout, opt, tstat);
    }
    else
    {
      if (opt.res("def") || opt.res("r")) put_rratio(fout, opt, sa, sr, st, g - 1, dfr);
      if (opt.res("vt")) put_vtest(fout, tstat);
      if (opt.res("def") || opt.res("at")) // 3群以上の時は省略時も出力
      {
        fout << "\n<ONE-WAY LAYOUT ANALYSIS OF VARIANCE(ANOVA1)>\n";
        putatab(fout, "GROUP(A)", sa, g - 1, sr, dfr, st, dft);
      }
      if (opt.res("m"))
      {
        double vr = (dfr > 0) ? sr / dfr : 0.0;
        if (opt.m() == mcdun)   // Dunnett型
        {
          put_mcopd(fout, opt, tstat, vr);
          if (opt.res("cc")) put_mdcid(fout, opt, tstat, vr);
        }
        else
        {
          put_mcop(fout, opt, tstat, vr);
          if (opt.res("cc")) put_mdci(fout, opt, tstat, vr);
        }
      }
    }
    fout << "\n" << endl;
  }
  if (opt.ngpname() != "")      // Ngraph用データ出力
  {
    plot_ngpid(foutngp, nd);
    plot_ngp(foutngp, opt);
  }
}

void anovamclass::put_resid(ostream& fout, int nd)
// 分析結果のタイトルとデータ識別情報を出力する
{
  if (id.ni() < 2) id.set(fname, 2);
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM(A):" << id.item(0) << "\n";
  fout << "ANALYSIS ITEM:" << id.item(1) << "\n";
}

double anovamclass::put_stat(ostream& fout, option& opt, statv<double>& tstat)
// 群別基礎統計量の出力と累積集計
{
  constexpr int LENline = 76;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  if (opt.res("stat"))
  {
    fout << "\nGROUP     :BASIC STATISTICS\n";
    putline(fout, LENline);
  }
  double sa = 0.0;
  for (list<gstat>::const_iterator gitr = gstatlst.begin(); gitr != gstatlst.end(); ++gitr)
  {
    int n = gitr->ystat().n();
    double m = gitr->ystat().mean();
    int npad = gitr->code().size() - strclen(gitr->code(), CHARwide);
    if (opt.res("stat"))
    {
      fout << left << setw(LENhed + npad) << gitr->code() << right << ":";
      putmsd(fout, n, m, gitr->ystat().vr());
    }
    tstat += gitr->ystat();  sa += n * m * m;
  }
  if (opt.res("stat"))
  {
    putline(fout, LENline);
    string tstr("TOTAL");
    int npad = tstr.size() - strclen(tstr, CHARwide);
    fout << left << setw(LENhed + npad) << tstr << right << ":";
    putmsd(fout, tstat.n(), tstat.mean(), tstat.vr());
  }
  sa -= tstat.n() * tstat.mean() * tstat.mean();
  const double Seps = DBL_EPSILON * fmax(5.0, tstat.ss());
  if (sa < Seps) sa = 0.0;
  return sa;                    // 要因Aの平方和を返す
}

void anovamclass::put_rratio(ostream& fout, option& opt, double sa, double sr, double st, int dfa, int dfr)
// 相関比の出力
{
  double r2 = sa / st;
  fout << "\nCorrelation Ratio:r=" << sqrt(r2) << "  r-squared=" << r2 << "\n";
  if (opt.res("cc"))
  {
    double va = (dfa > 0) ? sa / (double)dfa : 0.0;
    double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
    double f = (vr > 0.0) ? va / vr : 0.0;
    double crl = 0.0, cru = 0.0;
    calcrci(crl, cru, opt.cc(), dfa, dfr, f);
    fout << opt.cc() * 100.0 << "% CI of r:Lower=" << crl << "  Upper=" << cru << "\n";
  }
}

void anovamclass::put_ttest(ostream& fout, option& opt, const statv<double>& tstat)
// 2標本t検定とWelchの検定
{
  list<gstat>::const_iterator gitr = gstatlst.begin();
  const statv<double>* ystat1 = &(gitr->ystat());
  ++gitr;
  const statv<double>* ystat2 = &(gitr->ystat());
  int n1 = ystat1->n(), n2 = ystat2->n();
  double ss1 = ystat1->ss(), ss2 = ystat2->ss();
  if (n1 < 2 || n2 < 2 || (ss1 <= 0.0 && ss2 <= 0.0)) return; // 出力なし
  double v1 = ystat1->vr(), v2 = ystat2->vr();
  double f = (v2 > 0.0) ? v1 / v2 : 0.0;
  int df1 = n1 - 1, df2 = n2 - 1;
  if (v1 < v2)
  {
    f = (v1 > 0.0) ? v2 / v1 : 0.0;
    df1 = n2 - 1;  df2 = n1 - 1;
  }
  ldbl pr = put_ftest(fout, opt, f, df1, df2);
  fout << "\n<TWO SAMPLE T-TEST(->:Adopt This Result)>\n";
  string sel1 = "->", sel2 = "  ";
  if (pr <= 0.05L) { sel1 = "  ";  sel2 = "->"; }
  double d = ystat2->mean() - ystat1->mean();
  int df = n1 + n2 - 2;
  double vd = (ss1 + ss2) / df * (double)(n1 + n2) / (double)n1 / (double)n2;
  double t = d / sqrt(vd);
  pr = ttop(t, df);
  fout << sel1 << "t-test of Equal Variances(=ANOVA1):t=" << t;
  fout << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
  double ci = calmcci(mcfis, opt.cc(), 2, 1, df, 1, vd);
  if (opt.res("cc"))
  {
    fout << sel1 << opt.cc() * 100.0 << "% CI of Difference in Means:";
    fout << "Lower=" << d - ci << "  Upper=" << d + ci << "\n";
  }
  double vn1 = v1 / (double)n1, vn2 = v2 / (double)n2;
  vd = vn1 + vn2;
  df = (int)(vd * vd / (vn1 * vn1 / (double)(n1 - 1) + vn2 * vn2 / (double)(n2 -1)));
  t = d / sqrt(vd);
  pr = ttop(t, df);
  fout << sel2 << "Aspin-Welch Test:t=" << t;
  fout << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
  if (opt.res("cc"))
  {
    ci = calmcci(mcfis, opt.cc(), 2, 1, df, 1, vd);
    fout << sel2 << opt.cc() * 100.0 << "% CI of Difference in Means:";
    fout << "Lower=" << d - ci << "  Upper=" << d + ci << "\n";
  }
}

ldbl anovamclass::put_ftest(ostream& fout, option& opt, double f, int df1, int df2)
// 分散比Fの両側検定と推定結果を出力し両側p値を返す
{
  ldbl pr = ftop(f, df1, df2);
  pr = (pr < 0.5L) ? 2.0L * pr : 1.0L; // 両側有意確率にする
  if (opt.res("ft"))
  {
    fout << "\n<TWO-TAILED F-TEST TO COMPARE TWO VARIANCES>\n";
    fout << "  F=" << f << " DF1=" << df1 << " DF2=" << df2 << "  two-tailed p=" << pr << pind(pr) << "\n";
    if (opt.res("cc"))
    {
      double fl = 0.0, fu = 0.0;
      calfci(fl, fu, opt.cc(), df1, df2, f);
      fout << "  " << opt.cc() * 100.0 << "% CI of F:Lower=" << fl << "  Upper=" << fu << "\n";
    }
  }
  return pr;                    // 両側p値を返す
}

void anovamclass::put_vtest(ostream& fout, const statv<double>& tstat) const
// 分散の均一性の検定結果出力
{
  int g = gstatlst.size();
  int d = -g;
  double w = 0.0, ss = 0.0, lnss = 0.0;
  for (list<gstat>::const_iterator gitr = gstatlst.begin(); gitr != gstatlst.end(); ++gitr)
  {
    int df = gitr->ystat().n() - 1;
    d += gitr->ystat().n();  w += 1.0 / (double)df;
    ss += (double)df * gitr->ystat().vr();
    lnss += (double)df * log(gitr->ystat().vr());
  }
  double m = (double)d * log(ss / (double)d) - lnss;
  double c = 1.0 + 1.0 / 3.0 / (double)(g - 1) * (w - 1.0 / (double)d);
  double csqr = m / c;
  int df = g - 1;
  ldbl p = ctop(csqr, df);
  fout << "\nBartlett Test:chi-squared=" << csqr;
  fout << "  DF=" << df << "  p=" << p << pind(p) << endl;
}

void anovamclass::put_mcop(ostream& fout, const option& opt, const statv<double>& tstat, double vr) const
// 2群間の多重比較出力
{
  constexpr int LENline = 76;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  if (vr <= 0.0) return;        // 出力なし
  int g = gstatlst.size();
  int df1 = 1, df2 = tstat.n() - g, dc = 1;
  string method = "TUKEY";
  switch (opt.m())
  {
  case mcbon : method = "BONFERRONI";  dc = g * (g - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISONS OF MEANS>\n";
  if (opt.m() == mctuk)
  {
    fout << "GROUP     - GROUP          q-value      Groups          DF       p-value\n";
  }
  else
  {
    fout << "GROUP     - GROUP          F-value         DF1         DF2       p-value\n";
  }
  putline(fout, LENline);
  list<gstat>::const_iterator gitr1 = gstatlst.begin();
  for (int i = 0; i < g - 1; ++i, ++gitr1)
  {
    string codei(gitr1->code());
    int npadi = codei.size() - strclen(codei, CHARwide);
    int ni = (gitr1->ystat()).n();
    double mi = (gitr1->ystat()).mean();
    list<gstat>::const_iterator gitr2 = gitr1;
    ++gitr2;
    while (gitr2 != gstatlst.end())
    {
      int nj = (gitr2->ystat()).n();
      double dm = (gitr2->ystat()).mean() - mi;
      double f = dm * dm / df1 / (1.0 / (double)ni + 1.0 / (double)nj) / vr;
      double pr = calmcpr(opt.m(), f, g, df1, df2, dc);
      int npad = gitr2->code().size() - strclen(gitr2->code(), CHARwide);
      fout << left << setw(LENhed + npad) << gitr2->code() << "- ";
      fout << setw(LENhed + npadi) << codei << right << setw(12) << f;
      if (opt.m() == mctuk) fout << setw(12) << g;
      else fout << setw(12) << df1;
      fout << setw(12) << df2 << setw(14) << pr << pind(pr) << "\n";
      ++gitr2;
    }
  }
  putline(fout, LENline);
}

void anovamclass::put_mcopd(ostream& fout, const option& opt, const statv<double>& tstat, double vr) const
// Dunnett型多重比較出力
{
  constexpr int LENline = 76;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  if (vr <= 0.0) return;        // 出力なし
  int g = gstatlst.size();
  int df1 = 1, df2 = tstat.n() - g, dc = 1;
  fout << "\n<DUNNETT TYPE MULTIPLE COMPARISONS OF MEANS>\n";
  fout << "GROUP     - GROUP          d-value      Groups          DF       p-value\n";
  putline(fout, LENline);
  list<gstat>::const_iterator gitr = gstatlst.begin();
  string codei(gitr->code());
  int npadi = codei.size() - strclen(codei, CHARwide);
  double ni = (double)((gitr->ystat()).n());
  double mi = (gitr->ystat()).mean();
  ++gitr;
  for (int j = 1; j < g; ++j, ++gitr)
  {
    double nj = (double)((gitr->ystat()).n());
    double dm = (gitr->ystat()).mean() - mi;
    double f = dm * dm / df1 / (1.0 / ni + 1.0 / nj) / vr;
    double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
    double pr = calmcpr(opt.m(), f, g, df1, df2, dc, r);
    int npad = gitr->code().size() - strclen(gitr->code(), CHARwide);
    fout << left << setw(LENhed + npad) << gitr->code() << "- ";
    fout << setw(LENhed + npadi) << codei << right << setw(12) << f;
    fout << setw(12) << g << setw(12) << df2 << setw(14) << pr << pind(pr) << "\n";
  }
  putline(fout, LENline);
}

void anovamclass::put_mdci(ostream& fout, const option& opt, const statv<double>& tstat, double vr) const
// 2群間の平均値の差の信頼区間出力
{
  constexpr int LENline = 70;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  if (vr <= 0.0) return;        // 出力なし
  int g = gstatlst.size();
  int df1 = 1, df2 = tstat.n() - g, dc = 1;
  string method = "TUKEY";
  switch (opt.m())
  {
  case mcbon : method = "BONFERRONI";  dc = g * (g - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df1 = g - 1;
  }
  fout << "\n<" << method << " TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CONFIDENCE INTERVALS>\n";
  fout << "GROUP     - GROUP    Diff of Means   Precision   Lower-Lim   Upper-Lim\n";
  putline(fout, LENline);
  list<gstat>::const_iterator gitr1 = gstatlst.begin();
  for (int i = 0; i < g - 1; ++i, ++gitr1)
  {
    string codei(gitr1->code());
    int npadi = codei.size() - strclen(codei, CHARwide);
    int ni = (gitr1->ystat()).n();
    double mi = (gitr1->ystat()).mean();
    list<gstat>::const_iterator gitr2 = gitr1;
    ++gitr2;
    while (gitr2 != gstatlst.end())
    {
      int nj = (gitr2->ystat()).n();
      double dm = (gitr2->ystat()).mean() - mi;
      double vd = (1.0 / (double)ni + 1.0 / (double)nj) * vr;
      double ci = calmcci(opt.m(), opt.cc(), g, df1, df2, dc, vd);
      int npad = gitr2->code().size() - strclen(gitr2->code(), CHARwide);
      fout << left << setw(LENhed + npad) << gitr2->code() << "- ";
      fout << setw(LENhed + npadi) << codei << right << setw(12) << dm;
      fout << setw(12) << ci << setw(12) << dm - ci << setw(12) << dm + ci << "\n";
      ++gitr2;
    }
  }
  putline(fout, LENline);
}

void anovamclass::put_mdcid(ostream& fout, const option& opt, const statv<double>& tstat, double vr) const
// 2群間の平均値の差のDunnett型信頼区間出力
{
  constexpr int LENline = 70;   // 区切り線の長さ
  constexpr int LENhed = 10;    // ヘッダの長さ
  if (vr <= 0.0) return;        // 出力なし
  int g = gstatlst.size();
  int df1 = 1, df2 = tstat.n() - g, dc = 1;
  fout << "\n<DUNNETT TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CONFIDENCE INTERVALS>\n";
  fout << "GROUP     - GROUP    Diff of Means   Precision   Lower-Lim   Upper-Lim\n";
  putline(fout, LENline);
  list<gstat>::const_iterator gitr = gstatlst.begin();
  string codei(gitr->code());
  int npadi = codei.size() - strclen(codei, CHARwide);
  int ni = (gitr->ystat()).n();
  double mi = (gitr->ystat()).mean();
  ++gitr;
  for (int j = 1; j < g; ++j, ++gitr)
  {
    int nj = (gitr->ystat()).n();
    double dm =  (gitr->ystat()).mean() - mi;
    double vd = (1.0 / ni + 1.0 / nj) * vr;
    double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
    double ci = calmcci(opt.m(), opt.cc(), g, df1, df2, dc, vd, r);
    int npad = gitr->code().size() - strclen(gitr->code(), CHARwide);
    fout << left << setw(LENhed + npad) << gitr->code() << "- ";
    fout << setw(LENhed + npadi) << codei << right << setw(12) << dm;
    fout << setw(12) << ci << setw(12) << dm - ci << setw(12) << dm + ci << "\n";
  }
  putline(fout, LENline);
}

void anovamclass::plot_ngpid(ofstream& fout, int nd)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  if (id.ni() < 2) id.set(fname, 2);
  if (nd > 1)
  { fout << NGPdelm << "\n"; }
  else
  { fout << "-s1 -zx0," << gstatlst.size() + 1 << ",1 -dmark,78 -w80 -o400\n"; }
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%GROUP ITEM(A):" << id.item(0) << "\n";
  fout << "%ANALYSIS ITEM:" << id.item(1) << "\n";
}

void anovamclass::plot_ngp(ofstream& fout, option& opt) // Ngraph用データ出力
{
  fout << "%<MEAN & SE(-zx0," << gstatlst.size() + 1 << ",1 -derrorbar_y -w40 -o500)>\n";
  int i = 0;
  for (list<gstat>::const_iterator gitr = gstatlst.begin(); gitr != gstatlst.end(); ++gitr, ++i)
  {
    double se = gitr->ystat().se();
    fout << i + 1 << " " << gitr->ystat().mean();
    fout << " " << se << " " << -se << "\n";
  }
  fout << "= =\n";
  fout << "%<MEAN & SD>\n";
  i = 0;
  for (list<gstat>::const_iterator gitr = gstatlst.begin(); gitr != gstatlst.end(); ++gitr, ++i)
  {
    double sd = gitr->ystat().sd();
    fout << i + 1 << " " << gitr->ystat().mean();
    fout << " " << sd << " " << -sd << endl;
  }
  if (opt.res("cc"))
  {
    fout << "= =\n";
    fout << "%<MEAN & " << opt.cc() * 100.0 << "% CONFIDENCE INTERVALS>\n";
    i = 0;
    for (list<gstat>::const_iterator gitr = gstatlst.begin(); gitr != gstatlst.end(); ++gitr, ++i)
    {
      double t = ptot(1.0 - opt.cc(), gitr->ystat().n() - 1);
      double ci = t * gitr->ystat().se();
      double sd = gitr->ystat().sd();
      fout << i + 1 << " " << gitr->ystat().mean();
      fout << " " << ci << " " << -ci << endl;
    }
  }
}

void anovamclass::getput(const char* c_fname, option& opt)
// データを読み込んで集計結果を出力する
{
  if (opt.check() != 0) throw option_error("Run-time option error !");
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  istream* finp = nullptr;
  if (fname.empty())
  { finp = &cin;  fname = "stdin"; }
  else
  {
    finp = new ifstream(fname.c_str());
    if (!(*finp)) throw runtime_error(string("No such file->") + fname);
  }
  ofstream foutngp;
  if (opt.ngpname() != "")      // Ngraph用データ出力
  {
    if (ofopen(foutngp, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  }
  string sline;                 // データ入力用
  datac cdata(5);               // データオブジェクト
  int nd = 1;                   // データの数
  while (getsline(sline, *finp) != EOF)
  {
    strchomp(sline);
    if (sline.empty())          // ブランク行
    {
      if (gstatlst.size() > 0)   // 群統計量がある時
      {
        put(cout, foutngp, opt, nd);
        id.clear();  gstatlst.clear();  ++nd;
      }
    }
    else if (sline[0] == CHRtags) // データ識別情報
    {
      if (gstatlst.size() > 0)   // 群統計量がある時
      {
        put(cout, foutngp, opt, nd);
        id.clear();  gstatlst.clear();  ++nd;
      }
      if (id.append(sline) < 0) throw runtime_error(string("Wrong ID information->") + sline);
      if (id.ni() > 2) throw runtime_error(string("Too many items->") + sline);
    }
    else if (cdata.getdata(sline) > 0) // 統計量データ読み込み
    {
      gstat tgstat;
      get_stat(tgstat, cdata);
      gstatlst.push_back(tgstat);
    }
  }
  if (finp != &cin) delete finp;
  if (gstatlst.size() > 0) put(cout, foutngp, opt, nd);
}

int main(int argc, char *argv[]) // メインルーチン
{
  anovamclass anovam;           // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) anovam.getput(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { anovam.getput(argv[i], opt);  fok = false; }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
