// corege.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-24 09:43:09 tac>
//
// 相関係数と回帰直線を計算する。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W corege corege.C -I/usr/local/include
//              -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "corege  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:corege [<OPTION> ...] [<DATA FILE> ...]",
  "  The correlation and regression analysis with the first item as the",
  "  explanatory variable(X) and the next item as the criterion variable(Y).",
  "<OPTION>",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={ce|cl|data|line|ma|pe|pl|te|tl}[,...]:select the data for Ngraph",
  "    ce:confidence ellipse  cl:confidence limit  data:raw data",
  "    line:regression line  ma:major axis regression line",
  "    pe:prediction ellipse  pl:prediction limit  te:tolerance ellipse",
  "    tl:tolerance limit  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=corege_ngp.dat",
  "  -res=[-][{def|all|cc[<C-COEF>]|ce|cl|dw[s]|icc|r|reg|sdr|stat",
  "      |x[<VALUE>]|y[<VALUE>]}][,...]:select output results",
  "    -:with -, no output(-res=-:no output at all)",
  "    def:same output as default  all:output all",
  "    cc<C-COEF>:confidence intervals(0=<C-COEF<1)  ce:confidence ellipse",
  "    cl:confidence limit  dw[s]:Durbin-Watson ratio(dws:sort by x)",
  "    icc:intraclass correlation coefficient  r:correlation coefficient",
  "    reg:regression line  sdr:regression residual  stat:basic statistics",
  "    x[<VALUE>]:estimated value of Y when x is the mean or VALUE",
  "    y[<VALUE>]:inverse estimated value of X when y is the mean or VALUE",
  "    default:auto judgment according to data"
};

const string Title = "=== Correlation and Regression Analysis ==="; // 表題
const string NGPNAMEdef = "corege_ngp.dat"; // 省略時Ngraphファイル名
constexpr uint RDdef = 50;      // グラフの描画範囲分割数省略時値

enum dwtype                     // ダービン・ワトソン比の種類
{
  dwnon,                        // なし
  dwput,                        // 出力
  dwsort,                       // x順にソートした値を出力
  dwetc                         // その他
};

class option                    // オプション集合体
{
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  dwtype vdw;                   // ダービン・ワトソン比出力フラッグ
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  double vx;                    // yの値を推定するxの値
  double vy;                    // xの値を逆推定するyの値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  dwtype dw() const          // ダービン・ワトソン比の種類を返す
  { return vdw; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double x() const              // yの値を推定するxの値を返す
  { return vx; }
  double y() const              // xの値を逆推定するyの値を返す
  { return vy; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdw(dwsort), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vx(DBL_MAX), vy(DBL_MAX)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni != 2) throw runtime_error("Number of items must be 2 !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["ce"] = vngpmap["cl"] = vngpmap["line"] = vngpmap["ma"] = vngpmap["pe"] = vngpmap["pl"] = vngpmap["data"] = vngpmap["te"] = vngpmap["tl"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["ce"] = vngpmap["cl"] = vngpmap["line"] = vngpmap["ma"] = vngpmap["pe"] = vngpmap["pl"] = vngpmap["data"] = vngpmap["te"] = vngpmap["tl"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["r"] = vresmap["reg"] = vresmap["stat"] = true;
  vresmap["ce"] = vresmap["cl"] = vresmap["dw"] = vresmap["icc"] = vresmap["sdr"] = vresmap["x"] = vresmap["y"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["r"] = vresmap["reg"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["r"] = vresmap["reg"] = vresmap["stat"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["ce"] = vresmap["cl"] = vresmap["dw"] = vresmap["icc"] = vresmap["r"] = vresmap["reg"] = vresmap["sdr"] = vresmap["stat"] = vresmap["x"] = vresmap["y"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2));
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "dw")     // ソートしない時のダービン・ワトソン比
      { vresmap[res] = true;  vdw = dwput; }
      else if (res == "dws")    // ソートした時のダービン・ワトソン比
      { vresmap["dw"] = true;  vdw = dwsort; }
      else if (res[0] == 'x' && res.size() > 1) // x=指定値の時のy推定値を出力
      {
        vx = strtodf(res.substr(1));
        if (vx == BLANKf) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["x"] = true;
      }
      else if (res[0] == 'y' && res.size() > 1) // y=指定値の時のx逆推定値を出力
      {
        vy = strtodf(res.substr(1));
        if (vy == BLANKf) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["y"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class data_less                 // ケースデータの比較用関数オブジェクト
{
private:
  int kidx;                     // ソートキー項目のインデックス
public:
  data_less(int kidx0) : kidx(kidx0) // コンストラクタ
  {}
  ~data_less()                  // デストラクタ
  {}
  bool operator()(datac& cdata1, datac& cdata2) const; // ケースデータを比較する
};

bool data_less::operator()(datac& cdata1, datac& cdata2) const
// ケースデータを比較する
{
  int comp = 0;
  double fdata1 = cdata1.fdata(kidx);
  double fdata2 = cdata2.fdata(kidx);
  if (fdata1 < fdata2) comp = -1;
  else if (fdata1 > fdata2) comp = 1;
  return (comp < 0) ? true : false; // 小さいときだけtrueを返す
}

class varclass                  // 変数クラス
{
private:
  pairdbl vrange;               // 最小値と最大値
  statcor<double> vstatc;       // 相関基礎統計量
public:
  varclass() : vrange(DBL_MAX, -DBL_MAX), vstatc()
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int cal(ldata& datalst);      // データを集計を有効例数を返す
  int n() const                 // 例数を返す
  { return vstatc.n(); }
  double mx() const             // Xの平均値を返す
  { return vstatc.mx(); }
  double my() const             // Yの平均値を返す
  { return vstatc.my(); }
  double sxx() const            // Xの平方和散を返す
  { return vstatc.sxx(); }
  double syy() const            // Yの平方和散を返す
  { return vstatc.syy(); }
  double sxy() const            // XとYの積和を返す
  { return vstatc.sxy(); }
  double vxx() const            // Xの不偏分散を返す
  { return vstatc.vxx(); }
  double vyy() const            // Yの不偏分散を返す
  { return vstatc.vyy(); }
  double vxy() const            // XとYの共分散を返す
  { return vstatc.vxy(); }
  double rxy() const            // XとYの相関係数を返す
  { return vstatc.rxy(); }
  double xmin() const           // Xの最小値を返す
  { return vrange.first; }
  double xmax() const           // Xの最大値を返す
  { return vrange.second; }
};

int varclass::cal(ldata& datalst) // データを集計して有効例数を返す
{
  int nc = 0;
  ldata_itr ditr = datalst.begin();
  while (ditr != datalst.end())
  {
    double x = ditr->fdata(0), y = ditr->fdata(1);
    if (vstatc.sigma(x, y) < 1) // 欠損データはリストから削除する
    { ditr = datalst.erase(ditr); }
    else
    {
      if (x < vrange.first) vrange.first = x;
      if (x > vrange.second) vrange.second = x;
      ++nc;  ++ditr;
    }
  }
  return nc;                    // 有効例数を返す
}

class coregclass                // coregメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 分析結果のタイトルとデータ識別情報を出力する
  void put_r(ostream& fout, option& opt, const varclass& var) const;
  // 相関係数を出力する
  void put_icc(ostream& fout, const varclass& var) const;
  // Ebelの級内相関係数を出力する
  void put_reg(ostream& fout, option& opt, const varclass& var) const;
  // 回帰直線を出力する
  void put_esty(ostream& fout, const option& opt, int n, double b0, double b1, double mx, double xy, double sxx, double vr, int dfr, double t) const;
  // yの推定値を出力する
  void put_estx(ostream& fout, const option& opt, int n, double b0, double b1, double mx, double xy, double sxx, double vr, int dfr, double t) const;
  // xの逆推定値を出力する
  void put_regatab(ostream& fout, const varclass& var) const;
  // 回帰直線の分散分析結果を出力する
  void put_dw(ostream& fout, const option& opt, ldata& datalst, const varclass& var) const;
  // ダービン・ワトソン比を出力する
  void plot_ngp(option& opt, idd& id, ldata& datalst, const varclass& var) const;
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpdata(ofstream& fout, ldata& datalst) const;
  // Ngraph用に個々のデータを出力する
  void plot_ngpline(ofstream& fout, const varclass& var) const;
  // Ngraph用に回帰直線を出力する
  void plot_ngpcl(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に信頼限界を出力する
  void plot_ngptl(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に許容限界を出力する
  void plot_ngppl(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に予測限界を出力する
  void plot_clim(ofstream& fout, const string& title, const option& opt, double mx, double my, double b0, double b1, double t, double vc, double vb, double xmin, double xmax) const;
  // 信頼区間のデータを出力する
  void plot_ngpce(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に信頼楕円出力する
  void plot_ngpte(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に許容楕円を出力する
  void plot_ngppe(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に予測楕円を出力する
  void plot_ngpma(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に主軸回帰直線を出力する
  void plot_ngpcelim(ofstream& fout, double mx, double my, double a, double b, double h) const;
  // 楕円のデータを出力する
public:
  coregclass() : fname()        // コンストラクタ
  {}
  ~coregclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み累積集計する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void coregclass::put_resid(ostream& fout, idd& id) const
// 分析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "EXPLANATORY VARIABLE X:" << id.item(0) << "\n";
  fout << "CRITERION VARIABLE   Y:" << id.item(1) << "\n";
}

void coregclass::put_r(ostream& fout, option& opt, const varclass& var) const
// 相関係数を出力する
{
  int n = var.n();
  int dfr = n - 2;
  double r = var.rxy();
  fout << "\nCorrelation Coefficient:r=" << r << "\n";
  fout << "Coefficeint of Determination:r-squared=" << r * r << "\n";
  if (dfr < 1) return;
  if (n > 3)
  {
    double zr = 0.0;
    double ci = calrci(zr, opt.cc(), n, r);
    if (opt.res("cc")) fout << opt.cc() * 100.0 << "% CI of r(Fisher's z-transform):Lower=" << tanh(zr - ci) << "  Upper=" << tanh(zr + ci) << "\n";
    zr *= sqrt((double)n - 3.0);
    ldbl pr = ztop((ldbl)zr);
    fout << "Test of r(Fisher's z-transform):z=" << zr << "  p=" << pr << pind(pr) << "\n";
  }
  if (opt.res("ce") == false) return;
  double vxx = var.vxx(), vyy = var.vyy(), vxy = var.vxy();
  double cf = 1.0 / (double)n, rf = 1.0 + 1.0 / (double)n;
  double ccpr = opt.cc() * 100.0;
  double a = 0.0, b = 0.0, h = 0.0, deg = 180.0 / M_PI;
  calelp(a, b, h, opt.cc(), 2, dfr, vxx * cf, vyy * cf, vxy * cf);
  double ba = tan(h);
  fout << "Long Axis of Confidence Ellipse=Major Axis Regression:y=" << var.my() - ba * var.mx() << showpos << ba << noshowpos << "x\n";
  fout << ccpr << "% Confidence Ellipse:Long Axis=" << a << "  Short Axis=" << b << "\n";
  fout << "  Rotation Angle of Long Axis(degrees)=" << h * deg << "°\n";
  calelp(a, b, h, opt.cc(), 2, dfr, vxx, vyy, vxy);
  fout << ccpr << "% Tolerance Ellipse:Long Axis=" << a << "  Short Axis=" << b << "\n";
  fout << "  Rotation Angle of Long Axis(degrees)=" << h * deg << "°\n";
  calelp(a, b, h, opt.cc(), 2, dfr, vxx * rf, vyy * rf, vxy * rf);
  fout << ccpr << "% Prediction Ellipse:Long Axis=" << a << "  Short Axis=" << b << "\n";
  fout << "  Rotation Angle of Long Axis(degrees)=" << h * deg << "°\n";
}

void coregclass::put_icc(ostream& fout, const varclass& var) const
// Ebelの級内相関係数を出力する
{
  int n = var.n();
  int dfr = n - 2;
  double r = var.rxy();
  double sxx = var.sxx(), syy = var.syy(), sxy = var.sxy();
  double r11 = (sxx + syy > 0.0) ? 2.0 * sxy / (sxx + syy) : 0.0;
  double r22 = (sxx + syy + 2.0 * sxy > 0.0) ? 4.0 * sxy / (sxx + syy + 2.0 * sxy) : 0.0;
  double rv = (sxx + syy > 0.0) ? 2.0 * sqrt(sxx * syy) / (sxx + syy) : 0.0;
  fout << "\nEbel's Intraclass Correlation Coefficient\n";
  fout << "  r11=" << r11 << "  r22=" << r22 << "  Concordance Coefficient of V:rv=" << rv << endl;
}

void coregclass::put_reg(ostream& fout, option& opt, const varclass& var) const
// 回帰直線を出力する
{
  int n = var.n();
  int dfr = n - 2;
  double mx = var.mx(), my = var.my();
  double sxx = var.sxx(), syy = var.syy(), sxy = var.sxy();
  double b1 = (sxx > 0.0) ? sxy / sxx : 0.0;
  double b0 = my - b1 * mx;
  double sr = syy - sxy * b1;
  const double Seps = DBL_EPSILON * fmax(5.0, syy);
  if (sr < Seps) sr = 0.0;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double vb = (sxx > 0.0) ? vr / sxx : 0.0;
  double se0 = (sxx > 0.0) ? sqrt(vr * (1.0 / n + mx * mx / sxx)) : 0.0;
  double seb = (sxx > 0.0) ? sqrt(vr / sxx) : 0.0;
  fout << "\nRegression Line:y=" << b0 << showpos << b1 << noshowpos << "x\n";
  fout << "SE of Intercept=" << se0 << "  SE of Regression Coefficient=" << seb << "\n";
  if (opt.res("sdr")) fout << "Regression Residuals=" << vr << "  SD of Regression Residuals=" << sqrt(vr) << "\n";
  if (dfr < 1) return;
  double cf = 1.0 / (double)n, rf = 1.0 + 1.0 / (double)n;
  double t = (dfr > 0) ? (double)(ptot((ldbl)(1.0 - opt.cc()), dfr)) : 0.0;
  double ccpr = opt.cc() * 100.0;
  if (opt.res("cc"))
  {
    double ci = calmcci(mcfis, opt.cc(), 1, 1, dfr, 1, vb);
    fout << ccpr << "% CI of Regression Coefficient:";
    fout << "Lower=" << b1 - ci << "  Upper=" << b1 + ci << "\n";
  }
  if (opt.res("cl"))
  {
    fout << ccpr << "% Confidence Limit of Regression Line(CI of Mean):\n";
    fout << "  y=" << b0 << showpos << b1 << noshowpos << "x+/-";
    fout << t << "√{" << vr * cf << showpos << vb << noshowpos << "(x" << showpos << -mx << noshowpos << ")^2}\n";
  }
  if (opt.res("x")) put_esty(fout, opt, n, b0, b1, mx, my, sxx, vr, dfr, t);
  if (opt.res("y")) put_estx(fout, opt, n, b0, b1, mx, my, sxx, vr, dfr, t);
  if (opt.res("cl"))
  {
    fout << ccpr << "% Tolerance Limit of Regression Line(CI of Data):\n";
    fout << "  y=" << b0 << showpos << b1 << noshowpos << "x+/-";
    fout << t << "√(" << vr << ")\n";
    fout << ccpr << "% Prediction Limit of Regression Line(CI of Mean and Data):\n";
    fout << "  y=" << b0 << showpos << b1 << noshowpos << "x+/-";
    fout << t << "√{" << vr * rf << showpos << vb << noshowpos << "(x" << showpos << -mx << noshowpos << ")^2}\n";
  }
}

void coregclass::put_esty(ostream& fout, const option& opt, int n, double b0, double b1, double mx, double my, double sxx, double vr, int dfr, double t) const
// yの推定値を出力する
{
  double x = (opt.x() == DBL_MAX) ? mx : opt.x();
  double y = b0 + b1 * x, dx = x - mx;
  double sey = sqrt(vr * (1.0 / (double)n + dx * dx / sxx));
  double ci = t * sey;
  fout << "  Estimated Y when x=" << x << ":Lower=" << y - ci << "  Upper=" << y + ci << endl;
}

void coregclass::put_estx(ostream& fout, const option& opt, int n, double b0, double b1, double mx, double my, double sxx, double vr, int dfr, double t) const
// xの逆推定値を出力する
{
  if (sgn(b1) == 0) return;     // 逆推定できず
  double y = (opt.y() == DBL_MAX) ? my : opt.y();
  double cf = (y - my) / b1;
  double x = mx + cf;
  double g = t / b1 * t / b1 * vr / sxx;
  if (g >= 1.0) g = 0.0;        // g>1だと計算できないため
  double gr = 1.0 - g;
  double v = vr * (gr / (double)n + cf * cf / sxx);
  double ci = t / b1 * sqrt(v);
  double xl = (x - g * mx - ci) / gr, xu = (x - g * mx + ci) / gr;
  fout << "  Inverse Estimated X when y=" << y << ":Lower=" << xl << "  Upper=" << xu << endl;
}

void coregclass::put_regatab(ostream& fout, const varclass& var) const
// 回帰直線の分散分析結果を出力する
{
  int n = var.n();
  int dfr = n - 2;
  double mx = var.mx(), my = var.my();
  double sxx = var.sxx(), syy = var.syy(), sxy = var.sxy();
  double b1 = (sxx > 0.0) ? sxy / sxx : 0.0;
  double b0 = my - b1 * mx;
  double sr = syy - sxy * b1;
  const double Seps = DBL_EPSILON * fmax(5.0, syy);
  if (sr < Seps) sr = 0.0;
  fout << "\n<ANOVA OF REGRESSION LINE(=TEST OF REGRESSION COEFFICIENT)>\n";
  putatab(fout, "REGRESSION", sxy * b1, 1, sr, dfr, syy, n - 1);
}

void coregclass::put_dw(ostream& fout, const option& opt, ldata& datalst, const varclass& var) const
// ダービン・ワトソン比を出力する
{
  int n = var.n();
  if (n < 2) return;            // 計算できず
  double mx = var.mx(), my = var.my();
  double sxx = var.sxx(), syy = var.syy(), sxy = var.sxy();
  double b1 = (sxx > 0.0) ? sxy / sxx : 0.0;
  double b0 = my - b1 * mx;
  double sr = syy - sxy * b1;
  const double Seps = DBL_EPSILON * fmax(5.0, syy);
  if (sr < Seps) return;                              // 計算できず
  if (opt.dw() == dwsort) datalst.sort(data_less(0)); // x順にソートする
  ldata_itr ditr = datalst.begin();
  double e1 = ditr->fdata(1) - (b0 + b1 * ditr->fdata(0));
  double sume2 = e1 * e1, sumed = 0.0, epre = e1;
  ++ditr;
  while (ditr != datalst.end())
  {
    double ei = ditr->fdata(1) - (b0 + b1 * ditr->fdata(0));
    double ed = ei - epre;
    sume2 += ei * ei;  sumed += ed * ed;
    epre = ei;  ++ditr;
  }
  double dw = sumed / sume2;
  fout << "\nDurbin-Watson Ratio";
  if (opt.dw() == dwsort) fout << "(sort by x)";
  fout << ":DW=" << dw << "\n";
}

void coregclass::plot_ngp(option& opt, idd& id, ldata& datalst, const varclass& var) const
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("data"))          // データ
  {
    fout << "-s1 -dmark,00\n";
    plot_ngpid(fout, id);
    plot_ngpdata(fout, datalst);
    delml = true;
  }
  if (opt.ngp("line"))          // 回帰直線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_ngpline(fout, var);
  }
  if (opt.ngp("cl"))            // 信頼限界
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_ngpcl(fout, opt, var);
  }
  if (opt.ngp("tl"))            // 許容限界
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_ngptl(fout, opt, var);
  }
  if (opt.ngp("pl"))            // 予測限界
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_ngppl(fout, opt, var);
  }
  if (opt.ngp("ce"))            // 信頼楕円
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dpolygon\n";
    plot_ngpid(fout, id);
    plot_ngpce(fout, opt, var);
  }
  if (opt.ngp("te"))            // 許容楕円
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dpolygon\n";
    plot_ngpid(fout, id);
    plot_ngpte(fout, opt, var);
  }
  if (opt.ngp("pe"))            // 予測楕円
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dpolygon\n";
    plot_ngpid(fout, id);
    plot_ngppe(fout, opt, var);
  }
  if (opt.ngp("ma"))            // 主軸回帰直線
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline\n";
    plot_ngpid(fout, id);
    plot_ngpma(fout, opt, var);
  }
}

void coregclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%EXPLANATORY VARIABLE X:" << id.item(0) << "\n";
  fout << "%CRITERION VARIABLE   Y:" << id.item(1) << "\n";
}

void coregclass::plot_ngpdata(ofstream& fout, ldata& datalst) const
// Ngraph用に個々のデータを出力する
{
  fout << "%<RAW DATA>\n";
  for (auto& cdata : datalst) fout << cdata.fdata(0) << " " << cdata.fdata(1) << "\n";
}

void coregclass::plot_ngpline(ofstream& fout, const varclass& var) const
// Ngraph用に回帰直線を出力する
{
  fout << "%<REGRESSION LINE>\n";
  int n = var.n();
  if (n < 1) return;
  double mx = var.mx(), my = var.my();
  double sxx = var.sxx(), syy = var.syy(), sxy = var.sxy();
  double b = (sxx > 0.0) ? sxy / sxx : 0.0;
  double a = my - b * mx;
  double xmin = var.xmin(), xmax = var.xmax();
  double xrange = xmax - xmin;
  double x = xmin - xrange * 0.5;
  fout << x << " " << a + b * x << "\n";
  x = xmax + xrange * 0.5; 
 fout << x << " " << a + b * x << endl;
}

void coregclass::plot_ngpcl(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に信頼限界を出力する
{
  int n = var.n();
  if (n < 1) return;
  double mx = var.mx(), my = var.my();
  double sxx = var.sxx(), syy = var.syy(), sxy = var.sxy();
  int dfr = n - 2;
  double b1 = (sxx > 0.0) ? sxy / sxx : 0.0;
  double b0 = my - b1 * mx;
  double sr = syy - sxy * b1;
  const double Seps = DBL_EPSILON * fmax(5.0, syy);
  if (sr < Seps) sr = 0.0;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double vb = (sxx > 0.0) ? vr / sxx : 0.0;
  double cf = 1.0 / (double)n;
  double t = ptot(1.0 - opt.cc(), dfr);
  plot_clim(fout, "CONFIDENCE LIMIT(CI OF MEAN)", opt, mx, my, b0, b1, t, vr * cf, vb, var.xmin(), var.xmax());
}

void coregclass::plot_ngptl(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に許容限界を出力する
{
  int n = var.n();
  if (n < 1) return;
  double mx = var.mx(), my = var.my();
  double sxx = var.sxx(), syy = var.syy(), sxy = var.sxy();
  int dfr = n - 2;
  double b1 = (sxx > 0.0) ? sxy / sxx : 0.0;
  double b0 = my - b1 * mx;
  double sr = syy - sxy * b1;
  const double Seps = DBL_EPSILON * fmax(5.0, syy);
  if (sr < Seps) sr = 0.0;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double vb = (sxx > 0.0) ? vr / sxx : 0.0;
  double t = ptot(1.0 - opt.cc(), dfr);
  plot_clim(fout, "TOLETANCE LIMIT(CI OF DATA)", opt, mx, my, b0, b1, t, vr, 0.0, var.xmin(), var.xmax());
}

void coregclass::plot_ngppl(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に予測限界を出力する
{
  int n = var.n();
  if (n < 1) return;
  double mx = var.mx(), my = var.my();
  double sxx = var.sxx(), syy = var.syy(), sxy = var.sxy();
  int dfr = n - 2;
  double b1 = (sxx > 0.0) ? sxy / sxx : 0.0;
  double b0 = my - b1 * mx;
  double sr = syy - sxy * b1;
  const double Seps = DBL_EPSILON * fmax(5.0, syy);
  if (sr < Seps) sr = 0.0;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double vb = (sxx > 0.0) ? vr / sxx : 0.0;
  double rf = 1.0 + 1.0 / (double)n;
  double t = ptot(1.0 - opt.cc(), dfr);
  plot_clim(fout, "PREDICTION LIMIT(CI OF MEAN AND DATA)", opt, mx, my, b0, b1, t, vr * rf, vb, var.xmin(), var.xmax());
}

void coregclass::plot_clim(ofstream& fout, const string& title, const option& opt, double mx, double my, double b0, double b1, double t, double vc, double vb, double xmin, double xmax) const
// 信頼区間のデータを出力する
{
  double ccpr = opt.cc() * 100.0;
  fout << "%<UPPER OF " << ccpr << "% " << title << " OF REGRESSION LINE>\n";
  double xrange = xmax - xmin;
  double xstart = xmin - xrange * 0.5, xend = xmax + xrange * 0.5;
  double dx = (xend - xstart) / (double)RDdef;
  double x = xstart, dm = 0.0, y = 0.0;
  if (vb <= 0.0)                // 許容限界
  {
    y = b0 + b1 * xstart + t * sqrt(vc);
    fout << xstart << " " << y << "\n";
    y = b0 + b1 * xend + t * sqrt(vc);
    fout << xend << " " << y << "\n";
  }
  else
  {
    for (uint i = 0; i <= RDdef; ++i, x += dx)
    {
      dm = x - mx;
      y = b0 + b1 * x + t * sqrt(vc + vb * dm * dm);
      fout << x << " " << y << "\n";
    }
  }
  fout << "= =\n";
  fout << "%<LOWER OF " << ccpr << "% " << title << " OF REGRESSION LINE>\n";
  if (vb <= 0.0)                // 許容限界
  {
    y = b0 + b1 * xstart - t * sqrt(vc);
    fout << xstart << " " << y << "\n";
    y = b0 + b1 * xend - t * sqrt(vc);
    fout << xend << " " << y << "\n";
  }
  else
  {
    x = xstart;
    for (uint i = 0; i <= RDdef; ++i, x += dx)
    {
      dm = x - mx;
      y = b0 + b1 * x - t * sqrt(vc + vb * dm * dm);
      fout << x << " " << y << "\n";
    }
  }
}

void coregclass::plot_ngpce(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に信頼楕円を出力する
{
  int n = var.n();
  if (n < 1) return;
  double mx = var.mx(), my = var.my();
  double vxx = var.vxx(), vyy = var.vyy(), vxy = var.vxy();
  int dfr = n - 2;
  double cf = 1.0 / (double)n;
  double ccpr = opt.cc() * 100.0;
  fout << "%<" << ccpr << "% CONFIDENCE ELLIPSE>\n";
  double a = 0.0, b = 0.0, h = 0.0;
  calelp(a, b, h, opt.cc(), 2, dfr, vxx * cf, vyy * cf, vxy * cf);
  plot_ngpcelim(fout, mx, my, a, b, h);
}

void coregclass::plot_ngpte(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に許容楕円を出力する
{
  int n = var.n();
  if (n < 1) return;
  double mx = var.mx(), my = var.my();
  double vxx = var.vxx(), vyy = var.vyy(), vxy = var.vxy();
  int dfr = n - 2;
  double ccpr = opt.cc() * 100.0;
  fout << "%<" << ccpr << "% TOLERANCE ELLIPSE>\n";
  double a = 0.0, b = 0.0, h = 0.0;
  calelp(a, b, h, opt.cc(), 2, dfr, vxx, vyy, vxy);
  plot_ngpcelim(fout, mx, my, a, b, h);
}

void coregclass::plot_ngppe(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に予測楕円を出力する
{
  int n = var.n();
  if (n < 1) return;
  double mx = var.mx(), my = var.my();
  double vxx = var.vxx(), vyy = var.vyy(), vxy = var.vxy();
  int dfr = n - 2;
  double rf = 1.0 + 1.0 / (double)n;
  double ccpr = opt.cc() * 100.0;
  fout << "%<" << ccpr << "% PREDICTION ELLIPSE>\n";
  double a = 0.0, b = 0.0, h = 0.0;
  calelp(a, b, h, opt.cc(), 2, dfr, vxx * rf, vyy * rf, vxy * rf);
  plot_ngpcelim(fout, mx, my, a, b, h);
}

void coregclass::plot_ngpma(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に主軸回帰直線を出力する
{
  int n = var.n();
  if (n < 1) return;
  double mx = var.mx(), my = var.my();
  double vxx = var.vxx(), vyy = var.vyy(), vxy = var.vxy();
  int dfr = n - 2;
  double cf = 1.0 / (double)n;
  double ccpr = opt.cc() * 100.0;
  double a = 0.0, b = 0.0, h = 0.0;
  calelp(a, b, h, opt.cc(), 2, dfr, vxx * cf, vyy * cf, vxy * cf);
  fout << "%<LONG AXIS OF CONFIDENCE ELLIPSE=MAJOR AXIS REGRESSION LINE>\n";
  double xmin = var.xmin(), xmax = var.xmax();
  double xrange = (xmax - xmin) * 2.0;
  double xstart = xmin - xrange * 0.5, xend = xmax + xrange * 0.5;
  double ba = tan(h);
  double aa = var.my() - ba * var.mx();
  double y = aa + ba * xstart;
  fout << xstart << " " << y << "\n";
  y = aa + ba * xend;
  fout << xend << " " << y << endl;
}

void coregclass::plot_ngpcelim(ofstream& fout, double mx, double my, double a, double b, double h) const
// 楕円のデータを出力する
{
  constexpr uint RD = 60;       // 楕円の分解数
  double dt = 2.0 * M_PI / (double)RD;
  double x0 = 0.0, y0 = 0.0, x = 0.0, y = 0.0, t = 0.0;
  for (uint i = 0; i<= RD; ++i, t += dt)
  {
    x0 = a * cos(t);  y0 = b * sin(t);
    x = mx + x0 * cos(h) - y0 * sin(h);  y = my + x0 * sin(h) + y0 * cos(h);
    fout << x << " " << y << "\n";
  }
}

int coregclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み累積集計する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;                    // 読み込んだ例数を返す
}

void coregclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var;                 // 計算用変数クラス
  if (var.cal(datalst) < 1) return; // 有効なデータがない
  if (opt.res("res"))           // 計算結果出力
  {
    put_resid(cout, id);
    if (opt.res("stat"))
    {
      cout << "\n<BASIC STATISTICS OF VARIABLES>\n";
      constexpr int LENline = 69; // 区切り線の長さ
      putline(cout, LENline);
      cout << "X:";  putmsd(cout, var.n(), var.mx(), var.vxx());
      cout << "Y:";  putmsd(cout, var.n(), var.my(), var.vyy());
      putline(cout, LENline);
    }
    if (opt.res("r")) put_r(cout, opt, var);
    if (opt.res("icc")) put_icc(cout, var);
    if (opt.res("reg")) put_reg(cout, opt, var);
    if (opt.res("dw")) put_dw(cout, opt, datalst,var);
    if (opt.res("r") || opt.res("reg")) put_regatab(cout, var);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(opt, id, datalst, var); // Ngraph用データ出力
}

int main(int argc, char* argv[]) // メインルーチン
{
  coregclass coreg;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) coreg.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { coreg.get(argv[i], opt, id, datalst);  fok = false; }
    }
    coreg.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
