// crosove.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-24 16:32:35 tac>
//
// クロスオーバー・デザインによる分散分析。
// 主として生物学的同等性試験解析用。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W crosove crosove.C -I/usr/local/include
//              -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "crosove  Version 7.42e       Copyleft(Vo\\oV) 2023 Sugimoto Norio",
  "USAGE:crosove [<OPTION> ...] [<DATA FILE> ...]",
  "  The ANOVA for cross-over design with the first item as the group",
  "  and the subsequent items as the period.",
  "<OPTION>",
  "  -al[=<ALPHA>]:specify the significace level(0<ALPHA<1, -al:-al=0.05)",
  "  -dl[=<DIFFERENCE>]:specify the scientific significant difference",
  "    value after logarithmic transformation(-dl:-dl=log(1.25))",
  "  -dp[=<DIFFERENCE>]:specify the scientific significant difference",
  "    ratio to the mean of drag-1(-dp:-dp=0.2)",
  "  -dg=<DRAG-1>,<DRAG-2>:specify the drag name",
  "  -f=<FILE>:reads options from the specified file",
  "  -g=<CODE-1>,<CODE-2>:specify the drug code",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -log={10|e}:select the logarithm type(e:natural(default) 10:common)",
  "  -po=<1-BETA>:specify the power(0<1-BETA<1,default:0.8)",
  "  -res=[-][{def|all|at|be|cc[<C-COEF>]|pa|stat|tt|tt2|ttw|ut|utc|uto}]",
  "       [,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    at:ANOVA  be:settings for BE study  cc[<C-COEF>]:confidence intervals",
  "    pa:power analysis  stat:basic statistics  tt:auto t-test",
  "    tt2:2-sample t test  ttw:Welch test  ut:auto U-test",
  "    utc:U-test with correction  uto:U-test with no correction",
  "    default:auto judgment according to data"
};

const string Title = "=== ANOVA for Cross-Over Design ===";
constexpr double DLdef = 1.25;  // 省略時検出差(対数変換前の値)
constexpr double DPdef = 0.2;   // 省略時検出差(％)

using vecstatv = vector<statv<double>>; // 基礎統計量ベクトルの短名定義

enum sumtype                    // 度数集計の種類
{
  sumnon,                       // 集計なし
  sumorder,                     // 順序効果用集計
  sumperiod,                    // 時期効果用集計
  sumtreat,                     // 処置効果用集計
  sumetc                        // その他
};

enum logtype                    // 対数の種類
{
  logtnon,                      // 指定なし
  logte,                        // 自然対数
  logt10,                       // 常用対数
  logtetc                       // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double val;                   // 有意水準
  double vcc;                   // 信頼係数
  double vdl;                   // 検出差
  double vdp;                   // 検出差(比率)
  logtype vlogt;                // 対数の種類フラッグ
  double vpo;                   // 検出力
  vecstr vdgvec;                // 処置名ベクトル
  vecstr vgvec;                 // 群コードベクトル
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  testtype vtt;                 // t検定の種類
  testtype vut;                 // U検定の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double al() const             // 有意水準を返す
  { return val; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double dl() const             // 検出差を返す
  { return vdl; }
  double dp() const             // 検出差(%)を返す
  { return vdp; }
  logtype logt() const          // 対数の種類を返す
  { return vlogt; }
  double po() const             // 検出力を返す
  { return vpo; }
  const string& dg(int idx) const // 処置名を返す
  { return vdgvec[idx]; }
  const string& g(int idx) const // 群コードを返す
  { return vgvec[idx]; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  testtype tt() const           // t検定の種類を返す
  { return vtt; }
  testtype ut() const           // U検定の種類を返す
  { return vut; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int g);             // オプションのチェックとセット
  int set_res();                // -resオプションのセット
  int search_g(const string& gcode) const; // 群コードを検索する
};

option::option(const char* c_pname) : vpath(), vpname(), val(ALdef), vcc(CCdef), vdl(DLdef), vdp(DPdef), vlogt(logte), vpo(POdef), vdgvec(), vreslst(), vresmap(), vtt(testaut), vut(testaut)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "al")           // 有意水準指定
  {
    val = opt[1].empty() ? ALdef : strtodf(opt[1]);
    if (val <= 0.0 || 1.0 <= val) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "dl")      // 検出差の指定
  {
    if (opt[1].empty()) vdl = (vlogt == logt10) ? log10(DLdef) : log(DLdef);
    else vdl = strtodf(opt[1]);
    if (vdl < 0.0 || FLT_MAX < vdl) err = badopt(vpname, optstr, USAGE, 0.0, FLT_MAX);
  }
  else if (opt[0] == "dp")      // 検出差(%)の指定
  {
    vdp = opt[1].empty() ? DPdef : strtodf(opt[1]);
    if (vdp < 0.0 || FLT_MAX < vdp) err = badopt(vpname, optstr, USAGE, 0.0, FLT_MAX);
  }
  else if (opt[0] == "dg")      // 処置名の指定
  {
    if (strsplit(vdgvec, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "g")       // 群コードの指定
  {
    if (strsplit(vgvec, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "log")     // 対数の種類の指定
  {
    if (opt[1].empty() || opt[1] == "e") vlogt = logte;
    else if (opt[1] == "10") vlogt = logt10;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "po")      // 検出力の指定
  {
    vpo = opt[1].empty() ? POdef : strtodf(opt[1]);
    if (vpo <= 0.0 || 1.0 <= vpo) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(int g)        // オプションのチェック
{
  if (g != 2) return disperr(vpname, "Number of groups must be 2 !");
  if ((int)(vdgvec.size()) < g) // 処置名が足らない場合
  {
    int i = vdgvec.size();
    vdgvec.resize(g);
    while (i < g)
    {
      ostringstream ostr;
      ostr << "DG" << i + 1;
      vdgvec[i] = ostr.str();
      ++i;
    }
  }
  if (g < (int)(vdgvec.size())) return disperr(vpname, "Wrong number of analysis items !");
  if ((int)(vgvec.size()) < g)  // 群コードが足らない場合
  {
    int i = vgvec.size();
    vgvec.resize(g);
    while (i < g)
    {
      ostringstream ostr;
      ostr << i + 1;
      vgvec[i] = ostr.str();
      ++i;
    }
  }
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["stat"] = true;
  vresmap["be"] = vresmap["pa"] = vresmap["tt"] = vresmap["ut"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      { vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["stat"] = true; }
      else if (res == "all")
      {
        vresmap["at"] = vresmap["be"] = vresmap["cc"] = vresmap["pa"] = vresmap["stat"] = vresmap["tt"] = vresmap["ut"] = true;
        val = 0.1;  vcc = 0.9;  vdp = 0.2;  vpo = 0.8;
        vdl = (vlogt == logt10) ? log10(DLdef) : log(DLdef);
      }
      else if (res == "be")     // 生物学的同等性試験用設定
      {
        val = 0.1;  vcc = 0.9;  vdp = 0.2;  vpo = 0.8;
        vdl = (vlogt == logt10) ? log10(DLdef) : log(DLdef);
        vresmap["pa"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "tt2")
      { vresmap["tt"] = true;  vtt = testorg; } // 2標本t検定
      else if (res == "ttw")
      { vresmap["tt"] = true;  vtt = testcor; } // Welchの検定
      else if (res == "utc")
      { vresmap["ut"] = true;  vut = testcor; } // 連続修正有り
      else if (res == "uto")
      { vresmap["ut"] = true;  vut = testorg; } // 連続修正無し
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::search_g(const string& gcode) const // 群コードを検索する
{
  bool hit = false;
  int g = vgvec.size(), idx = 0;
  while (hit == false && idx < g)
  {
    if (gcode == vgvec[idx]) hit = true;
    else ++idx;
  }
  if (hit) return idx;          // 発見
  return -1;
}

class varclass                  // 変数クラス
{
private:
  vector<vecstatv> vgsvec;      // 群別統計量2重ベクトル
  set<double> vsumset;          // クロス前後の和データセット
  set<double> vdifset;          // クロス前後の差データセット
  vecstatv vtgstatvec;          // 群統計量ベクトル
  vecstatv vtpstatvec;          // 時期統計量ベクトル
  vecstatv vtdstatvec;          // 薬剤統計量ベクトル
  statv<double> vtstat;         // 全体統計量
  double vsg;                   // 群(順序)平方和
  int vdfg;                     // 群(順序)自由度
  double vsbr;                  // 群(順序)残差平方和
  int vdfbr;                    // 群(順序)残差自由度
  double vss;                   // 個体平方和
  int vdfs;                     // 個体自由度
  double vsp;                   // 時期平方和
  double vsd;                   // 処置平方和
  double vsr;                   // 残差平方和
  int vdfr;                     // 残差自由度
  int vdft;                     // 全体自由度
  int cal_crosov();        // クロスオーバー用各種パラメータを計算する
public:
  varclass(size_t g0) : vgsvec(g0), vsumset(), vdifset(), vtgstatvec(g0), vtpstatvec(g0), vtdstatvec(g0), vtstat(), vsg(0.0), vdfg(0), vsbr(0.0), vdfbr(0), vss(0.0), vdfs(0), vsp(0.0), vsd(0.0), vsr(0.0), vdfr(0), vdft(0)
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int cal(const option& opt, ldata& datalst); // データを集計して有効例数を返す
  size_t g() const              // 群数を返す
  { return vgsvec.size(); }
  int tgn(size_t idx) const     // 群別例数を返す
  { return (idx < vtgstatvec.size()) ? vtgstatvec[idx].n() : 0; }
  double tgmean(size_t idx) const // 群別平均値を返す
  { return (idx < vtgstatvec.size()) ? vtgstatvec[idx].mean() : 0.0; }
  double tgvr(size_t idx) const // 群別不偏分散を返す
  { return (idx < vtgstatvec.size()) ? vtgstatvec[idx].vr() : 0.0; }
  int tpn(size_t idx) const     // 時期別例数を返す
  { return (idx < vtpstatvec.size()) ? vtpstatvec[idx].n() : 0; }
  double tpmean(size_t idx) const // 時期別平均値を返す
  { return (idx < vtpstatvec.size()) ? vtpstatvec[idx].mean() : 0.0; }
  double tpvr(size_t idx) const // 時期別不偏分散を返す
  { return (idx < vtpstatvec.size()) ? vtpstatvec[idx].vr() : 0.0; }
  int tdn(size_t idx) const     // 処置別例数を返す
  { return (idx < vtdstatvec.size()) ? vtdstatvec[idx].n() : 0; }
  double tdmean(size_t idx) const // 処置別平均値を返す
  { return (idx < vtdstatvec.size()) ? vtdstatvec[idx].mean() : 0.0; }
  double tdvr(size_t idx) const // 処置別不偏分散を返す
  { return (idx < vtdstatvec.size()) ? vtdstatvec[idx].vr() : 0.0; }
  int n(size_t iidx, size_t jidx) const // 群別時期別例数を返す
  { return (iidx < vgsvec.size() && jidx < vgsvec.size()) ? (vgsvec[iidx])[jidx].n() : 0; }
  double mean(size_t iidx, size_t jidx) const // 群別時期別平均値を返す
  { return (iidx < vgsvec.size() && jidx < vgsvec.size()) ? (vgsvec[iidx])[jidx].mean() : 0.0; }
  double vr(size_t iidx, size_t jidx) const // 群別時期別不偏分散を返す
  { return (iidx < vgsvec.size() && jidx < vgsvec.size()) ? (vgsvec[iidx])[jidx].vr() : 0.0; }
  int sn(size_t idx) const      // 群別クロス和例数を返す
  { return (idx < vgsvec.size()) ? (vgsvec[idx])[2].n() : 0; }
  double smean(size_t idx) const // 群別クロス和平均値を返す
  { return (idx < vgsvec.size()) ? (vgsvec[idx])[2].mean() : 0.0; }
  double sss(size_t idx) const // 群別クロス和平方和を返す
  { return (idx < vgsvec.size()) ? (vgsvec[idx])[2].ss() : 0.0; }
  double svr(size_t idx) const // 群別クロス和不偏分散を返す
  { return (idx < vgsvec.size()) ? (vgsvec[idx])[2].vr() : 0.0; }
  int dn(size_t idx) const      // 群別クロス差例数を返す
  { return (idx < vgsvec.size()) ? (vgsvec[idx])[3].n() : 0; }
  double dmean(size_t idx) const // 群別クロス差平均値を返す
  { return (idx < vgsvec.size()) ? (vgsvec[idx])[3].mean() : 0.0; }
  double dss(size_t idx) const // 群別クロス差平方和を返す
  { return (idx < vgsvec.size()) ? (vgsvec[idx])[3].ss() : 0.0; }
  double dvr(size_t idx) const // 群別クロス差不偏分散を返す
  { return (idx < vgsvec.size()) ? (vgsvec[idx])[3].vr() : 0.0; }
  int tn() const                // 全体の例数を返す
  { return vtstat.n(); }
  double tmean() const          // 全体の平均値を返す
  { return vtstat.mean(); }
  double tvr() const            // 全体の不偏分散を返す
  { return vtstat.vr(); }
  double sg() const             // 群(順序)平方和を返す
  { return vsg; }
  int dfg() const               // 群(順序)自由度を返す
  { return vdfg; }
  double sbr() const            // 群(順序)残差平方和を返す
  { return vsbr; }
  int dfbr() const              // 群(順序)残差自由度を返す
  { return vdfbr; }
  double ss() const             // 個体平方和を返す
  { return vss; }
  int dfs() const               // 個体自由度を返す
  { return vdfs; }
  double sp() const             // 時期平方和を返す
  { return vsp; }
  double sd() const             // 処置平方和を返す
  { return vsd; }
  double sr() const             // 残差平方和を返す
  { return vsr; }
  int dfr() const               // 残差自由度を返す
  { return vdfr; }
  double vr() const             // 残差分散を返す
  { return (vdfr > 0) ? vsr / vdfr : 0.0; }
  int dft() const               // 全体自由度を返す
  { return vdft; }
  double st() const             // 全体の平方和を返す
  { return vtstat.ss(); }
  const set<double>& sumset() const // クロス前後の和データセットを返す
  { return vsumset; }
  const set<double>& difset() const // クロス前後の差データセットを返す
  { return vdifset; }
};

int varclass::cal(const option& opt, ldata& datalst)
// データを集計して有効例数を返す
{
  size_t g = vgsvec.size();
  for (size_t i = 0; i < g; ++i) vgsvec[i].resize(g + 2);
  vecdbl fdatavec(g);
  int nc = 0;
  vss = 0.0;                    // 個体平方和
  for (auto& cdata : datalst)
  {
    int idx = opt.search_g(cdata.sdata(0));
    if (idx >= 0)               // 群コード該当
    {
      bool blank = false;
      size_t j = 0;
      while (blank == false && j < g)
      {
        fdatavec[j] = cdata.fdata(j + 1);
        if (fdatavec[j] == BLANKf) blank = true;
        else ++j;
      }
      if (blank == false)
      {
        double sum = fdatavec[0] + fdatavec[1]; // クロス前後の和
        double dif = fdatavec[1] - fdatavec[0]; // クロス前後の差
        (vgsvec[idx])[0].sigma(fdatavec[0]);
        (vgsvec[idx])[1].sigma(fdatavec[1]);
        (vgsvec[idx])[2].sigma(sum);
        (vgsvec[idx])[3].sigma(-dif); // 符号を逆転した前後差を累積する
        vsumset.insert(sum);          // 新規データだけ挿入する
        vdifset.insert(dif);
        if (idx == 1) vdifset.insert(-dif); // 群2は符合を逆転した差も挿入する
        vss += sum * sum / 2.0; // 個体平方和用の値を累積する
        ++nc;
      }
    }
  }
  return cal_crosov();          // 全体の例数を返す
}

int varclass::cal_crosov() // クロスオーバー用各種パラメータを計算する
{
  size_t g = vgsvec.size();
  for (size_t i = 0; i < g; ++i)
  {
    for (size_t j = 0; j < g; ++j)
    {
      vtgstatvec[i] += (vgsvec[i])[j]; // 群データ累積
      vtpstatvec[j] += (vgsvec[i])[j]; // 時期データ累積
      size_t k = i + j;
      if (k >= g) k -= g;
      vtdstatvec[k] += (vgsvec[i])[j]; // 薬剤データ累積
    }
  }
  if (vtgstatvec[0].n() < (int)g) return disperr("calcros", "not enough data !");
  statv<double> gstat;
  statv<double> dstat;
  for (size_t i = 0; i < g; ++i)
  {
    vsg += (vgsvec[i])[2].mean() * (vgsvec[i])[2].mean() * (vgsvec[i])[2].n();
    vsd += (vgsvec[i])[3].mean() * (vgsvec[i])[3].mean() * (vgsvec[i])[3].n();
    gstat += (vgsvec[i])[2];  dstat += (vgsvec[i])[3];
    vtstat += vtgstatvec[i];
  }
  double cf = vtstat.mean() * vtstat.mean() * vtstat.n();
  vss -= cf;
  cf = gstat.mean() * gstat.mean() * gstat.n();
  vsg = (vsg  - cf) * 0.5;  vsp = vsd;
  cf = dstat.mean() * dstat.mean() * dstat.n();
  vsd = (vsd - cf) * 0.5;
  double mt = ((vgsvec[0])[3].sum() - (vgsvec[1])[3].sum()) / dstat.n();
  cf = mt * mt * dstat.n();
  vsp = (vsp - cf) * 0.5;
  vsbr = vss - vsg;
  vsr = ((vgsvec[0])[3].ss() + (vgsvec[1])[3].ss()) * 0.5;
  const double Seps = DBL_EPSILON * fmax(5.0, vss);
  if (vss < Seps) vss = 0.0;
  if (vsg < Seps) vsg = 0.0;
  if (vsp < Seps) vsp = 0.0;
  if (vsd < Seps) vsd = 0.0;
  if (vsr < Seps) vsr = 0.0;
  vdft = vtstat.n() - 1;  vdfg = g - 1;  vdfs = gstat.n() - 1;
  vdfbr = vdfs - vdfg;  vdfr = vdft - vdfs - vdfg * 2;
  return vtstat.n();            // 全体の例数を返す
}

class crosovclass               // crosovメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout); // 分析結果のタイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, const option& opt, const varclass& var) const;
  // 基礎統計量を出力する
  void put_coatab(ostream& fout, const varclass& var) const;
  // クロスオーバー用分散分析表を出力する
  void put_dgest(ostream& fout, option& opt, const varclass& var);
  // 処置に関する検定特性を出力する
  void put_orderest(ostream& fout, const option& opt, const varclass& var) const;
  // 順序効果の信頼区間を出力する
  void put_dgd(ostream& fout, const option& opt, int iidx, int ni, double mdi, double dij, double sed, double ci, double sedm, double dmin) const;
  // 処置差とその信頼区間と検出差を出力する
  void put_logdn(ostream& fout, const option& opt, double ta, double za, double zb, int ni, double dij, double sed, int dfr, double vr) const;
  // 対数検出差に関する検出力と必要例数を出力する
  void put_prodn(ostream& fout, const option& opt, double ta, double za, double zb, int ni, double mdi, double dij, double sed, int dfr, double vr) const;
  // 割合検出差に関する検出力と必要例数を出力する
  int cal_nci(const option& opt, double& ci, ldbl& po0, double dd, double d, double vr) const;
  // 信頼区間による必要例数を計算して返す
  void put_ttest(ostream& fout, const option& opt, const varclass& var) const;
  // クロスオーバー用t検定結果を出力する
  void put_tstat(ostream& fout, const option& opt, string& tstr, int n1, double m1, double ss1, double v1, int n2, double m2, double ss2, double v2) const;
  // 2標本t検定の基礎統計量と検定結果を出力する
  void put_mdci(ostream& fout, const option& opt, double md, int df, double vmd) const;
  // 平均値の差の信頼区間を出力する
  void put_utest(ostream& fout, option& opt, const varclass& var);
  // クロスオーバー用U検定結果を出力する
  void put_umrdci(ostream& fout, const option& opt, ftable& ftab, double u, double vu) const;
  // 順位平均値の差の信頼区間とU値の信頼区間を出力する
  void sum_ftab(option& opt, ftable& ftab, sumtype stype, const varclass& var);
  // クロスオーバー用U検定用用にクロス前後の和／差／クロス差を度数集計する
  int search_data(const set<double>& targset, double fdata) const;
  // 数字データをサーチして番号を返す
public:
  crosovclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~crosovclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み保存・累積する
  void put(option& opt);                     // 集計結果を出力する
};

void crosovclass::put_resid(ostream& fout)
// 分析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM   :" << id.item(0) << "\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    fout << "PERIOD ITEM" << setw(2) << i;
    fout << ":" << id.item(i) << "\n";
  }
}

void crosovclass::put_stat(ostream& fout, const option& opt, const varclass& var) const
// 基礎統計量を出力する
{
  constexpr int LENline = 75;   // 区切り線の長さ
  constexpr int LENhed = 9;     // ヘッダの長さ
  size_t g = var.g();
  fout << "\nGROUP    :BASIC STATISTICS OF ALL PERIOD BY GROUP\n";
  putline(fout, LENline);
  for (size_t i = 0; i < g; ++i)
  {
    fout << left << setw(LENhed) << opt.g(i) << right << ":";
    putmsd(fout, var.tgn(i), var.tgmean(i), var.tgvr(i));
  }
  putline(fout, LENline);
  fout << "\nPERIOD   :BASIC STATISTICS OF ALL GROUP BY PERIOD\n";
  putline(fout, LENline);
  for (size_t i = 0; i < g; ++i)
  {
    fout << left << setw(9) << i + 1 << right << ":";
    putmsd(fout, var.tpn(i), var.tpmean(i), var.tpvr(i));
  }
  putline(fout, LENline);
  fout << "\nDRAG     :BASIC STATISTICS OF ALL PERIOD BY DRAG\n";
  putline(fout, LENline);
  for (size_t i = 0; i < g; ++i)
  {
    int npad = opt.dg(i).size() - strclen(opt.dg(i), CHARwide);
    fout << left << setw(LENhed + npad) << opt.dg(i) << right << ":";
    putmsd(fout, var.tdn(i), var.tdmean(i), var.tdvr(i));
  }
  putline(fout, LENline);
  fout << "\nGROUP-PER:BASIC STATISTICS BY PERIOD AND DRAG\n";
  putline(fout, LENline);
  for (size_t i = 0; i < g; ++i)
  {
    for (size_t j = 0; j < g; ++j)
    {
      int npad = opt.g(i).size() - strclen(opt.g(i), CHARwide);
      fout << left << setw(LENhed - 4 + npad) << opt.g(i) << right;
      fout << setw(4) << j + 1 << ":";
      putmsd(fout, var.n(i, j), var.mean(i, j), var.vr(i, j));
    }
  }
  putline(fout, LENline);
  fout << "TOTAL    :";
  putmsd(fout, var.tn(), var.tmean(), var.tvr());
}

void crosovclass::put_coatab(ostream& fout, const varclass& var) const
// クロスオーバー用分散分析表を出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  fout << "\n";
  string tstr("ANOVA table");
  fout << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr << "\n";
  fout << "FACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, var.sg(), var.dfg(), var.sbr(), var.dfbr());
  tstr = "ORDER";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sg();
  fout << setw(10) << var.dfg() << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sbr();
  fout << setw(10) << var.dfbr() << setw(14) << vr << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.ss(), var.dfs(), var.sr(), var.dfr());
  tstr = "SUBJECT";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.ss();
  fout << setw(10) << var.dfs() << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, var.sp(), var.dfg(), var.sr(), var.dfr());
  tstr = "PERIOD";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sp();
  fout << setw(10) << var.dfg() << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  vr = calvfp(v, f, pr, var.sd(), var.dfg(), var.sr(), var.dfr());
  tstr = "DRUG";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sd();
  fout << setw(10) << var.dfg() << setw(14) << v << setw(12) << f;
  fout << setw(14) << pr << pind(pr) << "\n";
  tstr = "RESIDUALS";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.sr();
  fout << setw(10) << var.dfr() << setw(14) << vr << "\n";
  putline(fout, LENline);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(10 + npad) << tstr << right << setw(14) << var.st();
  fout << setw(10) << var.dft() << endl;
}

void crosovclass::put_dgest(ostream& fout, option& opt, const varclass& var)
// 処置に関する検定特性を出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  size_t g = var.g();
  int nt = var.tdn(0);
  double za = ptoz(opt.al()), zb = ptoz((1.0 - opt.po()) * 2.0);
  double ta = ptot(opt.al(), var.dfr()), tb = ptot((1.0 - opt.po()) * 2.0);
  double tc = ptot(1.0 - opt.cc(), var.dfr());
  fout << "\n<POWER ANALYSIS>\n";
  fout << "  Significance Level=" << opt.al() * 100.0 << "%  Power=" << opt.po() * 100.0 << "%  Confidence Coefficient=" << opt.cc() * 100.0 << "%\n";
  putline(fout, LENline);
  put_orderest(fout, opt, var);
  for (size_t i = 0; i < g - 1; ++i)
  {
    int ni = var.n(i, 0);
    double mdi = var.tdmean(i);
    double mddi = var.dmean(i);
    for (size_t j = i + 1; j < g; ++j)
    {
      fout << "Drug Effect:" << opt.dg(i) << " vs " << opt.dg(j);
      fout << " (Difference=" << opt.dg(j) << " - " << opt.dg(i) << ")\n";
      int nj = var.n(j, 0);
      double mdj = var.tdmean(j);
      double mddj = var.dmean(j);
      double dij = (mddj - mddi) * 0.5;
      double sed = sqrt(var.vr() / (double)g * (1.0 / (double)ni + 1.0 / (double)nj));
      double ci = tc * sed;
      double sedm = sqrt(var.vr() / (double)g * (1.0 / (double)(ni - 1) + 1.0 / (double)(nj - 1)));
      double dmin = (ta + tb) * sedm;
      put_dgd(fout, opt, i, ni, mdi, dij, sed, ci, sedm, dmin);
      if (opt.res("pa"))
      {
        if (opt.dl() > 0.0) put_logdn(fout, opt, ta, za, zb, ni, dij, sed, var.dfr(), var.vr());
        if (opt.dp() > 0.0) put_prodn(fout, opt, ta, za, zb, ni, mdi, dij, sed, var.dfr(), var.vr());
      }
      putline(fout, LENline);
    }
  }
}

void crosovclass::put_orderest(ostream& fout, const option& opt, const varclass& var) const
// 順序効果の信頼区間を出力する
{
  int n1 = var.sn(0), n2 = var.sn(1);
  int nt = n1 + n2, df = n1 + n2 - 2;
  double ss1 = var.sss(0), ss2 = var.sss(1);
  double v1 = var.svr(0), v2 = var.svr(1);
  double m1 = var.smean(0), m2 = var.smean(1);
  double md = m2 - m1;
  double vmd = (ss1 + ss2) / (double)df * (double)nt / (double)n1 / (double)n2;
  fout << "Ordering Effect:Difference=" << md << "\n  ";
  put_mdci(fout, opt, md, df, vmd);
}

void crosovclass::put_dgd(ostream& fout, const option& opt, int iidx, int ni, double mdi, double dij, double sed, double ci, double sedm, double dmin) const
// 処置差とその信頼区間と検出差を出力する
{
  fout << "  Difference=" << dij << "=";
  if (opt.logt() == logt10) fout << "log10(" << pow(10.0, dij);
  else fout << "ln(" << exp(dij);
  fout << ")  Ratio to " << opt.dg(iidx) << "=" << dij / mdi * 100.0 << "%\n";
  double dlow = dij - ci, dup = dij + ci;
  fout << "  " << opt.cc() * 100.0 << "%CI:Lower=" << dlow << "=";
  if (opt.logt() == logt10) fout << "log10(" << pow(10.0, dlow) << ")";
  else fout << "ln(" << exp(dlow) << ")";
  fout << "  Upper=" << dup << "=";
  if (opt.logt() == logt10) fout << "log10(" << pow(10.0, dup) << ")\n";
  else fout << "ln(" << exp(dup) << ")\n";
  fout << "    Ratio to " << opt.dg(iidx) << ":Lower=" << dlow / mdi * 100.0 << "%  Upper=" << dup / mdi * 100.0 << "%\n";
  fout << "  At " << ni << "cases/group:Diff=" << dmin << "=";
  if (opt.logt() == logt10) fout << "log10(" << pow(10.0, fabs(dmin));
  else fout << "ln(" << exp(dmin);
  fout << ")  Ratio to " << opt.dg(iidx) << "=" << fabs(dmin / mdi * 100.0) << "%\n";
}

void crosovclass::put_logdn(ostream& fout, const option& opt, double ta, double za, double zb, int ni, double dij, double sed, int dfr, double vr) const
// 対数検出差に関する検出力と必要例数を出力する
{
  double dd = opt.dl();
  double di = (za + zb) / dd;
  int n = (int)(di * di * vr + 1.9);
  double t1 = (dd / sed) - ta, t2 = 0.0;
  double pb = (t1 > 0.0) ? ttop(t1, dfr) * 0.5 : 1.0 - ttop(-t1, dfr) * 0.5;
  fout << "  At Diff=";
  if (opt.logt() == logt10) fout << "log10(" << pow(10.0, dd);
  else fout << "ln(" << exp(dd);
  fout << ")=" << dd << " and " << ni << "Cases/Group:Power=" << (1.0 - pb) * 100.0 << "%\n";
  fout << "  At Diff=";
  if (opt.logt() == logt10) fout << "log10(" << pow(10.0, dd);
  else fout << "ln(" << exp(dd);
  fout << ")=" << dd << ":Number of Cases/Group=" << n << "\n";
  di = (ptoz(1.0 - opt.po()) + za) / dd;
  n = (int)(di * di * vr + 1.9);
  int df = 2 * n - 2;
  double ci = ptot(opt.al(), df) * sqrt(vr / (double)n);
  ldbl po0 = 0.0L;
  fout << "  At Diff=0:Absolute Precision=" << ci << "  Number of Cases/Group=" << n << "\n";
  double d = fabs(dij);
  if (d < dd)                  // 観測差が検出差よりも小さい時だけ出力
  {
    n = cal_nci(opt, ci, po0, dd, d, vr);
    fout << "    At Diff=" << d << ":Absolute Precision=" << ci << " Number of Cases/Group=" << n << "\n";
  }
}

void crosovclass::put_prodn(ostream& fout, const option& opt, double ta, double za, double zb, int ni, double mdi, double dij, double sed, int dfr, double vr) const
// 割合検出差に関する検出力と必要例数を出力する
{
  double dd = fabs(mdi * opt.dp());
  double di = (za + zb) / dd;
  int n = int(di * di * vr + 1.9);
  double t1 = (dd / sed) - ta, t2 = 0.0;
  double pb = (t1 > 0.0) ? ttop(t1, dfr) * 0.5 : 1.0 - ttop(-t1, dfr) * 0.5;
  fout << "  Diff" << fabs(dd / mdi * 100.0) << "%=" << dd << " and " << ni << "Cases/Group:Power=" << (1.0 - pb) * 100.0 << "%\n";
  fout << "  Diff" << fabs(dd / mdi * 100.0) << "%=" << dd << ":Number of Cases/Group=" << n << "\n";
  di = (ptoz(1.0 - opt.po()) + za) / dd;
  n = int(di * di * vr + 1.9);
  int df = 2 * n - 2;
  double ci = ptot(opt.al(), df) * sqrt(vr / double(n));
  ldbl po0 = 0.0L;
  fout << "    At Diff=0:Absolute Precision=" << ci << "  Number of Cases/Group=" << n << "\n";
  double d = fabs(dij);
  if (d < dd)                  // 観測差が検出差よりも小さい時だけ出力
  {
    n = cal_nci(opt, ci, po0, dd, d, vr);
    fout << "    At Diff=" << d << ":Absolute Precision=" << ci << "  Number of Cases/Group=" << n << "\n";
  }
}

int crosovclass::cal_nci(const option& opt, double& ci, ldbl& po0, double dd, double d, double vr) const
// 信頼区間による必要例数を計算して返す
{
  if (dd < d) return 0;       // 計算不可能
  constexpr int Nmax = 10000; // 最大例数(繰り返し計算が終わらないため)
  double de = dd - d;
  double bci = 1.0 - opt.po() * de / dd;
  double di = (ptoz(opt.al()) + ptoz(bci)) / de;
  int n = int(di * di * vr + 0.9);
  int df = 2 * n - 2;
  double t0 = ptot(opt.al(), df);
  double sed = sqrt(vr / double(n));
  double tl = (-dd - d) / sed + t0, tu = (dd - d) / sed - t0;
  po0 = ncttolp(tu, df) - ncttolp(tl, df);
  while (po0 < (ldbl)(opt.po()) && n < Nmax)
  {
    n++;
    df = 2 * n - 2;
    t0 = ptot(opt.al(), df);  sed = sqrt(vr / double(n));
    tl = (-dd - d) / sed + t0;  tu = (dd - d) / sed - t0;
    po0 = ncttolp(tu, df) - ncttolp(tl, df);
  }
  ci = t0 * sed;                // 絶対精度
  return n;
}


void crosovclass::put_ttest(ostream& fout, const option& opt, const varclass& var) const
// クロスオーバー用t検定結果を出力する
{
  int n1 = var.sn(0), n2 = var.sn(1);
  double ss1 = var.sss(0), ss2 = var.sss(1);
  double v1 = var.svr(0), v2 = var.svr(1);
  double m1 = var.smean(0), m2 = var.smean(1);
  fout << "\n<PARAMETRIC TEST>\n";
  string tstr("BASIC STATISTICS OF THE SUM OF 2 PERIODS(ORDERING EFFECT)");
  put_tstat(fout, opt, tstr, n1, m1, ss1, v1, n2, m2, ss2, v2);
  n1 = var.dn(0);  n2 = var.dn(1);
  ss1 = var.dss(0);  ss2 = var.dss(1);
  v1 = var.dvr(0);  v2 = var.dvr(1);
  m1 = var.dmean(0);  m2 = -var.dmean(1);
  tstr = "BASIC STATISTICS OF THE CROSS DIFF OF 2 PERIODS(PERIOD EFFECT)";
  fout << "\n";
  put_tstat(fout, opt, tstr, n1, m1, ss1, v1, n2, m2, ss2, v2);
  m1 = var.dmean(0);  m2 = var.dmean(1);
  tstr = "BASIC STATISTICS OF THE DIFFERENCE OF 2 PERIODS(DRUG EFFECT)";
  fout << "\n";
  put_tstat(fout, opt, tstr, n1, m1, ss1, v1, n2, m2, ss2, v2);
}

void crosovclass::put_tstat(ostream& fout, const option& opt, string& tstr, int n1, double m1, double ss1, double v1, int n2, double m2, double ss2, double v2) const
// 2標本t検定の基礎統計量と検定結果を出力する
{
  constexpr int LENline = 77;   // 区切り線の長さ
  constexpr int LENhed = 9;     // ヘッダの長さ
  fout << "GROUP    :" << tstr << "\n";
  putline(fout, LENline);
  fout << left << setw(LENhed) << opt.g(0) << right << ":";
  putmsd(fout, n1, m1, v1);
  fout << left << setw(LENhed) << opt.g(1) << right << ":";
  putmsd(fout, n2, m2, v2);
  putline(fout, LENline);
  int nt = n1 + n2, df = n1 + n2 - 2;
  ldbl pr = 0.0L;
  bool welch = false;           // Welchの検定使用フラッグ
  string tname("Two Samples t-test(Equal Variances)");
  if (opt.tt() == testcor)      // Welchの検定使用
  { welch = true; }
  else if (opt.tt() == testaut) // 等分散性の両側F検定実行
  {
    double f = (v2 > 0.0) ? v1 / v2 : 0.0;
    int df1 = n1 - 1, df2 = n2 - 1;
    if (v1 < v2)
    {
      f = (v1 > 0.0) ? v2 / v1 : 0.0;
      df1 = n2 - 1;  df2 = n1 - 1;
    }
    pr = ftop(f, df1, df2);
    if (2.0L * pr <= (ldbl)(opt.al())) welch = true;
  }
  double md = m2 - m1;
  double vmd = (ss1 + ss2) / (double)df * (double)nt / (double)n1 / (double)n2;
  if (welch)
  {
    v1 /= (double)n1;  v2 /= (double)n2;
    vmd = v1 + v2;
    df = (int)(vmd * vmd / (v1 * v1 / (double)(n1 - 1) + v2 * v2 / (double)(n2 -1)));
    tname = "Aspin-Welch Test";
  }
  double t = md / sqrt(vmd);
  pr = ttop(t, df);
  fout << tname << ":t=" << t;
  fout << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
  put_mdci(fout, opt, md, df, vmd);
}

void crosovclass::put_mdci(ostream& fout, const option& opt, double md, int df, double vmd) const
// 平均値の差の信頼区間を出力する
{
  fout << opt.cc() * 100.0 << "% CI of Diff:";
  double ci = calmcci(mcnon, opt.cc(), 2, 1, df, 1, vmd) * 0.5;
  md *= 0.5;
  fout << "Lower=" << md - ci << "  Upper=" << md + ci << "\n";
}

void crosovclass::put_utest(ostream& fout, option& opt, const varclass& var)
// クロスオーバー用U検定結果を出力する
{
  fout << "\n<NONPARAMETIC TEST>\n";
  ftable ftab(2, (var.sumset()).size()); // クロス前後の和と差の度数表
  sum_ftab(opt, ftab, sumorder, var);
  double u = 0.0, vu = 0.0, z = 0.0;
  testtype utype = utest(u, vu, z, ftab, opt.ut());
  ldbl pr = ztop(z);
  fout << "Ordering Effect:Wilcoxon's Rank Sum Test(Mann-Whitney's U-test";
  if (utype == testcor) fout << ",corected";
  fout << ")\n";
  fout << "  z=" << z << "  p=" << pr << pind(pr) << "\n";
  put_umrdci(fout, opt, ftab, u, vu);
  ftab.resize(2, (var.difset()).size());
  sum_ftab(opt, ftab, sumperiod, var);
  utype = utest(u, vu, z, ftab, opt.ut());
  pr = ztop(z);
  fout << "\nPeriod Effect:Wilcoxon's Rank Sum Test(Mann-Whitney's U-test";
  if (utype == testcor) fout << ",corrected";
  fout << ")\n";
  fout << "  z=" << z << "  p=" << pr << pind(pr) << "\n";
  put_umrdci(fout, opt, ftab, u, vu);
  sum_ftab(opt, ftab, sumtreat, var);
  utype = utest(u, vu, z, ftab, opt.ut());
  pr = ztop(z);
  fout << "\nDrug Effect:Wilcoxon's Rank Sum Test(Mann-Whitney's U-test";
  if (utype == testcor) fout << ",corrected";
  fout << ")\n";
  fout << "  z=" << z << "  p=" << pr << pind(pr) << "\n";
  put_umrdci(fout, opt, ftab, u, vu);
}

void crosovclass::put_umrdci(ostream& fout, const option& opt, ftable& ftab, double u, double vu) const
// 順位平均値の差の信頼区間とU値の信頼区間を出力する
{
  int n1 = ftab.frtotal(1), n2 = ftab.frtotal(2);
  int nt = n1 + n2;
  double ta = u + (double)n2 * (double)(n2 + 1) * 0.5;
  double r2 = ta / (double)n2, tt = (double)nt * (double)(nt + 1) * 0.5;
  double r1 = (tt - ta) / (double)n1;
  double drm = r2 - r1;
  const double Reps = DBL_EPSILON * fmax(5.0, ta);
  if (fabs(drm) < Reps) drm = 0.0;
  double vd = vu / (double)n1 / (double)n2 * (double)nt * (1.0 / (double)n1 + 1.0 / (double)n2);
  double ci = calmcci(mcnon, opt.cc(), ftab.nr(), 1, 0, 1, vd);
  double dl = drm - ci, du = drm + ci;
  double pro = 100.0 / (double)nt;
  fout << "  Diff of Rank Sum=" << drm << "(" << drm * pro << "%)\n";
  fout << "    " << opt.cc() * 100.0 << "% CI:Lower=" << dl << "(" << dl * pro << "%)  Upper=" << du << "(" << du * pro << "%)\n";
  ci = calmcci(mcfis, opt.cc(), ftab.nr(), 1, 0, 1, vu);
  dl = u - ci;  du = u + ci;
  if (dl < 0.0) dl = 0.0;
  if (du > (double)n1 * (double)n2) du = (double)n1 * (double)n2;
  pro = 100.0 / (double)n1 / (double)n2;
  fout << "  U-Value=" << u << "(" << u * pro << "%)\n";
  fout << "    " << opt.cc() * 100.0 << "% CI:Lower=" << dl << "(" << dl * pro << "%)  Upper=" << du << "(" << du * pro << "%)\n";
}

void crosovclass::sum_ftab(option& opt, ftable& ftab, sumtype stype, const varclass& var)
// クロスオーバー用U検定用用にクロス前後の和／差／クロス差を度数集計する
{
  ftab.fclear();
  for (auto& cdata : datalst)
  {
    int rno = opt.search_g(cdata.sdata(0)) + 1;
    if (rno == 1 || rno == 2)   // 群1または2だけ集計
    {
      double fdpre = cdata.fdata(1), fdpost = cdata.fdata(2);
      if (fdpre != BLANKf && fdpost != BLANKf)
      {
        double fdata = 0.0;
        int cno = 0;
        switch (stype)
        {
        case sumorder : fdata = fdpre + fdpost;
          cno = search_data(var.sumset(), fdata);
          break;
        case sumperiod : fdata = (rno == 1) ? fdpost - fdpre : fdpre - fdpost;
          cno = search_data(var.difset(), fdata);
          break;
        case sumtreat : fdata = fdpost - fdpre;
          cno = search_data(var.difset(), fdata);
        }
        if (cno > 0) ftab.f(rno, cno, ftab.f(rno, cno) + 1);
      }
    }
  }
  ftab.set();
}

int crosovclass::search_data(const set<double>& targset, double fdata) const
// 数字データをサーチして番号を返す
{
  int sno = 1;
  bool hit = false;
  set<double>::const_iterator sitr = targset.begin();
  while (hit == false && sitr != targset.end())
  {
    if (sgndiff(fdata, *sitr) == 0) hit = true; // 誤差範囲なら一致
    else { ++sno;  ++sitr; }
  }
  return hit ? sno : -1;        // -1:なし
}

int crosovclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni() - 1) != 0) throw option_error("Run-time option error !");
  return nc;
}

void crosovclass::put(option& opt) // 集計結果を出力する
{
  if (opt.res("res") == false) return;
  put_resid(cout);
  if (id.nc() < 1)              // データなし
  {
    cout << "\n" << endl;
    return;
  }
  varclass var(id.ni() - 1);        // 計算用変数クラス
  if (var.cal(opt, datalst) < 1) return; // 有効なデータがない
  if (opt.res("stat")) put_stat(cout, opt, var);
  if (opt.res("at")) put_coatab(cout, var);
  if (opt.res("cc")) put_dgest(cout, opt, var);
  if (opt.res("tt")) put_ttest(cout, opt, var);
  if (opt.res("ut")) put_utest(cout, opt, var);
  cout << "\n" << endl;
}

int main(int argc, char *argv[]) // メインルーチン
{
  crosovclass crosov;           // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) crosov.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { crosov.get(argv[i], opt);  fok = false; }
    }
    crosov.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
