// dlae.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-25 09:33:19 tac>
//
// 微分ロジット分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W dlae dlae.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "dlae  Version 7.42e          Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:dlae [<OPTION> ...] [<DATA FILE> ...]",
  "  The differential logit analysis with the first item as the explanatory",
  "  variable(x) and the second item as the criterion variable(y).",
  "<OPTION>",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=dlae_dns.d",
  "    output item:x, y, estimate of y, logit(y), estimated value of logit(y)",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={logc|regc|regd|regl}[,...]:select the data for Ngraph",
  "    logc:x and y and logistic curve",
  "    regc:y and dy and quadratic regression curve",
  "    logd:x and dy and differential logistic curve",
  "    regl:x and logit(y) and logit line  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=dlae_ngp.dat",
  "  -res=[-][{def|all|cc[<C-COEF>]|regc|regl|stat|y[{Mh|m|<y value>}]}][,...]",
  "      :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    cc[<C-COEF>]:confidence intervals(0<=C-COEF<1)",
  "    regc:results of quadratic regression analysis",
  "    regl:results of logit regression analysis  stat:basic statistics",
  "    y[{Mh|m|<y value>}]:inverse estimate of x"
};

const string Title = "=== Differential Logit Analysis ==="; // 表題
const string DNSNAMEdef = "dlae_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "dlae_ngp.dat"; // 省略時Ngraphファイル名
constexpr int RDdef = 50;       // グラフの描画範囲分割数省略時値
constexpr int Pdef = 2;         // 回帰式の次数

class option                    // オプション集合体
{
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  string vdnsname;              // DANSファイル名
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // NgraphS出力の有無マップ
  double vcc;                   // 信頼係数
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  double vy;                    // xを逆算するyの値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  double cc() const             // 信頼係数を返す
  { return vcc; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  double y() const              // xを逆算するyの値
  { return vy; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vdnsname(), vngpname(), vngplst(), vngpmap(), vcc(CCdef), vreslst(), vresmap(), vy(BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dnsname")      // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni != 2) return disperr(vpname, "Number of items must be 2 !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["logc"] = vngpmap["regc"] = vngpmap["regd"] = vngpmap["regl"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["logc"] = vngpmap["regc"] = vngpmap["regd"] = vngpmap["regl"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["stat"] = vresmap["regc"] = vresmap["regl"] = vresmap["y"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["stat"] = vresmap["regc"] = vresmap["regl"] = vresmap["y"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all") // 省略時出力=全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["stat"] = vresmap["regc"] = vresmap["regl"] = vresmap["y"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;   // 出力不形式の都合でccは常に出力
      }
      else if (res == "yMh")    // vy=BLANKf:M/2に置き換える
      { vresmap["y"] = true; }
      else if (res[0] == 'y' && res.size() > 1)
      {
        if (res == "ym") vy = DBL_MAX; // 平均値に置き換える
        else vy = strtodf(res.substr(1));
        if (vy <= 0.0) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["y"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class data_less                 // ケースデータの比較用関数オブジェクト
{
private:
  int kidx;                     // ソートキー項目のインデックス
public:
  data_less(int kidx0) : kidx(kidx0) // コンストラクタ
  {}
  ~data_less()                  // デストラクタ
  {}
  bool operator()(datac& cdata1, datac& cdata2) const; // ケースデータを比較する
};

bool data_less::operator()(datac& cdata1, datac& cdata2) const
// ケースデータを比較する
{
  int comp = 0;
  double fdata1 = cdata1.fdata(kidx);
  double fdata2 = cdata2.fdata(kidx);
  if (fdata1 < fdata2) comp = -1;
  else if (fdata1 > fdata2) comp = 1;
  return (comp < 0) ? true : false; // 小さいときだけtrueを返す
}

class varclass                  // 変数クラス
{
private:
  pairdbl vxrange;              // xの最小値と最大値
  pairdbl vyrange;              // yの最小値と最大値
  statcor<double> vstatcxy;     // xとyの相関基礎統計量
  statcor<double> vstatcydy;    // yとdyの相関基礎統計量
  statcor<double> vstatcxly;    // xとlyの相関基礎統計量
  int vnp;                      // 回帰式の次数
  int vn0;                      // 回帰分析の例数
  int vncomp;                   // 補完例数
  int vnr;                      // 繰り返し計算数
  mat<double> vsmat;            // yとdyの回帰分析用積和行列
  double vy00;                  // y0初期値
  double vc0;                   // c初期値
  double vm0;                   // M初期値
  double vy0;                   // y0
  double vc;                    // c
  double vm;                    // M
  int cal_regc(ldata& datalst);
  // yとdyの回帰分析によってパラメータ初期値を推定する
  double cal_y0(double c, double m, ldata& datalst);
  // y0をロジット分析によって求める
  int cal_reglog(ldata& datalst);
  // ロジット分析によってパラメータを推定する(Mは更新しない)
  int cal_replog(ldata& datalst);
  // 繰り返しロジット分析によってパラメータを更新する
public:
  varclass() : vxrange(DBL_MAX, -DBL_MAX), vyrange(DBL_MAX, -DBL_MAX), vstatcxy(), vstatcydy(), vstatcxly(), vnp(Pdef), vn0(0), vncomp(0), vnr(0), vsmat(), vy00(0.0), vc0(0.0), vm0(0.0), vy0(0.0), vc(0.0), vm(0.0)
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int cal(ldata& datalst);      // データを集計して有効例数を返す
  void cal_mci(double& mlow, double& mup, double cc0) const;
  // Mの信頼区間を計算する(Newton法)
  double cal_dyest(double y) const; // dy推定値を返す
  double cal_dycr(double y, double x, bool plim) const; // dyの限界幅を返す
  int np() const                // 回帰式の次数を返す
  { return vnp; }
  int n0() const                // 回帰分析の例数を返す
  { return vn0; }
  int ncomp() const             // 補完例数を返す
  { return vncomp; }
  int nr() const                // 繰り返し計算数を返す
  { return vnr; }
  double xmin() const           // xの最小値を返す
  { return vxrange.first; }
  double xmax() const           // xの最大値を返す
  { return vxrange.second; }
  double ymin() const           // yの最小値を返す
  { return vyrange.first; }
  double ymax() const           // yの最大値を返す
  { return vyrange.second; }
  int xyn() const               // xとyの例数を返す
  { return vstatcxy.n(); }
  double xymx() const           // xの平均値を返す
  { return vstatcxy.mx(); }
  double xymy() const           // yの平均値を返す
  { return vstatcxy.my(); }
  double xyvxx() const          // xの不偏分散を返す
  { return vstatcxy.vxx(); }
  double xyvyy() const          // yの不偏分散を返す
  { return vstatcxy.vyy(); }
  int ydyn() const              // yとdyの例数を返す
  { return vstatcydy.n(); }
  int ydydft() const            // yとdyの全体自由度を返す
  { return vstatcydy.n(); }
  double ydymy() const          // yの平均値を返す
  { return vstatcydy.mx(); }
  double ydymdy() const         // dyの平均値を返す
  { return vstatcydy.my(); }
  double ydyvyy() const         // yの不偏分散を返す
  { return vstatcydy.vxx(); }
  double ydyvdydy() const       // dyの不偏分散を返す
  { return vstatcydy.vyy(); }
  double ydysyy() const         // yの平方和を返す
  { return vstatcydy.sxx(); }
  double ydysdydy() const       // dyの平方和を返す
  { return vstatcydy.syy(); }
  double ydysr() const          // yとdyの回帰分析の残差を返す
  { return vsmat(vnp, vnp); }
  int ydydfr() const            // yとdyの残差自由度を返す
  { return vstatcydy.n() - vnp; }
  double ydyvr() const          // yとdyの回帰分析の残差分散を返す
  { return (ydydfr() > 0) ? vsmat(vnp, vnp) / (double)ydydfr() : 0.0; }
  double ydyb(int idx) const // yとdyの偏回帰係数を返す
  { return (idx < vnp) ? vsmat(idx, vnp) : 0.0; }
  double ydys(int iidx, int jidx) const // yとdyの積和・平方和を返す
  { return (iidx < vnp && jidx < vnp) ? vsmat(iidx, jidx) : 0.0; }
  int xlyn() const              // xとlyの例数を返す
  { return vstatcxly.n(); }
  double xlymx() const          // xの平均値を返す
  { return vstatcxly.mx(); }
  double xlymly() const         // lyの平均値を返す
  { return vstatcxly.my(); }
  double xlyvxx() const         // xの不偏分散を返す
  { return vstatcxly.vxx(); }
  double xlyvlyly() const       // lyの不偏分散を返す
  { return vstatcxly.vyy(); }
  double xlysxx() const         // xの平方和を返す
  { return vstatcxly.sxx(); }
  double xlyslyly() const       // lyの平方和を返す
  { return vstatcxly.syy(); }
  double xlysxly() const        // xとlyの積和を返す
  { return vstatcxly.sxy(); }
  double y00() const            // y0初期値を返す
  { return vy00; }
  double c0() const             // c初期値を返す
  { return vc0; }
  double m0() const             // M初期値を返す
  { return vm0; }
  double y0() const             // y0を返す
  { return vy0; }
  double c() const              // cを返す
  { return vc; }
  double m() const              // Mを返す
  { return vm; }
};

int varclass::cal(ldata& datalst) // データを集計して有効例数を返す
{
  datalst.sort(data_less(0));   // datalstをx順にソートする
  int d = 1 + vnp;              // 回帰式は定数項なし
  vsmat.resize(d, d);
  vecdbl fdatavec(d);
  double x1 = 0.0, x2 = 0.0, y1 = 0.0, y2 = 0.0;
  int nxy = 0;
  ldata_itr ditrpre = datalst.begin();
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double x = ditr->fdata(0), y = ditr->fdata(1);
    ditr->resize(3);            // dyを追加する
    ditr->assign(2, BLANKf);
    if (x != BLANKf && y != BLANKf)
    {
      vstatcxy.sigma(x, y);
      if (x < vxrange.first) vxrange.first = x;
      if (x > vxrange.second) vxrange.second = x;
      if (y < vyrange.first) vyrange.first = y;
      if (y > vyrange.second) vyrange.second = y;
      if (nxy < 1)
      { x1 = x;  y1 = y;  nxy = 1; }
      else if (nxy == 1)
      {
        x2 = x;  y2 = y;  nxy = 2;
        ditrpre = ditr;
      }
      else
      {
        double dx1 = x2 - x1, dx2 = x - x2;
        double eps = DBL_EPSILON * fmax(5.0, fabs(dx1)); // 計算誤差
        double w1 = (fabs(dx1) < eps) ? 0.0 : 1.0 / dx1;
        double w2 = (fabs(dx2) < eps) ? 0.0 : 1.0 / dx2;
        double b1 = (y2 - y1) * w1, b2 = (y - y2) * w2;
        w1 = fabs(w1);  w2 = fabs(w2);
        double dy = (fabs(w1 + w2) < eps) ? 0.0 : (b1 * w1 + b2 * w2) / (w1 + w2);
        double fd = y2;
        vstatcydy.sigma(y2, dy);
        for (int i = 0; i < vnp; ++i)
        {
          fdatavec[i] = fd;
          fd *= y2;
       }
        fdatavec[d - 1] = dy;
        for (int i = 0; i < d; ++i)
        {
          fd = fdatavec[i];
          for (int j = i; j < d; ++j) vsmat.add(i, j, fd * fdatavec[j]);
        }
        ditrpre->assign(2, dy); // dyを1つ前のケースに保存する
        x1 = x2;  y1 = y2;
        x2 = x;  y2 = y;  ++nxy;
        ditrpre = ditr;
      }
    }
  }
  vn0 = cal_regc(datalst);      // 回帰分析によるパラメーター推定
  vncomp = cal_reglog(datalst); // ロジット分析によるパラメータ推定
  if (vncomp > 0) vnr = cal_replog(datalst); // 補完データがある時
  return nxy;                   // 有効例数を返す
}

int varclass::cal_regc(ldata& datalst)
// yとdyの回帰分析によってパラメータ初期値を推定する
{
  vy00 = (vyrange.first > 0.0) ? 0.5 * vyrange.first : 0.1;
  vm0 = (vyrange.second > vy00) ? vyrange.second : vy00 * 2.0;
  double b = (vstatcxy.sxx() > 0.0) ? vstatcxy.sxy() / vstatcxy.sxx() : 0.0;
  vc0 = (sgn(vm0) > 0) ? 8.0 * b / vm0 : 0.0; // 初期値を回帰直線によって設定
  vy0 = vy00;  vc = vc0;  vm = vm0;           // y0とcとMに初期値を代入
  size_t d = vsmat.r();
  int n = vstatcydy.n();
  if (n < vnp) return -1;       // 回帰分析不可能
  size_t dyidx = vnp;
  for (size_t i = 0; i < dyidx; ++i)
  {
    for (size_t j = i + 1; j <= dyidx; ++j) vsmat.assign(j, i, vsmat(i, j));
  }
  for (size_t i = 0; i < dyidx; ++i) vsmat.sweep(i);
  double b1 = vsmat(0, dyidx), b2 = vsmat(1, dyidx);
  double c = b1, m = -b1 / b2;
  if (sgn(c) < 0 || sgn(m) <= 0) return 0; // cまたはM不適切
  double y0 = cal_y0(c, m, datalst); // y0初期値をロジット分析によって求める
  if (y0 == BLANKf) return 0;  // y0不適切
  vy0 = vy00 = y0;  vc = vc0 = c;  vm = vm0 = m; // y0とcとMを更新
  return n;                     // 2次回帰分析の例数を返す
}

double varclass::cal_y0(double c, double m, ldata& datalst)
// y0をロジット分析によって求める
{
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (y != BLANKf && x != BLANKf)
    {
      if (0.0 < y && y < m)     // 範囲内(外れ値を補完しない)
      {
        double ly = log(y / (m - y)); // 実測値のロジット
        vstatcxly.sigma(x, ly);
      }
    }
  }
  if (vstatcxly.n() < 1) return BLANKf; // ロジットがない時
  double ly0 = vstatcxly.my() - c * vstatcxly.mx();
  double y0 = m / (1.0 + exp(-ly0));
  if (y0 >= m) return BLANKf; // 不適切なy0
  return y0;                  // y0を返す
}

int varclass::cal_reglog(ldata& datalst)
// ロジット分析によってパラメータを推定する(Mは更新しない)
{
  vstatcxly.clear();             // statcxlyを再使用するので初期化する
  double ly0 = log(vy0 / (vm - vy0)); // ロジット回帰式の切片
  int ncomp = 0;                   // 外れ値補完数
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (y != BLANKf && x != BLANKf)
    {
      double ly = ly0 + vc * x;  // ロジット推定値(外れ値補完用)
      if (0.0 < y && y < vm) ly = log(y / (vm - y)); // 範囲内
      else ++ncomp;
      vstatcxly.sigma(x, ly);
    }
  }
  if (vstatcxly.n() < 2) return -1; // ロジット分析不可能
  double mx = vstatcxly.mx(), my = vstatcxly.my();
  double sxx = vstatcxly.sxx(), sxy = vstatcxly.sxy();
  vc = (sgn(sxx) > 0) ? sxy / sxx : 0.0;
  ly0 = my - vc * mx;
  vy0 = vm / (1.0 + exp(-ly0));
  if (sgn(vc) < 0 || vy0 >= vm) return -2; // パラメータ不適切
  return ncomp;                 // 外れ値補完数を返す
}

int varclass::cal_replog(ldata& datalst)
// 繰り返しロジット分析によってパラメータを更新する
{
  constexpr int Rmax = 100;     // 繰り返し計算数最大値
  double syypre = vstatcxly.syy();
  const double Seps = 1.0e-5 * fmax(5.0, syypre);
  int nr = 0, ncomp = 0;
  bool comp = true;
  while (comp)
  {
    ncomp = cal_reglog(datalst);
    double syy = vstatcxly.syy();  ++nr;
    if (ncomp < 0 || fabs(syy - syypre) < Seps || nr >= Rmax) comp = false;
    else syypre = syy;
  }
  if (ncomp < 0) return ncomp;  // 繰り返し計算発散
  return nr;                    // 繰り返し計算数を返す
}

void varclass::cal_mci(double& mlow, double& mup, double cc0) const
// Mの信頼区間を計算する(Newton法)
{
  constexpr double EPSdef = 1.0e-5; // 収束誤差省略時値
  double t = ptot(1.0 - cc0, ydydfr());
  double dy0 = 0.0, y = vm * 1.1, dey = 1.0, slope = 1.0;
  double vpre = cal_dyest(vm * 0.5);
  const double Rerror = EPSdef * vpre;
  constexpr double Delta = EPSdef;
  constexpr double Mmax = DBL_MAX * 0.01;
  constexpr int RMAXdef = 100;
  const double Yeps = DBL_EPSILON * fmax(5.0, fabs(y));
  int nr = 0;
  while (fabs((vpre = cal_dyest(y) + cal_dycr(y, t, false)) - dy0) >= Rerror && nr < RMAXdef)
  {
    dey = Delta * y;
    if (fabs(dey) < Yeps) break;
    slope = (cal_dyest(y + dey) + cal_dycr(y + dey, t, false) - vpre) / dey;
    if (fabs(slope) < Yeps) break;
    y += (dy0 - vpre) / slope;
    if (y <= 0.0) y = dey;
    ++nr;
  }
  mup = y;
  if (mup < vm || Mmax < mup) // 上限サーチできず
  { mup = mlow = BLANKf; }
  else
  {
    y =  vm - (mup - vm);  nr = 0;
    while (fabs((vpre = cal_dyest(y) - cal_dycr(y, t, false)) - dy0) >= Rerror && nr < RMAXdef)
    {
      dey = Delta * y;
      if (fabs(dey) < Yeps) break;
      slope = (cal_dyest(y + dey) - cal_dycr(y + dey, t, false) - vpre) / dey;
      if (fabs(slope) < Yeps) break;
      y += (dy0 - vpre) / slope;
      if (y <= 0.0) y = dey;
      ++nr;
    }
    mlow = y;
    if (mlow < 0.0) mlow = 0.0;
  }
}

double varclass::cal_dyest(double y) const // dy推定値を返す
{
  double yp = y;
  double dyest = 0.0;
  for (int i = 0; i < vnp; ++i)
  {
    dyest += vsmat(i, vnp) * yp;
    yp *= y;
  }
  return dyest;                 // dy推定値を返す
}

double varclass::cal_dycr(double y, double x, bool plim) const
// dyの限界幅を返す
{
  vec<double> yv(vnp + 1);
  double yp = y;
  for (int i = 0; i < vnp; ++i)
  {
    yv.assign(i, yp);
    yp *= y;
  }
  yv.assign(vnp, 0.0);
  vec<double> tv(yv.trn() * vsmat);
  double w = inp(tv, yv);
  if (plim) w += 1.0;           // 予測限界幅
  return x * sqrt(ydyvr() * w); // vの限界幅を返す
}

class dlaclass                  // dlaメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout);
  // 分析結果のタイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, const varclass& var) const;
  // 基礎統計量を出力する
  void put_regc(ostream& fout, const option& opt, const varclass& var) const;
  // yとdyの2次回帰分析の結果を出力する
  void put_coef2(ostream& fout, const option& opt, const varclass& var) const;
  // 2次回帰式の偏回帰係数とパラメータを出力する
  void put_reglog(ostream& fout, option& opt, const varclass& var);
  // ロジット回帰分析の結果を出力する
  void put_coef1(ostream& fout, const option& opt, double ly0, double vb0, double c, double vb1, int dfr, double m) const;
  // 1次回帰式の回帰係数を出力する
  void put_dns(option& opt, const varclass& var);
  // xとyとy推測値とlyとly推測値をデータファイル形式で出力する
  void put_dnsid(ofstream& fout);
  // DANSデータのタイトルとデータ識別情報を出力する
  void plot_ngp(option& opt, const varclass& var); // Ngraph用データ出力
  void plot_ngpid(ofstream& fout);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpxy(ostream& fout); // Ngraph用にxとyのデータを出力する
  void plot_ngplog(ostream& fout, const varclass& var) const;
  // Ngraph用にxとyのロジスティック曲線を出力する
  int set_xrange(double& xmin, double& xmax, double& dx, double y0, double c, double m, double xmin0, double xmax0) const;
  // xの最小値と最大値とdxを返す
  void plot_ngploglim(ostream& fout, const option& opt, const varclass& var) const;
  // Ngraph用にxとyのロジスティック曲線の信頼＆予測限界を出力する
  void plot_loglim(ostream& fout, const string& title, const varclass& var, double ccpr, double tv, double vc, double vb, double mx) const;
  // Ngraph用にロジスティック曲線の信頼／予測限界を出力する
  void plot_ngpydy(ostream& fout);
  // Ngraph用にyとdyのデータを出力する
  void plot_ngpregc(ostream& fout, const varclass& var) const;
  // Ngraph用にyとdyの2次回帰曲線を出力する
  void plot_ngpregclim(ostream& fout, const option& opt, const varclass& var) const;
  // Ngraph用にyとdyの2次回帰曲線の信頼＆予測限界を出力する
  void plot_regclim(ostream& fout, const varclass& var, double ccpr, double tv, bool plim) const;
  // Ngraph用にyとdyの2次回帰曲線の信頼／予測限界を出力する
  void plot_ngpxdy(ostream& fout);
  // Ngraph用にxとdyのデータを出力する
  void plot_ngpregd(ostream& fout, const varclass& var) const;
  // Ngraph用にxとdyの微分ロジスティック曲線を出力する
  void plot_ngpxly(ostream& fout, const varclass& var);
  // Ngraph用にxとlyのデータを出力する
  void plot_ngpregl(ostream& fout, const varclass& var) const;
  // Ngraph用にxとlyのロジット直線を出力する
  void plot_ngpregllim(ostream& fout, const option& opt, const varclass& var) const;
  // Ngraph用にxとlyのロジット直線の信頼＆予測限界を出力する
  void plot_regllim(ostream& fout, const string& title, const varclass& var, double ccpr, double tv, double vc, double vb, double mx) const;
  // Ngraph用にロジット直線の信頼／予測限界を出力する
public:
  dlaclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~dlaclass()                   // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

void dlaclass::put_resid(ostream& fout)
// 分析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "EXPLANATORY VARIABLE X:" << id.item(0) << "\n";
  fout << "CRITERION VARIABLE   Y:" << id.item(1) << endl;
}

void dlaclass::put_stat(ostream& fout, const varclass& var) const
// 基礎統計量を出力する
{
  constexpr int LENline = 70; // 区切り線の長さ
  fout << "\n<BASIC STATISTICS OF VARIABLES>\n";
  putline(fout, LENline);
  fout << "X :";  putmsd(fout, var.xyn(), var.xymx(), var.xyvxx());
  fout << "Y :";  putmsd(fout, var.xyn(), var.xymy(), var.xyvyy());
  fout << "DY:";  putmsd(fout, var.ydyn(), var.ydymdy(), var.ydyvdydy());
  putline(fout, LENline);
}

void dlaclass::put_regc(ostream& fout, const option& opt, const varclass& var) const
// yとdyの2次回帰分析の結果を出力する
{
  fout << "\n<QUADRATIC REGRESSION ANALYSYS>\n";
  fout  << "  Model:dy=b1y^1 + b2y^2\n";
  size_t dyidx = var.np();      // 次数(2)
  int n = var.ydyn();
  double st = var.ydysdydy() + (double)n * var.ydymdy() * var.ydymdy();
  double sr = var.ydysr();
  int dfr = var.ydydfr(), dft = var.ydydft();
  const double Seps = DBL_EPSILON * fmax(5.0, st);
  if (dfr < 1 || sr < Seps) sr = 0.0;
  int dfb = var.np();
  double sb = st - sr;
  if (sb < Seps) sb = 0.0;
  put_coef2(fout, opt, var);
  double ly0 = log(var.y00() / (var.m0() - var.y00()));
  double r2 = (st > 0.0) ? sb / st : 0.0;
  fout << "Intercept of the Logit Model:y0=" << var.y00() << "  ly0=" << ly0 << "\n";
  fout << "Multiple R-squared=" << r2 << "  Multiple R=" << sqrt(r2) << "\n\n";
  if (dfr >= 0)
  {
    putatab(fout, "REGRESSION", sb, dfb, sr, dfr, st, dft);
    fout << endl;
  }
}

void dlaclass::put_coef2(ostream& fout, const option& opt, const varclass& var) const
// 2次回帰式の偏回帰係数とパラメータを出力する
{
  constexpr int Llen = 64;      // 区切り線の長さ
  fout << " Partial Reg-Coeff            SE";
  fout << setw(5) << opt.cc() * 100.0 << "%CI:Lower-Lim     Upper-Lim\n";
  putline(fout, Llen);
  int didx = var.np(), idx = 0;
  int offset = 1;
  vecdbl blow(3);
  vecdbl bup(3);
  for (int i = 0; i < didx; ++i)
  {
    double bi = var.ydyb(i), sii = var.ydys(i, i);
    double vbi = var.ydyvr() * sii;
    fout << "b" << left << setw(3) << offset + i;
    fout << right << setw(14) << bi << setw(14) << sqrt(vbi);
    double ci = calmcci(mcfis, opt.cc(), 1, 1, var.ydydfr(), 1, vbi);
    blow[idx] = bi - ci;  bup[idx] = bi + ci;
    fout << setw(18) << blow[idx] << setw(14) << bup[idx] << "\n";
    ++idx;
  }
  putline(fout, Llen);
  double clow = blow[0], cup = bup[0], mlow = 0.0, mup = 0.0;
  var.cal_mci(mlow, mup, opt.cc());
  fout << "c   " << right << setw(14) << var.c0();
  if (var.n0() > 0) fout << setw(32) << clow << setw(14) << cup;
  else fout << " (Estimated by Regression Equation)";
  fout << "\n";
  fout << "M   " << right << setw(14) << var.m0();
  if (var.n0() > 0)
  {
    if (mlow != BLANKf) fout << setw(32) << mlow << setw(14) << mup;
  }
  else fout << " (Estimated by Regression Equation)";
  fout << "\n";
  putline(fout, Llen);
}

void dlaclass::put_reglog(ostream& fout, option& opt, const varclass& var)
// ロジット回帰分析結果を出力する
{
  constexpr int LENline = 70; // 区切り線の長さ
  fout << "<BASIC STATISTICS OF VARIABLES OF LOGIT ANALYSIS>\n";
  if (var.nr() > 0) fout << "  Number of Complementary Repetitions=" << var.nr() << ")\n";
  putline(fout, LENline);
  fout << "X :";  putmsd(fout, var.xlyn(), var.xlymx(), var.xlyvxx());
  fout << "LY:";  putmsd(fout, var.xlyn(), var.xlymly(), var.xlyvlyly());
  putline(fout, LENline);
  fout << "\n<LOGIT ANALYSIS>\n";
  fout << "  Logit Model:ly=ly0 + cx\n";
  fout << "  logit:ly=ln(y/{M-y})  ly0=ln(y0/{M-y0})  M=" << var.m() << "(constant)\n";
   int dft = var.xlyn() - var.ncomp() - 1, dfb = 1; // 補完値は自由度を増やさない
  int dfr = dft - dfb;
  double mx = var.xlymx(), my = var.xlymly();
  double sxx = var.xlysxx(), st = var.xlyslyly(), sxy = var.xlysxly();
  double b1 = (sxx > 0.0) ? sxy / sxx : 0.0; // =c
  double ly0 = my - b1 * mx;
  double sb = sxy * b1;
  double sr = st - sb;
  const double Seps = DBL_EPSILON * fmax(5.0, st);
  if (dfr < 1 || sr < Seps) sr = 0.0;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double vb0 = vr * (1.0 / (double)(dft + 1) + mx * mx / sxx);
  double vb1 = (sxx > 0.0) ? vr / sxx : 0.0;
  put_coef1(fout, opt, ly0, vb0, var.c(), vb1, dfr, var.m());
  double r2 = (st > 0.0) ? sb / st : 0.0;
  if (opt.res("y"))
  {
    double y = opt.y();
    if (opt.y() == BLANKf) y = var.m() * 0.5;
    else if (opt.y() == DBL_MAX) y = var.xymy();
    else if (opt.y() >= var.m()) y = var.m() * 0.5;
    double ly = log(y / (var.m() - y));
    double x = (ly - ly0) / var.c();
    fout << "At y=" << y << ":ly=" << ly << "  x=" << x << "\n";
  }
  fout << "Multiple R-squared=" << r2 << "  Multiple R=" << sqrt(r2) << "\n\n";
  if (dfr >= 0)
  {
    putatab(fout, "REGRESSION", sb, dfb, sr, dfr, st, dft);
    fout << endl;
  }
}

void dlaclass::put_coef1(ostream& fout, const option& opt, double ly0, double vb0, double c, double vb1, int dfr, double m) const
// 1次回帰式の回帰係数を出力する
{
  constexpr int Llen = 64;      // 区切り線の長さ
  fout << " Partial Reg-Coeff            SE";
  fout << setw(5) << opt.cc() * 100.0 << "%CI:Lower-Lim     Upper-Lim\n";
  putline(fout, Llen);
  fout << "ly0 " << right << setw(14) << ly0 << setw(14) << sqrt(vb0);
  double ci = calmcci(mcfis, opt.cc(), 1, 1, dfr, 1, vb0);
  double ly0l = ly0 - ci, ly0u = ly0 + ci;
  fout << setw(18) << ly0l << setw(14) << ly0u << "\n";
  fout << "c   " << right << setw(14) << c << setw(14) << sqrt(vb1);
  ci = calmcci(mcfis, opt.cc(), 1, 1, dfr, 1, vb1);
  fout << setw(18) << c - ci << setw(14) << c + ci << "\n";
  putline(fout, Llen);
  double y0 = m / (1.0 + exp(-ly0));
  double y0l = m / (1.0 + exp(-ly0l)), y0u = m / (1.0 + exp(-ly0u));
  fout << "y0  " << right << setw(14) << y0 << setw(14) << " ";
  fout << setw(18) << y0l << setw(14) << y0u << "\n";
  putline(fout, Llen);
}

void dlaclass::put_dns(option& opt, const varclass& var)
// xとyとy推測値とlyとly推測値をデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  put_dnsid(fout);
  datac tdata(5);
  double ly0 = log(var.y0() / (var.m() - var.y0()));
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double x = ditr->fdata(0), y = ditr->fdata(1);
    tdata.assign(0, x);  tdata.assign(1, y);
    if (y == BLANKf || x == BLANKf)
    {
      tdata.assign(2, BLANKs);  tdata.assign(3, BLANKs);
      tdata.assign(4, BLANKs);
    }
    else
    {
      double lye = ly0 + var.c() * x; // ロジット推測値
      double ye = var.m() / (1.0 + exp(-lye));
      tdata.assign(2, ye);  // y推測値
      double ly = lye;
      if (0.0 < y && y < var.m()) ly = log(y / (var.m() - y)); // 範囲内
      else ly = BLANKf;
      tdata.assign(3, ly);
      tdata.assign(4, lye);
    }
    tdata.put(fout);
  }
}

void dlaclass::put_dnsid(ofstream& fout)
// DANSデータのタイトルとデータ識別情報を出力する
{
  idd tid(id.title(), 5);
  ostringstream ostr;
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  tid.assign_item(0, id.item(0));
  tid.assign_item(1, id.item(1));
  ostr << "Estimate of " << id.item(1);
  tid.assign_item(2, ostr.str());
  ostr.str("");
  ostr << "Logit of " << id.item(1);
  tid.assign_item(3, ostr.str());
  tid.assign_item(4, "Estimate of Logit");
  tid.put_item(fout, formtag);
}

void dlaclass::plot_ngp(option& opt, const varclass& var)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("logc"))
  {
    fout << "-s1 -dline\n";
    plot_ngpid(fout);
    plot_ngpxy(fout);
    fout << NGPdelm << "\n";
    plot_ngplog(fout, var);
    fout << NGPdelm << "\n";
    plot_ngploglim(fout, opt, var);
    delml = true;
  }
  if (opt.ngp("regc"))
  {
    if (delml)                  // 区切り行を出力する
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
      delml = true;
    }
    plot_ngpid(fout);
    plot_ngpydy(fout);
    fout << NGPdelm << "\n";
    plot_ngpregc(fout, var);
    fout << NGPdelm << "\n";
    plot_ngpregclim(fout, opt, var);
  }
  if (opt.ngp("regd"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline\n";
      delml = true;
    }
    plot_ngpid(fout);
    plot_ngpxdy(fout);
    fout << NGPdelm << "\n";
    plot_ngpregd(fout, var);
  }
  if (opt.ngp("regl"))
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline\n";
    plot_ngpid(fout);
    plot_ngpxly(fout, var);
    fout << NGPdelm << "\n";
    plot_ngpregl(fout, var);
    fout << NGPdelm << "\n";
    plot_ngpregllim(fout, opt, var);
  }
}

void dlaclass::plot_ngpid(ofstream& fout)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%EXPLANATORY VARIABLE X:" << id.item(0) << "\n";
  fout << "%CRITERION VARIABLE   Y:" << id.item(1) << endl;
}

void dlaclass::plot_ngpxy(ostream& fout)
// Ngraph用にxとyのデータを出力する
{
 fout << "%<RAW DATA OF X AND Y>\n";
 for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (y != BLANKf && x != BLANKf) fout << x << " " << y << endl;
  }
}

void dlaclass::plot_ngplog(ostream& fout, const varclass& var) const
// Ngraph用にxとyのロジスティック曲線を出力する
{
  fout << "%<LOGISTIC CURVE OF X AND Y>\n";
  double xmin = 0.0, xmax = 1.0, dx = 1.0;
  int rd = set_xrange(xmin, xmax, dx, var.y0(), var.c(), var.m(), var.xmin(), var.xmax());
  double x = xmin;
  double ly0 = log(var.y0() / (var.m() - var.y0()));
  for (int i = 0; i <= rd; ++i, x += dx) fout << x << " " << var.m() / (1.0 + exp(-ly0 - var.c() * x)) << endl;
}

int dlaclass::set_xrange(double& xmin, double& xmax, double& dx, double y0, double c, double m, double xmin0, double xmax0) const
// xの最小値と最大値とdxを返す
{
  constexpr double EPSYdef = 0.001; // yの微小値省略時値
  double ly0 = log(y0 / (m - y0));
  double y = m * EPSYdef;       // y最小値
  xmin = (log(y / (m - y)) - ly0) / c;
  if (xmin > xmin0) xmin = xmin0;
  if (sgn(xmin) < 0) xmin = 0.0;
  y = m * (1.0 - EPSYdef);      // y最大値
  xmax = (log(y / (m - y)) - ly0) / c;
  if (xmax < xmax0) xmax = xmax0;
  int rd = (int)(xmax - xmin);
  dx = 1.0;
  if (rd < RDdef)               // 分割数が省略時値未満の時
  {
    int r = (int)((double)RDdef / (double)rd + 0.9);
    rd *= r;  dx /= (double)r;
  }
  else if (rd > RDdef * 10)     // 分割数が省略時値の10倍以上の時
  {
    int r = (int)((double)rd / (double)(RDdef * 10) + 0.9);
    rd /= r;  dx *= (double)r;
  }
  return rd;                    // 分割数を返す
}

void dlaclass::plot_ngploglim(ostream& fout, const option& opt, const varclass& var) const
// Ngraph用にxとyのロジスティック曲線の信頼＆予測限界を出力する
{
  int dft = var.xlyn() - var.ncomp() - 1, dfb = 1; // 補完値は自由度を増やさない
  int dfr = dft - dfb;
  double mx = var.xlymx(), my = var.xlymly();
  double sxx = var.xlysxx(), st = var.xlyslyly(), sxy = var.xlysxly();
  double b1 = (sxx > 0.0) ? sxy / sxx : 0.0;
  double sb = sxy * b1;
  double sr = st - sb;
  const double Seps = DBL_EPSILON * fmax(5.0, st);
  if (dfr < 1 || sr < Seps) sr = 0.0;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double cf = 1.0 / (double)(dft + 1), rf = 1.0 + 1.0 / (double)(dft + 1);
  double vb = (sxx > 0.0) ? vr / sxx : 0.0;
  double tv = ptot(1.0 - opt.cc(), dfr);
  plot_loglim(fout, "CONFIDENCE LIMITS", var, opt.cc() * 100.0, tv, vr * cf, vb, mx);
  fout << NGPdelm << "\n";
  plot_loglim(fout, "PREDICTION LIMITS", var, opt.cc() * 100.0, tv, vr * rf, vb, mx);
}

void dlaclass::plot_loglim(ostream& fout, const string& title, const varclass& var, double ccpr, double tv, double vc, double vb, double mx) const
// Ngraph用にロジスティック曲線の信頼／予測限界を出力する
{
  fout << "%<" << ccpr << "% " << title << " OF LOGISTIC CURVE:UPPER>\n";
  double xmin = 0.0, xmax = 1.0, dx = 1.0;
  int rd = set_xrange(xmin, xmax, dx, var.y0(), var.c(), var.m(), var.xmin(), var.xmax());
  double ly0 = log(var.y0() / (var.m() - var.y0()));
  double x = xmin, dm = 0.0, ly = 0.0, y = 0.0;
  for (int i = 0; i <= rd; ++i, x += dx)
  {
    dm = x - mx;
    ly = ly0 + var.c() * x + tv * sqrt(vc + vb * dm * dm);
    fout << x << " " << var.m() / (1.0 + exp(-ly)) << "\n";
  }
  fout << "= =\n"; 
  fout << "%<" << ccpr << "% " << title << " OF LOGISTIC CURVE:LOWER>\n";
  x = xmin;
  for (int i = 0; i <= rd; ++i, x += dx)
  {
    dm = x - mx;
    ly = ly0 + var.c() * x - tv * sqrt(vc + vb * dm * dm);
    fout << x << " " << var.m() / (1.0 + exp(-ly)) << endl;
  }
}

void dlaclass::plot_ngpydy(ostream& fout)
// Ngraph用にyとdyのデータを出力する
{
  fout << "%<RAW DATA OF Y AND DY>\n";
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1), dy = cdata.fdata(2);
    if (x != BLANKf && y != BLANKf && dy != BLANKf) fout << y << " " << dy << endl;
  }
}

void dlaclass::plot_ngpregc(ostream& fout, const varclass& var) const
// Ngraph用にyとdyの2次回帰曲線を出力する
{
  fout << "%<QUADRATIC REGRESSION CURVE OF Y AND DY>\n";
  double yrange = var.m() - var.y0();
  double ystart = var.y0() - yrange * 0.1, yend = var.m() + yrange * 0.1;
  double dy = (yend - ystart) / (double)RDdef;
  double y = ystart;
  for (int i = 0; i <= RDdef; ++i, y += dy) fout << y << " " << var.cal_dyest(y) << endl;
}

void dlaclass::plot_ngpregclim(ostream& fout, const option& opt, const varclass& var) const
// Ngraph用にyとdyの2次回帰曲線の信頼＆予測限界を出力する
{
  double tv = ptot(1.0 - opt.cc(), var.ydydfr());
  plot_regclim(fout, var, opt.cc() * 100.0, tv, false);
  fout << NGPdelm << "\n";
  plot_regclim(fout, var, opt.cc() * 100.0, tv, true);
}

void dlaclass::plot_regclim(ostream& fout, const varclass& var, double ccpr, double tv, bool plim) const
// Ngraph用にyとdyの2次回帰曲線の信頼／予測限界を出力する
{
  fout << "%" << ccpr << "% ";
  if (plim) fout << "PREDICTION";
  else fout << "CONFIDENCE";
  fout << " LIMITS OF QUADRATIC REGRESSION CURVE OF Y AND DY:UPPER>\n";
  double yrange = var.m() - var.y0();
  double ystart = var.y0() - yrange * 0.1, yend = var.m() + yrange * 0.1;
  double dy = (yend - ystart) / (double)RDdef;
  double y = ystart, v = 0.0;
  for (int i = 0; i <= RDdef; ++i, y += dy)
  {
    v = var.cal_dyest(y) + var.cal_dycr(y, tv, plim);
    fout << y << " " << v << "\n";
  }
  fout << "= =\n";
  fout << "%" << ccpr << "% ";
  if (plim) fout << "PREDICTION";
  else fout << "CONFIDENCE";
  fout << " LIMITS OF QUADRATIC REGRESSION CURVE OF Y AND DY:LOWER>\n";
  y = ystart;
  for (int i = 0; i <= RDdef; ++i, y += dy)
  {
    v = var.cal_dyest(y) - var.cal_dycr(y, tv, plim);
    fout << y << " " << v << endl;
  }
}

void dlaclass::plot_ngpxdy(ostream& fout)
// Ngraph用にxとdyのデータを出力する
{
  fout << "%<RAW DATA OF X AND DY>\n";
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1), dy = cdata.fdata(2);
    if (x != BLANKf && y != BLANKf && dy != BLANKf) fout << x << " " << dy << endl;
  }
}

void dlaclass::plot_ngpregd(ostream& fout, const varclass& var) const
// Ngraph用にxとdyの微分ロジスティック曲線を出力する
{
  fout << "%<DIFFERENTIAL LOGISTIC CURVE OF X AND DY>\n";
  double xmin = 0.0, xmax = 1.0, dx = 1.0;
  int rd = set_xrange(xmin, xmax, dx, var.y0(), var.c(), var.m(), var.xmin(), var.xmax());
  double x = xmin;
  for (int i = 0; i <= rd; ++i, x += dx) fout << x << " " << logden(x, var.m(), var.c(), var.y0()) << endl;
}

void dlaclass::plot_ngpxly(ostream& fout, const varclass& var)
// Ngraph用にxとlyのデータを出力する
{
  fout << "%<RAW DATA OF X AND LOGIT(Y)>\n";
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (y != BLANKf && x != BLANKf)
    {
      if (0.0 < y && y < var.m()) fout << x << " " << log(y / (var.m() - y)) << endl;
    }
  }
}

void dlaclass::plot_ngpregl(ostream& fout, const varclass& var) const
// Ngraph用にxとlyのロジット直線を出力する
{
  fout << "%<LOGIT LINE OF X AND LOGIT(Y)>\n";
  double xmin = 0.0, xmax = 0.0, dx = 0.0;
  set_xrange(xmin, xmax, dx, var.y0(), var.c(), var.m(), var.xmin(), var.xmax());
  double ly0 = log(var.y0() / (var.m() - var.y0()));
  fout << xmin << " " << ly0 + var.c() * xmin << "\n";
  fout << xmax << " " << ly0 + var.c() * xmax << endl;
}

void dlaclass::plot_ngpregllim(ostream& fout, const option& opt, const varclass& var) const
// Ngraph用にxとlyのロジット直線の信頼＆予測限界を出力する
{
  int dft = var.xlyn() - var.ncomp() - 1, dfb = 1; // 補完値は自由度を増やさない
  int dfr = dft - dfb;
  double mx = var.xlymx(), my = var.xlymly();
  double sxx = var.xlysxx(), st = var.xlyslyly(), sxy = var.xlysxly();
  double b1 = (sxx > 0.0) ? sxy / sxx : 0.0;
  double sb = sxy * b1;
  double sr = st - sb;
  const double Seps = DBL_EPSILON * fmax(5.0, st);
  if (dfr < 1 || sr < Seps) sr = 0.0;
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double cf = 1.0 / (double)(dft + 1), rf = 1.0 + 1.0 / (double)(dft + 1);
  double vb = (sxx > 0.0) ? vr / sxx : 0.0;
  double tv = ptot(1.0 - opt.cc(), dfr);
  plot_regllim(fout, "CONFIDENCE LIMITS", var, opt.cc() * 100.0, tv, vr * cf, vb, mx);
  fout << NGPdelm << "\n";
  plot_regllim(fout, "PREDUCTION LIMITS", var, opt.cc() * 100.0, tv, vr * rf, vb, mx);
}

void dlaclass::plot_regllim(ostream& fout, const string& title, const varclass& var, double ccpr, double tv, double vc, double vb, double mx) const
// Ngraph用にロジット直線の信頼／予測限界を出力する
{
  fout << "%<" << ccpr << "% " << title << " OF LOGIT LINE OF X AND LOGIT(Y):UPPER>\n";
  double xmin = 0.0, xmax = 1.0, dx = 1.0;
  int rd = set_xrange(xmin, xmax, dx, var.y0(), var.c(), var.m(), var.xmin(), var.xmax());
  double ly0 = log(var.y0() / (var.m() - var.y0()));
  double x = xmin, dm = 0.0, ly = 0.0;
  for (int i = 0; i <= rd; ++i, x += dx)
  {
    dm = x - mx;
    ly = ly0 + var.c() * x + tv * sqrt(vc + vb * dm * dm);
    fout << x << " " << ly << "\n";
  }
  fout << "= =\n"; 
  fout << "%<" << ccpr << "% " << title << " OF LOGIT LINE OF X AND LOGIT(Y):LOWER>\n";
  x = xmin;
  for (int i = 0; i <= rd; ++i, x += dx)
  {
    dm = x - mx;
    ly = ly0 + var.c() * x - tv * sqrt(vc + vb * dm * dm);
    fout << x << " " << ly << endl;
  }
}

int dlaclass::get(const char* c_fname, option& opt)
// データを読み込み累積集計する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;                    // ケース数を返す
}

void dlaclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var;                 // 計算用変数クラス
  int nxy = var.cal(datalst);
  if (nxy < 1) return;       // 有効なデータがない
  if (opt.res("res"))                       // 分析結果出力
  {
    put_resid(cout);
    if (opt.res("stat")) put_stat(cout, var);
    if (var.xlyn() >= 2)        // ロジット分析実行
    {
      if (opt.res("regc")) put_regc(cout, opt, var);
      if (opt.res("regl")) put_reglog(cout, opt, var);
    }
    cout << "\n" << endl;
  }
  if (var.xlyn() < 2) return;
  if (opt.dnsname() != "") put_dns(opt, var); // データファイル出力
  if (opt.ngpname() != "") plot_ngp(opt, var); // Ngraphデータ出力
}

int main(int argc, char* argv[]) // メインルーチン
{
  dlaclass dla;                 // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) dla.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { dla.get(argv[i], opt);  fok = false; }
    }
    dla.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occured !");
    return 255;
  }
  return 0;
}
