// dpcve.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-25 09:36:52 tac>
//
// 最初の項目を群項目、次の項目を集計項目としてDP-plot分析の交差検証を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W dpcve dpcve.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "dpcve  Version 7.41e         Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:dpcv [<OPTION> ...] [<DATA FILE> ...]",
  "  The cross validation method of DP-plot analysis as the first item as",
  "  the group and the next item as the analysis data.",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -g{1|2}[=<CODE>]:specify the code of group1 or group2",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -pi[=<PRIOR-P>]:specify the prior probabillity of group1(0<PRIOR-P<1)",
  "  -r=<NUMBER>:specify the number of data in the validation group",
  "  -roc[={ac|cm|rp|yp|<LIMIT>}]:specify limits for ROC analysis",
  "    ac:maximum point of accuracy  cm:maimum point of chi-square",
  "    rp:optimum point of ROC curve  yp:maximum point of Youden index",
  "    LIMIT:specified value (-roc:-roc=rp)"
};

constexpr int Nlim = 5;         // 境界値数
constexpr int Nh = Nlim + 1;    // パラメータ数
constexpr double PIdef = 0.5;   // 省略時事前確率

using vecstatv = vector<statv<double>>; // 基礎統計量ベクトルの別名定義
using lditr = list<ldata_itr>;  // データリスト反復子リストの別名定義
using pairvec = pair<vecdbl, vecint>; // ベクトルペアの別名定義
using ldpair = list<pairvec>;   // ベクトルペアリストの別名定義
using ldpair_itr = ldpair::iterator; // 反復子の別名定義
using ldpair_citr = ldpair::const_iterator; // const反復子の別名定義

double fsn(double z, bool uporder) // 理論的感度を返す
{
  return (uporder) ? 1.0 - (double)(nedp(z)) : (double)(nedp(z));
}

double fsp(double z, bool uporder) // 理論的特異度を返す
{
  return (uporder) ? nedp(z) : 1.0 - nedp(z);
}

double fac(double sn, double sp, double pi1) // 理論的正診率を返す関数
{
  return pi1 * sn + (1.0 - pi1) * sp;
}

double fppv(double sn, double sp, double pi1) // 理論的PPVを返す関数
{
  double ppv = (sn < DBL_EPSILON || pi1 < DBL_EPSILON) ? 1.0 : 1.0 / (1.0 + (1.0 - pi1) / pi1 * (1.0 - sp) / sn);
  if (ppv < DBL_EPSILON) ppv = 0.0;
  return (ppv > 1.0) ? 1.0 : ppv;
}

double fnpv(double sn, double sp, double pi1) // 理論的NPVを返す関数
{
  double npv = (sp < DBL_EPSILON || (1.0 - pi1) < DBL_EPSILON) ? 1.0 : 1.0 / (1.0 + pi1 / (1.0 - pi1) * (1.0 - sn) / sp);
  if (npv < DBL_EPSILON) npv = 0.0;
  return (npv > 1.0) ? 1.0 : npv;
}

enum roctype                    // ROC分析用境界値の種類
{
  rocnon,                       // ROC分析結果非出力
  rocac,                        // 正診率最大点
  roccm,                        // 最大χ2乗法
  rocrp,                        // ROC曲線最適点(R-point)
  rocyp,                        // Youden index(感度+特異度-1)最大点(Y-point)
  roclim,                       // 境界値(上下限は自動判定)
  rocetc                        // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  int vr;                       // 検証群のデータ数
  vecstr vgvec;                 // 群コード
  double vpi;                   // 群1の事前確率
  roctype vroc;                 // ROC分析用境界値の種類
  double vlim;                  // 境界値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  int r() const                 // 検証群データ数を返す
  { return vr; }
  const string& g(int idx) const // 群コードを返す
  { return vgvec[idx]; }
  double pi() const             // 群1の事前確率を返す
  { return vpi; }
  roctype roc() const           // ROC分析結果用境界値の種類を返す
  { return vroc; }
  double lim() const            // 境界値を返す
  { return vlim; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
};

option::option(const char* c_pname) : vpath(), vpname(), vr(1), vgvec(2), vpi(BLANKf), vroc(rocrp), vlim(BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
  vgvec[0] = "1";
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "g1")      // 群1のコード指定
  { vgvec[0] = opt[1]; }
  else if (opt[0] == "g2")      // 群2のコード指定
  { vgvec[1] = opt[1]; }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "pi")      // 群1の事前確率指定
  {
    vpi = opt[1].empty() ? PIdef : strtodf(opt[1]);
    if (vpi <= 0.0 || 1.0 <= vpi) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "r")       // 検証群データ数の指定
  {
    char* ptr = nullptr;
    if (opt[1].empty()) vr = 1;
    else vr = int(strtol(opt[1].c_str(), &ptr, 10));
    if (vr < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0] == "roc")     // ROC分析結果の出力と境界値の種類の指定
  {
    if (opt[1].empty() || opt[1] == "rp") vroc = rocrp;
    else if (opt[1] == "ac") vroc = rocac;
    else if (opt[1] == "cm") vroc = roccm;
    else if (opt[1] == "yp") vroc = rocyp;
    else
    {
      vroc = roclim;
      vlim = strtodf(opt[1]);
      if (vlim == BLANKf) err = badopt(vpname, optstr, USAGE);
    }
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "Number of items must be 2 !");
  if (vgvec[0].empty() && vgvec[1].empty()) return disperr(vpname, "The code of the group is not specified !");
  return 0;                     // エラーなし
}

class data_less                 // ケースデータの比較用関数オブジェクト
{
private:
  int kidx;                     // ソートキー項目のインデックス
public:
  data_less(int kidx0) : kidx(kidx0) // コンストラクタ
  {}
  ~data_less()                  // デストラクタ
  {}
  bool operator()(datac& cdata1, datac& cdata2) const; // ケースデータを比較する
};

bool data_less::operator()(datac& cdata1, datac& cdata2) const
// ケースデータを比較する
{
  int comp = 0;
  double fdata1 = cdata1.fdata(kidx);
  double fdata2 = cdata2.fdata(kidx);
  if (fdata1 < fdata2) comp = -1;
  else if (fdata1 > fdata2) comp = 1;
  return (comp < 0) ? true : false; // 小さいときだけtrueを返す
}

class varclass                  // 変数クラス
{
private:
  int vnlim;                    // 境界値数
  int vnh;                      // パラメータ数
  vecint ndpvec;                // 検証結果例数ベクトル
  vecdbl hvec;                  // パラメータベクトル
  double vpi10;                 // 群1の事前確率(全例)
  double vpi1;                  // 群1の事前確率
  bool vuporder;                // 昇順フラッグ(m1>=m2の時true)
  ldpair vcvdpairlst;           // 交差検証法の結果リスト
  int cal_dp(option& opt, lditr& lditrlstt);
  // DP-plot分析を行って総例数を返す
  double cal_acpoint(double lim0, double m1, double v1, double m2, double v2) const;
  // AC-pointを返す
  double cal_pnpoint(double lim0, double m1, double sd1, double m2, double sd2) const;
  // PN-pointを返す
  double cal_roc(double& auc, int n1, int n2, option& opt, lditr& lditrlstt) const;
  // ROC分析の境界を返す
  void cv_dp(lditr& lditrlstv);
  // DP-plot分析結果でデータリストを検証して正判別例数を合計する
  void set_cvdlst(int vno);     // 検証結果をリストに追加する
  int cal_dpcv(option& opt, ldata& datalst, int r, int k);
  // 交差検証を行って検証回数を返す
public:
  varclass(int nlim0, int nh0) : vnlim(nlim0), vnh(nh0), ndpvec(nlim0 * 4, 0), hvec(nh0, 0.0), vpi1(0.0), vuporder(true), vcvdpairlst()
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int cal(option& opt, ldata& datalst); // データを集計して変数をセットする
  int nlim() const                      // 境界値数を返す
  { return vnlim; }
  int nh() const                // パラメータ数を返す
  { return vnh; }
  double pi10() const           // 群1の事前確率を返す(全例)
  { return vpi10; }
  const ldpair& cvdpairlst() const // 交差検証法の結果リストを返す
  { return vcvdpairlst; }
};

int varclass::cal_dp(option& opt, lditr& lditrlstt)
// DP-plot分析を行って総例数を返す
{
  vecstatv statvec(2);          // 基礎統計量ベクトル
  for (ldata_itr ditr : lditrlstt) statvec[(size_t)(ditr->fdata(0))].sigma(ditr->fdata(1)); // 群ごとに統計量を集計する
  int n1 = statvec[0].n(), n2 = statvec[1].n();
  int nt = n1 + n2;
  double m1 = statvec[0].mean(), m2 = statvec[1].mean();
  double v1 = statvec[0].vr(), v2 = statvec[1].vr();
  double sd1 = statvec[0].sd(), sd2 = statvec[1].sd();
  double sdt = sd1 + sd2;
  double auc = 0.0;
  vuporder = (m1 >= m2) ? true : false;
  vpi1 = (opt.pi() == BLANKf) ? (double)n1 / (double)nt : opt.pi();
  double lim0 = (m1 + m2) * 0.5;  // 判別分析の境界値
  const double SDeps = DBL_EPSILON * fmax(5.0, fabs(lim0));
  hvec[0] = (fabs(sdt) < SDeps) ? lim0 : (m1 * sd2 + m2 * sd1) / sdt;
  hvec[1] = cal_acpoint(lim0, m1, v1, m2, v2);
  hvec[2] = cal_pnpoint(lim0, m1, sd1, m2, sd2);
  hvec[3] = lim0;
  hvec[4] = cal_roc(auc, n1, n2, opt, lditrlstt);
  hvec[5] = auc;
  return n1 + n2;               // 総例数を返す
}

double varclass::cal_acpoint(double lim0, double m1, double v1, double m2, double v2) const
// AC-pointを返す
{
  if (sgndiff(m1, m2) == 0) return lim0; // 計算不可能、初期値を返す
  double vd = v2 - v1;
  if (sgndiff(v2, v1) == 0)     // 2群の分散が等しい時
  {
    return (m1 + m2) * 0.5 + v1 * log((1.0 - vpi1) / vpi1) / (m1 - m2);
  }
  const double Veps = DBL_EPSILON * fmax(5.0, fabs(lim0));
  if (v1 < Veps || v2 < Veps) return lim0; // 計算不可能、初期値を返す
  double md = m1 * v2 - m2 * v1;
  double sd1 = sqrt(v1), sd2 = sqrt(v2);
  double tmp = md * md - vd * (m1 * m1 * v2 - m2 * m2 * v1 + 2.0 * v1 * v2 * log((1.0 - vpi1) / vpi1 * sd1 / sd2));
  double limm = (md - sqrt(tmp)) / vd, limp = (md + sqrt(tmp)) / vd;
  return (vuporder) ? limm : limp; // 理論値を返す
}

double varclass::cal_pnpoint(double lim0, double m1, double sd1, double m2, double sd2) const
// PN-pointを返す
{
  const double SDeps = DBL_EPSILON * fmax(5.0, fabs(lim0));
  if (sd1 < SDeps || sd2 < SDeps) return lim0; // 初期値を返す
  constexpr double Rerror = 1.0e-6;            // 収束限界
  constexpr double Delta = Rerror * 10.0; // 刻み幅
  constexpr double SDx = 5.0;   // 範囲を決めるためのSDの倍率
  double lim = lim0;
  double z1 = (lim - m1) / sd1, z2 = (lim - m2) / sd2;
  double sn = fsn(z1, vuporder), sp = fsp(z2, vuporder);
  double pr = fppv(sn, sp, vpi1) - fnpv(sn, sp, vpi1);
  while (fabs(pr) >= Rerror)
  {
    double dlim = Delta * lim;
    double dlim1 = dlim / sd1, dlim2 = dlim / sd2;
    double snd = fsn(z1 + dlim1, vuporder), spd = fsp(z2 + dlim2, vuporder);
    double prd = fppv(snd, spd, vpi1) - fnpv(snd, spd, vpi1);
    double g = (prd - pr) / dlim;
    if (fabs(g) < SDeps)
    { pr = 0.0; }
    else
    {
      lim -= pr / g;
      z1 = (lim - m1) / sd1;  z2 = (lim - m2) / sd2;
      sn = fsn(z1, vuporder);  sp = fsp(z2, vuporder);
      pr = fppv(sn, sp, vpi1) - fnpv(sn, sp, vpi1);
    }
  }
  double lmin = m2 - sd1 * SDx, lmax = m1 + sd2 * SDx;
  if (vuporder == false) { lmin = m1 - sd1 * SDx;  lmax = m2 + sd2 * SDx; }
  if (lim < lmin || lmax < lim) return lim0; // 範囲外の時は初期値を返す
  return lim;                   // 範囲内
}

double varclass::cal_roc(double& auc, int n1, int n2, option& opt, lditr& lditrlstt) const
// ROC分析の境界を返す
{
  int nt = n1 + n2, tp = n1, dfn = 0, tn = 0, dtn = 0;
  int tp_max = 0, tn_max = n2;
  if (vuporder == false) lditrlstt.reverse(); // 群1の平均値の方が小さい時
  roctype roc = opt.roc();
  if (roc == rocnon) roc = rocrp; // ROC分析結果非出力時も分析だけはしておく
  double ac_max = 0.0, csqr_max = 0.0, yi_max = 0.0, dist_min = FLT_MAX;
  double w = 0.0, lim = BLANKf;
  double fd = 0.0, fdpre = BLANKf;
  for (ldata_itr ditr : lditrlstt)
  {
    fd = ditr->fdata(1);
    if (fd != fdpre)
    {
      tp -= dfn;  tn += dtn;
      double sn = (double)tp / (double)n1;
      double sp = (double)tn / (double)n2;
      double ac = (double)(tp + tn) / (double)nt;
      w += (double)(tp * dtn) + 0.5 * (double)(dfn * dtn);
      double csqr = 0.0, yi = 0.0, x = 0.0, y = 0.0, dist = 0.0;
      switch (roc)
      {
      case rocac :              // 正診率最大
        if (ac > ac_max)
        { ac_max = ac;  lim = fd; }
        break;
      case roccm : csqr = csqr22(tn, n2 - tn, n1 - tp, tp); // 最大χ2乗法
        if (csqr > csqr_max)
        { csqr_max = csqr;  lim = fd; }
        break;
      case rocyp : yi = sn + sp - 1.0; // Youden index(Y-point)
        if (yi > yi_max)
        { yi_max = yi;  lim = fd; }
        break;
      default : x = 1.0 - sp;  y = 1.0 - sn; // ROC曲線最適点(R-point)
        dist = x * x + y * y;
        if (dist < dist_min)
        { dist_min = dist;  lim = fd; }
        break;
      }
      fdpre = fd;  dfn = dtn = 0;
    }
    if ((int)(ditr->fdata(0)) == 0) ++dfn;
    else ++dtn;
  }
  if (roc == roclim) lim = opt.lim();
  else if (lim == BLANKf) lim = fd; // 境界値がサーチできていない時
  w += 0.5 * (double)(dfn * dtn);
  auc = w / (double)n1 / (double)n2;
  return lim;                   // 境界値を返す
}

void varclass::cv_dp(lditr& lditrlstv)
// DP-plot分析結果でデータリストを検証して正判別例数を合計する
{
  ndpvec.assign(ndpvec.size(), 0);
  for (ldata_itr ditr : lditrlstv)
  {
    int gidx = (int)(ditr->fdata(0));
    double f = ditr->fdata(1);
    for (int i = 0; i < vnlim; ++i)
    {
      int j = i * 4;
      if (gidx == 0)              // 群1
      {
        ndpvec[j] += 1;
        if (vuporder)
        {
          if (f >= hvec[i]) ndpvec[j + 1] += 1; // 正判別
        }
        else
        {
          if (f <= hvec[i]) ndpvec[j + 1] += 1; // 正判別
        }
      }
      else if (gidx == 1)         // 群2
      {
        ndpvec[j + 2] += 1;
        if (vuporder)
        {
          if (f < hvec[i]) ndpvec[j + 3] += 1; // 正判別
        }
        else
        {
          if (f > hvec[i]) ndpvec[j + 3] += 1; // 正判別
        }
      }
    }
  }
}

void varclass::set_cvdlst(int vno) // 検証結果をリストに追加する
{
  vecdbl cvdhvec(1 + vnh);     // 検証結果パラメータベクトル
  vecint cvdndpvec(vnlim * 6); // 検証結果例数ベクトル
  cvdhvec[0] = (double)((vno < 1) ? vnh : vno);
  for (int i = 0; i < vnh; ++i) cvdhvec[1 + i] = hvec[i];
  for (int i = 0; i < vnlim; ++i)
  {
    int j = i * 4, k = i * 6;
    cvdndpvec[k] = ndpvec[j];
    cvdndpvec[k + 1] = ndpvec[j + 1];
    cvdndpvec[k + 2] = ndpvec[j + 2];
    cvdndpvec[k + 3] = ndpvec[j + 3];
    cvdndpvec[k + 4] = ndpvec[j] + ndpvec[j + 2];
    cvdndpvec[k + 5] = ndpvec[j + 1] + ndpvec[j + 3];
  }
  vcvdpairlst.push_back(make_pair(cvdhvec, cvdndpvec));
}

int varclass::cal_dpcv(option& opt, ldata& datalst, int r, int k)
// 交差検証を行って検証回数を返す
{
  lditr lditrlstt, lditrlstv; // トレーニングデータとバリデーションデータ反復子
  for (int i = 0; i < k; ++i)
  {
    int sidx = r * i, eidx = r * (i + 1) - 1;
    int j = 0;
    for (ldata_itr dpos = datalst.begin(); dpos != datalst.end(); ++dpos, ++j)
    {
      if (sidx <= j && j <= eidx) lditrlstv.push_back(dpos);
      else lditrlstt.push_back(dpos);
    }
    if (cal_dp(opt, lditrlstt) > 0)
    {
      cv_dp(lditrlstv);
      set_cvdlst(i + 1);
    }
    lditrlstt.clear();  lditrlstv.clear();
  }
  return k;                     // 検証回数を返す
}

int varclass::cal(option& opt, ldata& datalst)
// データを集計して変数をセットする
{
  lditr lditrlst;               // データリスト反復子リスト
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr) lditrlst.push_back(ditr);
  int nt = cal_dp(opt, lditrlst);     // 全例のDP-plot分析(nt:総例数)
  vpi10 = vpi1;                       // 群1の確率(全例)
  int r = opt.r();                    // 検証群のデータ数
  if (nt < 1 || r > nt - 2) return 0; // 交差検証実行せず
  cv_dp(lditrlst);                    // 全例の検証
  set_cvdlst(0);                      // 全例の検証結果をリストに追加する
  lditrlst.clear();                   // 反復子リストを削除する
  int k = nt / r;                     // 検証群数
  if (nt % r > 0) ++k;                // 検証群数を切り上げる
  return cal_dpcv(opt, datalst, r, k); // 交差検証を行って検証回数を返す
}

class dpcvclass                 // dpcvメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int search_g(const option& opt, const string& gcode) const;
  // 群コードを検索して群インデックスを返す
  void put_dpcv(ostream& fout, const varclass& var);
  // 判別分析の交差検証結果をデータファイル形式で出力する
public:
  dpcvclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~dpcvclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

int dpcvclass::search_g(const option& opt, const string& gcode) const
// 群コードを検索して群インデックスを返す
{
  if (gcode == BLANKs) return -1; // 欠損データはヒットさせない
  int idx = -1;
  if ((opt.g(0)).empty())       // 群1のコード指定がない場合
  { idx = (gcode == opt.g(1)) ? 1 : 0; }
  else if (gcode == opt.g(0))   // 群1のコード
  { idx = 0; }
  else if ((opt.g(1)).empty())  // 群2のコード指定がない場合
  { idx = 1; }
  else if (gcode == opt.g(1))   // 群2のコード
  { idx = 1; }
  return idx;                   // -1:ヒットせず,0:群1,1:群2
}

void dpcvclass::put_dpcv(ostream& fout, const varclass& var)
// 判別分析の交差検証結果をデータファイル形式で出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  idd tid(id.title(), 1 + var.nh() + var.nlim() * 6);
  tid.append_stra(id.item(0), atype, avalnon);
  tid.append_stra(id.item(1), atype, avalnon);
  tid.put_stra(fout, formtag);
  tid.assign_item(0, "N of Parameters/Validation Group Number");
  tid.assign_item(1, "SS-point");
  tid.assign_item(2, "AC-point");
  tid.assign_item(3, "PN-point");
  tid.assign_item(4, "Limits for LDA");
  tid.assign_item(5, "Limits for ROC");
  tid.assign_item(6, "AUC of ROC");
  for (int i = 0; i < var.nlim(); ++i)
  {
    int j = 2 + var.nlim() + i * 6;
    tid.assign_item(j, tid.item(1 + i) + ":n in Group1");
    tid.assign_item(j + 1, tid.item(1 + i) + ":n of TP");
    tid.assign_item(j + 2, tid.item(1 + i) + ":n in Group2");
    tid.assign_item(j + 3, tid.item(1 + i) + ":n of FN");
    tid.assign_item(j + 4, tid.item(1 + i) + ":Total n");
    tid.assign_item(j + 5, tid.item(1 + i) + ":n of AC");
  }
  ostringstream ostr;
  ostr << "AC-point(Prior pr of Group1=" << var.pi10() << ")";
  tid.assign_item(2, ostr.str());
  ostr.str("");
  ostr << "PN-point(Prior pr of Group1=" << var.pi10() << ")";
  tid.assign_item(3, ostr.str());
  tid.put_item(fout, formtag);
  int prec0 = fout.precision(); // 現在の不動小数点精度(有効数字)を保存する
  for (ldpair_citr cvitr = (var.cvdpairlst()).begin(); cvitr != (var.cvdpairlst()).end(); ++cvitr)
  {
    fout << cvitr->first[0];
    fout.precision(16);
    for (size_t i = 1; i < cvitr->first.size(); ++i) fout << CHRdelm << cvitr->first[i];
    fout.precision(prec0);
    for (size_t i = 0; i < cvitr->second.size(); ++i) fout << CHRdelm << cvitr->second[i];
    fout << endl;
  }
}

int dpcvclass::get(const char* c_fname, option& opt) // データを読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  nc = (int)(id.nc());          // ケース数変更用
  ldata_itr ditr = datalst.begin();
  while (ditr != datalst.end())
  {
    int idx = search_g(opt, ditr->sdata(0));
    double fdata = ditr->fdata(1);
    if (idx < 0 || fdata == BLANKf) // 欠損データはリストから削除する
    { ditr = datalst.erase(ditr);  --nc; }
    else
    {
      ditr->assign(0, (double)idx); // 群コードを群インデックスに変える
      ++ditr;
    }
  }
  if (nc < (int)(id.nc())) id.nc(nc); // ケース数を減らす
  return nc;
}

void dpcvclass::put(option& opt) // 集計結果を出力する
{
  datalst.sort(data_less(1)); // データリストを項目データ順にソートする
  varclass var(Nlim, Nh);     // 変数オブジェクト
  if (var.cal(opt, datalst) > 0) put_dpcv(cout, var);
}

int main(int argc, char* argv[]) // メインルーチン
{
  dpcvclass dpcv;               // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) dpcv.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { dpcv.get(argv[i], opt);  fok = false; }
    }
    dpcv.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
