// dpplote.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-25 11:16:13 tac>
//
// 群項目と集計項目からDP-plot分析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W dpplote dpplote.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "dpplote  Version 7.41e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:dpplote [<OPTION> ...] [<DATA FILE> ...]",
  "  The DP-plot analysis as the first item as the group and the next item",
  "  as the analysis data.",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -g{1|2}[=<CODE>]:specify the code of group1(disease) or group2(normal)",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -lim={auto|low|up}:specify upper/lower limit",
  "    auto:automatic judgment(default)  low:lower  up:upper",
  "  -ngp={ac|dpp|log|pnf|pv|snp}[,...]:select the data for Ngraph",
  "    ac:accuracy curve  dpp:common DP-plot  log:logistic curve",
  "    pnf:positive/negative rate curve  pv:PPV/NPV curve",
  "    snp:sensitivity/specificity curve  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=dpplote_ngp.dat",
  "  -pi[=<PRIOR-P>]:specify the prior probabillity of group1(0<PRIOR-P<1)",
  "  -res=[-][{def|all|ac[<ACCURACY>]|cc[<C-COEF>]|dpa|stat|uda}][,...]",
  "      :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    ac[<ACCURACY>]:limits when the specified accuracy",
  "    cc[<C-COEF>]:confidence intervals(0=<C-COEF<1)  dpa:DP-plot analysis",
  "    stat:basic statistics  uda:univariate linear discriminant analysis",
  "    default:auto judgment according to data"
};

const string Title = "=== DP-plot(Diagnostic Performance Plot) Analysis ==="; // 表題
const string NGPNAMEdef = "dpplote_ngp.dat"; // 省略時Ngraphファイル名
constexpr uint RDdef = 50;      // 座標軸の分割数
constexpr size_t Nco = 8;       // 境界値の個数
constexpr double Rerror = 1.0e-8; // 収束限界
constexpr double ACdef = 0.7;     // 省略時正診率
constexpr double PIdef = 0.5;     // 省略時事前確率

enum limtype                    // 境界値の種類
{
  limnon,                       // 指定なし
  limauto,                      // 自動判定
  limlow,                       // 下限
  limup,                        // 上限
  limetc                        // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vac;                   // 正診率
  double vcc;                   // 信頼係数
  vecstr vgvec;                 // 群コードベクトル
  limtype vlim;                 // 境界値の種類
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  double vpi;                   // 群1の事前確率
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double ac() const             // 正診率を返す
  { return vac; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& g(int idx) const // 群コードを返す
  { return vgvec[idx]; }
  limtype lim() const           // 境界値の種類を返す
  { return vlim; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  double pi() const             // 群1の事前確率を返す
  { return vpi; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int search_g(const string& gcode) const; // 群コードを検索する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vac(ACdef), vcc(CCdef), vgvec(2), vlim(limauto), vngpname(), vngplst(), vngpmap(), vpi(BLANKf), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == 0) return;     // 初期設定せず
  vgvec[0] = "1";               // 群1のコードだけ初期設定
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "g1")      // 群1のコード指定
  {
    if (opt[1].empty()) vgvec[0].erase();
    else vgvec[0] = opt[1];
  }
  else if (opt[0] == "g2")      // 群2のコード指定
  {
    if (opt[1].empty()) vgvec[1].erase();
    else vgvec[1] = opt[1];
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "lim")
  {
    if (opt[1].empty() || opt[1] == "auto") vlim = limauto;
    else if (opt[1] == "low") vlim = limlow;
    else if (opt[1] == "up") vlim = limup;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pi")      // 群1の事前確率指定
  {
    vpi = opt[1].empty() ? PIdef : strtodf(opt[1]);
    if (vpi <= 0.0 || 1.0 <= vpi) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni != 2) return disperr(vpname, "Number of items must be 2 !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["ac"] = vngpmap["dpp"] = vngpmap["log"] = vngpmap["pnf"] = vngpmap["pv"] = vngpmap["snp"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["ac"] = vngpmap["dpp"] = vngpmap["log"] = vngpmap["pnf"] = vngpmap["pv"] = vngpmap["snp"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["dpa"] = vresmap["stat"] = vresmap["uda"] = true;
  vresmap["ac"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["dpa"] = vresmap["stat"] = vresmap["uda"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["dpa"] = vresmap["stat"] = vresmap["uda"] = true;
      }
      else if (res == "all")
      {
        vresmap["def"] = vresmap["ac"] = vresmap["cc"] = vresmap["dpa"] = vresmap["stat"] = vresmap["uda"] = true;
      }
      else if (res[0] == 'a' && res[1] == 'c' && res.size() > 2)
      {
        vac = strtodf(res.substr(2)); // 正診率
        if (vac < 0.0 || 1.0 < vac) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["ac"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::search_g(const string& gcode) const // 群コードを検索する
{
  if (gcode == BLANKs) return -1; // 欠損データはヒットさせない
  int idx = -1;
  if (vgvec[0].empty())         // 群1のコード指定がない場合
  {
    idx = 0;
    if (gcode == vgvec[1]) idx = 1;
  }
  else if (gcode == vgvec[0])
  { idx = 0; }
  else if (vgvec[1].empty())    // 群2のコード指定がない場合
  { idx = 1; }
  else if (gcode == vgvec[1])
  { idx = 1; }
  return idx;                   // -1:ヒットせず,0:群1,1:群2
}

class data_less                 // ケースデータの比較用関数オブジェクト
{
private:
  int kidx;                     // ソートキー項目のインデックス
public:
  data_less(int kidx0) : kidx(kidx0) // コンストラクタ
  {}
  ~data_less()                  // デストラクタ
  {}
  bool operator()(datac& cdata1, datac& cdata2) const; // ケースデータを比較する
};

bool data_less::operator()(datac& cdata1, datac& cdata2) const
// ケースデータを比較する
{
  int comp = 0;
  double fdata1 = cdata1.fdata(kidx);
  double fdata2 = cdata2.fdata(kidx);
  if (fdata1 < fdata2) comp = -1;
  else if (fdata1 > fdata2) comp = 1;
  return (comp < 0) ? true : false; // 小さいときだけtrueを返す
}

class diag                      // 診断率クラス
{
private:
  double vpi;                   // 群1の事前確率
  bool vuporder;                // データの昇順/下順の区別
  vecdbl vcovec;                // 境界値
  vecdbl vsnvec;                // 感度
  vecdbl vspvec;                // 特異度
  vecdbl vacvec;                // 正診率
  vecdbl vppvvec;               // PPV
  vecdbl vnpvvec;               // NPV
  double vd2;                   // マハラノビスの汎距離
  double vppr;                  // 誤判別確率
  double vaucn;                 // 理論的AUC
  vector<statv<double>> statvec; // 基礎統計量ベクトル
  void search_acc(ldata& datalst, double acc); // 実測値正診率点の探索
  void cal_uda(double acc);      // 単変量線形判別分析
  double cal_cospmin(double m1, double sd1, double m2, double sd2);
  // 理論的|SN-SP|≒0の時の境界値を計算する
  double cal_coacmax(double m1, double sd1, double m2, double sd2);
  // 理論的正診率最大値の時の境界値を計算する
  double cal_copnmin(double m1, double sd1, double m2, double sd2);
  // 理論的|PPV-NPV|≒0の時の境界値を計算する
  int search_coacc(vecdbl& covec, double acc, double m1, double sd1, double m2, double sd2);
  // 理論的正診率点の探索
  void set_spvec(size_t ncp, double m1, double sd1, double m2, double sd2);
  // 診断率配列をセットする
public:
  diag() : vpi(BLANKf), vuporder(true), vcovec(Nco, BLANKf), vsnvec(Nco, BLANKf), vspvec(Nco, BLANKf), vacvec(Nco, BLANKf), vppvvec(Nco, BLANKf), vnpvvec(Nco, BLANKf), vd2(BLANKf), vppr(BLANKf), vaucn(BLANKf), statvec(2)
  // デフォールトコンストラクタ
  {}
  explicit diag(double pi0) : vpi(pi0), vuporder(true), vcovec(Nco, BLANKf), vsnvec(Nco, BLANKf), vspvec(Nco, BLANKf), vacvec(Nco, BLANKf), vppvvec(Nco, BLANKf), vnpvvec(Nco, BLANKf), vd2(BLANKf), vppr(BLANKf), vaucn(BLANKf), statvec(2)
  // コンストラクタ
  {}
  virtual ~diag()               // デストラクタ
  {}
  diag(const diag& diag0) : vpi(diag0.vpi), vuporder(diag0.vuporder), vcovec(diag0.vcovec), vsnvec(diag0.vsnvec), vspvec(diag0.vspvec), vacvec(diag0.vacvec), vppvvec(diag0.vppvvec), vnpvvec(diag0.vnpvvec), vd2(diag0.vd2), vppr(diag0.vppr), vaucn(diag0.vaucn), statvec(diag0.statvec)
  // コピーコンストラクタ
  {}
  const diag& operator=(const diag& diag0); // コピー代入演算子
  double pi() const             // 群1の事前確率を返す
  { return vpi; }
  double pi(double pi0)         // 群1の事前確率を設定する
  { return vpi = pi0; }
  bool uporder() const          // データの昇順/下順の区別を返す
  { return vuporder; }
  double co(size_t idx = 0) const // 理論的境界値を返す
  { return (idx < Nco) ? vcovec[idx] : 0.0; }
  double sn(size_t idx = 0) const // 理論的感度を返す
  { return (idx < Nco) ? vsnvec[idx] : 0.0; }
  double sp(size_t idx = 0) const // 理論的特異度を返す
  { return (idx < Nco) ? vspvec[idx] : 0.0; }
  double ac(size_t idx = 0) const // 理論的正診率を返す 
  { return (idx < Nco) ? vacvec[idx] : 0.0; }
  double ppv(size_t idx = 0) const // 理論的PPVを返す 
  { return (idx < Nco) ? vppvvec[idx] : 0.0; }
  double npv(size_t idx = 0) const // 理論的NPVを返す 
  { return (idx < Nco) ? vnpvvec[idx] : 0.0; }
  double d2() const             // マハラノビスの汎距離を返す
  { return vd2; }
  double ppr() const            // 誤判別確率を返す
  { return vppr; }
  double aucn() const           // 理論的AUCを返す
  { return vaucn; }
  double fsn(double z) const    // 理論的感度を返す
  { return (vuporder) ? 1.0 - (double)(sndis(z)) : (double)(sndis(z)); }
  double fsp(double z) const    // 理論的特異度を返す
  { return (vuporder) ? (double)(sndis(z)) : 1.0 - (double)(sndis(z)); }
  double fac(double z1, double z2) const // 理論的正診率を返す
  { return vpi * fsn(z1) + (1.0 - vpi) * fsp(z2); }
  double fppv(double sn, double sp) const; // 感度と特異度から陽性予測値を返す
  double fnpv(double sn, double sp) const; // 感度と特異度から陰性予測値を返す
  statv<double>& dstat(size_t idx) // 診断項目の基礎統計量を返す
  { return (idx < 2) ? statvec[idx] : statvec[0]; }
  int cal_dpa(const option& opt, ldata& datalst); // DP-plot分析
  int cal_wtest(double& dev, double& vd, double& t, int& df); // Welchの検定
};

const diag& diag::operator=(const diag& diag0) // コピー代入演算子
{
  if (&diag0 == this) return *this; // 自分自身への代入を防ぐ
  vpi = diag0.vpi;  vuporder = diag0.vuporder;
  for (size_t i = 0; i < Nco; ++i)
  {
    vcovec[i] = diag0.vcovec[i];  vsnvec[i] = diag0.vsnvec[i];
    vspvec[i] = diag0.vspvec[i];  vacvec[i] = diag0.vacvec[i];
    vppvvec[i] = diag0.vppvvec[i];  vnpvvec[i] = diag0.vnpvvec[i];
  }
  vd2 = diag0.vd2;  vppr = diag0.vppr;  vaucn = diag0.vaucn;
  statvec.resize(diag0.statvec.size());
  statvec[0] = diag0.statvec[0];  statvec[1] = diag0.statvec[1];
  return *this;
}

double diag::fppv(double sn, double sp) const // 感度と特異度から陽性予測値を返す
{
  if (sgn(vpi) <= 0) return BLANKf;
  if (sgn(sn) <= 0) return vuporder ? 1.0 : 0.0;
  double ppv = 1.0 / (1.0 + (1.0 - vpi) / vpi * (1.0 - sp) / sn);
  if (sgn(ppv) <= 0) return 0.0;
  return (ppv > 1.0) ? 1.0 : ppv;
}

double diag::fnpv(double sn, double sp) const // 感度と特異度から陰性予測値を返す
{
  if (sgn(1.0 - vpi) <= 0) return BLANKf;
  if (sgn(sp) <= 0) return vuporder ? 1.0 : 0.0;
  double npv = 1.0 / (1.0 + vpi / (1.0 - vpi) * (1.0 - sn) / sp);
  if (sgn(npv) <= 0) return 0.0;
  return (npv > 1.0) ? 1.0 : npv;
}

int diag::cal_dpa(const option& opt, ldata& datalst) // DP-plot分析
{
  for (auto& cdata : datalst) statvec[(size_t)(cdata.fdata(0))].sigma(cdata.fdata(1)); // 群ごとに統計量を集計する
  // for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr) statvec[(size_t)(ditr->fdata(0))].sigma(ditr->fdata(1)); // 群ごとに統計量を集計する
  int n1 = statvec[0].n(), n2 = statvec[1].n();
  int nt = n1 + n2;
  if (nt < 1 || n1 < 1 || n2 < 1) return 0; // 計算できず
  if (opt.lim() == limlow)      // 境界値が下限の時
  { vuporder = true; }
  else if (opt.lim() == limup)  // 境界値が上限の時
  { vuporder = false; }
  else                          // 自動判定
  { vuporder = (statvec[0].mean() >= statvec[1].mean()) ? true : false; }
  if (vuporder == false) datalst.reverse(); // データリストを逆順にする
  if (vpi == BLANKf) vpi = (double)n1 / (double)nt; // vpiを現在のデータの割合で設定する
  if (opt.ac() != BLANKf) search_acc(datalst, opt.ac());
  cal_uda(opt.ac());
  return nt;                    // 全例数を返す
}

void diag::search_acc(ldata& datalst, double acc) // 実測値正診率点の探索
{
  int n1 = statvec[0].n(), n2 = statvec[1].n();
  int n = n1 + n2, tp = n1, dfn = 0, tn = 0, dtn = 0;
  double f = 0.0, fpre = BLANKf;
  int idx0 = 6, didx = 1;
  if (vuporder == false)        // 境界値が上限の時
  { idx0 = 7;  didx = -1; }
  int idx = idx0;
  for (auto& cdata : datalst)
  {
    f = cdata.fdata(1);
    if (f != fpre)
    {
      tp -= dfn;  tn += dtn;
      double sn = (double)tp / (double)n1;
      double sp = (double)tn / (double)n2;
      double ac = (double)(tp + tn) / (double)n;
      if (ac >= acc)
      {
        vcovec[idx] = f;
        vsnvec[idx] = sn;  vspvec[idx] = sp;  vacvec[idx] = ac;
        vppvvec[idx] = (n2 - tn + tp > 0) ? (double)tp / (double)(n2 - tn + tp) : 0.0;
        vnpvvec[idx] = (tn + n1 - tp > 0) ? (double)tn / (double)(tn + n1 - tp) : 0.0;
        if (idx == idx0) idx += didx;
      }
      fpre = f;  dfn = dtn = 0;
    }
    if ((int)(cdata.fdata(0)) == 0) ++dfn;
    else ++dtn;
  }
}

void diag::cal_uda(double acc)  // 単変量線形判別分析
{
  int n1 = statvec[0].n(), n2 = statvec[1].n();
  double m1 = statvec[0].mean(), m2 = statvec[1].mean();
  double v1 = statvec[0].vr(), v2 = statvec[1].vr();
  double sd1 = statvec[0].sd(), sd2 = statvec[1].sd();
  const double Veps = DBL_EPSILON * fmax(5.0, v1 + v2);
  if (v1 + v2 < Veps) return; // データが定数の時
  vcovec[0] = (m1 + m2) * 0.5;   // 判別分析の境界値
  if (sgndiff(m1, m2) == 0 || (sgndiff(vpi, 0.5) == 0 && sgndiff(sd1, sd2) == 0))
  { vcovec[1] = vcovec[2] = vcovec[3] = vcovec[0]; } // 判別分析の境界値
  else
  {
    vcovec[1] = cal_cospmin(m1, sd1, m2, sd2);
    vcovec[2] = cal_coacmax(m1, sd1, m2, sd2);
    vcovec[3] = cal_copnmin(m1, sd1, m2, sd2);
    if (acc != BLANKf)
    {
      double acmax = (vcovec[2] == BLANKf) ? 0.0 : fac((vcovec[2] - m1) / sd1, (vcovec[2] - m2) / sd2);
      if (acc <= acmax)
      {
        vecdbl covec(2);
        if (search_coacc(covec, acc, m1, sd1, m2, sd2) > 0)
        { vcovec[4] = covec[0];  vcovec[5] = covec[1]; }
      }
    }
  }
  vd2 = 2.0 * (m1 - m2) * (m1 - m2) / (v1 + v2);
  double z = sqrt(vd2 * 0.5);
  vaucn = (double)(sndis(z));   // 理論AUC
  z = sqrt(vd2) * 0.5;
  vppr = 1.0 - (double)(sndis(z)); // 誤判別確率
  if ((m1 < m2 && vuporder) || (m1 > m2 && vuporder == false)) vppr = 1.0 - vppr;
  set_spvec(Nco - 2, m1, sd1, m2, sd2); // 感度・特異度配列を設定する
}

double diag::cal_cospmin(double m1, double sd1, double m2, double sd2)
// 理論的|SN-SP|≒0の時の境界値を返す
{
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  if (sd1 < SDeps || sd2 < SDeps) return (m1 + m2) * 0.5;
  return (m1 * sd2 + m2 * sd1) / (sd1 + sd2);
}

double diag::cal_coacmax(double m1, double sd1, double m2, double sd2)
// 理論的正診率最大値の時の境界値を計算する
{
  double v1 = sd1 * sd1, v2 = sd2 * sd2;
  double vd = v2 - v1;
  if (sgndiff(v2, v1) == 0)     // 2群の分散が等しい時
  {
    return (m1 + m2) * 0.5 + v1 * log((1.0 - vpi) / vpi) / (m1 - m2);
  }
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  if (sd1 < SDeps || sd2 < SDeps) return BLANKf; // 計算不可能
  double md = m1 * v2 - m2 * v1;
  double tmp = md * md - vd * (m1 * m1 * v2 - m2 * m2 * v1 + 2.0 * v1 * v2 * log((1.0 - vpi) / vpi * sd1 / sd2));
  double com = (sgn(tmp) >= 0 && sgn(vd) != 0) ? (md - sqrt(tmp)) / vd : BLANKf;
  double cop = (sgn(tmp) >= 0 && sgn(vd) != 0) ? (md + sqrt(tmp)) / vd : BLANKf;
  return (vuporder) ? com : cop; // 理論値を返す
}

int diag::search_coacc(vecdbl& covec, double acc, double m1, double sd1, double m2, double sd2)
// 理論的正診率点の探索
{
  constexpr double SDx = 4.0;         // SDの倍数
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  if (sd1 < SDeps || sd2 < SDeps) return 0; // データが定数の時は計算できず
  double comin = m2 - sd2 * SDx, comax = m1 + sd1 * SDx;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { comin = m1 - sd1 * SDx;  comax = m2 + sd2 * SDx; }
  constexpr uint RDdef = 1000;  // 分割数
  const double Dco = (comax - comin) / (double)RDdef;
  covec[0] = covec[1] = BLANKf;
  int idx = 0;
  double co = comin;
  for (uint i = 0; i <= RDdef; ++i, co += Dco)
  {
    double ac = fac((co - m1) / sd1, (co - m2) / sd2);
    if (ac >= acc)
    {
      covec[idx] = co;
      if (idx == 0) ++idx;
    }
  }
  int nco = 0;
  if (covec[0] == BLANKf) covec[0] = comin;
  else nco = 1;
  if (covec[1] == BLANKf) covec[1] = comax;
  else ++nco;
  return nco;                   // 境界値数を返す
}

double diag::cal_copnmin(double m1, double sd1, double m2, double sd2)
// 理論的|PPV-NPV|≒0の時の境界値を計算する
{
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  if (sd1 < SDeps || sd2 < SDeps) return BLANKf; // 計算不可能
  constexpr double SDx = 5.0;   // 範囲を決めるためのSDの倍率
  double co = (m1 + m2) * 0.5;
  double z1 = (co - m1) / sd1, z2 = (co - m2) / sd2;
  double sn = fsn(z1), sp = fsp(z2);
  double pr = fppv(sn, sp) - fnpv(sn, sp);
  const double COeps = DBL_EPSILON * fmax(5.0, fabs(co));
  const double Delta = fabs(co) * 0.00001; // 刻み幅
  double comin = m2 - sd1 * SDx, comax = m1 + sd2 * SDx;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { comin = m1 - sd1 * SDx;  comax = m2 + sd2 * SDx; }
  while (fabs(pr) >= Rerror)
  {
    double dco = Delta * co;
    double dco1 = dco / sd1, dco2 = dco / sd2;
    double snd = fsn(z1 + dco1), spd = fsp(z2 + dco2);
    double prd = fppv(snd, spd) - fnpv(snd, spd);
    double slope = (prd - pr) / dco;
    if (fabs(slope) < COeps) break; // 収束
    co -= pr / slope;
    if (co < comin || comax < co) // 範囲外
    {
      co = BLANKf;
      break;
    }
    z1 = (co - m1) / sd1;  z2 = (co - m2) / sd2;
    sn = fsn(z1);  sp = fsp(z2);
    pr = fppv(sn, sp) - fnpv(sn, sp);
  }
  if (co == BLANKf) return co;  // 計算できない時
  return co;                    // 境界値を返す
}

void diag::set_spvec(size_t ncp, double m1, double sd1, double m2, double sd2)
// 診断率配列をセットする
{
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  for (size_t i = 0; i < ncp; ++i)
  {
    if (vcovec[i] != BLANKf)
    {
      double sn = 0.0, sp = 0.0, z1 = 0.0, z2 = 0.0;
      if (sd1 > SDeps)
      {
        z1 = (vcovec[i] - m1) / sd1;
        sn = fsn(z1);
      }
      else
      {
        z1 = m1;
        if (vuporder) sn = (m1 < vcovec[i]) ? 0.0 : 1.0;
        else sn = (vcovec[i] < m1) ? 0.0 : 1.0;
      }
      if (sd2 > SDeps)
      {
        z2 = (vcovec[i] - m2) / sd2;
        sp = fsp(z2);
      }
      else
      {
        z2 = m2;
        if (vuporder) sp = (vcovec[i] < m2) ? 0.0 : 1.0;
        else sp = (m2 < vcovec[i]) ? 0.0 : 1.0;
      }
      vsnvec[i] = sn;  vspvec[i] = sp;  vacvec[i] = fac(z1, z2);
      vppvvec[i] = fppv(sn, sp);  vnpvvec[i] = fnpv(sn, sp);
    }
  }
}

int diag::cal_wtest(double& dev, double& vd, double& t, int& df)
// Welchの検定結果を返す
{
  dev = fabs(statvec[0].mean() - statvec[1].mean());
  int n1 = statvec[0].n(), n2 = statvec[1].n();
  double v1 = statvec[0].vr() / (double)n1;
  double v2 = statvec[1].vr() / (double)n2;
  vd = v1 + v2;
  df = (vd > 0.0) ? (int)(vd * vd / (v1 * v1 / (n1 - 1) + v2 * v2 / (n2 -1))) : n1 + n2  - 2;
  t = (vd > 0.0) ? dev / sqrt(vd) : 0.0;
  return df;                    // 自由度を返す
}

class dpplotclass               // dpplotメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 分析結果のタイトルとデータ識別情報を出力する
  void put_result(ostream& fout, option& opt, idd& id, ldata& datalst, diag& dg) const;
  // 分析結果を出力する
  void put_stat(ostream& fout, const option& opt, diag& dg) const;
  // 基礎統計量を出力する
  void put_dpa(ostream& fout, const diag& dg) const;
  // DP-plot分析の結果を出力する
  void put_acc(ostream& fout, const option& opt, const diag& dg) const;
  // 指定した正診率になる時の診断率を出力する
  void put_uda(ostream& fout, option& opt, diag& dg) const;
  // 単変量線型判別分析の結果とWelchの検定結果を出力する
  void plot_ngp(option& opt, idd& id, ldata& datalst, diag& dg) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ac(ofstream& fout, const option& opt, ldata& datalst, diag& dg) const;
  // Ngraph用正診率曲線を出力する
  void plot_acn(ofstream& fout, const option& opt, diag& dg) const;
  // Ngraph用理論的正診率曲線を出力する
  void plot_log(ofstream& fout, diag& dg) const;
  // Ngraph用ロジスティック曲線を出力する
  void plot_logci(ofstream& fout, const option& opt, diag& dg) const;
  // Ngraph用ロジスティック曲線の信頼区間を出力する
  void plot_pnfn(ofstream& fout, diag& dg) const;
  // Ngraph用理論的陽性率/陰性率曲線を出力する 
  void plot_pv(ofstream& fout, ldata& datalst, diag& dg) const;
  // Ngraph用的PPV/NPV曲線を出力する
  void plot_pvn(ofstream& fout, diag& dg) const;
  // Ngraph用理論的PPV/NPV曲線を出力する
  void plot_pvnci(ofstream& fout, const option& opt, diag& dg) const;
  // Ngraph用理論的PPV/NPV曲線の信頼区間を出力する
  void plot_snp(ofstream& fout, const option& opt, ldata& datalst, diag& dg) const;
  // Ngraph用感度/特異度曲線を出力する
  void plot_snpn(ofstream& fout, const option& opt, diag& dg) const;
  // Ngraph用理論的感度/特異度曲線を出力する
public:
  dpplotclass() : fname()       // コンストラクタ
  {}
  ~dpplotclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み累積する
  void put(option& opt, idd& id, ldata& datalst) const; // 集計結果を出力する
};

void dpplotclass::put_resid(ostream& fout, idd& id) const
// 分析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM   :" << id.item(0) << "\n";
  fout << "ANALYSIS ITEM:" << id.item(1) << "\n";
}

void dpplotclass::put_result(ostream& fout, option& opt, idd& id, ldata& datalst, diag& dg) const
// 分析結果を出力する
{
  constexpr int LENline = 74;   // 区切り線の長さ
  put_resid(fout, id);
  bool putres = false;
  if (opt.res("stat"))
  {
    fout << "\n";
    putline(fout, LENline);
    put_stat(fout, opt, dg);
    putres = true;
  }
  if (opt.res("dpa"))
  {
    if (putres == false) fout << "\n";
    putline(fout, LENline);
    put_dpa(fout, dg);
    putres = true;
  }
  if (opt.res("ac"))
  {
    if (putres == false) fout << "\n";
    putline(fout, LENline);
    put_acc(fout, opt, dg);
    putres = true;
  }
  if (opt.res("uda"))
  {
    if (putres == false) fout << "\n";
    putline(fout, LENline);
    put_uda(fout, opt, dg);
    putres = true;
  }
  if (putres) putline(fout, LENline);
  fout << "\n" << endl;
}

void dpplotclass::put_stat(ostream& fout, const option& opt, diag& dg) const
// 基礎統計量を出力する
{
  fout << "GROUP1(Disease):Code=";
  if ((opt.g(0)).empty()) fout << "other then " << opt.g(1) << "\n";
  else fout << opt.g(0) << "\n";
  fout << "  ";
  putmsd(fout, (dg.dstat(0)).n(), (dg.dstat(0)).mean(), (dg.dstat(0)).vr());
  fout << "GROUP2(Normal) :Code=";
  if ((opt.g(1)).empty()) fout << "other then " << opt.g(0) << "\n";
  else fout << opt.g(1) << "\n";
  fout << "  ";
  putmsd(fout, (dg.dstat(1)).n(), (dg.dstat(1)).mean(), (dg.dstat(1)).vr());
}

void dpplotclass::put_dpa(ostream& fout, const diag& dg) const
// DP-plot分析の結果を出力する
{
  fout << "SS-point(sensitivity=specificity)=";
  if (dg.co(1) == BLANKf)
  { fout << "not exist\n"; }
  else
  {
    string tstr("More than ");
    if (dg.uporder() == false) tstr = "Less than ";
    fout << tstr << dg.co(1) << " is GROUP1\n";
    fout << "  sensitivity=" << dg.sn(1) << "  specificity=" << dg.sp(1) << "  accuracy=" << dg.ac(1) << "\n";
    fout << "  PPV=" << dg.ppv(1) << "  NPV=" << dg.npv(1) << "\n";
  }
  fout << "AC-point(Maximum Accuracy)=";
  if (dg.co(2) == BLANKf)
  { fout << "not exist\n"; }
  else
  {
    fout << dg.co(2) << "(Prior Probability of GROUP1=" << dg.pi() << ")\n";
    fout << "  sensitivity=" << dg.sn(2) << "  specificity=" << dg.sp(2) << "  accuracy=" << dg.ac(2) << "\n";
    fout << "  PPV=" << dg.ppv(2) << "  NPV=" << dg.npv(2) << "\n";
  }
  fout << "PN-point(PPV=NPV)=";
  if (dg.co(3) == BLANKf)
  { fout << "not exist" << endl; }
  else
  {
    fout << dg.co(3) << "(Prior Probability of GROUP1=" << dg.pi() << ")\n";
    fout << "  sensitivity=" << dg.sn(3) << "  specificity=" << dg.sp(3) << "  accuracy=" << dg.ac(3) << "\n";
    fout << "  PPV=" << dg.ppv(3) << "  NPV=" << dg.npv(3) << endl;
  }
}

void dpplotclass::put_acc(ostream& fout, const option& opt, const diag& dg) const
// 指定した正診率になる時の診断率を出力する
{
  fout << "Minimum Point at Which the Theoretical Accuracy is more than " << opt.ac() << "=";
  if (dg.co(4) == BLANKf)
  { fout << "not exist\n"; }
  else
  {
    fout << dg.co(4) << "\n";
    fout << "  sensitivity=" << dg.sn(4) << "  specificity=" << dg.sp(4) << "  accuracy=" << dg.ac(4) << "\n";
    fout << "  PPV=" << dg.ppv(4) << "  NPV=" << dg.npv(4) << "\n";
  }
  fout << "Maximum Point at Which the Theoretical Accuracy is more then " << opt.ac() << "=";
  if (dg.co(5) == BLANKf)
  { fout << "not exist\n"; }
  else
  {
    fout << dg.co(5) << "\n";
    fout << "  sensitivity=" << dg.sn(5) << "  specificity=" << dg.sp(5) << "  accuracy=" << dg.ac(5) << "\n";
    fout << "  PPV=" << dg.ppv(5) << "  NPV=" << dg.npv(5) << "\n";
  }
  fout << "Minimum Point at Which the Accuracy is more than " << opt.ac() << "=";
  if (dg.co(6) == BLANKf)
  { fout << "not exist\n"; }
  else
  {
    fout << dg.co(6) << "\n";
    fout << "  sensitivity=" << dg.sn(6) << "  specificty=" << dg.sp(6) << "  accuracy=" << dg.ac(6) << "\n";
    fout << "  PPV=" << dg.ppv(6) << "  NPV=" << dg.npv(6) << "\n";
  }
  fout << "Maximum Point at Which the Accuracy is more then " << opt.ac() << "=";
  if (dg.co(7) == BLANKf)
  { fout << "not exist" << endl; }
  else
  {
    fout << dg.co(7) << "\n";
    fout << "  sensitivity=" << dg.sn(7) << "  specificity=" << dg.sp(7) << "  accuracy=" << dg.ac(7) << "\n";
    fout << "  PPV=" << dg.ppv(7) << "  NPV=" << dg.npv(7) << endl;
  }
}

void dpplotclass::put_uda(ostream& fout, option& opt, diag& dg) const
// 単変量線型判別分析の結果とWelchの検定結果を出力する
{
  string tstr("More then ");
  if (dg.uporder() == false) tstr = "Less then ";
  fout << "Univariate Linear Discriminant Analysis:" << tstr << dg.co(0) << " is GROUP1\n";
  fout << "  sencitivity=" << dg.sn(0) << "  specificity=" << dg.sp(0) << "  accuracy=" << dg.ac(0) << "\n";
  fout << "  PPV=" << dg.ppv(0) << "  NPV=" << dg.npv(0) << "\n";
  double n1 = (double)((dg.dstat(0)).n()), n2 = (double)((dg.dstat(1)).n());
  double f = dg.d2() * n1 * n2 / (n1 + n2);
  double r2 = f / (f + n1 + n2 - 2.0);
  fout << "Mahalanobis's D-squared=" << dg.d2() << "  Correlation Ratio=" << sqrt(r2) << "\n";
  fout << "  discriminant-p=phi(D/2)=phi(" << sqrt(dg.d2()) * 0.5 << ")=" << 1.0 - dg.ppr() << "(Probrem-p=1-p=" << dg.ppr() << ")\n";
  fout << "  AUC of ROC Curve=phi(D/square root 2)=phi(" << sqrt(dg.d2() * 0.5) << ")=" << dg.aucn() << "\n";
  double dev = 0.0, vd = 0.0, t = 0.0;
  int df = 0;
  dg.cal_wtest(dev, vd, t, df);
  ldbl pr = ttop(t, df);
  fout << "Welch's Test:t=" << t << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
  if (opt.res("cc"))
  {
    double ci = calmcci(mcfis, opt.cc(), 2, 1, df, 1, vd);
    fout << "  " << opt.cc() * 100.0 << "%CI of the Mean Difference=";
    fout << dev << " +/- " << ci << "(" << dev - ci << "-" << dev + ci << ")\n";
  }
}

void dpplotclass::plot_ngp(option& opt, idd& id, ldata& datalst, diag& dg) const
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("dpp") || opt.ngp("snp"))
  {
    plot_ngpid(fout, id);
    plot_snp(fout, opt, datalst, dg);
    fout << NGPdelm << "\n";
    plot_snpn(fout, opt, dg);
    delml = true;
  }
  if (opt.ngp("dpp") || opt.ngp("ac"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    if (opt.ngp("ac"))
    {
      plot_ac(fout, opt, datalst, dg);
      fout << NGPdelm << "\n";
    }
    plot_acn(fout, opt, dg);
  }
  if (opt.ngp("pv"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_pv(fout, datalst, dg);
    fout << NGPdelm << "\n";
    plot_pvn(fout, dg);
    if (0.0 <= opt.cc() && opt.cc() < 1.0)
    {
      fout << NGPdelm << "\n";
      plot_pvnci(fout, opt, dg);
    }
  }
  if (opt.ngp("pnf"))
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_pnfn(fout, dg);
  }
  if (opt.ngp("log"))
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80 -zy0,1,0.2\n";
    plot_ngpid(fout, id);
    plot_log(fout, dg);
    if (0.0 <= opt.cc() && opt.cc() < 1.0)
    {
      fout << NGPdelm << "\n";
      plot_logci(fout, opt, dg);
    }
  }
}

void dpplotclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用タイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%GROUP ITEM   :" << id.item(0) << "\n";
  fout << "%ANALYSIS ITEM:" << id.item(1) << "\n";
}

void dpplotclass::plot_ac(ofstream& fout, const option& opt, ldata& datalst, diag& dg) const
// Ngraph用正診率曲線を出力する
{
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  int tp = n1, tn = 0;
  double n = (double)(n1 + n2);
  double pi1 = (double)n1 / (double)n;
  fout << "%<ACCURACY CURVE:PRIOR PROBABILITY OF GROUP1=" << pi1 << ">\n";
  double f = 0.0, ac = 0.0;
  for (auto& cdata : datalst)
  {
    f = cdata.fdata(1);
    ac = (double)(tp + tn) / n;
    fout << f << " " << ac << "\n";
    if ((int)(cdata.fdata(0)) == 0) --tp;
    else ++tn;
  }
  ac = (double)n2 / n;
  fout << f << " " << ac << "\n";
  fout << "= =\n";
  pi1 = (0.0 < opt.pi() && opt.pi() < 1.0) ? opt.pi() : 0.5;
  fout << "%<ACCURACY CURVE:PRIOR PROBABILITY OF GROUP1=" << pi1 << ">\n"; // 指定された事前確率
  tp = n1;  tn = 0;
  for (auto& cdata : datalst)
  {
    f = cdata.fdata(1);
    double sn = (double)tp / (double)n1;
    double sp = (double)tn / (double)n2;
    ac = pi1 * sn + (1.0 - pi1) * sp;
    fout << f << " " << ac << "\n";
    if ((int)(cdata.fdata(0)) == 0) --tp;
    else ++tn;
  }
  ac = 1.0 - pi1;
  fout << f << " " << ac << endl;
}

void dpplotclass::plot_acn(ofstream& fout, const option& opt, diag& dg) const
// Ngraph用理論的正診率曲線を出力する
{
  constexpr double SDx = 4.0;   // SDの倍数
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  double m1 = (dg.dstat(0)).mean(), sd1 = (dg.dstat(0)).sd();
  double m2 = (dg.dstat(1)).mean(), sd2 = (dg.dstat(1)).sd();
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 +sd2);
  if (sd1 < SDeps || sd2 < SDeps) return; // データが定数の時
  double xmin = m2 - sd2 * SDx, xmax = m1 + sd1 * SDx;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { xmin = m1 - sd1 * SDx;  xmax = m2 + sd2 * SDx; }
  double dx = (xmax - xmin) / (double)RDdef;
  double pi0 = dg.pi();         // 現在の事前確率を保存
  double pi1 = (double)n1 / (double)(n1 + n2);
  dg.pi(pi1);
  fout << "%<THEORETICAL ACCURACY CURVE:PRIOR PROBABILITY OF GROUP1=" << pi1 << ">\n";
  double x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  { fout << x << " " << dg.fac((x - m1) / sd1, (x - m2) / sd2) << "\n"; }
  fout << "= =\n";
  pi1 = (0.0 < opt.pi() && opt.pi() < 1.0) ? opt.pi() : 0.5;
  dg.pi(pi1);
  fout << "%<THEORETICAL ACCURACY CURVE:PRIOR PROBABILITY OF GROUP1=" << pi1 << ">\n"; // 指定された事前確率
  x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  { fout << x << " " << dg.fac((x - m1) / sd1, (x - m2) / sd2) << "\n"; }
  dg.pi(pi0);                   // 事前確率を元に戻す
}

void dpplotclass::plot_log(ofstream& fout, diag& dg) const
// Ngraph用ロジスティック曲線を出力する
{
  constexpr double SDx = 4.0;   // SDの倍数
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  double m1 = (dg.dstat(0)).mean(), sd1 = (dg.dstat(0)).sd();
  double m2 = (dg.dstat(1)).mean(), sd2 = (dg.dstat(1)).sd();
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  if (sd1 + sd2 < SDeps) return; // データが定数の時
  double xmin = m2 - sd2 * SDx, xmax = m1 + sd1 * SDx;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { xmin = m1 - sd1 * SDx;  xmax = m2 + sd2 * SDx; }
  double dx = (xmax - xmin) / (double)RDdef;
  double orpi = (1.0 - dg.pi()) / dg.pi();
  fout << "%<LOGISTIC CURVE:PRIOR PROBABILITY OF GROUP1=" << dg.pi() << ">\n";
  double x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double p1 = snden((x - m1) / sd1), p2 = snden((x - m2) / sd2);
    double pr = 1.0 / (1.0 + orpi * p2 / p1);
    if (dg.uporder() == false) pr = 1.0 - pr;
    fout << x << " " << pr << "\n";
  }
}

void dpplotclass::plot_logci(ofstream& fout, const option& opt, diag& dg) const
// Ngraph用ロジスティック曲線の信頼区間を出力する
{
  constexpr double SDx = 4.0;   // SDの倍数
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  double m1 = (dg.dstat(0)).mean(), sd1 = (dg.dstat(0)).sd();
  double m2 = (dg.dstat(1)).mean(), sd2 = (dg.dstat(1)).sd();
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  if (sd1 + sd2 < SDeps) return; // データが定数の時
  double xmin = m2 - sd2 * SDx, xmax = m1 + sd1 * SDx;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { xmin = m1 - sd1 * SDx;  xmax = m2 + sd2 * SDx; }
  double dx = (xmax - xmin) / (double)RDdef;
  double orpi = (1.0 - dg.pi()) / dg.pi();
  double t = ptoz(1.0 - opt.cc());
  fout << "%<LOWER LIMIT OF " << opt.cc() * 100.0 << "%CI OF THE LOGISTIC CURVE:PRIOR PROBABILITY OF GROUP1=" << dg.pi() << ">\n";
  double x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double p1 = snden((x - m1) / sd1), p2 = snden((x - m2) / sd2);
    double p = 1.0 / (1.0 + orpi * p2 / p1);
    double pl = p - t * sqrt(p * (1.0 - p) / (double)(n1 + n2));
    if (pl < 0.0) pl = 0.0;
    if (dg.uporder() == false) pl = 1.0 - pl;
    fout << x << " " << pl << "\n";
  }
  fout << "= =\n";
  fout << "%<UPPER LIMIT OF " << opt.cc() * 100.0 << "%CI OF THE LOGISTIC CURVE:PRIOR PROBABILITY OF GROUP1=" << dg.pi() << ">\n";
  x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double p1 = snden((x - m1) / sd1), p2 = snden((x - m2) / sd2);
    double p = 1.0 / (1.0 + orpi * p2 / p1);
    double pu = p + t * sqrt(p * (1.0 - p) / (double)(n1 + n2));
    if (pu > 1.0) pu = 1.0;
    if (dg.uporder() == false) pu = 1.0 - pu;
    fout << x << " " << pu << "\n";
  }
}

void dpplotclass::plot_pnfn(ofstream& fout, diag& dg) const
// Ngraph用理論的陽性率/陰性率曲線を出力する
{
  constexpr double SDx = 4.0;   // SDの倍数
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  double m1 = (dg.dstat(0)).mean(), sd1 = (dg.dstat(0)).sd();
  double m2 = (dg.dstat(1)).mean(), sd2 = (dg.dstat(1)).sd();
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  if (sd1 < SDeps || sd2 < SDeps) return; // データが定数の時
  double xmin = m2 - sd2 * SDx, xmax = m1 + sd1 * SDx;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { xmin = m1 - sd1 * SDx;  xmax = m2 + sd2 * SDx; }
  double dx = (xmax - xmin) / (double)RDdef;
  fout << "%<THEORETICAL POSITIVE RATE CURVE:PRIOR PROBABILITY OF GROUP1=" << dg.pi() << ">\n";
  double x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sn = dg.fsn((x - m1) / sd1);
    double sp = dg.fsp((x - m2) / sd2);
    double pf = dg.pi() * sn + (1.0 - dg.pi()) * (1.0 - sp);
    fout << x << " " << pf << "\n";
  }
  fout << "= =\n";
  fout << "%<THEORETICAL NEGATIVE RATE CURVE:PRIOR PROBABILITY OF GROUP2=" << 1.0 - dg.pi() << ">\n";
  x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sn = dg.fsn((x - m1) / sd1);
    double sp = dg.fsp((x - m2) / sd2);
    double nf = dg.pi() * (1.0 - sn) + (1.0 - dg.pi()) * sp;
    fout << x << " " << nf << "\n";
  }
  fout << "= =\n";
  fout << "%<THEORETICAL DISEASE RATE CURVE:PRIOR PROBABILITY OF GROUP1=" << dg.pi() << ">\n";
  x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sn = dg.fsn((x - m1) / sd1);
    fout << x << " " << dg.pi() * sn << "\n";
  }
  fout << "= =\n";
  fout << "%<THEORETICAL NORMAL RATE CURVE:PRIOR PROBABILITY OF GROUP2=" << 1.0 - dg.pi() << ">\n";
  x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sp = dg.fsp((x - m2) / sd2);
    fout << x << " " << (1.0 - dg.pi()) * sp << "\n";
  }
}

void dpplotclass::plot_pv(ofstream& fout, ldata& datalst, diag& dg) const
// Ngraph用PPV/NPV曲線を出力する
{
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  int n = n1 + n2, tp = n1, tn = 0;
  double f = 0.0, ppv = 0.0;
  fout << "%<PPV CURVE:PRIOR PROBABILITY OF GROUP1=" << dg.pi() << ">\n";
  for (auto& cdata : datalst)
  {
    f = cdata.fdata(1);
    double sn = (double)tp / (double)n1;
    double sp = (double)tn / (double)n2;
    ppv = dg.fppv(sn, sp);
    fout << f << " " << ppv << "\n";
    if ((int)(cdata.fdata(0)) == 0) --tp;
    else ++tn;
  }
  tp = n1;  tn = 0;
  double npv = 0.0;
  fout << "= =\n";
  fout << "%<NPV CURVE:PRIOR PROBABILITY OF GROUP2=" << 1.0 - dg.pi() << ">\n";
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    f = ditr->fdata(1);
    double sn = (double)tp / (double)n1;
    double sp = (double)tn / (double)n2;
    npv = dg.fnpv(sn, sp);
    fout << f << " " << npv << "\n";
    if ((int)(ditr->fdata(0)) == 0) --tp;
    else ++tn;
  }
}

void dpplotclass::plot_pvn(ofstream& fout, diag& dg) const
// Ngraph用理論的PPV/NPV曲線を出力する
{
  constexpr double SDx = 4.0;   // SDの倍数
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  double m1 = (dg.dstat(0)).mean(), sd1 = (dg.dstat(0)).sd();
  double m2 = (dg.dstat(1)).mean(), sd2 = (dg.dstat(1)).sd();
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  if (sd1 < SDeps || sd2 < SDeps) return; // データが定数の時
  double xmin = m2 - sd2 * SDx, xmax = m1 + sd1 * SDx;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { xmin = m1 - sd1 * SDx;  xmax = m2 + sd2 * SDx; }
  double dx = (xmax - xmin) / (double)RDdef;
  fout << "%<THEORETICAL PPV CURVE:PRIOR PROBABILITY OF GROUP1=" << dg.pi() << ">\n";
  double x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sn = dg.fsn((x - m1) / sd1);
    double sp = dg.fsp((x - m2) / sd2);
    double ppv = dg.fppv(sn, sp);
    fout << x << " " << ppv << "\n";
  }
  fout << "= =\n";
  fout << "%<THEORETICAL NPV CURVE:PRIOR PROBABILITY OF GROUP2=" << 1.0 - dg.pi() << ">\n";
  x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sn = dg.fsn((x - m1) / sd1);
    double sp = dg.fsp((x - m2) / sd2);
    double npv = dg.fnpv(sn, sp);
    fout << x << " " << npv << "\n";
  }
}

void dpplotclass::plot_pvnci(ofstream& fout, const option& opt, diag& dg) const
// Ngraph用理論的PPV/NPV曲線の信頼区間を出力する
{
  constexpr double SDx = 4.0;   // SDの倍数
   int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  double m1 = (dg.dstat(0)).mean(), sd1 = (dg.dstat(0)).sd();
  double m2 = (dg.dstat(1)).mean(), sd2 = (dg.dstat(1)).sd();
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  if (sd1 < SDeps || sd2 < SDeps) return; // データが定数の時
  double xmin = m2 - sd2 * SDx, xmax = m1 + sd1 * SDx;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { xmin = m1 - sd1 * SDx;  xmax = m2 + sd2 * SDx; }
  double dx = (xmax - xmin) / (double)RDdef;
  double t = ptoz(1.0 - opt.cc());
  fout << "%<LOWER OF " << opt.cc() * 100.0 << "%CI OF THE PPV CURVE:PRIOR PROBABILITY OF GROUP1=" << dg.pi() << ">\n";
  double x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sn = dg.fsn((x - m1) / sd1);
    double sp = dg.fsp((x - m2) / sd2);
    double ppv = dg.fppv(sn, sp);
    double ppvl = ppv - t * sqrt(ppv * (1.0 - ppv) / (double)(n1 + n2));
    if (ppvl < 0.0) ppvl = 0.0;
    fout << x << " " << ppvl << "\n";
  }
  fout << "= =\n";
  fout << "%<UPPER OF " << opt.cc() * 100.0 << "%CI OF THE PPV CURVE:PRIOR PROBABILITY OF GROUP1=" << dg.pi() << ">\n";
  x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sn = dg.fsn((x - m1) / sd1);
    double sp = dg.fsp((x - m2) / sd2);
    double ppv = dg.fppv(sn, sp);
    double ppvu = ppv + t * sqrt(ppv * (1.0 - ppv) / (double)(n1 + n2));
    if (ppvu > 1.0) ppvu = 1.0;
    fout << x << " " << ppvu << "\n";
  }
  fout << "= =\n";
  fout << "%<LOWER OF " << opt.cc() * 100.0 << "%CI OF THE NPV CURVE:PRIOR PROBABILITY OF GROUP2=" << 1.0 - dg.pi() << ">\n";
  x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sn = dg.fsn((x - m1) / sd1);
    double sp = dg.fsp((x - m2) / sd2);
    double npv = dg.fnpv(sn, sp);
    double npvl = npv - t * sqrt(npv * (1.0 - npv) / (double)(n1 + n2));
    if (npvl < 0.0) npvl = 0.0;
    fout << x << " " << npvl << "\n";
  }
  fout << "= =\n";
  fout << "%<UPPER OF " << opt.cc() * 100.0 << "%CI OF THE NPV CURVE:PRIOR PROBABILITY OF GROUP2=" << 1.0 - dg.pi() << ">\n";
  x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sn = dg.fsn((x - m1) / sd1);
    double sp = dg.fsp((x - m2) / sd2);
    double npv = dg.fnpv(sn, sp);
    double npvu = npv + t * sqrt(npv * (1.0 - npv) / (double)(n1 + n2));
    if (npvu > 1.0) npvu = 1.0;
    fout << x << " " << npvu << "\n";
  }
}

void dpplotclass::plot_snp(ofstream& fout, const option& opt, ldata& datalst, diag& dg) const
// Ngraph用感度/特異度曲線を出力する
{
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  int n = n1 + n2, tp = n1, tn = 0;
  fout << "%<SENCITIVITY CURVE GROUP1:CODE=";
  if ((opt.g(0)).empty()) fout << "other then " << opt.g(1) << ">\n";
  else fout << opt.g(0) << ">\n";
  double f = 0.0;
  for (auto& cdata : datalst)
  {
    if ((int)(cdata.fdata(0)) == 0) // 群1
    {
      f = cdata.fdata(1);
      double sn = (double)tp / (double)n1;
      fout << f << " " << sn << "\n";
      --tp;
    }
  }
  fout << f << " 0\n";
  fout << "= =\n";
  fout << "%<SPECIFICITY CURVE GROUP2:CODE=";
  if ((opt.g(1)).empty()) fout << "other then " << opt.g(0) << ">\n";
  else fout << opt.g(1) << ">\n";
  tp = n1;  tn = 0;
  for (auto& cdata : datalst)
  {
    if ((int)(cdata.fdata(0)) == 1) // 群2
    {
      f = cdata.fdata(1);
      double sp = (double)tn / (double)n2;
      fout << f << " " << sp << "\n"; 
      ++tn;
    }
  }
  fout << f << " 1\n";
}

void dpplotclass::plot_snpn(ofstream& fout, const option& opt, diag& dg) const
// Ngraph用理論的感度／特異度曲線を出力する
{
  constexpr double SDx = 4.0;   // SDの倍数
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  double m1 = (dg.dstat(0)).mean(), sd1 = (dg.dstat(0)).sd();
  double m2 = (dg.dstat(1)).mean(), sd2 = (dg.dstat(1)).sd();
  const double SDeps = DBL_EPSILON * fmax(5.0, sd1 + sd2);
  if (sd1 < SDeps || sd2 < SDeps) return; // データが定数の時
  double xmin = m2 - sd2 * SDx, xmax = m1 + sd1 * SDx;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { xmin = m1 - sd1 * SDx;  xmax = m2 + sd2 * SDx; }
  double dx = (xmax - xmin) / (double)RDdef;
  fout << "%<THEORETICAL SENCITIVITY CURVE GROUP1:CODE=";
  if ((opt.g(0)).empty()) fout << "other then " << opt.g(1) << ">\n";
  else fout << opt.g(0) << ">\n";
  double x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sn = dg.fsn((x - m1) / sd1);
    fout << x << " " << sn << "\n";
  }
  fout << "= =\n";
  fout << "%<THEORETICAL SPECIFICITY CURVE GROUP2=";
  if ((opt.g(1)).empty()) fout << "other then " << opt.g(0) << ">\n";
  else fout << opt.g(1) << ">\n";
  x = xmin;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double sp = dg.fsp((x - m2) / sd2);
    fout << x << " " << sp << "\n";
  }
}

int dpplotclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  nc = (int)(id.nc());          // ケース数変更用
  ldata_itr ditr = datalst.begin();
  while (ditr != datalst.end()) // リストの要素を削除するのでforではなくwhile
  {
    int idx = opt.search_g(ditr->sdata(0));
    double fdata = ditr->fdata(1);
    if (idx < 0 || fdata == BLANKf) // 欠損データはリストから削除する
    { ditr = datalst.erase(ditr);  --nc; }
    else
    {
      ditr->assign(0, (double)idx); // 群コードを群インデックスに変える
      ++ditr;
    }
  }
  if (nc < (int)(id.nc())) id.nc(nc); // ケース数を減らす
  return nc;
}

void dpplotclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  constexpr int LENline = 74;   // 区切り線の長さ
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      cout << "\n" << endl;
    }
    return;
  }
  datalst.sort(data_less(1));    // 項目データ順にソートする
  diag dg(opt.pi());             // 診断率オブジェクト
  if (dg.cal_dpa(opt, datalst) < 1) // 計算できず
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      if (opt.res("stat"))
      {
        cout << "\n";
        putline(cout, LENline);
        put_stat(cout, opt, dg);
        putline(cout, LENline);
      }
      cout << "\n" << endl;
    }
    return;
  }
  if (opt.res("res")) put_result(cout, opt, id, datalst, dg);
  if (opt.ngpname() != "") plot_ngp(opt, id, datalst, dg);
}

int main(int argc, char* argv[]) // メインルーチン
{
  dpplotclass dpplot;           // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int nc = 0;                 // 読み込んだケース数
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) dpplot.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { dpplot.get(argv[i], opt, id, datalst);  fok = false; }
    }
    dpplot.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occured !");
    return 255;
  }
  return 0;
}
