// nlrae.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-25 11:57:38 tac>
//
// 非線形最小二乗法による回帰分析
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法:g++ -O -s -W nlrae nlrae.C
//             -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(mat.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "nlrae  Version 7.41e         Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:nlrae [<OPTION> ...] [<DATA FILE> ...]",
  "  The regression analysis by non-linear least-squared method with",
  "  the first item as the explanatory variable(X) and the second item",
  "  as the criterion variable(Y).",
  "<OPTION>",
  "  -eps=<EPS>:specify the convergence error(0<EPS)",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -h0=<VALUE-1>,...,<VALUE-p>:specify the initial values of the parameters",
  "  -meth={newton|non|simplex|etc}:specify the calculation method",
  "  -model={exp|log|log1|logc|etc}:specify the model",
  "  -ngp={data|func}[,…]:select the data for Ngraph",
  "    data:raw data of X and Y  func:model function  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=nlrae_ngp.dat",
  "  -res=[-][{def|all|atab|matab|model|stat}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    atab:ANOVA table  matab:model-specific ANOVA table  model:model function",
  "    stat:basic statistics  default:auto judgment according to data",
  "  -rmax=<NUMBER>:specify the maximum number of repetitions(0=<NUMBER)",
  "  -ym=<VALUE>:minimum value of Y(logc only)"
};

const string Title = "=== Regression Analysis by Non-linear LSM ===";
const string NGPNAMEdef = "nlrae_ngp.dat"; // 省略時Ngraphファイル名
constexpr double EPSdef = 1.0e-4; // 収束誤差省略時値
constexpr int RMAXdef = 100;      // 最大計算回数省略時値
constexpr int RDdef = 50;         // グラフの描画範囲分割数省略時値
constexpr size_t Pdef = 2;        // 回帰式の次数

enum methtype                   // 非線形最小2乗法の種類
{
  methnon,                      // 指定なし
  methnewton,                   // Newton-Raphson法
  methsimplex,                  // Simplex法
  methetc                       // その他
};

enum modeltype                  // モデルの種類
{
  modelnon,                     // 指定なし
  modelexp,                     // 指数関数
  modellog,                     // ロジスティック関数
  modellog1,                    // ロジスティック関数(M=1固定)
  modellogc,                    // ロジスティック関数(最小値有)
  modeletc                      // その他
};

double absmin(const vec<double>& hv) // 絶対値最小のパラメターの値を返す
{
  double hmin = DBL_MAX;
  for (size_t i = 0;  i < hv.dim();  ++i)
  {
    double hi = fabs(hv(i));
    if (DBL_MIN < hi && hi < hmin) hmin = hi;
  }
  return (hmin == DBL_MAX) ? 1.0 : hmin;
}

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double veps;                  // 非線形最小2乗法の収束誤差
  methtype vmeth;               // 非線形最小2乗法の種類
  modeltype vmodel;             // モデルの種類
  int vnh;                      // パラメーター数
  lstr h0lst;                   // 初期値リスト
  vecdbl vh0vec;                // 初期値ベクトル
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  int vrmax;                    // 最大計算回数
  double vym;                   // yの定数項
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double eps() const            // 非線形最小2乗法の収束誤差を返す
  { return veps; }
  double h0(size_t idx) const   // 初期値を返す
  { return (idx < vh0vec.size()) ? vh0vec[idx] : 0.0; }
  methtype meth() const         // 計算法の種類を返す
  { return vmeth; }
  const string methname() const; // 計算法の名称を返す
  modeltype model() const        // モデルの種類を返す
  { return vmodel; }
  const string modelname() const; // モデルの名称を返す
  int nh() const                // パラメーター数を返す
  { return vnh; }
  size_t nh0() const            // 指定された初期値数を返す
  { return vh0vec.size(); }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rmax() const              // 最大計算回数を返す
  { return vrmax; }
  double ym() const               // yの最小値を返す
  { return vym; }
  double ym(double ym0)         // yの最小値を設定する
  { return (vym = ym0); }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), veps(EPSdef), vmeth(methetc), vmodel(modellog), vnh(3), h0lst(), vh0vec(), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vrmax(RMAXdef), vym(BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "eps")          // 非線形最小二乗法の収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < DBL_EPSILON) err = badopt(vpname, optstr, USAGE, DBL_EPSILON, DBL_MAX);
  }
  else if (opt[0] == "f")       // オプション指定ファイルから読み込み
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "h0")      // 初期値の指定
  {
    if (strsplit(h0lst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "meth")    // 計算法の指定
  {
    if (opt[1].empty() || opt[1] == "etc") vmeth = methetc;
    else if (opt[1] == "newton") vmeth = methnewton;
    else if (opt[1] == "non") vmeth = methnon;
    else if (opt[1] == "simplex") vmeth = methsimplex;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "model")   // モデルの指定
  {
    if (opt[1].empty() || opt[1] == "log") { vmodel = modellog;  vnh = 3; }
    else if (opt[1] == "etc") { vmodel = modeletc;  vnh = 2; }
    else if (opt[1] == "exp") { vmodel = modelexp;  vnh = 2; }
    else if (opt[1] == "log1") { vmodel = modellog1;  vnh = 2; }
    else if (opt[1] == "logc") { vmodel = modellogc;  vnh = 3; }
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ym")      // yの最小値指定
  {
    vym = opt[1].empty() ? BLANKf : strtodf(opt[1]);
    if (vym == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "rmax")    // 最大計算回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else                          // 間違ったオプション
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni != 2) return disperr(vpname, "Number of items must be 2 !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  if (h0lst.size() < 1) return 0; // 初期値指定なし
  if ((int)(h0lst.size()) > vnh) return disperr(vpname, "Too many initial values !");
  vh0vec.resize(vnh);
  vh0vec.assign(vnh, 0.0);      // 初期値を0にする
  size_t i = 0;
  for (string h0str : h0lst)
  {
    vh0vec[i] = strtodf(h0str);
    if (vh0vec[i] == BLANKf) return disperr(vpname, "Wrong specification of initial value !");
    ++i;
  }
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["data"] = vngpmap["func"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["data"] = vngpmap["func"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["atab"] = vresmap["model"] = vresmap["stat"] = true;
  vresmap["matab"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["atab"] = vresmap["model"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["atab"] = vresmap["model"] = vresmap["stat"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["atab"] = vresmap["matab"] = vresmap["model"] = vresmap["stat"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

const string option::methname() const // 計算法の名称を返す
{
  switch (vmeth)
  {
  case methnewton : return "Newton Method";
  case methnon : return "Non";
  case methsimplex : return "Downdill-simplex Method";
  default : return "Downhill-simplex+Newton Method";
  }
}

const string option::modelname() const // モデルの名称を返す
{
  switch (vmodel)
  {
  case modelexp : return "Exponential Function";
  case modellog : return "Logistic Function";
  case modellog1 : return "Logistic Function(M=1:fixed)";
  case modellogc : return "Logistic Function(Minimum Value Available)";
  case modelnon : return "Non";
  default : return "Other Model(Linear Function)";
  }
}

class data_less                 // ケースデータの比較用関数オブジェクト
{
private:
  int kidx;                     // ソートキー項目のインデックス
public:
  data_less(int kidx0) : kidx(kidx0) // コンストラクタ
  {}
  ~data_less()                  // デストラクタ
  {}
  bool operator()(datac& cdata1, datac& cdata2) const; // ケースデータを比較する
};

bool data_less::operator()(datac& cdata1, datac& cdata2) const
// ケースデータを比較する
{
  int comp = 0;
  double fdata1 = cdata1.fdata(kidx);
  double fdata2 = cdata2.fdata(kidx);
  if (fdata1 < fdata2) comp = -1;
  else if (fdata1 > fdata2) comp = 1;
  return (comp < 0) ? true : false; // 小さいときだけtrueを返す
}

class varclass                  // 変数クラス
{
private:
  int vnh;                      // パラメータ数
  pairdbl xrange;               // xの最小値と最大値
  pairdbl yrange;               // yの最小値と最大値
  statcor<double> statcxy;      // xとyの相関基礎統計量
  vec<double> vh0v;             // パラメーター初期値
  vec<double> vhv;              // パラメーター推測値
  void init_h0(statcor<double>& statcxln, statcor<double>& statcxly, option& opt, ldata& datalst);
  // パラメータの初期値を設定する
  int init_log(option& opt, ldata& datalst); // 微分回帰分析による初期値の設定
  int cal_smat(mat<double>& smat, double ym, ldata& datalst) const;
  // 微分回帰分析用に積和行列を計算する
public:
  varclass(int nh0) : vnh(nh0), xrange(DBL_MAX, -DBL_MAX), yrange(DBL_MAX, -DBL_MAX), statcxy(), vh0v(nh0), vhv(nh0)
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int init(option& opt, ldata& datalst); // データを読み込み変数を初期化する
  int nh() const                // パラメータ数を返す
  { return vnh; }
  double xmin() const           // xの最小値を返す
  { return xrange.first; }
  double xmax() const           // xの最大値を返す
  { return xrange.second; }
  double ymin() const           // yの最小値を返す
  { return yrange.first; }
  double ymax() const           // yの最大値を返す
  { return yrange.second; }
  int xyn() const               // xとyの例数を返す
  { return statcxy.n(); }
  double xymx() const           // xの平均値を返す
  { return statcxy.mx(); }
  double xyvxx() const          // xの不偏分散を返す
  { return statcxy.vxx(); }
  double xymy() const           // yの平均値を返す
  { return statcxy.my(); }
  double xyvyy() const          // yの不偏分散を返す
  { return statcxy.vyy(); }
  double xysyy() const          // yの平方和を返す
  { return statcxy.syy(); }
  const vec<double>& h0v() const // パラメーター初期値を返す
  { return vh0v; }
  double h0v(size_t idx) const  // 特定のパラメーター初期値を返す
  { return (idx < vh0v.size()) ? vh0v(idx) : 0.0; }
  const vec<double>& hv() const // パラメーターを返す
  { return vhv; }
  vec<double>& hv(const vec<double>& hv0); // パラメーターを設定する
  double hv(size_t idx) const   // 特定のパラメーターを返す
  { return (idx < vhv.size()) ? vhv(idx) : 0.0; }
};

void varclass::init_h0(statcor<double>& statcxln, statcor<double>& statcxly, option& opt, ldata& datalst)
// パラメータの初期値を設定する
{
  double y0 = 0.5 * yrange.first, m = yrange.second;
  double b = (statcxy.sxx() > 0.0) ? statcxy.sxy() / statcxy.sxx() : 0.0;
  double c = (sgn(m) > 0) ? 8.0 * b / m : 0.0; // ロジスティック関数の回帰係数
  double a = statcxy.my() - b * statcxy.mx();
  double bln = (statcxln.sxx() > 0.0) ? statcxln.sxy() / statcxln.sxx() : 0.0;
  double aln = statcxln.my() - bln * statcxln.mx();
  double bly = (statcxly.sxx() > 0.0) ? statcxly.sxy() / statcxly.sxx() : 0.0;
  double aly = bly * statcxly.mx() - statcxly.my();
  switch (opt.model())
  {
  case modellog :               // ロジスティック関数
    if (init_log(opt, datalst) < 0) // 微分回帰分析による初期値設定失敗
    {
      if (sgn(y0) <= 0) y0 = 0.1;
      if (sgn(c) < 0) c = 0.0;
      vh0v.assign(0, y0);
      vh0v.assign(1, c);
      vh0v.assign(2, m);
    }
    break;
  case modellog1 :              // ロジスティック関数(M=1固定)
    if (statcxly.n() > 1)       // ロジット回帰分析による初期値設定
    {
      y0 = 1.0 / (0.5 * exp(aly) + 1.0);
      c = (sgn(bly) > 0) ? bly : 0.0;
      vh0v.assign(0, y0);
      vh0v.assign(1, bly);
    }
    else
    {
      y0 = 0.5 * a;
      if (sgn(y0) <= 0 || 1.0 <= y0) y0 = 0.1;
      c = (sgn(b) > 0) ? 8.0 * b : 0.0;
      vh0v.assign(0, y0);
      vh0v.assign(1, c);
    }
    break;
  case modellogc :              // ロジスティック関数(最小値有)
    if (init_log(opt, datalst) < 0) // 微分回帰分析による初期値設定失敗
    {
      if (y0 <= 0.0) y0 = 0.1;
      if (sgn(c) < 0) c = 0.0;
      vh0v.assign(0, y0);
      vh0v.assign(1, c);
      vh0v.assign(2, m);
      if (opt.ym() == BLANKf) opt.ym(0.5 * y0); // 最小値自動設定
    }
    break;
  case modelexp : vh0v.assign(0, exp(aln)); // 指数関数
    vh0v.assign(1, bln);
    break;
  default : vh0v.assign(0, a);  vh0v.assign(1, b); // その他の関数
  }
  if (opt.nh0() > 0)            // 初期値が指定されている場合
  {
    for (size_t i = 0; i < opt.nh0(); ++i) vh0v.assign(i, opt.h0(i));
  }
  vhv = vh0v;
}

int varclass::init_log(option& opt, ldata& datalst)
// 微分回帰分析による初期値の設定
{
  size_t p = Pdef;
  bool initym = false;
  double ym = 0.0;
  if (opt.model() == modellogc) // 最小値があるロジスティック関数
  {
    if (opt.ym() == BLANKf) { initym = true;  p = Pdef + 1; }// 最小値自動設定
    else ym = opt.ym();
  }
  size_t d = p + 1;
  mat<double> smat(d);          // 積和行列
  if (cal_smat(smat, ym, datalst) < 0) return -1;
  size_t dyidx = p;             // dyのインデックス
  double b0 = 0.0, b1 = 0.0, b2 = 0.0, c = 0.0, m = 0.0;
  if (initym)
  {
    b0 = smat(0, dyidx);  b1 = smat(1, dyidx);  b2 = smat(2, dyidx);
    double sqrtb = sqrt(b1 * b1 - 4.0 * b2 * b0);
    ym = (-b1 + sqrtb) / (2.0 * b2);
    double ymax = (-b1 - sqrtb) / (2.0 * b2);
    m = ymax - ym;
    c = -b2 * m;  opt.ym(ym);
  }
  else
  {
    b1 = smat(0, dyidx);  b2 = smat(1, dyidx);
    c = b1;  m = -b1 / b2;
  }
  if (sgn(c) < 0) return -1;    // 2次回帰分析結果不適切
  statcor<double> statcxly;     // xとlogit(y)の相関基礎統計量
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (x != BLANKf && y != BLANKf)
    {
      double yi = y - ym;
      if (0.0 < yi && yi < m) statcxly.sigma(x, log(yi / (m - yi)));
    }
  }
  double ly0 = statcxly.my() - c * statcxly.mx();
  double y0 = ym + m / (1.0 + exp(-ly0));
  if (y0 - ym > m) return -1;   // 2次回帰分析結果不適切
  vh0v.assign(0, y0);  vh0v.assign(1, c);  vh0v.assign(2, m);
  return 0;                     // 初期値設定OK
}

int varclass::cal_smat(mat<double>& smat, double ym, ldata& datalst) const
// 微分回帰分析用に積和行列を計算する
{
  size_t d = smat.r();
  size_t p = d - 1;             // 2次回帰式のパラメータ数
  vecdbl fdatavec(d);
  size_t offset = 0;
  if (d == Pdef + 2) { fdatavec[0] = 1.0;  offset = 1; }
  size_t dyidx = p;             // dyのインデックス
  double x1 = 0.0, x2 = 0.0, y1 = 0.0, y2 = 0.0;
  size_t nxy = 0;
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (x != BLANKf && y != BLANKf)
    {
      if (nxy < 1)
      { x1 = x;  y1 = y;  nxy = 1; }
      else if (nxy == 1)
      { x2 = x;  y2 = y;  nxy = 2; }
      else
      {
        double dx1 = x2 - x1, dx2 = x - x2;
        double eps = DBL_EPSILON * fmax(5.0, fabs(dx1)); // 計算誤差
        double w1 = (fabs(dx1) < eps) ? 0.0 : 1.0 / dx1;
        double w2 = (fabs(dx2) < eps) ? 0.0 : 1.0 / dx2;
        double b1 = (y2 - y1) * w1, b2 = (y - y2) * w2;
        w1 = fabs(w1);  w2 = fabs(w2);
        double dy = (fabs(w1 + w2) < eps) ? 0.0 : (b1 * w1 + b2 * w2) / (w1 + w2);
        double yi = y2 - ym;    // 定数項を引く
        double fd = yi;
        for (size_t i = 0; i < Pdef; ++i)
        {
          fdatavec[offset + i] = fd;
          fd *= yi;
        }
        fdatavec[dyidx] = dy;
        for (size_t i = 0; i < d; ++i)
        {
          fd = fdatavec[i];
          for (size_t j = i; j < d; ++j) smat.add(i, j, fd * fdatavec[j]);
        }
        x1 = x2;  y1 = y2;
        x2 = x;  y2 = y;  ++nxy;
      }
    }
  }
  if (nxy - 2 < p) return -1;   // 微分回帰分析できず
  for (size_t i = 0; i < dyidx; ++i)
  {
    for (size_t j = i + 1; j <= dyidx; ++j) smat.assign(j, i, smat(i, j));
  }
  for (size_t i = 0; i < dyidx; ++i) smat.sweep(i);
  return nxy - 2;               // 集計した例数を返す
}

int varclass::init(option& opt, ldata& datalst)
// データを読み込み変数を初期化する
{
  statcor<double> statcxln;     // xとln(y)の相関基礎統計量
  statcor<double> statcxly;     // xとlogit(y)の相関基礎統計量
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (x != BLANKf && y != BLANKf)
    {
      if (x < xrange.first) xrange.first = x;
      if (x > xrange.second) xrange.second = x;
      if (y < yrange.first) yrange.first = y;
      if (y > yrange.second) yrange.second = y;
      statcxy.sigma(x, y);
      if (0.0 < y) statcxln.sigma(x, log(y));
      if (0.0 < y && y < 1.0) statcxly.sigma(x, log(y / (1.0 - y)));
    }
  }
  if (statcxy.n() > opt.nh()) init_h0(statcxln, statcxly, opt, datalst);
  // パラメータ初期値設定
  return statcxy.n();           // 集計例数を返す
}

vec<double>& varclass::hv(const vec<double>& hv0) // パラメーターを設定する
{
  vhv = hv0;
  return vhv;
}

class nlraclass                 // メインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int cal_nlra(const option& opt, double& qerr, varclass& var);
  // 非線形最小2乗法によってパラメータを求める
  int cal_simplex(const option& opt, varclass& var);
  // Simplex法によってパラメータを求める
  bool cal_simplex(int& nr, vec<double>& hv); // Simplex法による非線形最小2乗計算
  void init_simplex(mat<double>& simplexmat, const vec<double>& hv);
  // シンプレックスの初期設定
  int check_simplex(const option& opt, mat<double>& simplexmat, vec<double>& hlv, vec<double>& hmv, vec<double>& hhv, double& fl, double& fs, double& fh, int& lidx, int& hidx, int nr);
  // fiの計算と終了基準のチェック
  bool check_hv(const option& opt, const vec<double>& hv) const;
  // パラメータの発散チェック
  int cal_newton(const option& opt, varclass& var); // Newton-Raphson法によってパラメータを求める
  void cal_gh(const option& opt, vec<double>& gradv, mat<double>& hessmat, const vec<double>& hv);
  // 傾斜ベクトルとHessの行列の計算
  int check_q(const option& opt, double qnew, double qold, const vec<double>& hv, int nr);
  // Q(h)の計算と終了基準のチェック
  double fq(const option& opt, const vec<double>& hv); // 誤差関数
  double fqd(const option& opt, int hidx, const vec<double>& hv);
  // 誤差の偏導関数
  double fqd2(const option& opt, int hiidx, int hjidx, const vec<double>& hv); // 誤差の2階偏導関数
  double func(const option& opt, double x, const vec<double>& hv); // モデル関数
  double funcd(const option& opt, double x, int hidx, const vec<double>& hv); // モデル偏導関数
  double funcd2(const option& opt, double x, int hiidx, int hjidx, const vec<double>& hv);
  // モデル2階偏導関数
  void put_resid(ostream& fout); // 分析結果のタイトルとデータ識別情報を出力する
  void put_model(ostream& fout, const option& opt, int nr, double qerr, const varclass& var) const;
  // モデル関数の非線形最小2乗法計算結果の出力
  void put_nlatab(ostream& fout, option& opt, int nr, double qerr, const varclass& var);
  // 非線形最小2乗法計の分散分析表出力
  void put_exp(ostream& fout, const option& opt, const varclass& var) const;
  // 指数関数の結果出力
  void put_log(ostream& fout, const option& opt, const varclass& var) const;
  // ロジスティック関数の結果出力
  void put_log1(ostream& fout, const option& opt, const varclass& var) const;
  // ロジスティック関数(M=1に固定)の結果出力
  void put_logc(ostream& fout, const option& opt, const varclass& var) const;
  // ロジスティック関数(定数項有)の結果出力
  void put_etc(ostream& fout, const option& opt, const varclass& var) const;
  // 1次関数の結果出力
  void plot_ngp(option& opt, int nr, const varclass& var);
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_data(ofstream& fout); // Ngraph用に個々のデータを出力する
  void plot_func(ofstream& fout, const option& opt, int nr, const varclass& var);
  // Ngraph用関数曲線の出力
public:
  nlraclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~nlraclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み保存・累積する
  void put(option& opt);                     // 計算結果の出力
};

int nlraclass::cal_nlra(const option& opt, double& qerr, varclass& var)
// 非線形最小2乗法によってパラメータを求める
{
  int nr1 = 0;  // 1次反復計算数(収束を良くするためSimplex法で計算)
  if (opt.meth() == methetc || opt.meth() == methsimplex)
  {
    if ((nr1 = cal_simplex(opt, var)) < 0) // 収束失敗
    {
      qerr = fq(opt, var.hv());
      return nr1;
    }
  }
  int nr2 = 0;                  // 2次反復計算数
  if (opt.meth() == methetc || opt.meth() == methnewton)
  {
    if ((nr2 = cal_newton(opt, var)) < 0) // 収束失敗
    {
      qerr = fq(opt, var.hv());
      return nr2;
    }
  }
  qerr = fq(opt, var.hv());
  return nr1 + nr2;             // 1次と2次の合計反復数を返す
}

int nlraclass::cal_simplex(const option& opt, varclass& var)
// Simplex法によってパラメータを求める
{
  int p = var.nh();                 // パラメーター数
  mat<double> simplexmat(p, p + 1); // シンプレックスを行列で表す
  vec<double> hlv(var.hv()), hmv(var.hv()), hhv(var.hv());
  double fl = 0.0, fs = 0.0, fh = 0.0; // 最小値、2番目に大きい値、最大値
  int lidx = 0, hidx = 0, ok = 0, nr = 0;
  init_simplex(simplexmat, var.hv());
  while ((ok = check_simplex(opt, simplexmat, hlv, hmv, hhv, fl, fs, fh, lidx, hidx, nr)) == 0)
  {
    vec<double> hrv(hmv);
    hrv *= 2.0;
    hrv -= hhv;
    double fr = fq(opt, hrv);
    if (fr < fl)                // 折り返し成功
    {
      vec<double> hev(hrv);
      hev *= 2.0;
      hev -= hmv;
      double fe = fq(opt, hev);
      if (fe < fr)              // 拡大成功
      { for (int i = 0; i < p; ++i) simplexmat.assign(i, hidx, hev(i)); }
      else                      // 拡大失敗(行き過ぎ)
      { for (int i = 0; i < p; ++i) simplexmat.assign(i, hidx, hrv(i)); }
    }
    else if (fr <= fs)          // 斜め方向の折り返し
    { for (int i = 0; i < p; ++i) simplexmat.assign(i, hidx, hrv(i)); }
    else                        // 縮小失敗、Simplex移動
    {
      if (fr < fh) { fh = fr;  hhv = hrv; } // 折り返し後のSimplex中に最小点
      vec hcv(hmv + hhv);
      hcv *= 0.5;
      double fc = fq(opt, hcv);
      if (fc < fh)              // 縮小成功
      { for (int i = 0; i < p; ++i) simplexmat.assign(i, hidx, hcv(i)); }
      else                      // 縮小失敗、Simplex移動
      {
        for (int j = 0; j <= p; ++j)
        {
          if (j != lidx)
          {
            for (int i = 0; i < p; ++i) simplexmat.assign(i, j, (simplexmat(i, j) + hlv(i)) * 0.5);
          }
        }
      }
    }
    ++nr;
  }
  if (ok < 0 || fl > var.xysyy()) return -1; // 発散
  var.hv(hlv);
  return nr;                    // 収束(繰り返し数を返す)
}

void nlraclass::init_simplex(mat<double>& simplexmat, const vec<double>& hv)
// シンプレックスの初期設定
{
  constexpr double Sqrt2 = sqrt(2.0); // コンパイル時定数
  constexpr double Rs = 0.1;          // シンプレックスの1辺の相対割合
  int p = hv.dim();
  double s = absmin(hv) * Rs;
  double v = (sqrt((double)(p + 1)) - 1.0 + (double)p) / (double)p / Sqrt2 * s;
  double w = (sqrt((double)(p + 1)) - 1.0) / (double)p / Sqrt2 * s;
  for (int i = 0; i < p; ++i) simplexmat.assign(i, 0, hv(i));
  for (int j = 1; j <= p; ++j)
  {
    for (int i = 0; i < p; ++i) simplexmat.assign(i, j, (i == j - 1) ? v + hv(i) : w + hv(i));
  }
}

int nlraclass::check_simplex(const option& opt, mat<double>& simplexmat, vec<double>& hlv, vec<double>& hmv, vec<double>& hhv, double& fl, double& fs, double& fh, int& lidx, int& hidx, int nr)
// fiの計算と終了基準のチェック
{
  if (opt.rmax() > 0 && nr >= opt.rmax()) return 2; // 繰り返し回数最大値
  int p = hlv.dim();
  int eidx = p - 1;
  vecdbl fvec(p + 1, 0.0);
  fl = FLT_MAX;  fh = -FLT_MAX;
  for (int j = 0; j <= p; ++j)
  {
    vec<double> hjv(simplexmat.sub(0, j, eidx, j));
    double fj = fq(opt, hjv);
    fvec[j] = fj;
    if (fj < fl) { fl = fj;  lidx = j; }
    if (fj > fh) { fh = fj;  hidx = j; }
  }
  if (fl > FLT_MAX || isnan(fl)) return -1;  // 発散
  hlv = simplexmat.sub(0, lidx, eidx, lidx); // 最小値の時のhv
  hhv = simplexmat.sub(0, hidx, eidx, hidx); // 最大値の時のhv
  if (check_hv(opt, hlv) == false) return -2;     // 発散
  double ferr = fabs(fh - fl);
  if (sgndiff(fh, fl) == 0 || ferr < fabs(fl * opt.eps())) return 1; // 収束
  fs = fl;
  hmv = 0.0;
  for (int j = 0; j <= p; ++j)
  {
    if (j != hidx)
    {
      if (fvec[j] > fs) fs = fvec[j];
      hmv += simplexmat.sub(0, j, eidx, j);
    }
  }
  hmv *= 1.0 / (double)p;       // hhv以外のp個のhvの重心
  return 0;                     // 収束せず
}

bool nlraclass::check_hv(const option& opt, const vec<double>& hv) const
// パラメータの発散チェック
{
  bool ok = true;
  for (size_t i = 0; ok && i < hv.dim(); ++i)
  {
    if (fabs(hv(i)) > FLT_MAX || isnan(hv(i))) ok = false;
  }
  if (ok == false) return ok;   // 発散
  if (opt.model() == modellog || opt.model() == modellog1 || opt.model() == modellogc)
  {
    if (sgn(hv(1)) < 0)
    { ok = false; }
    else if (opt.model() == modellog1)
    {
      if (sgn(hv(0)) <= 0 || 1.0 <= hv(0)) ok = false;
    }
    else if (opt.model() == modellog)
    {
      if (hv(0) >= hv(2)) ok = false;
    }
    else if (opt.model() == modellogc)
    {
      if(hv(0) - opt.ym() >= hv(2)) ok = false;
    }
  }
  return ok;                    // false:発散
}

int nlraclass::cal_newton(const option& opt, varclass& var)
// Newton-Raphson法によってパラメータを求める
{
  int p = var.nh();
  vec<double> gradv(p), htv(var.hv());
  mat<double> hessmat(p, p);
  double qnew = fq(opt, htv), qold = 0.0;
  int ok = 0, nr = 0;
  do
  {
    cal_gh(opt, gradv, hessmat, htv);
    hessmat.inv();
    htv -= hessmat * gradv;
    qold = qnew;
    qnew = fq(opt, htv);  ++nr;
  } while ((ok = check_q(opt, qnew, qold, htv, nr)) == 0);
  if (ok < 0 || qnew > var.xysyy()) return -1; // 発散
  var.hv(htv);
  return nr;                    // 収束(繰り返し数を返す)
}

void nlraclass::cal_gh(const option& opt, vec<double>& gradv, mat<double>& hessmat, const vec<double>& hv)
// 傾斜ベクトルとHessの行列の計算
{
  for (size_t i = 0; i < hv.dim(); ++i)
  {
    gradv.assign(i, fqd(opt, i, hv));
    for (size_t j = i; j < hv.dim(); ++j)
    { hessmat.assign(i, j, fqd2(opt, i, j, hv)); }
  }
  for (size_t i = 0; i < hv.dim() - 1; ++i)
  {
    for (size_t j = i + 1; j < hv.dim(); ++j) hessmat.assign(j, i, hessmat(i, j));
  }
}

int nlraclass::check_q(const option& opt, double qnew, double qold, const vec<double>& hv, int nr)
// Q(h)の計算と終了基準のチェック
{
  if (check_hv(opt, hv) == false) return -1; // 発散
  if (qnew > FLT_MAX || isnan(qnew)) return -1; // 発散
  double qerr = fabs(qold - qnew);
  if (sgndiff(qold, qnew) == 0 || qerr < fabs(qnew * opt.eps())) return 1; // 収束
  if (opt.rmax() > 0 && nr >= opt.rmax()) return 2; // 繰り返し回数最大値
  return 0;                     // 収束せず
}

double nlraclass::fq(const option& opt, const vec<double>& hv) // 誤差関数
{
  double q = 0.0;
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (x != BLANKf && y != BLANKf)
    {
      double ei = y - func(opt, x, hv);
      q += ei * ei;
    }
  }
  return q;                     // 平方誤差合計を返す
}

double nlraclass::fqd(const option& opt, int hidx, const vec<double>& hv)
// 誤差の偏導関数
{
  double dq = 0.0;
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (x != BLANKf && y != BLANKf)
    {
      double dei = y * funcd(opt, x, hidx, hv) - func(opt, x, hv) * funcd(opt, x, hidx, hv);
      dq -= 2.0 * dei;
    }
  }
  return dq;                    // 誤差導関数値を返す
}

double nlraclass::fqd2(const option& opt, int hiidx, int hjidx, const vec<double>& hv)
// 誤差の2階偏導関数
{
  double d2q = 0.0;
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (x != BLANKf && y != BLANKf)
    {
      double qd2 = funcd2(opt, x, hiidx, hjidx, hv);
      double d2ei = y * qd2 - funcd(opt, x, hiidx, hv) * funcd(opt, x, hjidx, hv) - func(opt, x, hv) * qd2;
      d2q -= 2.0 * d2ei;
    }
  }
  return d2q;                   // 誤差2階導関数値を返す
}

double nlraclass::func(const option& opt, double x, const vec<double>& hv)
// モデル関数
{
  double f = 0.0;
  switch (opt.model())
  {
  case modelexp : f = hv(0) * exp(hv(1) * x); // 指数関数
    break;
  case modellog : f = logdis(x, hv(2), hv(1), hv(0)); // ロジスティック関数
    break;
  case modellog1 : f = logdis(x, 1.0, hv(1), hv(0)); // ロジスティック関数(M=1固定)
    break;
  case modellogc : f = opt.ym() + logdis(x, hv(2), hv(1), hv(0) - opt.ym()); // ロジスティック関数(最小値有)
    break;
  default : f = hv(0) + hv(1) * x; // 1次関数
  }
  return f;                     // 関数値を返す
}

double nlraclass::funcd(const option& opt, double x, int hidx, const vec<double>& hv)
// モデル偏導関数
{
  constexpr double Delta = EPSdef; // 微小変動の割合(差分法用)
  vec<double> hdv(hv);
  double dh = 0.0, grad = 0.0;
  switch (opt.model())
  {
  case modelexp :               // 指数関数
    switch (hidx)
    {
    case 0 : grad = exp(hv(1) * x); // hv(0)で偏微分
      break;
    case 1 : grad = x * exp(hv(1) * x); // hv(1)で偏微分
    }
    break;
  case modeletc :               // 1次関数
    switch (hidx)
    {
    case 0 : grad = 1.0;        // hv(0)で偏微分
      break;
    case 1 : grad = x;          // hv(1)で偏微分
    }
    break;
  default : dh = hv(hidx) * Delta; // 差分法
    hdv.assign(hidx, hdv(hidx) + dh);
    grad = (func(opt, x, hdv) - func(opt, x, hv)) / dh;
  }
  return grad;                  // 偏導関数値を返す
}

double nlraclass::funcd2(const option& opt, double x, int hiidx, int hjidx, const vec<double>& hv)
// モデル2階偏導関数
{
  constexpr double Delta = EPSdef; // 微小変動の割合(差分法用)
  vec hijv(hv), hjv(hv);
  double fhi = 0.0, fhj = 0.0, dhi = 0.0, dhj = 0.0, grad2 = 0.0;
  switch (opt.model())
  {
  case modelexp :               // 指数関数
    switch (hiidx)
    {
    case 0 :
      switch (hjidx)
      {
      case 0 : grad2 = 0.0;
        break;
      case 1 : grad2 = x * exp(hv(1) * x);
      }
      break;
    case 1 :
      switch (hjidx)
      {
      case 0 : grad2 = x * exp(hv(1) * x);
        break;
      case 1 : grad2 = x * x * hv(0) * exp(hv(1) * x);
      }
    }
    break;
  case modeletc : grad2 = 0.0;  // 1次関数
    break;
  default : dhi = hv(hiidx) * Delta; // 差分法
    dhj = hv(hjidx) * Delta;
    hijv.assign(hiidx, hv(hiidx) + dhi);
    fhi = func(opt, x, hijv);
    hjv.assign(hjidx, hv(hjidx) + dhj);
    fhj = func(opt, x, hjv);
    hijv.assign(hjidx, hijv(hjidx) + dhj);
    grad2 = (func(opt, x, hijv) - fhi - fhj + func(opt, x, hv)) / dhi / dhj;
  }
  return grad2;                 // 2階偏導関数値を返す
}

void nlraclass::put_resid(ostream& fout) // 分析結果のデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "EXPLANATORY VARIABLE X:" << id.item(0) << "\n";
  fout << "CRITERION VARIABLE   Y:" << id.item(1) << "\n";
}

void nlraclass::put_model(ostream& fout, const option& opt, int nr, double qerr, const varclass& var) const
// モデル関数の非線形最小2乗法計算結果の出力
{
  switch (opt.model())
  {
  case modelexp : put_exp(fout, opt, var);
    break;
  case modellog : put_log(fout, opt, var);
    break;
  case modellog1 : put_log1(fout, opt, var);
    break;
  case modellogc : put_logc(fout, opt, var);
    break;
  default : put_etc(fout, opt, var);
  }
  int n = var.xyn(), p = var.nh();
  double aic = (sgn(qerr) > 0) ? (double)n * log(qerr) + 2.0 * (double)p : 0.0;
  double sa = var.xysyy(), sl = qerr;
  const double Seps = DBL_EPSILON * fmax(5.0, sa);
  if (sl < Seps) sl = 0.0;
  double sc = sa - sl;
  if (sc < Seps) sc = 0.0;
  double r2 = (sa > 0.0) ? sc / sa : 0.0;
  fout << "\n<Method:" << opt.methname() << "  Number of Repetitions=";
  if (nr > 0) fout << nr;
  else fout << "NON-LINEAR LSM DIVERGED !";
  fout << ">\n  SSresiduals(Q)=" << qerr << "  AIC=" << aic << "\n";
  fout << "  Multiple R-squared=" << r2 << "  Multiple R=" << sqrt(r2) << "\n";
}

void nlraclass::put_nlatab(ostream& fout, option& opt, int nr, double qerr, const varclass& var)
// 非線形最小2乗法計の分散分析表出力
{
  int n = var.xyn(), p = var.nh();
  double sa = var.xysyy(), sl = qerr;
  const double Seps = DBL_EPSILON * fmax(5.0, sa);
  if (sl < Seps) sl = 0.0;
  int dfa = n - 1, dfc = p - 1;
  double sc = sa - sl;
  if (sc < Seps) sc = 0.0;
  int dfl = dfa - dfc;
  if (opt.res("atab"))
  {
    fout << "\n";
    putatab(fout, "REGRESSION", sc, dfc, sl, dfl, sa, dfa);
  }
  if (opt.res("def") == false && opt.res("matab") == false) return;
  vec htv(var.hv());
  switch (opt.model())
  {
  case modellog : htv.assign(0, var.hv(2) * 0.5);  htv.assign(1, 0.0);
    sa = fq(opt, htv);
    break;
  case modellog1 : htv.assign(0, 0.5);  htv.assign(1, 0.0);
    sa = fq(opt, htv);
    break;
  case modellogc : htv.assign(0, var.hv(2) * 0.5 + opt.ym());  htv.assign(1, 0.0);
    sa = fq(opt, htv);
    break;
  default : return;
  }
  sc = sa - sl;
  if (sc < Seps) sc = 0.0;
  double r2 = (sa > 0.0) ? sc / sa : 0.0;
  fout << "\n<Total Sum of Squares is Model-specific>\n";
  fout << "  Multiple R-squared=" << r2 << "  Multiple R=" << sqrt(r2) << "\n\n";
  putatab(fout, "REGRESSION", sc, dfc, sl, dfl, sa, dfa);
}

void nlraclass::put_exp(ostream& fout, const option& opt, const varclass& var) const
// 指数関数の結果出力
{
  fout << "\n<Model:" << opt.modelname() << ">\n";
  fout << "  Model function :y=A exp(c x)\n";
  fout << "  Initial value  :A=" << var.h0v(0) << "  c=" << var.h0v(1) << "\n";
  fout << "  Estimated value:A=" << var.hv(0) << "  c=" << var.hv(1) << endl;
}

void nlraclass::put_log(ostream& fout, const option& opt, const varclass& var) const
// ロジスティック関数の結果出力
{
  double y00 = var.h0v(0), c0 = var.h0v(1), m0 = var.h0v(2);
  double y0 = var.hv(0), c = var.hv(1), m = var.hv(2);
  double ly00 = log(y00 / (m0 - y00)), ly0 = log(y0 / (m - y0));
  fout << "\n<Model:" << opt.modelname() << ">\n";
  fout << "  Model Function :y=M/[1+{(M-y0)/y0}exp(-c x)]=M/{1+exp(-ly0-c x)}\n";
  fout << "  Initial Value  :y0=" << y00 << "  c=" << c0 << "  M=" << m0 << "  ly0=" << ly00 << "\n";
  fout << "  Estimated Value:y0=" << y0 << "  c=" << c << "  M=" << m << "  ly0=" << ly0 << endl;
}

void nlraclass::put_log1(ostream& fout, const option& opt, const varclass& var) const
// ロジスティック関数(M=1固定)の結果出力
{
  double y00 = var.h0v(0), c0 = var.h0v(1);
  double y0 = var.hv(0), c = var.hv(1);
  double ly00 = log(y00 / (1.0 - y00)), ly0 = log(y0 / (1.0 - y0));
  fout << "\n<Model:" << opt.modelname() << ">\n";
  fout << "  Model Function :y=1/[1+{(1-y0)/y0}exp(-c x)]=1/{1+exp(-ly0-c x)}\n";
  fout << "  Initial Value  :y0=" << y00 << "  c=" << c0 << "  ly0=" << ly00 << "\n";
  fout << "  Estimated Value:y0=" << y0 << "  c=" << c << "  ly0=" << ly0 << endl;
}

void nlraclass::put_logc(ostream& fout, const option& opt, const varclass& var) const
// ロジスティック関数(最小値有)の結果出力
{
  double y00 = var.h0v(0), c0 = var.h0v(1), m0 = var.h0v(2);
  double y0 = var.hv(0), c = var.hv(1), m = var.hv(2);
  double ly00 = log((y00 - opt.ym()) / (m0 - (y00 - opt.ym())));
  double ly0 = log((y0 - opt.ym()) / (m - (y0 - opt.ym())));
  fout << "\n<Model:" << opt.modelname() << ">\n";
  fout << "  Model Function :y=ym+M/[1+[{M-(y0-ym)}/(y0-ym)]exp(-c x)]\n";
  fout << "                   =ym+M/{1+exp(-ly0-c x)}\n";
  fout << "  Initial Value  :y0=" << y00 << "  c=" << c0 << "  m=" << m0 << "\n";
  fout << "                  ym=" << opt.ym() << "  ly0=" << ly00 << "\n";
  fout << "  Estimated Value:y0=" << y0 << "  c=" << c << "  m=" << m << "\n";
  fout << "                  ym=" << opt.ym() << "  ly0=" << ly0 << endl;
}

void nlraclass::put_etc(ostream& fout, const option& opt, const varclass& var) const
// 1次関数の結果出力
{
  fout << "\n<Model:" << opt.modelname() << ">\n";
  fout << "  Model Function :y=a + b・x\n";
  fout << "  Initial Value  :a=" << var.h0v(0) << "  b=" << var.h0v(1) << "\n";
  fout << "  Estimated Value:a=" << var.hv(0) << "  b=" << var.hv(1) << endl;
}

void nlraclass::plot_ngp(option& opt, int nr, const varclass& var)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("data"))
  {
    fout << "-s1 -dmark,00\n";
    plot_ngpid(fout);
    plot_data(fout);
    delml = true;
  }
  if (opt.ngp("func"))
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline\n";
    plot_ngpid(fout);
    plot_func(fout, opt, nr, var);
  }
}

void nlraclass::plot_ngpid(ofstream& fout)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%EXPLANATORY VARIABLE X:" << id.item(0) << "\n";
  fout << "%CRITERION VARIABLE   Y:" << id.item(1) << "\n";
}

void nlraclass::plot_data(ofstream& fout)
// Ngraph用に個々のデータを出力する
{
  fout << "%<RAW DATA>\n";
  for (auto& cdata : datalst)
  {
    double x = cdata.fdata(0), y = cdata.fdata(1);
    if (x != BLANKf && y != BLANKf) fout << x << " " << y << endl;
  }
}

void nlraclass::plot_func(ofstream& fout, const option& opt, int nr, const varclass& var)
// Ngraph用関数曲線の出力
{
  fout << "%<Model:" << opt.modelname() << ">\n";
  fout << "%Method:" << opt.methname() << "  Number of Repetitions=";
  if (nr > 0) fout << nr;
  else fout << "NON-LINEAR LSM DIVERGED !";
  fout << "\n";
  double xmgn = (var.xmax() - var.xmin()) * 0.05;
  double xs = var.xmin() - xmgn, xe = var.xmax() + xmgn;
  double dx = (xe - xs) / (double)RDdef;
  double x = xs;
  for (int i = 0; i <= RDdef; ++i, x += dx) fout << x << " " << func(opt, x, var.hv()) << endl;
}

int nlraclass::get(const char* c_fname, option& opt)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void nlraclass::put(option& opt) // 計算結果の出力
{
  if (id.nc() < 1)              // データ無し
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  datalst.sort(data_less(0));   // 説明変数順にソートする
  varclass var(opt.nh());       // 変数オブジェクト
  if (opt.nh() > var.init(opt, datalst)) throw option_error("例数が足りません！");
  double qerr = 0.0;
  int nr = cal_nlra(opt, qerr, var);
  constexpr int LENline = 75;   // 区切り線の長さ
  if (opt.res("res"))           // 計算結果出力
  {
    put_resid(cout);
    if (opt.res("stat"))
    {
      cout << "\n<BASIC STATISTICS OF VARIABLES>\n";
      putline(cout, LENline);
      int n = var.xyn();
      cout << "X:";  putmsd(cout, n, var.xymx(), var.xyvxx());
      cout << "Y:";  putmsd(cout, n, var.xymy(), var.xyvyy());
      putline(cout, LENline);
    }
    if (opt.res("model")) put_model(cout, opt, nr, qerr, var);
    put_nlatab(cout, opt, nr, qerr, var);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(opt, nr, var);
}

int main(int argc, char* argv[]) // メインルメーチン
{
  nlraclass nlra;               // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) nlra.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { nlra.get(argv[i], opt);  fok = false; }
    }
    nlra.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occured !");
    return 255;
  }
  return 0;
}
