// roce.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-26 10:07:04 tac>
//
// 群項目と集計項目からROC分析を行う。
// カットオフポイントはROC曲線最適点、つまりROC曲線が感度1・特異度1の点
// (左上隅)に最も近い点か、感度と特異度の平均が最大となる点、
// 正診率が最大となる点または指定された点のどれかで計算。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W roce roce.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "roce  Version 7.41e          Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:roce  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The ROC analysis with the first item as the group and the second item",
  "  as the analysis data.",
  "<OPTION>",
  "  -co[{low|up}]={a|c|r|y|<VALUE>}:specify the cut-off point",
  "    co:automatic judgement  colow:lower limit  coup:upper limit",
  "    a:maximum accuracy point  c:maximum chi-squared point",
  "    r:optomum ROC curve point(default)",
  "    y:maximum Youden index point  <VALUE>:specified point",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=roce_dns.d",
  "    output data:diagnostic rate",
  "  -f=<FILE>:read options from the specified file",
  "  -g{1|2}=<CODE>:specify the code of group1(disease) or group2(normal)",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={roc|roci}[,...]:select the data for Ngraph",
  "    roc:ROC curve  roci:CI of ROC curve  default:all",
  "  -ngpname[=<FILE>]:specify the file to outpu Ngraph data",
  "    default:no output  -ngpname:-ngpname=roce_ngp.dat",
  "  -res=[-][{def|all|auc|cc[<C-COEF>]|or|roc|stat|uda}][,...]",
  "       :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    auc:AUC of ROC curve  cc[<C-COEF>]:confidence intervals(0=<C-COEF<1)",
  "    or:odds ratio  roc:ROC analysis  stat:basic statistics",
  "    uda:univariate linear discriminant analysis",
  "    default:automatic judgement according to data",
  "  -pi=<PRIOR-P>:specify the prior probability of group1(0<PRIOR-P<1)"
};

const string Title = "=== Receiver Operating Characteristic Analysis ==="; // 表題
const string DNSNAMEdef = "roce_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "roce_ngp.dat"; // 省略時Ngraphファイル名
constexpr int RDdef = 50;       // 座標軸の分割数
constexpr double PIdef = 0.5;   // 省略時事前確率

enum cotype                     // カットオフポイントの種類
{
  conon,                        // 指定なし
  coac,                         // 正診率最大点
  cocm,                         // 最大χ2乗法
  coroc,                        // ROC曲線最適点
  coyi,                         // Youden index:(感度+特異度-1)最大点
  colim,                        // 境界値(上下限は自動判定)
  lowlim,                       // 下限境界値
  uplim,                        // 上限境界値
  coetc                         // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  cotype vco;                   // カットオフポイントの種類
  double vlim;                  // 境界値
  string vdnsname;              // DANSファイル名
  vecstr vgvec;                 // 群コードベクトル
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  double vpi;                   // 群1の事前確率
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  cotype co() const             // カットオフポイントの種類を返す
  { return vco; }
  double lim() const            // 境界値を返す
  { return vlim; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  const string& g(int idx) const // 群コードを返す
  { return vgvec[idx]; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  double pi() const             // 群1の事前確率を返す
  { return vpi; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int search_g(const string& gcode) const; // 群コードを検索する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vco(coroc), vlim(BLANKf), vgvec(2), vdnsname(), vngpname(), vngplst(), vngpmap(), vpi(BLANKf), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == 0) return;     // 初期設定せず
  vgvec[0] = "1";               // 群1のコードだけ初期設定
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0][0] == 'c' && opt[0][1] == 'o') // カットオフポイントの種類の指定
  {
    if (opt[0] == "co")
    {
      if (opt[1].empty() || opt[1] == "r") vco = coroc;
      else if (opt[1] == "a") vco = coac;
      else if (opt[1] == "c") vco = cocm;
      else if (opt[1] == "y") vco = coyi;
      else vco = colim;
    }
    else if (opt[0] == "colow") vco = lowlim;
    else if (opt[0] == "coup") vco = uplim;
    else err = badopt(vpname, optstr, USAGE);
    if (vco == colim || vco == lowlim || vco == uplim) // 指定境界値の時
    {
      vlim = strtodf(opt[1]);
      if (vlim == BLANKf) err = badopt(vpname, optstr, USAGE);
    }
  }
  else if (opt[0] == "dnsname") // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "g1")      // 群1のコード指定
  {
    if (opt[1].empty()) vgvec[0].erase();
    else vgvec[0] = opt[1];
  }
  else if (opt[0] == "g2")      // 群2のコード指定
  {
    if (opt[1].empty()) vgvec[1].erase();
    else vgvec[1] = opt[1];
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "pi")      // 群1の事前確率指定
  {
    vpi = opt[1].empty() ? PIdef : strtodf(opt[1]);
    if (vpi <= 0.0 || 1.0 <= vpi) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::search_g(const string& gcode) const // 群コードを検索する
{
  if (gcode == BLANKs) return -1; // 欠損データはヒットさせない
  int idx = -1;
  if (vgvec[0].empty())         // 群1のコード指定がない場合
  {
    idx = 0;
    if (gcode == vgvec[1]) idx = 1;
  }
  else if (gcode == vgvec[0])
  { idx = 0; }
  else if (vgvec[1].empty())    // 群2のコード指定がない場合
  { idx = 1; }
  else if (gcode == vgvec[1])
  { idx = 1; }
  return idx;                   // -1:ヒットせず,0:群1,1:群2
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni != 2) return disperr(vpname, "Number of items must be 2 !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["roc"] = vngpmap["roci"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["roc"] = vngpmap["roci"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["auc"] = vresmap["cc"] = vresmap["or"] = vresmap["roc"] = vresmap["stat"] = true;
  vresmap["uda"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["auc"] = vresmap["cc"] = vresmap["or"] = vresmap["roc"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["auc"] = vresmap["cc"] = vresmap["or"] = vresmap["roc"] = vresmap["stat"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["auc"] = vresmap["cc"] = vresmap["or"] = vresmap["roc"] = vresmap["stat"] = vresmap["uda"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class data_less                 // ケースデータの比較用関数オブジェクト
{
private:
  int kidx;                     // ソートキー項目のインデックス
public:
  data_less(int kidx0) : kidx(kidx0) // コンストラクタ
  {}
  ~data_less()                  // デストラクタ
  {}
  bool operator()(datac& cdata1, datac& cdata2) const; // ケースデータを比較する
};

bool data_less::operator()(datac& cdata1, datac& cdata2) const
// ケースデータを比較する
{
  int comp = 0;
  double fdata1 = cdata1.fdata(kidx);
  double fdata2 = cdata2.fdata(kidx);
  if (fdata1 < fdata2) comp = -1;
  else if (fdata1 > fdata2) comp = 1;
  return (comp < 0) ? true : false; // 小さいときだけtrueを返す
}

class diag                      // 診断率とROCクラス
{
private:
  cotype vco;                   // 境界値の種類
  cotype vcolu;                 // 境界値の上限／下限フラッグ
  double vpi;                   // 群1の事前確率
  double vlim;                  // 境界値
  int vtp;                      // True-Positive例数
  int vtn;                      // True-Negative例数
  int vfp;                      // False-Positive例数
  int vfn;                      // False-Negative例数
  double vauc;                  // ROC曲線のAUC(曲線下面積)
  double vaucv;                 // ROC曲線のAUCの分散
  double vu;                    // Mann-WhitneyのU値
  double vuv;                   // Mann-WhitneyのU値の分散
  double vnlim;                 // 理論的境界値
  double vnsn;                  // 理論的境界値による感度
  double vnsp;                  // 理論的境界値による特異度
  double vnac;                  // 理論的境界値による正診率
  double vnppv;                 // 理論的境界値によるPPV
  double vnnpv;                 // 理論的境界値によるNPV
  double vnauc;                 // 理論的AUC
  double vd2;                   // マハラノビスの汎距離
  double vppr;                  // 誤判別確率
  vector<statv<double>> statvec; // 基礎統計量ベクトル
  void cal_roc(ldata& datalst); // ROC分析を行う
  void cal_dis();               // 単変量判別分析を行う
public:
  diag() : vco(conon), vcolu(lowlim), vpi(BLANKf), vlim(BLANKf), vtp(0), vtn(0), vfp(0), vfn(0), vauc(BLANKf), vaucv(BLANKf), vu(BLANKf), vuv(BLANKf), vnlim(BLANKf), vnsn(BLANKf), vnsp(BLANKf), vnac(BLANKf), vnppv(BLANKf), vnnpv(BLANKf), vnauc(BLANKf), vd2(BLANKf), vppr(BLANKf), statvec(2)
  // デフォールトコンストラクタ
  {}
  explicit diag(cotype co0, double pi0 = BLANKf, double lim0 = BLANKf) : vco(co0), vcolu(lowlim), vpi(pi0), vlim(lim0), vtp(0), vtn(0), vfp(0), vfn(0), vauc(BLANKf), vaucv(BLANKf), vu(BLANKf), vuv(BLANKf), vnlim(BLANKf), vnsn(BLANKf), vnsp(BLANKf), vnac(BLANKf), vnppv(BLANKf), vnnpv(BLANKf), vnauc(BLANKf), vd2(BLANKf), vppr(BLANKf), statvec(2)
  // コンストラクタ
  {}
  virtual ~diag()               // デストラクタ
  {}
  diag(const diag& diag0) : vco(diag0.vco), vcolu(diag0.vcolu), vpi(diag0.vpi), vlim(diag0.vlim), vtp(diag0.vtp), vtn(diag0.vtn), vfp(diag0.vfp), vfn(diag0.vfn), vauc(diag0.vauc), vaucv(diag0.vaucv), vu(diag0.vu), vuv(diag0.vuv), vnlim(diag0.vnlim), vnsn(diag0.vnsn), vnsp(diag0.vnsp), vnac(diag0.vnac), vnppv(diag0.vnppv), vnnpv(diag0.vnnpv), vnauc(diag0.vnauc), vd2(diag0.vd2), vppr(diag0.vppr), statvec(diag0.statvec)
  // コピーコンストラクタ
  {}
  const diag& operator=(const diag& diag0); // コピー代入演算子
  cotype co() const             // 境界値の種類を返す
  { return vco; }
  cotype co(cotype co0)         // 境界値の種類を設定する
  { return (vco = co0); }
  cotype colu() const           // 境界値の上限／下限フラッグを返す
  { return vcolu; }
  cotype colu(cotype colu0)    // 境界値の上限／下限フラッグを設定する
  { return (vcolu = colu0); }
  double pi()                   // 群1の事前確率を返す
  { return vpi; }
  double pi(double pi0)         // 群1の事前確率を設定する
  { return (vpi = pi0); }
  double lim() const            // 境界値を返す
  { return vlim; }
  double lim(double lim0)       // 境界値を設定する
  { return (vlim = lim0); }
  int tp() const                // True-Positive例数を返す
  { return vtp; }
  int tn() const                // True-Negative例数を返す
  { return vtn; }
  int fp() const                // False-Positive例数を返す
  { return vfp; }
  int fn() const                // False-Negative例数を返す
  { return vfn; }
  double sn() const             // 感度を返す
  { return ((vtp + vfn) < 1) ? 0.0 : double(vtp) / double(vtp + vfn); }
  double sp() const             // 特異度を返す
  { return ((vfp + vtn) < 1) ? 0.0 : double(vtn) / double(vfp + vtn); }
  double ac() const             // 正診率を返す
  { return ((vtp + vfn + vfp + vtn) < 1) ? 0.0 : double(vtp + vtn) / double(vtp + vfn + vfp + vtn); }
  double plr() const;           // 陽性尤度比を返す
  double nlr() const;           // 陰性尤度比を返す
  double ppv();                 // 陽性予測値を返す
  double npv();                 // 陰性予測値を返す
  double auc() const            // ROC曲線のAUC(曲線下面積)を返す
  { return vauc; }
  double aucv() const           // ROC曲線AUCの分散を返す
  { return vaucv; }
  double u() const              // Mann-WhitneyのU値を返す
  { return vu; }
  double uv() const             // Mann-WhitneyのU値の分散を返す
  { return vuv; }
  double nlim() const           // 理論的境界値を返す
  { return vnlim; }
  double nsn() const            // 理論的境界値による感度を返す
  { return vnsn; }
  double nsp() const            // 理論的境界値による特異度を返す
  { return vnsp; }
  double nac() const            // 理論的境界値による正診率を返す 
  { return vnac; }
  double nppv() const           // 理論的境界値によるPPVを返す 
  { return vnppv; }
  double nnpv() const           // 理論境界値によるNPVを返す 
  { return vnnpv; }
  double nauc() const           // 理論的AUCを返す
  { return vnauc; }
  double d2() const             // マハラノビスの汎距離を返す
  { return vd2; }
  double ppr() const            // 誤判別確率を返す
  { return vppr; }
  statv<double>& dstat(size_t idx) // 診断項目の基礎統計量を返す
  { return statvec[idx]; }
  int cal_diag(ldata& datalst); // ROC分析と単変量判別分析で診断率を計算する
  int cal_wtest(double& dev, double& vd, double& t, int& df);
  // Welchの検定結果を返す
};

const diag& diag::operator=(const diag& diag0) // コピー代入演算子
{
  if (&diag0 == this) return *this; // 自分自身への代入を防ぐ
  vco = diag0.vco;  vcolu = diag0.vcolu;  vpi = diag0.vpi;  vlim = diag0.vlim;
  vtp = diag0.vtp;  vtn = diag0.vtn;  vfp = diag0.vfp;  vfn = diag0.vfn;
  vauc = diag0.vauc;  vaucv = diag0.vaucv;  vu = diag0.vu;  vuv = diag0.vuv;
  vnlim = diag0.vnlim;  vnsn = diag0.vnsn;
  vnsp = diag0.vnsp;  vnac = diag0.vnac;
  vnppv = diag0.vnppv;  vnnpv = diag0.vnnpv;
  vnac = diag0.vnac;  vnauc = diag0.vnauc;
  vd2 = diag0.vd2;  vppr = diag0.vppr;
  statvec.resize(diag0.statvec.size());
  statvec[0] = diag0.statvec[0];  statvec[1] = diag0.statvec[1];
  return *this;
}

double diag::plr() const        // 陽性尤度比を返す
{
  if (sgn(1.0 - sp()) <= 0) return (sn() > 0.0) ? BLANKf : 0.0; // BLANKf:∞
  return sn() / (1.0 - sp());
}

double diag::nlr() const        // 陰性尤度比を返す
{
  if (sgn(sp()) <= 0) return (sn() < 1.0) ? BLANKf : 0.0;
  return (1.0 - sn()) / sp();
}

double diag::ppv()              // 陽性予測値を返す
{
  if (sgn(vpi) <= 0) return BLANKf;
  if (sgn(sn()) <= 0) return (sp() < 1.0) ? 0.0 : 1.0;
  return 1.0 / (1.0 + (1.0 - vpi) / vpi * (1.0 - sp()) / sn());
}

double diag::npv()              // 陰性予測値を返す
{
  if (sgn(1.0 - vpi) <= 0) return BLANKf;
  if (sgn(sp()) <= 0) return (sn() < 1.0) ? 0.0 : 1.0;
  return 1.0 / (1.0 + vpi / (1.0 - vpi) * (1.0 - sn()) / sp());
}

int diag::cal_diag(ldata& datalst)
// ROC分析と単変量判別分析で診断率を計算する
{
  for (auto& cdata : datalst) statvec[(size_t)(cdata.fdata(0))].sigma(cdata.fdata(1));
  int n = statvec[0].n() + statvec[1].n();
  if (n < 1) return n;          // データなし
  if (statvec[0].n() < 1 || statvec[1].n() < 1) return 0; // 計算できず
  if (vpi == BLANKf) vpi = (double)(statvec[0].n()) / (double)n; // -pi指定なしの時
  cal_roc(datalst);             // ROC分析
  cal_dis();                    // 単変量判別分析
  return n;                     // データ数を返す
}

void diag::cal_roc(ldata& datalst) // ROC分析を行う
{
  int n1 = statvec[0].n(), n2 = statvec[1].n();
  int n = n1 + n2, tp = n1, dfn = 0, tn = 0, dtn = 0;
  int ac_max = 0, tp_max = 0, tn_max = n2;
  double csqr_max = 0.0, yi_max = 0.0;
  double dist_min = 2.0, m1 = statvec[0].mean(), m2 = statvec[1].mean();
  cotype co = vco;
  vcolu = lowlim;
  if (vco == uplim)             // 上限境界値指定の場合
  { datalst.reverse();  vcolu = uplim; } // データリストを逆順にする
  else if (vco != lowlim)       // 上下限自動判定の場合
  {
    co = vcolu = lowlim;
    if (m1 < m2) { co = vcolu = uplim;  datalst.reverse(); }
  }
  double w = 0.0, q1 = 0.0, q2 = 0.0, k = 0.0;
  double fpre = (datalst.begin())->fdata(1), lim = vlim;
  bool firstf = true;
  for (auto& cdata : datalst)
  {
    double f = cdata.fdata(1);
    if (f != fpre || firstf)
    {
      tp -= dfn;  tn += dtn;
      w += (double)(tp * dtn) + 0.5 * (double)(dfn * dtn);
      q1 += (double)dtn * ((double)(tp * (tp + dfn)) + (double)(dfn * dfn) / 3.0);
      q2 += (double)dfn * ((double)(tn * (tn - dtn)) + (double)(dtn * dtn) / 3.0);
      int g = dfn + dtn;
      k += (double)g * (double)(g * g - 1);
      int ac = tp + tn;
      int fn = n1 - tp, fp = n2 - tn;
      double csqr = 0.0;
      double yi = 0.0, x = 0.0, y = 0.0, dist = 0.0;
      fpre = f;  dfn = dtn = 0;
      switch (vco)
      {
      case lowlim :             // 下限境界値指定
        if (tp_max == 0 && lim <= f) { tp_max = tp;  tn_max = tn; }
        break;
      case uplim :              // 上限境界値指定
        if (tp_max == 0 && f <= lim) { tp_max = tp;  tn_max = tn; }
        break;
      case colim :              // 境界値指定(上下限は自動判定)
        if (co == lowlim)
        {
          if (tp_max == 0 && lim <= f) { tp_max = tp;  tn_max = tn; }
        }
        else
        {
          if (tp_max == 0 && f <= lim) { tp_max = tp;  tn_max = tn; }
        }
        break;
      case coac :             // 正診率基準
        if (ac > ac_max)
        { ac_max = ac;  tp_max = tp;  tn_max = tn;  lim = f; }
        break;
      case cocm : csqr = csqr22(tn, fp, fn, tp);
        if (csqr > csqr_max)    // 最大χ2乗法
        { csqr_max = csqr;  tp_max = tp;  tn_max = tn;  lim = f; }
        break;
      case coyi : yi = (double)tp / (double)n1 - (double)fp / (double)n2;
        // Youden index
        if (yi > yi_max)
        { yi_max = yi;  ac_max = ac;  tp_max = tp;  tn_max = tn;  lim = f; }
        break;
      case coroc : x = (double)fn / (double)n1; // ROC曲線最適点基準
        y = (double)fp / (double)n2;
        dist = x * x + y * y;
        if (dist < dist_min)
        {
          dist_min = dist;
          ac_max = ac;  tp_max = tp;  tn_max = tn;  lim = f;
        }
        break;
      }
      firstf = false;
    }
    if ((int)(cdata.fdata(0)) == 0) ++dfn;
    else ++dtn;
  }
  vco = co;  vlim = lim;
  vtp = tp_max;  vfn = n1 - tp_max;  vfp = n2 - tn_max;  vtn = tn_max;
  w += 0.5 * (double)(dfn * dtn);
  q1 += (double)dtn * ((double)(dfn * dfn) / 3.0);
  q2 += (double)dfn * ((double)(n2 * tn) + (double)(dtn * dtn) / 3.0);
  k = 1.0 - k / (double)(n - 1) / (double)n / (double)(n + 1);
  vu = w;  vuv = k * (double)n1 * (double)n2 * (double)(n + 1) / 12.0;
  w = w / (double)n1 / (double)n2;
  q1 = q1 / (double)n1 / (double)n1 / (double)n2;
  q2 = q2 / (double)n1 / (double)n2 / (double)n2;
  vauc = w;
  vaucv = (w * (1.0 - w) + (n1 - 1) * (q1 - w * w) + (n2 - 1) * (q2 - w * w)) / (double)n1 / (double)n2;
  const double Veps = DBL_EPSILON * 5.0;
  if (vaucv < Veps) vaucv = 0.0;
}

void diag::cal_dis()            // 単変量判別分析を行う
{
  int n1 = statvec[0].n(), n2 = statvec[1].n();
  double m1 = statvec[0].mean(), m2 = statvec[1].mean();
  double sd1 = statvec[0].sd(), sd2 = statvec[1].sd();
  double v1 = statvec[0].vr(), v2 = statvec[1].vr();
  vd2 = (v1 + v2 > 0.0) ? vd2 = 2.0 * (m1 - m2) * (m1 - m2) / (v1 + v2) : BLANKf;
  vnlim = (m1 + m2) * 0.5;      // 判別分析の境界値
  double dev = m1 - vnlim;
  if (vcolu == uplim) dev = -dev;
  vnsn = (sd1 > 0.0) ? sndis(dev / sd1) : 1.0;
  dev = vnlim - m2;
  if (vcolu == uplim) dev = -dev;
  vnsp = (sd2 > 0.0) ? sndis(dev / sd2) : 1.0;
  vnac = (n1 + n2 > 0) ? ((double)n1 * vnsn + (double)n2 * vnsp) / (double)(n1 + n2) : 1.0;
  double orpi = (1.0 - pi()) / pi();
  vnppv = (vnsn > 0.0) ? 1.0 / (1.0 + orpi * (1.0 - vnsp) / vnsn) : 0.0;
  vnnpv = (vnsp > 0.0) ? 1.0 / (1.0 + (1.0 - vnsn) / vnsp / orpi) : 0.0;
  vnauc = 1.0;  vppr = 0.0;
  if (vd2 == BLANKf) return;
  double z = sqrt(vd2 * 0.5);
  vnauc = sndis(z);              // 理論AUC
  z = sqrt(vd2) * 0.5;
  vppr = 1.0 - (double)(sndis(z));
}

int diag::cal_wtest(double& dev, double& vd, double& t, int& df)
// Welchの検定結果を返す
{
  dev = fabs(statvec[0].mean() - statvec[1].mean());
  int n1 = statvec[0].n(), n2 = statvec[1].n();
  double v1 = statvec[0].vr() / (double)n1;
  double v2 = statvec[1].vr() / (double)n2;
  vd = v1 + v2;
  df = (vd > 0.0) ? (int)(vd * vd / (v1 * v1 / (n1 - 1) + v2 * v2 / (n2 -1))) : n1 + n2  - 2;
  t = (vd > 0.0) ? dev / sqrt(vd) : 0.0;
  return df;                    // 自由度を返す
}

class rocclass                  // rocメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout); // 分析結果のタイトルとデータ識別情報を出力する
  void put_result(ostream& fout, option& opt, diag& dg);
  // 診断率の計算結果を出力する
  void put_stat(ostream& fout, const option& opt, diag& dg) const;
  // 基礎統計量を出力する
  void put_roca(ostream& fout, option& opt, diag& dg); // ROC分析の結果を出力する
  void put_pnpv(ostream& fout, option& opt, diag& dg);
  // 陽性予測値と陰性予測値を出力する
  void put_binci(ostream& fout, int n, int nt, double cc) const;
  // F分布を利用した出現率の信頼区間を出力する
  void put_lrci(ostream& fout, double& lrl, double& lru, int n1, int nt1, int n2, int nt2, double cc) const;
  // 尤度比の信頼区間を出力し、値を返す
  void put_pvci(ostream& fout, double& pvl, double& pvu, bool ppv, double p0, double lrl, double lru, double cc) const;
  // 陽性/陰性予測値の信頼区間を出力し、値を返す
  void put_auc(ostream& fout, option& opt, diag& dg);
  // ROC曲線のAUC(曲線下面積)とMann-WhitneyのU値と検定結果を出力する
  void put_ozr(ostream& fout, option& opt, diag& dg);
  // オッズ比(True群/False群)とその検定結果を出力する
  void put_uda(ostream& fout, option& opt, diag& dg);
  // 単変量判別分析結果と理論的AUCを出力する
  void put_dns(const option& opt, diag& dg);
  // データファイル形式の診断率を出力する
  void put_dnsid(ofstream& fout);
  // DANSデータのタイトルとデータ識別情報を出力する
  void plot_ngraph(option& opt, diag& dg); // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_roc(ofstream& fout, diag& dg); // Ngraph用ROC曲線を出力する
  void plot_roci(ofstream& fout, const option& opt, diag& dg);
  // Ngraph用ROC曲線信頼区間を出力する
  void plot_nroc(ofstream& fout, diag& dg) const;
  // Ngraph用理論ROC曲線を出力する
  void plot_nroci(ofstream& fout, const option& opt, diag& dg) const;
  // Ngraph用理論ROC曲線信頼区間を出力する
public:
  rocclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~rocclass()                   // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);                     // 集計結果を出力する
};

void rocclass::put_resid(ostream& fout)
// 分析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM   :" << id.item(0) << "\n";
  fout << "ANALYSIS ITEM:" << id.item(1) << "\n";
}

void rocclass::put_result(ostream& fout, option& opt, diag& dg)
// 診断率の計算結果を出力する
{
  constexpr int LENline = 74;   // 区切り線の長さ
  put_resid(fout);
  bool putres = false;
  if (opt.res("stat"))
  {
    fout << "\n";
    putline(fout, LENline);
    put_stat(fout, opt, dg);
    putres = true;
  }
  if (opt.res("roc"))
  {
    if (putres == false) fout << "\n";
    else putres = true;
    putline(fout, LENline);
    put_roca(fout, opt, dg);
  }
  if (opt.res("auc"))
  {
    if (putres == false) fout << "\n";
    else putres = true;
    putline(fout, LENline);
    put_auc(fout, opt, dg);
  }
  if (opt.res("or"))
  {
    if (putres == false) fout << "\n";
    else putres = true;
    putline(fout, LENline);
    put_ozr(fout, opt, dg);
  }
  if (opt.res("uda"))
  {
    if (putres == false) fout << "\n";
    else putres = true;
    putline(fout, LENline);
    put_uda(fout, opt, dg);
  }
  if (putres) putline(fout, LENline);
  fout << "\n" << endl;
}

void rocclass::put_stat(ostream& fout, const option& opt, diag& dg) const
// 基礎統計量を出力する
{
  fout << "GROUP1(Disease):Code=";
  if ((opt.g(0)).empty()) fout << "other then " << opt.g(1) << "  ";
  else fout << opt.g(0) << "             ";
  fout << "True-Positive=" << dg.tp() << " False-Negative=" << dg.fn() << "\n";
  fout << "  ";
  putmsd(fout, (dg.dstat(0)).n(), (dg.dstat(0)).mean(), (dg.dstat(0)).vr());
  fout << "GROUP2(Normal) :Code=";
  if ((opt.g(1)).empty()) fout << "other then " << opt.g(0) << "  ";
  else fout << opt.g(1) << "             ";
  fout << "False-Positive=" << dg.fp() << " True-Negative=" << dg.tn() << "\n";
  fout << "  ";
  putmsd(fout, (dg.dstat(1)).n(), (dg.dstat(1)).mean(), (dg.dstat(1)).vr());
}

void rocclass::put_roca(ostream& fout, option& opt, diag& dg)
// ROC分析の結果を出力する
{
  string tstr("More than ");
  if (dg.colu() == uplim) tstr = "Less than ";
  fout << "ROC Analysis:" << tstr << dg.lim();
  if (opt.co() == coac) tstr = "(Maximum Acuuracy Point)";
  else if (opt.co() == cocm) tstr = "(Maximum Chi-squared Point)";
  else if (opt.co() == coyi) tstr = "(Maximum Youden Index Point)";
  else if (opt.co() == coroc) tstr = "(Optomum ROC Curve Point)";
  else if (opt.co() == colim) tstr = "(Specified Point)";
  else tstr.erase();
  fout << " is GROUP1" << tstr << "\n";
  int nt0 = dg.fn() + dg.tp();
  fout << "  sensitivity=" << dg.sn() << "(" << dg.tp() <<"/" << nt0 << ")";
  if (opt.res("cc")) put_binci(fout, dg.tp(), nt0, opt.cc());
  fout << "\n";
  int nt1 = dg.fp() + dg.tn();
  fout << "  specificity=" << dg.sp() << "(" << dg.tn() << "/" << nt1 << ")";
  if (opt.res("cc")) put_binci(fout, dg.tn(), nt1, opt.cc());
  fout << "\n";
  int nt = nt0 + nt1;
  fout << "  accuracy=" << dg.ac() << "(" << dg.tp() + dg.tn() <<"/" << nt << ")";
  if (opt.res("cc")) put_binci(fout, dg.tp() + dg.tn(), nt, opt.cc());
  fout << "\n";
  put_pnpv(fout, opt, dg);
}

void rocclass::put_pnpv(ostream& fout, option& opt, diag& dg)
// 陽性予測値と陰性予測値を出力する
{
  int nt0 = dg.fn() + dg.tp();
  int nt1 = dg.fp() + dg.tn();
  double lrpl = 1.0, lrpu = 1.0, lrnl = 1.0, lrnu = 1.0;
  fout << "  Positive Likelihood Ratio(LR+)=";
  if (dg.plr() == BLANKf)
  { fout << "∞\n"; }
  else
  {
    fout << dg.plr();
    if (dg.plr() > 0.0) put_lrci(fout, lrpl, lrpu, dg.tp(), nt0, dg.fp(), nt1, opt.cc());
  }
  fout << "  Negative Likelihood Ratio(LR-)=";
  if (dg.nlr() == BLANKf)
  { fout << "INFINITY\n"; }
  else
  {
    fout << dg.nlr();
    if (dg.nlr() > 0.0) put_lrci(fout, lrnl, lrnu, dg.fn(), nt0, dg.tn(), nt1, opt.cc());
  }
  fout << "Prior Probability:GROUP1=" << dg.pi() << "  GROUP2=" << 1.0 - dg.pi() << "\n";
  double ppvl = 1.0, ppvu = 1.0, npvl = 1.0, npvu = 1.0;
  fout << "  Positive Predictive Value(PPV)=" << dg.ppv();
  if (0.0 < dg.ppv() && dg.ppv() < 1.0) put_pvci(fout, ppvl, ppvu, true, dg.pi(), lrpl, lrpu, opt.cc());
  fout << "  Negative Predictive Value(NPV)=" << dg.npv();
  if (0.0 < dg.npv() && dg.npv() < 1.0) put_pvci(fout, npvl, npvu, false, dg.pi(), lrnl, lrnu, opt.cc());
  fout << "  Risk Ratio(RR)=";
  if (dg.npv() < 1.0)
  {
    double rr = dg.ppv() / (1.0 - dg.npv());
    fout << rr;
    if (opt.res("cc"))
    {
      double rrl = (npvl < 1.0) ? ppvl / (1.0 - npvl) : 1.0;
      double rru = (npvu < 1.0) ? ppvu / (1.0 - npvu) : 1.0;
      fout << " (" << opt.cc() * 100.0 << "%CI:" << rrl << "-" << rru << ")";
    }
  }
  else
  { fout << "infinity"; }
  fout << endl;
}

void rocclass::put_binci(ostream& fout, int n, int nt, double cc) const
// F分布を利用した出現率の信頼区間を出力する
{
  double plow = 0.0, pup = 1.0;
  calbexci(plow, pup, n, nt, cc);
  fout << " (" << cc * 100.0 << "%CI:" << plow << "-" << pup << ")";
}

void rocclass::put_lrci(ostream& fout, double& lrl, double& lru, int n1, int nt1, int n2, int nt2, double cc) const
// 尤度比の信頼区間を出力し、値を返す
{
  double lr = 1.0, b = 0.0, v = 0.0, ci = 0.0, bl = 0.0, bu = 0.0;
  lrl = 1.0;  lru = 1.0;
  if (n1 > 0 && nt1 > 0 && n2 > 0 && nt2 > 0)
  {
    lr = (double)n1 / (double)nt1 * (double)nt2 / (double)n2;
    b = log(lr);
    v = 1.0 / (double)n1 + 1.0 / (double)n2 - 1.0 / (double)nt1 - 1.0 / (double)nt2;
    const double Veps = DBL_EPSILON * fmax(5.0, lr);
    if (v < Veps) v = 0.0;
    ci = ptoz(1.0 - cc) * sqrt(v);
    bl = b - ci;  bu = b + ci;
    lrl = exp(bl);  lru = exp(bu);
  }
  fout << " (" << cc * 100.0 << "%CI:" << lrl << "-" << lru << ")\n";
}

void rocclass::put_pvci(ostream& fout, double& pvl, double& pvu, bool ppv, double p0, double lrl, double lru, double cc) const
// 陽性/陰性予測値の信頼区間を出力し、値を返す
{
  if (p0 <= 0.0 || 1.0 <= p0) return;
  pvl = pvu = 1.0;
  if (lrl > 0.0 && lru > 0.0)
  {
    if (ppv)                    // 陽性予測値
    {
      pvl = 1.0 / (1.0 + (1.0 - p0) / p0 / lrl);
      pvu = 1.0 / (1.0 + (1.0 - p0) / p0 / lru);
    }
    else                        // 陰性予測値
    {
      pvl = 1.0 / (1.0 + p0 / (1.0 - p0) * lru);
      pvu = 1.0 / (1.0 + p0 / (1.0 - p0) * lrl);
    }
  }
  if (pvl < 0.0) pvl = 0.0;
  if (pvu > 1.0) pvu = 1.0;
  fout << " (" << cc * 100.0 << "%CI:" << pvl << "-" << pvu << ")\n";
}

void rocclass::put_auc(ostream& fout, option& opt, diag& dg)
// ROC曲線のAUC(曲線下面積)とMann-WhitneyのU値と検定結果を出力する
{
  double se = sqrt(dg.aucv()), z = 0.0, ci = 0.0;
  fout << "AUC of ROC Curve=" << dg.auc();
  if (opt.res("cc"))
  {
    z = ptoz(1.0 - opt.cc());
    ci = z * se;
    double aucl = dg.auc() - ci, aucu = dg.auc() + ci;
    if (aucl < 0.0) aucl = 0.0;
    if (aucu > 1.0) aucu = 1.0;
    fout << " (" << opt.cc() * 100.0 << "%CI:" << aucl << "-" << aucu << ")";
  }
  fout << "  SE=" << se << "\n";
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  double um = (double)n1 * (double)n2 * 0.5;
  se = sqrt(dg.uv());
  fout << "Mann-Whitney's U-test:U=" << dg.u() << "=AUCx" << n1 << "x" << n2;
  if (opt.res("cc"))
  {
    ci = z * se;
    double ul = dg.u() - ci, uu = dg.u() + ci;
    if (ul < 0.0) ul = 0.0;
    if (uu > n1 * n2) uu = double(n1) * double(n2);
    fout << " (" << opt.cc() * 100.0 << "%CI:" << ul << "-" << uu << ")";
  }
  z = (sgn(se) > 0) ? (dg.u() - um) / se : 0.0;
  ldbl pr = ztop(z);
  fout << "\n  z=" << z << "  p=" << pr << pind(pr) << endl;
}

void rocclass::put_ozr(ostream& fout, option& opt, diag& dg)
// オッズ比(群1/群2)とその検定結果を出力する
{
  double n11 = dg.tp(), n12 = dg.fn(), n21 = dg.fp(), n22 = dg.tn();
  if (dg.tp() == 0 || dg.fn() == 0 || dg.fp() == 0 || dg.tn() == 0)
  { n11 += 0.5;  n12 += 0.5;  n21 += 0.5;  n22 += 0.5; }
  double ozr = n11 / n12 * n22 / n21;
  double b = log(ozr), v = 1.0 / n11 + 1.0 / n12 + 1.0 / n21 + 1.0 / n22;
  double csqr = b * b / v, se = sqrt(v), ci = 0.0;
  fout << "Odds Ratio(GROUP1/GROUP2):OR=" << ozr;
  if (opt.res("cc"))
  {
    ci = ptoz(1.0 - opt.cc()) * se;
    fout << " (" << opt.cc() * 100.0 << "%CI:" << exp(b - ci) << "-" << exp(b + ci) << ")\n";
  }
  fout << "  ln(OR)=" << b;
  if (opt.res("cc")) fout << " (" << opt.cc() * 100.0 << "%CI:" << b - ci << "-" << b + ci << ")";
  ldbl pr = ctop(csqr, 1);
  fout << "\n  chi-squared=" << csqr << "  DF=1  p=" << pr << pind(pr) << endl;
}

void rocclass::put_uda(ostream& fout, option& opt, diag& dg)
// 単変量判別分析結果と理論的AUCを出力する
{
  string tstr("More then ");
  if (dg.colu() == uplim) tstr = "Less then ";
  fout << "Univariate Linear Discriminant Analysis:" << tstr << dg.nlim();
  fout << " is GROUP1\n";
  fout << "  sencitivity=" << dg.nsn() << "  specificity=" << dg.nsp() << "  accuracy=" << dg.nac() << "\n";
  if (0.0 < opt.pi()) fout << "  PPV=" << dg.nppv() << "  NPV=" << dg.nnpv() << "\n";
  if (dg.d2() == BLANKf) return;
  double n1 = (double)((dg.dstat(0)).n()), n2 = (double)((dg.dstat(1)).n());
  double f = dg.d2() * n1 * n2 / (n1 + n2);
  double r2 = f / (f + n1 + n2 - 2.0);
  fout << "Mahalanobis's D-squared=" << dg.d2() << "  Correlation Ratio=" << sqrt(r2) << "\n";
  fout << "  Discriminant-p=phi(D/2)=phi(" << sqrt(dg.d2()) * 0.5 << ")=" << 1.0 - dg.ppr() << "(Probrem-p=1-p=" << dg.ppr() << ")\n";
  fout << "AUC of ROC Curve=phi(D/square root 2)=phi(" << sqrt(dg.d2() * 0.5) << ")=" << dg.nauc() << "\n";
  double dev = 0.0, vd = 0.0, t = 0.0;
  int df = 0;
  dg.cal_wtest(dev, vd, t, df);
  double pr = ttop(t, df);
  fout << "Welch's Test:t=" << t << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
  if (opt.res("cc"))
  {
    double ci = calmcci(mcfis, opt.cc(), 2, 1, df, 1, vd);
    fout << "  " << opt.cc() * 100.0 << "%CI of the Mean Difference=";
    fout << dev << " +/- " << ci << "(" << dev - ci << "-" << dev + ci << ")\n";
  }
}

void rocclass::put_dns(const option& opt, diag& dg) // データファイル形式の診断率を出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  put_dnsid(fout);
  datac tdata(3);
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  int tp = n1, dfn = 0, tn = 0, dtn = 0;
  double fpre = (datalst.begin())->fdata(1);
  for (auto& cdata : datalst)
  {
    double f = cdata.fdata(1);
    if (f != fpre)
    {
      tp -= dfn;  tn += dtn;
      fpre = f;  dfn = dtn = 0;
      double sn = (double)tp / (double)n1, sp = (double)tn / (double)n2;
      tdata.assign(0, f);  tdata.assign(1, sn);  tdata.assign(2, sp);
      tdata.put(fout);
    }
    if ((int)(cdata.fdata(0)) == 0) ++dfn;
    else ++dtn;
  }
}

void rocclass::put_dnsid(ofstream& fout)
// DANSデータのタイトルとデータ識別情報を出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  idd tid(id.title(), 3);
  tid.assign_item(0, "Cut-off Point");
  tid.assign_item(1, "Sensitivity");
  tid.assign_item(2, "Specificity");
  tid.put_item(fout, formtag);
}

void rocclass::plot_ngraph(option& opt, diag& dg) // Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("roc"))
  {
    fout << "-s1 -dline -w80 -zx1,0,0.2 -zy0,1,0.2\n";
    plot_ngpid(fout);
    plot_roc(fout, dg);
    fout << NGPdelm << "\n";
    plot_nroc(fout, dg);
    delml = true;
  }
  if (opt.ngp("roci"))
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80 -zx1,0,0.2 -zy0,1,0.2\n";
    plot_ngpid(fout);
    plot_roci(fout, opt, dg);
    fout << NGPdelm << "\n";
    plot_nroci(fout, opt, dg);
  }
}

void rocclass::plot_ngpid(ofstream& fout)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%GROUP ITEM   :" << id.item(0) << "\n";
  fout << "%ANALYSIS ITEM:" << id.item(1) << endl;
}

void rocclass::plot_roc(ofstream& fout, diag& dg) // Ngraph用ROC曲線を出力する
{
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  int tp = n1, dfn = 0, tn = 0, dtn = 0;
  fout << "%<ROC CURVE>\n";
  fout << "0 1\n";
  double fpre = (datalst.begin())->fdata(1);
  double f = 0.0, sn = 0.0, sp = 0.0;
  for (auto& cdata : datalst)
  {
    f = cdata.fdata(1);
    if (f != fpre)
    {
      tp -= dfn;  tn += dtn;
      fpre = f;  dfn = dtn = 0;
      sn = (double)tp / (double)n1;  sp = (double)tn / (double)n2;
      fout << sp << " " << sn << "\n";
    }
    if ((int)(cdata.fdata(0)) == 0) ++dfn;
    else ++dtn;
  }
  fout << "1 0\n";
}

void rocclass::plot_roci(ofstream& fout, const option& opt, diag& dg)
// Ngraph用ROC曲線の信頼区間を出力する
{
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  int tp = n1, dfn = 0, tn = 0, dtn = 0;
  double fpre = (datalst.begin())->fdata(1);
  double f = 0.0, sn = 0.0, sp = 0.0, plow = 0.0, pup = 1.0;
  fout << "%<" << opt.cc() * 100.0 << "% CI of ROC CURVE:LOWER LIMIT>\n"; // 信頼区間下限
  fout << "0 1\n";
  for (auto& cdata : datalst)
  {
    f = cdata.fdata(1);
    if (f != fpre)
    {
      tp -= dfn;  tn += dtn;
      fpre = f;  dfn = dtn = 0;
      sp = (double)tn / (double)n2;
      calbexci(plow, pup, tp, n1, opt.cc());
      fout << sp << " " << plow << "\n";
    }
    if ((int)(cdata.fdata(0)) == 0) ++dfn;
    else ++dtn;
  }
  fout << "1 0\n";
  tp = n1;  dfn = 0;  tn = 0;  dtn = 0;
  fpre = (datalst.begin())->fdata(1);
  fout << "= =\n";
  fout << "%<" << opt.cc() * 100.0 << "% CI of ROC CURVE:UPPER LIMIT>\n"; // 信頼区間上限
  fout << "0 1\n";
  for (auto& cdata : datalst)
  {
    f = cdata.fdata(1);
    if (f != fpre)
    {
      tp -= dfn;  tn += dtn;
      fpre = f;  dfn = dtn = 0;
      sp = (double)tn / (double)n2;
      calbexci(plow, pup, tp, n1, opt.cc());
      fout << sp << " " << pup << "\n";
    }
    if (int(cdata.fdata(0)) == 0) ++dfn;
    else ++dtn;
  }
  fout << "1 0\n";
}

void rocclass::plot_nroc(ofstream& fout, diag& dg) const
// Ngraph用理論的ROC曲線を出力する
{
  double m1 = (dg.dstat(0)).mean(), sd1 = (dg.dstat(0)).sd();
  double m2 = (dg.dstat(1)).mean(), sd2 = (dg.dstat(1)).sd();
  fout << "%<THEORETICAL ROC CURVE>\n";
  if (sgndiff(m1, m2) == 0 && sgn(sd1 + sd2) == 0) // データが定数の時
  {
    fout << "0 1\n";
    fout << "1 0\n";
    return;
  }
  double xmin = m2 - sd2 * 4.0, xmax = m1 + sd1 * 4.0;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { xmin = m1 - sd1 * 4.0;  xmax = m2 + sd2 * 4.0; }
  double dx = (xmax - xmin) / (double)RDdef;
  double sp = 0.0, sn = 0.0;
  if (dg.colu() == lowlim) fout << "0 1\n";
  else fout << "1 0\n";
  double x = xmin;
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    if (dg.colu() == lowlim)
    { sp = sndis((x - m2) / sd2);  sn = 1.0 - sndis((x - m1) / sd1); }
    else
    { sp = 1.0 - sndis((x - m2) / sd2);  sn = sndis((x - m1) / sd1); }
    fout << sp << " " << sn << "\n";
  }
  if (dg.colu() == lowlim) fout << "1 0\n";
  else fout << "0 1\n";
}

void rocclass::plot_nroci(ofstream& fout, const option& opt, diag& dg) const
// Ngraph用理論的ROC曲線の信頼区間を出力する
{
  int n1 = (dg.dstat(0)).n(), n2 = (dg.dstat(1)).n();
  double m1 = (dg.dstat(0)).mean(), sd1 = (dg.dstat(0)).sd();
  double m2 = (dg.dstat(1)).mean(), sd2 = (dg.dstat(1)).sd();
  if (sgndiff(m1, m2) == 0 && sgn(sd1 + sd2) == 0) // データが定数の時
  {
    fout << "%<" << opt.cc() * 100.0 << "% CI OF THEORETICAL ROC CURVE:LOWER LIMIT>\n";
    fout << "0 1\n";
    fout << "1 0\n";
    fout << "%<" << opt.cc() * 100.0 << "% CI OF THEORETICAL ROC CURVE:UPPER LIMIT>\n";
    fout << "0 1\n";
    fout << "1 0\n";
    return;
  }
  double xmin = m2 - sd2 * 4.0, xmax = m1 + sd1 * 4.0;
  if (m1 < m2)                  // 群1の平均値の方が小さい時
  { xmin = m1 - sd1 * 4.0;  xmax = m2 + sd2 * 4.0; }
  double dx = (xmax - xmin) / (double)RDdef;
  double sp = 0.0, sn = 0.0, plow = 0.0, pup = 1.0;
  fout << "%<" << opt.cc() * 100.0 << "% CI OF THEORETICAL ROC CURVE:LOWER LIMIT>\n"; // 信頼区間下限
  if (dg.colu() == lowlim) fout << "0 1\n";
  else fout << "1 0\n";
  double x = xmin;
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    if (dg.colu() == lowlim)
    { sp = sndis((x - m2) / sd2);  sn = 1.0 - sndis((x - m1) / sd1); }
    else
    { sp = 1.0 - sndis((x - m2) / sd2);  sn = sndis((x - m1) / sd1); }
    calbnmci(plow, pup, sn, n1, opt.cc());
    fout << sp << " " << plow << "\n";
  }
  if (dg.colu() == lowlim) fout << "1 0\n";
  else fout << "0 1\n";
  fout << "= =\n";
  fout << "%<" << opt.cc() * 100.0 << "% CI OF THEORETICAL ROC CURVE:UPPER LIMIT>\n"; // 信頼区間上限
  if (dg.colu() == lowlim) fout << "0 1\n";
  else fout << "1 0\n";
  x = xmin;
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    if (dg.colu() == lowlim)
    { sp = sndis((x - m2) / sd2);  sn = 1.0 - sndis((x - m1) / sd1); }
    else
    { sp = 1.0 - sndis((x - m2) / sd2);  sn = sndis((x - m1) / sd1); }
    calbnmci(plow, pup, sn, n1, opt.cc());
    fout << sp << " " << pup << "\n";
  }
  if (dg.colu() == lowlim) fout << "1 0\n";
  else fout << "0 1\n";
}

int rocclass::get(const char* c_fname, option& opt) // データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  nc = (int)(id.nc());          // ケース数変更用
  ldata_itr ditr = datalst.begin();
  while (ditr != datalst.end())
  {
    int idx = opt.search_g(ditr->sdata(0));
    double fdata = ditr->fdata(1);
    if (idx < 0 || fdata == BLANKf) // 欠損データはリストから削除する
    { ditr = datalst.erase(ditr);  --nc; }
    else
    {
      ditr->assign(0, (double)idx); // 群コードを群インデックスに変える
      ++ditr;
    }
  }
  if (nc < (int)(id.nc())) id.nc(nc); // ケース数を減らす
  return nc;
}

void rocclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  datalst.sort(data_less(1));    // 項目データ順にソートする
  diag dg(opt.co(), opt.pi(), opt.lim());
  if (dg.cal_diag(datalst) < 1)
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  if (opt.res("res")) put_result(cout, opt, dg);
  if (opt.dnsname() != "") put_dns(opt, dg);
  if (opt.ngpname() != "") plot_ngraph(opt, dg);
}

int main(int argc, char* argv[]) // メインルーチン
{
  rocclass roc;                 // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) roc.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { roc.get(argv[i], opt);  fok = false; }
    }
    roc.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occured !");
    return 255;
  }
  return 0;
}
