// sstate.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-26 10:58:15 tac>
//
// 実数をデータを層別集計し、基礎統計量を出力する。複数項目を一度に集計可能。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -w sstate sstate.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "sstate  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:sstate  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The stratify basic statistics for each item.",
  "<OPTION>",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=sstate_dns.d",
  "    output data:basic statistics",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ol[={-|md|mn|sd|<VALUE>}[<COEF>]]:specify a reference value for outliers",
  "    -ol:-ol=sd2  -:no output outliers(default)",
  "    md:median x COEF  mn:mean x COEF  sd:mean + COEFxSD  <VALUE>:reference value",
  "    <COEF>:coefficient(default:2)",
  "  -res=[-][{def|all|cc[<C-COEF>]|norm|si|st|stat}][,...]",
  "      :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default=all  all:output all",
  "    cc[<C-COEF>]:confidence intervals(0=<C-COEF<1)  norm:normality indicators",
  "    si:output the stratified results  st:output the overall results",
  "    stat:basic statistics  default:output all"
};

const string Title = "=== Basic Statistics ==="; // 表題
const string DNSNAMEdef = "sstate_dns.d"; // 省略時DANSファイル名

enum oltype                     // 外れ値の基準値の種類
{
  olnon,                        // 出力なし
  olmn,                         // 平均値
  olmd,                         // 中央値
  olsd,                         // SD
  oletc                         // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  string vdnsname;              // DANSファイル名
  oltype vol;                   // 外れ値の基準値
  double volc;                  // 外れ値の係数
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  oltype ol() const             // 外れ値の基準値を返す
  { return vol; }
  double olc() const            // 外れ値の係数を返す
  { return volc; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのセットとチェック
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdnsname(), vol(olnon), volc(0.0), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dnsname")      // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ol")      // 外れ値の基準値と係数の指定
  {
    if (opt[1].empty())
    { vol = olsd; }
    else if (opt[1] == "-")
    { vol = olnon; }
    else if (opt[1][0] == 'm')
    {
      if (opt[1][1] == 'd') vol = olmd;
      else if (opt[1][1] == 'n') vol = olmn;
      else err = badopt(vpname, optstr, USAGE);
    }
    else if (opt[1][0] == 's' && opt[1][1] == 'd')
    { vol = olsd; }
    else if (isdigit(opt[1][0]) || (opt[1][0] == '-' && isdigit(opt[1][1])))
    {
      vol = oletc;  volc = strtodf(opt[1]);
      if (volc == BLANKf) err = badopt(vpname, optstr, USAGE);
    }
    else
    { err = badopt(vpname, optstr, USAGE); }
    if (err == 0 && vol != oletc)
    {
      volc = (opt[1].size() < 3) ? 2.0 : strtodf(opt[1].substr(2));
      if (volc < 0.0 || 1000.0 < volc) err = badopt(vpname, optstr, USAGE);
    }
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "At least 2 items are required !");
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["norm"] = vresmap["si"] = vresmap["st"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["norm"] = vresmap["si"] = vresmap["st"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["cc"] = vresmap["norm"] = vresmap["si"] = vresmap["st"] = vresmap["stat"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, " Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class stats : public statk<double> // 基礎統計量+k統計量+四分位値クラス
{
private:
  double vmin;                  // 最小値
  double vq1;                   // 第1四分位値(25パーセンタイル)
  double vmed;                  // 中央値(第2四分位値)
  double vq3;                   // 第3四分位値(75パーセンタイル)
  double vmax;                  // 最大値
  int vnol;                     // 外れ値の例数
  list<double> vfdlst;          // 実数データリスト
public:
  stats() : statk(), vmin(BLANKf), vq1(BLANKf), vmed(BLANKf), vq3(BLANKf), vmax(BLANKf), vnol(0), vfdlst()
  // デフォールトコンストラクタ
  {}
  stats(int n0, double sig0, double sig20, double sigmax30, double sigmax40, double min0, double q10, double med0, double q30, double max0, double nol0, list<double>& fdls0) : statk(n0, sig0, sig20, sigmax30, sigmax40), vmin(min0), vq1(q10), vmed(med0), vq3(q30), vmax(max0), vnol(nol0), vfdlst(fdls0)
  // コンストラクタ
  {}
  stats(const stats& stats0) : statk(stats0.vn, stats0.sigmax, stats0.sigmax2, stats0.sigmax3, stats0.sigmax4), vmin(stats0.vmin), vq1(stats0.vq1), vmed(stats0.vmed), vq3(stats0.vq3), vmax(stats0.vmax), vnol(stats0.vnol), vfdlst(stats0.vfdlst)
  // コピーコンストラクタ
  {}
  virtual ~stats()              // デストラクタ
  {}
  const stats& operator=(const stats& stats0); // コピー代入演算子
  void clear();                                // データをクリアする
  int sigma(double x);                         // データの累積集計
  int sort(oltype ol, double olc);             // データをソートする
  const list<double>& fdlst() const // 実数データリストを返す
  { return vfdlst; }
  double min() const            // 最小値を返す
  { return vmin; }
  double q1() const             // 第1四分位値を返す
  { return vq1; }
  double med() const            // 中央値を返す
  { return vmed; }
  double q3() const             // 第3四分位値を返す
  { return vq3; }
  double max() const            // 最大値を返す
  { return vmax; }
  int nol() const               // 外れ値の例数を返す
  { return vnol; }
  double ptv(double pt);        // パーセンタイル値を返す
};

const stats& stats::operator=(const stats& stats0)
// コピー代入演算子
{
  if (&stats0 == this) return *this; // 自分自身への代入を防ぐ
  vn = stats0.vn;  sigmax = stats0.sigmax;  sigmax2 = stats0.sigmax2;
  sigmax3 = stats0.sigmax3;  sigmax4 = stats0.sigmax4;
  vmin = stats0.vmin;  vq1 = stats0.vq1;  vmed = stats0.vmed;
  vq3 = stats0.vq3;  vmax = stats0.vmax;  vnol = stats0.vnol;
  vfdlst = stats0.vfdlst;
  return *this;
}

void stats::clear()             // データをクリアする
{
  vn = 0;  sigmax = sigmax2 = sigmax3 = sigmax4 = 0.0;
  vmin = vq1 = vmed = vq3 = vmax = BLANKf;  vnol = 0;
  vfdlst.clear();
}

int stats::sigma(double x)      // データの累積集計
{
  if (x == BLANKf) return 0;    // 累積せず
  double x2 = x * x;
  ++vn;  sigmax += x;  sigmax2 += x2;
  sigmax3 += x * x2;  sigmax4 += x2 * x2;
  vfdlst.push_back(x);
  return 1;                     // 累積
}

int stats::sort(oltype ol, double olc) // データをソートする
{
  if (vfdlst.empty()) return 0; // データ無し
  if (vmin != BLANKf) return 1; // ソート済み
  vfdlst.sort();
  list<double>::const_iterator fditr = vfdlst.begin();
  vmin = vq1 = vmed = *fditr;
  fditr = vfdlst.end();
  --fditr;
  vmax = vq3 = *fditr;
  if (vfdlst.size() < 2) return 1;
  double fidx = (double)(vfdlst.size() - 1) * 0.25;
  int idx = (int)fidx;
  fditr = vfdlst.begin();
  for (int i = 0; i < idx; ++i) ++fditr;
  double xpre = *fditr;
  ++fditr;
  vq1 = xpre + (*fditr - xpre) * (fidx - (double)idx);
  fidx = (double)(vfdlst.size() - 1) * 0.5;
  idx = (int)fidx;
  fditr = vfdlst.begin();
  for (int i = 0; i < idx; ++i) ++fditr;
  xpre = *fditr;
  ++fditr;
  vmed = xpre + (*fditr - xpre) * (fidx - (double)idx);
  fidx = (double)(vfdlst.size() - 1) * 0.75;
  idx = (int)fidx;
  fditr = vfdlst.begin();
  for (int i = 0; i < idx; ++i) ++fditr;
  xpre = *fditr;
  ++fditr;
  vq3 = xpre + (*fditr - xpre) * (fidx - (double)idx);
  double lim = 0.0;
  switch (ol)
  {
  case olmn : lim = olc * mean();
    break;
  case olmd : lim = olc * vmed;
    break;
  case olsd : lim = mean() + olc * sd();
    break;
  default : lim = olc;
  }
  idx = 0;
  for (fditr = vfdlst.begin(); *fditr <= lim && fditr != vfdlst.end(); ++fditr) ++idx;
  vnol = this->n() - idx;       // 外れ値の例数
  return 1;
}

double stats::ptv(double pt)    // パーセンタイル値を返す
{
  if (vfdlst.empty()) return BLANKf; // データ無し
  if (vmin == BLANKf) sort(olnon, 0.0); // ソートする
  list<double>::const_iterator fditr = vfdlst.begin();
  if (vfdlst.size() < 2) return *fditr; // データが1つだけの時
  if (pt >= 100.0) return vmax;
  double fidx = (double)(vfdlst.size() - 1) * pt * 0.01;
  int idx = (int)fidx;
  fditr = vfdlst.begin();
  for (int i = 0; i < idx; ++i) ++fditr;
  double xpre = *fditr;
  ++fditr;
  return xpre + (*fditr - xpre) * (fidx - (double)idx);
}

using vecstats = vector<stats>; // 基礎統計量ベクトル配列の別名定義

class sstatclass                // sstatメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout, const string& sstr);
  // 分析結果のタイトルとデータ識別情報を出力する
  void put_tab(ostream& fout, option& opt, vecstats& statsvec);
  // 集計結果を集計表形式で出力する
  void put_stat(ostream& fout, option& opt, stats& statsj);
  // 基礎統計量を出力する
  void put_norm(ostream& fout, option& opt, stats& statsj);
  // 正規性の指標を出力する
  void cal_sk(double& g1, double& v1, string& skew, double& g2, double& v2, string& kur, stats& statj);
  // 歪度と尖度を計算する
  void put_dnsid(ofstream& fout, const option& opt);
  // DANSデータのタイトルとデータ識別情報を出力する
  void put_dns(ofstream& fout, const option& opt, map<string, vecstats>::iterator spos);
  // 集計結果をデータファイル形式で出力する
public:
  sstatclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~sstatclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込む
  void put(option& opt);                     // 基礎統計量を出力する
};

void sstatclass::put_resid(ostream& fout, const string& sstr)
// 分析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  idd tid(id);
  stra tstra;
  tstra.get('s', 1, sstr);
  lstra tstrals;
  tstrals.push_back(tstra);
  tid.append_stra(tstrals);     // 層別コードを追加する
  tid.put_title(fout);
  tid.put_stra(fout);
}

void sstatclass::put_tab(ostream& fout, option& opt, vecstats& statsvec)
// 集計結果を集計表形式で出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  int nss = (int)(id.ni()) - 1;
  for (int j = 0; j < nss; ++j)
  {
    if (statsvec[j].n() > 0)
    {
      statsvec[j].sort(opt.ol(), opt.olc());
      fout << "ANALYSIS ITEM:" << id.item(j + 1) << "\n";
      int n = statsvec[j].n();
      double m = statsvec[j].mean(), sd = statsvec[j].sd();
      bool putres = false;
      if (opt.res("stat"))
      {
        putline(fout, LENline);
        put_stat(fout, opt, statsvec[j]);
        putres = true;
      }
      if (opt.res("norm"))
      {
        putline(fout, LENline);
        put_norm(fout, opt, statsvec[j]);
        putres = true;
      }
      if (putres) putline(fout, LENline);
      fout << "\n";
    }
  }
  fout << endl;
}

void sstatclass::put_stat(ostream& fout, option& opt, stats& statsj)
// 基礎統計量を出力する
{
  int n = statsj.n();
  double m = statsj.mean(), sd = statsj.sd();
  putmsd(fout, n, m, statsj.vr());
  double cv = (m == 0.0) ? 0.0 : sd / m * 100.0;
  fout << "Min=" << statsj.min();
  fout << "  25%tile=" << statsj.q1();
  fout << "  Median=" << statsj.med();
  fout << "  75%tile=" << statsj.q3();
  fout << "  Max=" << statsj.max() << "\n";
  fout << "Range=" << statsj.max() - statsj.min();
  fout << "  Coefficient of Variation=" << cv << "%\n";
  if (opt.res("cc"))
  { 
    double t = ptot(1.0 - opt.cc(), n - 1);
    double ci = t * statsj.se();
    fout << "  " << opt.cc() * 100.0 << "%CI of Mean:" << m - ci << " - " << m + ci;
    fout << " (" <<  m << " +/- " << ci << ")\n";
    ci = t * sqrt((1.0 + 1.0 / (double)n) * statsj.vr());
    fout << "  " << opt.cc() * 100.0 << "%PI of Data:" << m - ci << " - " << m + ci;
    fout << " (" <<  m << " +/- " << ci << ")\n";
    t = ptoz(1.0 - opt.cc());
    int r = (int)((double)(n + 1) * 0.5 - t * sqrt((double)n) * 0.5);
    int i = 1;
    list<double>::const_iterator fditr = (statsj.fdlst()).begin();
    while (i < r && fditr != (statsj.fdlst()).end()) { ++i;  ++fditr; }
    double medlow = (fditr != (statsj.fdlst()).end()) ? *fditr : statsj.min();
    while (i < n - r + 1 && fditr != (statsj.fdlst()).end()) { ++i;  ++fditr; }
    double medup = (fditr != (statsj.fdlst()).end()) ? *fditr : statsj.max();
    fout << "  " << opt.cc() * 100.0 << "%CI of Median:" << medlow << " - " << medup;
    fout << " (" << r << "th - " << n - r + 1 << "th Data)\n";
    double ah = (1.0 - opt.cc()) * 50.0;
    medlow = statsj.ptv(ah);
    medup = statsj.ptv(100.0 - ah);
    fout << "  " << ah << "%tile - " << 100.0 - ah << "%tile of Data:" << medlow << " - " << medup << "\n";
  }
}

void sstatclass::put_norm(ostream& fout, option& opt, stats& statsj)
// 正規性の指標を出力する
{
  int n = statsj.n();
  double g1 = 0.0, v1 = 0.0, g2 = 0.0, v2 = 0.0;
  string skew, kur;
  cal_sk(g1, v1, skew, g2, v2, kur, statsj);
  double z = (v1 > 0.0) ? g1 / sqrt(v1) : 0.0;
  ldbl pr = ztop(z);
  fout << "Skewness:g1=" << left << setw(10) << g1;
  fout << "z=" << setw(10) << z << "p=" << setw(12) << pr;
  fout << setw(4) << pind(pr) << right << skew << "\n";
  z = (v2 > 0.0) ? g2 / sqrt(v2) : 0.0;
  pr = ztop(z);
  fout << "Kurtosis:g2=" << left << setw(10) << g2;
  fout << "z=" << setw(10) << z << "p=" << setw(12) << pr;
  fout << setw(4) << pind(pr) << right << kur << "\n";
  double jb = (double)(n) * (g1 * g1 + g2 * g2 / 4.0) / 6.0;
  pr = ctop(jb, 2);
  fout << "Jarque-Bera Test:JB=chi-squared=" <<  jb << "  DF=2  p=" << pr << pind(pr) << "\n";
  if (opt.res("cc"))
  {
    double zc = ptoz(1.0 - opt.cc());
    double ci = zc * sqrt(v1);
    fout << "  " << opt.cc() * 100.0 << "%CI of g1:" << g1 - ci << " - " << g1 + ci;
    fout << " (" <<  g1 << " +/- " << ci << ")\n";
    ci = zc * sqrt(v2);
    fout << "  " << opt.cc() * 100.0 << "%CI of g2:" << g2 - ci << " - " << g2 + ci;
    fout << " (" <<  g2 << " +/- " << ci << ")\n";
  }
}

void sstatclass::cal_sk(double& g1, double& v1, string& skew, double& g2, double& v2, string& kur, stats& statj)
// 歪度と尖度を計算する
{
  g1 = v1 = g2 = v2 = 0.0;  skew.erase();  kur.erase();
  int n = statj.n();
  if (n < 4) return;
  double k2 = statj.k2(), k3 = statj.k3(), k4 = statj.k4();
  if (k2 > 0.0) { g1 = k3 / k2 / sqrt(k2);  g2 = k4 / k2 / k2; }
  v1 = 6.0 * (double)n * (double)(n - 1) / (double)(n - 2) / (double)(n + 1) / (double)(n + 3);
  v2 = 24.0 * (double)n * (double)(n - 1) * (double)(n - 1) / (double)(n - 3) / (double)(n - 2) / (double)(n + 3) / (double)(n + 5);
  if (g1 < -1.0) skew = "Right Sloped";
  else if (fabs(g1) <= 1.0) skew = "Nearly Symmetrical";
  else skew = "Left Sloped";
  if (g2 < -3.0) kur = "Blunt Peak";
  else if (fabs(g2) <= 3.0) kur = "Nearly Normal";
  else kur = "sharp peak";
}

void sstatclass::put_dnsid(ofstream& fout, const option& opt)
// データ識別情報をデータファイル形式で出力する
{
  int nitem = 14;               // 出力項目数
  if (opt.ol() != olnon) nitem += 1;
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  idd tid(id.title(), 1 + nitem);
  tid.assign_item(0, id.item(0));
  tid.assign_item(1, "Number");
  tid.assign_item(2, "Mean");
  tid.assign_item(3, "SD");
  tid.assign_item(4, "SE");
  tid.assign_item(5, "Min");
  tid.assign_item(6, "25%tail");
  tid.assign_item(7, "Median");
  tid.assign_item(8, "75%tail");
  tid.assign_item(9, "Max");
  tid.assign_item(10, "Range");
  tid.assign_item(11, "CV(%)");
  tid.assign_item(12, "Skewness");
  tid.assign_item(13, "Kurtosis");
  tid.assign_item(14, "Total");
  if (opt.ol() != olnon)
  {
    stringstream ostr;
    switch (opt.ol())
    {
    case olmn : ostr << opt.olc() << " x Mean";
      break;
    case olmd : ostr << opt.olc() << " x Median";
      break;
    case olsd : ostr << "Mean + " << opt.olc() << "SD";
      break;
    default : ostr << opt.olc();
      break;
    }
    tid.assign_item(15, ostr.str()+" Proportion of Outliers");
  }
  tid.put_item(fout, formtag);
}

void sstatclass::put_dns(ofstream& fout, const option& opt, map<string, vecstats>::iterator spos)
// 集計結果をデータファイル形式で出力する
{
  int nitem = 14;               // 出力項目数
  if (opt.ol() != olnon) nitem += 1;
  for (size_t i = 0; i < id.ni() - 1; ++i)
  {
    datac cdata(1 + nitem);
    stringstream ostr;
    ostr << spos->first << "-" << id.item(i + 1);
    cdata.assign(0, ostr.str());
    if (spos->second[i].n() < 1)
    { for (int j = 1; j <= nitem; ++j) cdata.assign(j, BLANKs); }
    else
    {
      spos->second[i].sort(opt.ol(), opt.olc());
      cdata.assign(1, (double)(spos->second[i].n()));
      cdata.assign(2, spos->second[i].mean());
      cdata.assign(3, spos->second[i].sd());
      cdata.assign(4, spos->second[i].se());
      cdata.assign(5, spos->second[i].min());
      cdata.assign(6, spos->second[i].q1());
      cdata.assign(7, spos->second[i].med());
      cdata.assign(8, spos->second[i].q3());
      cdata.assign(9, spos->second[i].max());
      cdata.assign(10, spos->second[i].max() - spos->second[i].min());
      double cv = (spos->second[i].mean() == 0.0) ? 0.0 : spos->second[i].sd() / spos->second[i].mean() * 100.0;
      cdata.assign(11, cv);
      double g1 = 0.0, v1 = 0.0, g2 = 0.0, v2 = 0.0;
      string skew, kur;
      cal_sk(g1, v1, skew, g2, v2, kur, spos->second[i]);
      cdata.assign(12, g1);
      cdata.assign(13, g2);
      cdata.assign(14, spos->second[i].sum());
      if (opt.ol() != olnon)
      {
        double fn = spos->second[i].nol();
        cdata.assign(15, fn / (double)(spos->second[i].n()));
      }
    }
    cdata.put(fout);
  }
}

int sstatclass::get(const char* c_fname, option& opt) // データを読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void sstatclass::put(option& opt) // 基礎統計量を出力する
{
  if (id.nc() < 1) return;      // データなし
  ofstream foutdns;
  if (opt.dnsname() != "")      // DANSデータ出力
  {
    if (ofopen(foutdns, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
    put_dnsid(foutdns, opt);
  }
  map<string, vecstats> ssvecmap; // 基礎統計量ベクトルマップ
  vecstats ssvectot(id.ni() - 1); // 全体の基礎統計量ベクトル
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    if (ditr->sdata(0) != BLANKs) // 層別コードあり
    {
      map<string, vecstats>::iterator spos = ssvecmap.find(ditr->sdata(0));
      if (spos != ssvecmap.end()) // 既存コード
      {
        for (size_t j = 1; j < id.ni(); ++j)
        {
          double fdata = ditr->fdata(j);
          if (fdata != BLANKf)
          { spos->second[j - 1].sigma(fdata);  ssvectot[j - 1].sigma(fdata); }
        }
      }
      else                      // 新規コード
      {
        vecstats tssvec(id.ni() - 1);
        for (size_t j = 1; j < id.ni(); ++j)
        {
          double fdata = ditr->fdata(j);
          if (fdata != BLANKf)
          { tssvec[j - 1].sigma(fdata);  ssvectot[j - 1].sigma(fdata); }
        }
        ssvecmap.insert(map<string, vecstats>::value_type(ditr->sdata(0), tssvec));
      }
    }
  }
  if (ssvecmap.empty()) return; // データなし
  for (map<string, vecstats>::iterator sitr = ssvecmap.begin(); sitr != ssvecmap.end(); ++sitr)
  {
    if (opt.res("si"))
    {
      put_resid(cout, sitr->first);
      put_tab(cout, opt, sitr->second);
    }
    if (opt.dnsname() != "") put_dns(foutdns, opt, sitr);
  }
  if (opt.res("st"))
  {
    put_resid(cout, "ALL");
    put_tab(cout, opt, ssvectot);
  }
  if (opt.dnsname() != "")
  {
    ssvecmap.clear();
    ssvecmap.insert(map<string, vecstats>::value_type("ALL", ssvectot));
    map<string, vecstats>::iterator sitr = ssvecmap.begin();
    put_dns(foutdns, opt, sitr);
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  sstatclass sstat;             // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) sstat.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { sstat.get(argv[i], opt);  fok = false; }
    }
    sstat.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occured !");
    return 255;
  }
  return 0;
}
