// statiste.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-26 11:10:10 tac>
//
// 実数をデータを集計し、基礎統計量を出力する。複数項目を一度に集計可能。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -w statiste statiste.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <dans.H>               // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "statiste  Version 7.42e      Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:statiste  [<OPTON> ...]  [<DATA FILE> ...]",
  "  Compute the basic statistics for each item.",
  "<OPTION>",
  "  -ch=<LOWER>,<INTERVAL>:specify the lower limit of class and class interval",
  "  -dnsname[=<FILE>]:specify the file to output DANS data",
  "    default:no output  -dnsname:-dnsname=statiste_dns.d",
  "    output data:item name and basic statistcs",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={box|cd|cdp|d|ncd|nd|nrcd|nrd|rcd|rcdp|rd}[,...]",
  "      :select the data for Ngraph",
  "    box:box and whisker plot  cd:cumulative d  cdp:cd plot",
  "    d:frequency distribution  ncd:theoretical cd  nd:theoretical d",
  "    nrcd:theoretical rcd  nrd:theoretical rd  rcd:relative cd  rcdp:rcd plot",
  "    rd:relative d  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=statiste_ngp.dat",
  "  -res=[-][{def|all|cc[<C-COEF>]|norm|pt[<%TILE>]|stat}][,...]",
  "      :select output results",
  "    -:with -, no output(-res=-:no output at all)",
  "    def:same output as default  all:output at all",
  "    cc[<C-COEF>]:confidence intervals(0=<C-COEF<1)  norm:normality indicators",
  "    pt[<%TILE>]:percentiles(0<%TILE=<100)  stat:basic statistcs",
  "    default:automatic judgement according to data",
  "  -sep[={-|+}]:specify the item aggregation method(-sep:-sep=+)"
};

const string Title = "=== Basic Statistics ==="; // 表題
constexpr uint RDdef = 50;      // グラフの描画範囲分割数省略時値
const string DNSNAMEdef = "statiste_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "statiste_ngp.dat"; // 省略時Ngraphファイル名
const double PTdef = 50.0;                   // 省略時パーセンタイル

using listdbl = list<double>;   // 倍精度実数リストの別名定義

enum ngptype                    // Ngraph用出力データの種類
{
  ngpnon,                       // 出力なし
  ngpbox,                       // 箱ヒゲプロット
  ngpcd,                        // 累積度数分布
  ngpcdp,                       // 累積度数分布のプロット
  ngpd,                         // 度数分布
  ngpdns,                       // データファイル形式
  ngpncd,                       // 理論累積度数分布
  ngpnd,                        // 理論度数分布
  ngpnrcd,                      // 理論相対累積度数分布
  ngpnrd,                       // 理論相対度数分布
  ngprcd,                       // 相対累積度数分布
  ngprcdp,                      // 相対累積度数分布のプロット
  ngprd,                        // 相対度数分布
  ngpetc                        // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  double vcmin;                 // 度数分布の階級下限
  double vh;                    // 度数分布の級間
  string vdnsname;              // DANSファイル名
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  double vpt;                   // パーセンタイル
  bool vsep;                    // 項目ごと集計フラッグ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double cmin() const           // 度数分布の階級下限を返す
  { return vcmin; }
  double h() const              // 度数分布の級間を返す
  { return vh; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  double pt() const             // パーセンタイルを返す
  { return vpt; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  bool sep() const              // 項目ごと集計フラッグを返す
  { return vsep; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check();                  // オプションのセットとチェック
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vcmin(BLANKf), vh(BLANKf), vdnsname(), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vpt(PTdef), vsep(true)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dnsname")      // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "ch")      // 階級下限と級間の指定
  {
    datac cdata(2);
    cdata.getdata(opt[1]);
    vcmin = cdata.fdata(0);  vh = cdata.fdata(1);
    if (vcmin == BLANKf || vh == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "sep")     // 項目ごとの集計指定
  {
    if (opt[1].empty() || opt[1] == "+") vsep = true;
    else if (opt[1] == "-") vsep = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check()             // オプションのセットとチェック
{
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["box"] = vngpmap["cd"] = vngpmap["cdp"] = vngpmap["d"] = vngpmap["ncd"] = vngpmap["nd"] = vngpmap["nrcd"] = vngpmap["nrd"] = vngpmap["rcd"] = vngpmap["rcdp"] = vngpmap["rd"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["box"] = vngpmap["cd"] = vngpmap["cdp"] = vngpmap["d"] = vngpmap["ncd"] = vngpmap["nd"] = vngpmap["nrcd"] = vngpmap["nrd"] = vngpmap["rcd"] = vngpmap["rcdp"] = vngpmap["rd"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["norm"] = vresmap["stat"] = true;
  vresmap["pt"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["norm"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["norm"] = vresmap["stat"] = true;
      }
      else if (res == "all")
      {
        vresmap["def"] = vresmap["cc"] = vresmap["norm"] = vresmap["pt"] = vresmap["stat"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " optiion !");
        vresmap["cc"] = true;
      }
      else if (res[0] == 'p' && res[1] == 't' && res.size() > 2)
      {
        vpt = strtodf(res.substr(2)); // パーセンタイル
        if (vpt <= 0.0 || 100.0 < vpt) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["pt"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class stats                     // 基礎統計量+k統計量+四分位値クラス
{
private:
  statk<double> vfdstatk;       // 基礎統計量+k統計量クラス
  double vmin;                  // 最小値
  double vq1;                   // 第1四分位値(25パーセンタイル)
  double vmed;                  // 中央値(第2四分位値)
  double vq3;                   // 第3四分位値(75パーセンタイル)
  double vmax;                  // 最大値
  listdbl vfdlst;               // 実数データリスト
public:
  stats() : vfdstatk(), vmin(BLANKf), vq1(BLANKf), vmed(BLANKf), vq3(BLANKf), vmax(BLANKf), vfdlst()
  // デフォールトコンストラクタ
  {}
  stats(int n0, double sig0, double sig20, double sig30, double sig40, double min0, double q10, double med0, double q30, double max0, listdbl& fdlst0) : vfdstatk(n0, sig0, sig20, sig30, sig40), vmin(min0), vq1(q10), vmed(med0), vq3(q30), vmax(max0), vfdlst(fdlst0)
  // コンストラクタ
  {}
  stats(const stats& stats0) : vfdstatk(stats0.vfdstatk), vmin(stats0.vmin), vq1(stats0.vq1), vmed(stats0.vmed), vq3(stats0.vq3), vmax(stats0.vmax), vfdlst(stats0.vfdlst)
  // コピーコンストラクタ
  {}
  virtual ~stats()              // デストラクタ
  {}
  const stats& operator=(const stats& stats0); // コピー代入演算子
  void clear();                                // データをクリアする
  int sigma(double x);                         // データの累積集計
  int sort();                                  // データをソートする
  const statk<double>& fdstatk() const // 基礎統計量+k統計量クラスを返す
  { return vfdstatk; }
  const listdbl& fdlst() const  // 実数データリストを返す
  { return vfdlst; }
  int n() const                 // 係数を返す
  { return vfdstatk.n(); }
  double mean() const           // 平均値を返す
  { return vfdstatk.mean(); }
  double vr() const             // 不偏分散を返す
  { return vfdstatk.vr(); }
  double sd() const             // 不偏標準偏差を返す
  { return vfdstatk.sd(); }
  double se() const             // 標準誤差を返す
  { return vfdstatk.se(); }
  double k2() const             // k2統計量を返す
  { return vfdstatk.k2(); }
  double k3() const             // k3統計量を返す
  { return vfdstatk.k3(); }
  double k4() const             // k4統計量を返す
  { return vfdstatk.k4(); }
  double min() const            // 最小値を返す
  { return vmin; }
  double q1() const             // 第1四分位値を返す
  { return vq1; }
  double med() const            // 中央値を返す
  { return vmed; }
  double q3() const             // 第3四分位値を返す
  { return vq3; }
  double max() const            // 最大値を返す
  { return vmax; }
  double ptv(double pt);        // パーセンタイル値を返す
};

const stats& stats::operator=(const stats& stats0) // コピー代入演算子
{
  if (&stats0 == this) return *this; // 自分自身への代入を防ぐ
  vfdstatk = stats0.vfdstatk;
  vmin = stats0.vmin;  vq1 = stats0.vq1;  vmed = stats0.vmed;
  vq3 = stats0.vq3;  vmax = stats0.vmax;  vfdlst = stats0.vfdlst;
  return *this;
}

void stats::clear()             // データをクリアする
{
  vfdstatk.clear();
  vmin = vq1 = vmed = vq3 = vmax = BLANKf;  vfdlst.clear();
}

int stats::sigma(double x)      // データの累積集計
{
  if (x == BLANKf) return 0;    // 累積せず
  vfdstatk.sigma(x);  vfdlst.push_back(x);
  return 1;                     // 累積
}

int stats::sort()               // データをソートする
{
  if (vfdlst.empty()) return 0; // データ無し
  if (vmin != BLANKf) return 1; // ソート済み
  vfdlst.sort();
  listdbl::const_iterator fditr = vfdlst.begin();
  vmin = vq1 = vmed = *fditr;
  fditr = vfdlst.end();
  --fditr;
  vmax = vq3 = *fditr;
  if (vfdlst.size() < 2) return 1;
  double fidx = (double)(vfdlst.size() - 1) * 0.25;
  uint idx = (uint)fidx;
  fditr = vfdlst.begin();
  for (uint i = 0; i < idx; ++i) ++fditr;
  double xpre = *fditr;
  ++fditr;
  vq1 = xpre + (*fditr - xpre) * (fidx - (double)idx);
  fidx = (double)(vfdlst.size() - 1) * 0.5;
  idx = (uint)(fidx);
  fditr = vfdlst.begin();
  for (uint i = 0; i < idx; ++i) ++fditr;
  xpre = *fditr;
  ++fditr;
  vmed = xpre + (*fditr - xpre) * (fidx - (double)idx);
  fidx = (double)(vfdlst.size() - 1) * 0.75;
  idx = (uint)fidx;
  fditr = vfdlst.begin();
  for (uint i = 0; i < idx; ++i) ++fditr;
  xpre = *fditr;
  ++fditr;
  vq3 = xpre + (*fditr - xpre) * (fidx - (double)idx);
  return 1;
}

double stats::ptv(double pt)    // パーセンタイル値を返す
{
  if (vfdlst.empty()) return BLANKf; // データ無し
  if (vmin == BLANKf) this->sort(); // ソートする
  listdbl::const_iterator fditr = vfdlst.begin();
  if (vfdlst.size() < 2) return *fditr; // データが1つだけの時
  if (pt >= 100.0) return vmax;
  double fidx = (double)(vfdlst.size() - 1) * pt * 0.01;
  uint idx = (uint)fidx;
  fditr = vfdlst.begin();
  for (uint i = 0; i < idx; ++i) ++fditr;
  double xpre = *fditr;
  ++fditr;
  return xpre + (*fditr - xpre) * (fidx - (double)idx);
}

class statistclass              // statistメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout, const option& opt);
  // 分析結果のタイトルとデータ識別情報を出力する
  void put_stat(ostream& fout, option& opt, stats& statsj);
  // 基礎統計量を出力する
  void put_mci(ostream& fout, const option& opt, const string& itemj, stats& statsj) const;
  // 各種の信頼区間を出力する
  void put_norm(ostream& fout, option& opt, const string& itemj, stats& statsj);
  // 正規性の指標を出力する
  void cal_sk(double& g1, double& v1, string& skew, double& g2, double& v2, string& kur, stats& statsj) const;
  // 歪度と尖度を計算する
  void put_dns(ofstream& fout, const option& opt, const string& itemj, stats& statsj);
  // 計算結果をデータファイル形式で出力する
  void put_dnsid(ofstream& fout);
  // DANSデータのタイトルとデータ識別情報を出力する
  void plot_ngp(ofstream& fout, option& opt, const string& itemj, stats& statsj, bool delml);
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, const option& opt);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_box(ostream& fout, const option& opt, const string& itemj, stats& statsj) const;
  // Ngraph用箱ヒゲプロットデータを出力する
  void plot_drd(ostream& fout, const option& opt, ngptype ngp, const string& itemj, stats& statsj) const;
  // Ngraph用度数分布／相対度数分布データを出力する
  void plot_cdrcd(ostream& fout, const option& opt, ngptype ngp, const string& itemj, stats& statsj) const;
  // Ngraph用累積度数分布＆プロット／相対累積度数分布＆プロットデータを出力する
  void plot_ndrd(ostream& fout, const option& opt, ngptype ngp, const string& itemj, stats& statsj) const;
  // Ngraph用理論度数分布／理論相対度数分布データを出力する
  void plot_ncdrcd(ostream& fout, const option& opt, ngptype ngp, const string& itemj, stats& statsj) const;
  // Ngraph用理論累積度数分布／理論相対累積度数分布データを出力する
public:
  statistclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~statistclass()               // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込む
  void put(option& opt);                     // 基礎統計量を出力する
};

void statistclass::put_resid(ostream& fout, const option& opt)
// 分析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  if (opt.sep() == false) id.put_item(fout);
}

void statistclass::put_stat(ostream& fout, option& opt, stats& statsj)
// 基礎統計量を出力する
{
  int n = statsj.n();
  double m = statsj.mean();
  putmsd(fout, n, m, statsj.vr());
  double cv = (m == 0.0) ? 0.0 : statsj.sd() / m * 100.0;
  fout << "Min=" << statsj.min();
  fout << "  25%tile=" << statsj.q1();
  fout << "  Median=" << statsj.med();
  fout << "  75%tile=" << statsj.q3();
  fout << "  Max=" << statsj.max() << "\n";
  fout << "Range=" << statsj.max() - statsj.min();
  fout << "  Coefficient of Variation=" << cv << "%";
  if (opt.res("pt")) fout << "  " << opt.pt() << "%tile=" << statsj.ptv(opt.pt());
  fout << endl;
}

void statistclass::put_mci(ostream& fout, const option& opt, const string& itemj, stats& statsj) const
// 各種の信頼区間を出力する
{
  int n = statsj.n();
  double m = statsj.mean();
  double t = (double)(ptot((ldbl)(1.0 - opt.cc()), n - 1));
  double ci = t * statsj.se();
  fout << "  " << opt.cc() * 100.0 << "%CI of Mean:" << m - ci << " - " << m + ci;
  fout << " (" <<  m << " +/- " << ci << ")\n";
  ci = t * sqrt((1.0 + 1.0 / (double)n) * statsj.vr());
  fout << "  " << opt.cc() * 100.0 << "%PI of Data:" << m - ci << "〜" << m + ci;
  fout << " (" <<  m << " +/- " << ci << ")\n";
  t = (double)(ptoz((ldbl(1.0 - opt.cc()))));
  uint r = (uint)((double)(n + 1) * 0.5 - t * sqrt((double)n) * 0.5);
  uint i = 1;
  listdbl::const_iterator fditr = (statsj.fdlst()).begin();
  while (i < r && fditr != (statsj.fdlst()).end()) { ++i;  ++fditr; }
  double medlow = (fditr != (statsj.fdlst()).end()) ? *fditr : statsj.min();
  while (i < n - r + 1 && fditr != (statsj.fdlst()).end()) { ++i;  ++fditr; }
  double medup = (fditr != (statsj.fdlst()).end()) ? *fditr : statsj.max();
  fout << "  " << opt.cc() * 100.0 << "%CI of Median:" << medlow << " - " << medup;
  fout << " (" << r << "th - " << n - r + 1 << "th Data)\n";
  double ah = (1.0 - opt.cc()) * 50.0;
  medlow = statsj.ptv(ah);
  medup = statsj.ptv(100.0 - ah);
  fout << "  " << ah << "%tile - " << 100.0 - ah << "%tile of Data:" << medlow << " - " << medup << endl;
}

void statistclass::put_norm(ostream& fout, option& opt, const string& itemj, stats& statsj)
// 正規性の指標を出力する
{
  double g1 = 0.0, v1 = 0.0, g2 = 0.0, v2 = 0.0;
  string skew, kur;
  cal_sk(g1, v1, skew, g2, v2, kur, statsj);
  double z = (v1 > 0.0) ? g1 / sqrt(v1) : 0.0;
  double zc = 0.0, ci = 0.0;
  ldbl pr = ztop(z);
  fout << "Skewness:g1=" << left << setw(10) << g1;
  fout << "z=" << setw(10) << z << "p=" << setw(12) << pr;
  fout << setw(4) << pind(pr) << right << skew << "\n";
  z = (v2 > 0.0) ? g2 / sqrt(v2) : 0.0;
  pr = ztop((ldbl)z);
  fout << "Kurtosis:g2=" << left << setw(10) << g2;
  fout << "z=" << setw(10) << z << "p=" << setw(12) << pr;
  fout << setw(4) << pind(pr) << right << kur << "\n";
  double jb = (double)(statsj.n()) * (g1 * g1 + g2 * g2 / 4.0) / 6.0;
  pr = ctop(jb, 2);
  fout << "Jarque-Bera Test:JB=chi-squared=" <<  jb << "  DF=2  p=" << pr << pind(pr) << "\n";
  if (opt.res("cc"))
  {
    zc = (double)(ptoz((ldbl)(1.0 - opt.cc())));
    ci = zc * sqrt(v1);
    fout << "  " << opt.cc() * 100.0 << "%CI of g1:" << g1 - ci << " - " << g1 + ci;
    fout << " (" <<  g1 << " +/- " << ci << ")\n";
    ci = zc * sqrt(v2);
    fout << "  " << opt.cc() * 100.0 << "%CI of g2:" << g2 - ci << " - " << g2 + ci;
    fout << " (" <<  g2 << " +/- " << ci << ")\n";
  }
}

void statistclass::cal_sk(double& g1, double& v1, string& skew, double& g2, double& v2, string& kur, stats& statsj) const
// 歪度と尖度を計算する
{
  g1 = v1 = g2 = v2 = 0.0;  skew.erase();  kur.erase();
  int n = statsj.n();
  if (n < 4) return;
  double k2 = statsj.k2(), k3 = statsj.k3(), k4 = statsj.k4();
  if (k2 > 0.0) { g1 = k3 / k2 / sqrt(k2);  g2 = k4 / k2 / k2; }
  v1 = 6.0 * (double)n * (double)(n - 1) / (double)(n - 2) / (double)(n + 1) / (double)(n + 3);
  v2 = 24.0 * (double)n * (double)(n - 1) * (double)(n - 1) / (double)(n - 3) / (double)(n - 2) / (double)(n + 3) / (double)(n + 5);
  if (g1 < -1.0) skew = "Right Sloped";
  else if (fabs(g1) <= 1.0) skew = "Nearly Symmetrical";
  else skew = "Left Sloped";
  if (g2 < -3.0) kur = "Blunt Peak";
  else if (fabs(g2) <= 3.0) kur = "Nearly Normal";
  else kur = "sharp peak";
}

void statistclass::put_dns(ofstream& fout, const option& opt, const string& itemj, stats& statsj)
// 計算結果をデータファイル形式で出力する
{
  datac cdata(18);
  if (opt.sep() || id.ni() == 1) cdata.assign(0, itemj);
  else cdata.assign(0, "All items");
  cdata.assign(1, (double)(statsj.n()));
  cdata.assign(2, statsj.mean());
  cdata.assign(3, statsj.sd());
  cdata.assign(4, statsj.se());
  cdata.assign(5, statsj.min());
  cdata.assign(6, statsj.q1());
  cdata.assign(7, statsj.med());
  cdata.assign(8, statsj.q3());
  cdata.assign(9, statsj.max());
  cdata.assign(10, statsj.max() - statsj.min());
  double cv = (statsj.mean() == 0.0) ? 0.0 : statsj.sd() / statsj.mean() * 100.0;
  cdata.assign(11, cv);
  double g1 = 0.0, v1 = 0.0, g2 = 0.0, v2 = 0.0;
  string skew, kur;
  cal_sk(g1, v1, skew, g2, v2, kur, statsj);
  cdata.assign(12, g1);
  double z = (v1 > 0.0) ? g1 / sqrt(v1) : 0.0L;
  cdata.assign(13, z);
  cdata.assign(14, (double)(ztop(z)));
  cdata.assign(15, g2);
  z = (v2 > 0.0) ? g2 / sqrt(v2) : 0.0;
  cdata.assign(16, z);
  cdata.assign(17, (double)(ztop(z)));
  cdata.put(fout);
}

void statistclass::put_dnsid(ofstream& fout)
// DANSデータのタイトルとデータ識別情報を出力する
{
  idd tid(id.title(), 18);
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  tid.assign_item(0, "Item Name");
  tid.assign_item(1, "Number");
  tid.assign_item(2, "Mean");
  tid.assign_item(3, "SD");
  tid.assign_item(4, "SE");
  tid.assign_item(5, "Min");
  tid.assign_item(6, "25%tile");
  tid.assign_item(7, "Median");
  tid.assign_item(8, "75%tile");
  tid.assign_item(9, "Max");
  tid.assign_item(10, "Range");
  tid.assign_item(11, "CV(%)");
  tid.assign_item(12, "Skewness");
  tid.assign_item(13, "z-value of skewness");
  tid.assign_item(14, "p-value of skewness");
  tid.assign_item(15, "Krutois");
  tid.assign_item(16, "z-value of krutosis");
  tid.assign_item(17, "p-value of krutosis");
  tid.put_item(fout, formtag);
}

void statistclass::plot_ngp(ofstream& fout, option& opt, const string& itemj, stats& statsj, bool delml)
// Ngraph用データを出力する
{
  if (opt.ngp("box"))           // 箱ひげプロット
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -drectangle -x1 -y3 -w80\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_box(fout, opt, itemj, statsj);
  }
  if (opt.ngp("cd"))            // 累積度数分布
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_cdrcd(fout, opt, ngpcd, itemj, statsj);
  }
  if (opt.ngp("cdp"))           // 累積度数分布プロット
  {
    if (delml)
    { fout << NGPdelm << "\n";}
    else
    {
      fout << "-s1 -dmark,00\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_cdrcd(fout, opt, ngpcdp, itemj, statsj);
  }
  if (opt.ngp("d"))             // 度数分布
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -drectangle -x1 -y3 -w80\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_drd(fout, opt, ngpd, itemj, statsj);
  }
  if (opt.ngp("ncd"))           // 理論累積度数分布
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_ncdrcd(fout, opt, ngpncd, itemj, statsj);
  }
  if (opt.ngp("nd"))            // 理論度数分布
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_ndrd(fout, opt, ngpnd, itemj, statsj);
  }
  if (opt.ngp("nrcd"))          // 理論相対累積度数分布
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_ncdrcd(fout, opt, ngpnrcd, itemj, statsj);
  }
  if (opt.ngp("nrd"))           // 理論相対度数分布
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_ndrd(fout, opt, ngpnrd, itemj, statsj);
  }
  if (opt.ngp("rcd"))           // 相対累積度数分布
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_cdrcd(fout, opt, ngprcd, itemj, statsj);
  }
  if (opt.ngp("rcdp"))          // 相対累積度数分布プロット
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dmark,00\n";
      delml = true;
    }
    plot_ngpid(fout, opt);
    plot_cdrcd(fout, opt, ngprcdp, itemj, statsj);
  }
  if (opt.ngp("rd"))            // 相対度数分布
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -drectangle -x1 -y3 -w80\n";
    plot_ngpid(fout, opt);
    plot_drd(fout, opt, ngprd, itemj, statsj);
  }
}

void statistclass::plot_ngpid(ofstream& fout, const option& opt)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  if (opt.sep() == false) id.put_item(fout, formngp);
}

void statistclass::plot_box(ostream& fout, const option& opt, const string& itemj, stats& statsj) const
// Ngraph用箱ヒゲプロットデータを出力する
{
  if (opt.sep()) fout << "%ANALYSIS ITEM:" << itemj << "\n";
  fout << "%<BOX AND WHISKER PLOT:BOX\(-s1 zx0.5,1.5,0.5 -drectangle -x1 -y3 -w80)>\n";
  double x = 1.0;
  fout << x - 0.1 << " " << statsj.q1() << " " << x + 0.1 << " " << statsj.med() << "\n";
  fout << x - 0.1 << " " << statsj.med() << " " << x + 0.1 << " " << statsj.q3() << "\n";
  double hw = (statsj.q3() - statsj.q1()) * 1.5;
  double hlow = statsj.q1() - hw, hup = statsj.q3() + hw;
  double low = statsj.min(), up = statsj.max();
  bool shclow = true, shcup = true;
  listdbl::const_iterator fditr = (statsj.fdlst()).begin();
  while ((shclow || shcup) && fditr != (statsj.fdlst()).end())
  {
    if (shclow && hlow <= *(fditr)) { shclow = false;  low = *(fditr); }
    if (shcup && hup < *(fditr)) shcup = false;
    else up = *(fditr);
    ++fditr;
  }
  fout << "= =\n";
  fout << "%<BOX AND WHISKER PLOT:WHISKER(-zx0.5,1.5,0.5 -derrorbar_y -x1 -y2 -w40 -o500)>\n";
  fout << x << " " << statsj.q1() << " " << low - statsj.q1() << " 0\n";
  fout << x << " " << statsj.q3() << " 0 " << up - statsj.q3() << "\n";
  fout << "= =\n";
  fout << "%<BOX AND WHISKER PLOT:MEAN(-s1 -zx0.5,1.5,0.5 -dmark,72 -w80 -o400)>\n";
  fout << x << " " << statsj.mean() << "\n";
  fout << "= =\n";
  fout << "%<BOX AND WHISKER PLOT:OUTLIERS(-s1 -zx0.5,1.5,0.5 -dmark,2 -w80 -o400)>\n";
  fditr = (statsj.fdlst()).begin();
  while (fditr != (statsj.fdlst()).end())
  {
    if (*(fditr) < low) fout << x << " " << *(fditr) << endl;
    else if (up < *(fditr)) fout << x << " " << *(fditr) << endl;
    ++fditr;
  }
}

void statistclass::plot_drd(ostream& fout, const option& opt, ngptype ngp, const string& itemj, stats& statsj) const
// Ngraph用度数分布／相対度数分布データを出力する
{
  if (opt.sep())
  {
    if (ngp == ngpd) fout << "%<FREQUENCY DISTRIBUTION";
    else fout << "%<RELATIVE FREQUENCY DISTRIBUTION";
    fout << ">\n%ANALYSIS ITEM:" << itemj << "\n";
  }
  double cmin = opt.cmin();
  if (cmin == BLANKf) cmin = statsj.min();
  double h = opt.h();
  if (h == BLANKf) h = statsj.sd() / 3.0; // Charlierの級間基準
  double xlow = cmin;
  int i = 0;
  listdbl::const_iterator fditr = (statsj.fdlst()).begin();
  while (fditr != (statsj.fdlst()).end())
  {
    double xup = xlow + h;
    int ilow = i;
    while (fditr != (statsj.fdlst()).end() && *fditr < xup)
    { ++i;  ++fditr; }
    int f = i - ilow;
    if (f > 0)
    {
      double y = (ngp == ngpd) ? (double)f : (double)f / (double)(statsj.n());
      fout << xlow << " 0 " << xup << " " << y << "\n";
    }
    xlow = xup;
  }
}

void statistclass::plot_cdrcd(ostream& fout, const option& opt, ngptype ngp, const string& itemj, stats& statsj) const
// Ngraph用累積度数分布＆プロット／相対累積度数分布＆プロットデータを出力する
{
  if (opt.sep())
  {
    if (ngp == ngpcd) fout << "%<CUMURATIVE FREQUENCY DISTRIBUTION";
    else if (ngp == ngpcdp) fout << "%<CUMURATIVE FREQUENCY DISTRIBUTION PLOT";
    else if (ngp == ngprcd) fout << "%<RELATIVE CUMURATIVE FREQUENCY DISTRIBUTION";
    else fout << "%<RELATIVE CUMURATIVE FREQUENCY DISTRIBUTION PLOT";
    fout << ">\n%ANALYSIS ITEM:" << itemj << "\n";
  }
  double range = statsj.max() - statsj.min();
  double xmin = statsj.min() - range * 0.2, xmax = statsj.max() + range * 0.2;
  int n = statsj.n(), ni = 0;
  int di = (n > (int)RDdef) ? n / (int)RDdef : 0;
  double ypre = 0.0;
  if (ngp == ngpncd || ngp == ngpnrcd) fout << xmin << " 0\n";
  for (double fdata : statsj.fdlst())
  {
    double x = fdata;
    ++ni;
    double y = (double)ni;
    if (di == 0 || ni % di == 0) // プロット出力
    {
      if (ngp == ngpcd || ngp == ngprcd) fout << x << " " << ypre << "\n";
      if (ngp == ngprcd || ngp == ngprcdp) y /= (double)n;
      fout << x << " " << y << "\n";
      ypre = y;
    }
  }
  if (ngp == ngpncd || ngp == ngpnrcd) fout << xmax << " " << ypre << endl;
}

void statistclass::plot_ndrd(ostream& fout, const option& opt, ngptype ngp, const string& itemj, stats& statsj) const
// Ngraph用理論度数分布／理論相対度数分布データを出力する
{
  if (opt.sep())
  {
    if (ngp == ngpnd) fout << "%<THEORETICAL FREQUENCY DISTRIBUTION";
    else fout << "%<THEORETICAL RELATIVE FREQUENCY DISTRIBUTION";
    fout << ">\n%ANALYSIS ITEM:" << itemj << "\n";
  }
  double sd = statsj.sd();
  double h = opt.h();
  if (h == BLANKf) h = sd / 3.0; // Charlierの級間基準
  double range = sd * 8.0;
  double xmin = statsj.mean() - sd * 4.0, xmax = statsj.mean() + sd * 4.0;
  double dx = (xmax - xmin) / (double)RDdef, x = xmin;
  fout << xmin << " 0\n";
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    ldbl pr = sndis((ldbl)((x + h * 0.5 - statsj.mean()) / sd)) - sndis((ldbl)((x - h * 0.5 - statsj.mean()) / sd));
    ldbl y = (ngp == ngpnd) ? pr * (ldbl)(statsj.n()) : pr;
    fout << x << " " << y << "\n";
  }
}

void statistclass::plot_ncdrcd(ostream& fout, const option& opt, ngptype ngp, const string& itemj, stats& statsj) const
// Ngraph用理論累積度数分布／理論相対累積度数分布データを出力する
{
  if (opt.sep())
  {
    if (ngp == ngpncd) fout << "%<THEORETICAL CUMURATIVE FREQUENCY DISTRIBUTION";
    else fout << "%<THEORETICAL RELATIVE CUMURATIVE FREQUENCY DISTRIBUTION";
    fout << ">\n%ANALYSIS ITEM:" << itemj << "\n";
  }
  double sd = statsj.sd();
  double range = sd * 8.0;
  double xmin = statsj.mean() - sd * 4.0, xmax = statsj.mean() + sd * 4.0;
  double dx = range / (double)RDdef, x = xmin;
  ldbl y = 0.0L;
  fout << xmin << " 0\n";
  for (uint i = 0; i < RDdef; ++i, x += dx)
  {
    ldbl pr = sndis((x - statsj.mean()) / sd);
    y = (ngp == ngpncd) ? pr * (ldbl)statsj.n() : pr;
    fout << x << " " << y << "\n";
  }
  y = (ngp == ngpncd) ? (ldbl)(statsj.n()) : 1.0L;
  fout << xmax << " " << y << endl;
}

int statistclass::get(const char* c_fname, option& opt)
// データを読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check() != 0) throw option_error("Run-time option error !");
  return nc;
}

void statistclass::put(option& opt) // 基礎統計量を出力する
{
  constexpr int LENline = 78;   // 区切り線の長さ
  if (opt.res("res")) put_resid(cout, opt); // 分析結果出力
  ofstream foutdns;
  if (opt.dnsname() != "")      // DANSデータ出力
  {
    if (ofopen(foutdns, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
    put_dnsid(foutdns);
  }
  ofstream foutngp;
  if (opt.ngpname() != "")      // Ngraphデータ出力
  {
    if (ofopen(foutngp, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  }
  stats statsj;                 // 基礎統計量+k統計量+四分位値
  size_t jstart = 0, jend = id.ni();
  if (opt.sep() == false) { jstart = 0;  jend = 1; } // 全項目を一度に集計
  for (size_t j = jstart; j < jend; ++j)
  {
    statsj.clear();
    if (opt.sep())
    {
      for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr) statsj.sigma(ditr->fdata(j));
    }
    else
    {
      for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
      {
        for (size_t i = 0; i < id.ni(); ++i) statsj.sigma(ditr->fdata(i));
      }
    }
    if (statsj.n() > 0)
    {
      statsj.sort();
      if (opt.res("res"))       // 計算結果出力
      {
        if (opt.sep()) cout << "ANALYSIS ITEM:" << id.item(j) << "\n";
        bool putres = false;
        if (opt.res("stat"))
        {
          putline(cout, LENline);
          put_stat(cout, opt, statsj);
          if (opt.res("cc")) put_mci(cout, opt, id.item(j), statsj);
          putres = true;
        }
        if (opt.res("norm"))
        {
          putline(cout, LENline);
          put_norm(cout, opt, id.item(j), statsj);
          putres = true;
        }
        if (putres) putline(cout, LENline);
        cout << "\n";
      }
      if (opt.dnsname() != "") put_dns(foutdns, opt, id.item(j), statsj);
      if (opt.ngpname() != "") plot_ngp(foutngp, opt, id.item(j), statsj, (j > jstart) ? true : false);
    }
  }
  if (opt.res("res")) cout << endl;
}

int main(int argc, char* argv[]) // メインルーチン
{
  statistclass statist;         // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) statist.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { statist.get(argv[i], opt);  fok = false; }
    }
    statist.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occured !");
    return 255;
  }
  return 0;
}
