// tcorege.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-26 11:34:43 tac>
//
// 時期別相関係数と回帰直線、連動係数と連動回帰直線を計算する。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W tcorege tcorege.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "tcorege  Version 7.41e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:tcorege  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The correlation analysis and linear regression analysis and linkage analysis",
  "  and linkage regression analysis by time period",
  "<OPTION>",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={ce|data|ddata|line|lline}[,...]:select the data for Ngraph",
  "    ce:confidence,tolerance,prediction ellipse  data:raw data",
  "    ddata:difference of raw data  line:regression line",
  "    lline:linkage regression line  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=tcorege_ngp.dat",
  "  -res=[-][{def|all|cc[<C-COEF>]|l|r|stat}][,...]:select output results",
  "    -:with -, no output(-res=-:no output at all)",
  "    def:same output as default  all:output all",
  "    l:linkage coefficient and linkage regression line",
  "    cc[<C-COEF>]:confidence intervals(0=<C-COEF<1)",
  "    r:correlation coefficient and regression line  stat:basic statistics",
  "    default:auto judgement according to data",
  "  -ti[={-|+}]:specify whether to output results by time period(-ti:ti=+)",
  "  -tt[={-|+}]:specify whether to output time total results(-tt:tt=+)",
  "  -x=[<START NUMBER>][-[<END NUMBER>]]:specify the item number",
  "    of the explanatory variables",
  "  -y=<START NUMBER>:specify the start item number of the criterion variable"
};

const string Title = "=== Correlation and Linkage Analysis by Time Period ==="; // 表題
const string NGPNAMEdef = "tcorege_ngp.dat"; // 省略時Ngraphファイル名

constexpr int LENline = 75;     // 区切り線の長さ

class option                    // オプション集合体
{
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  bool vti;                     // 時期別集計結果出力フラッグ
  bool vtt;                     // 時期合計集計結果出力フラッグ
  int vnt;                      // 時期数
  int vxs;                      // 変数Xの開始項目番号
  int vys;                      // 変数Yの開始項目番号
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  bool ti() const               // 時期別集計結果出力フラッグを返す
  { return vti; }
  bool tt() const               // 時期合計集計結果出力フラッグを返す
  { return vtt; }
  int nt() const                // 時期数を返す
  { return vnt; }
  int xs() const                // 変数Xの開始項目番号を返す
  { return vxs; }
  int ys() const                // 変数Yの開始項目番号を返す
  { return vys; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vti(true), vtt(true), vnt(1), vxs(1), vys(2)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ti")      // 時期別集計結果の出力指定
  {
    if (opt[1].empty() || opt[1] == "+") vti = true;
    else if (opt[1] == "-") vti = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "tt")      // 時期合計集計結果の出力指定
  {
    if (opt[1].empty() || opt[1] == "+") vtt = true;
    else if (opt[1] == "-") vtt = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "x")       // 説明変数の項目指定
  {
    int xe = 0;
    if (getrange(vxs, xe, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
    else if (vxs < 1 || vxs > xe) err = badopt(vpname, optstr, USAGE);
    else vnt = xe - vxs + 1;
  }
  else if (opt[0] == "y")       // 目的変数の項目指定
  {
    char* ptr = nullptr;
    vys = (int)(strtol(opt[1].c_str(), &ptr, 10));
    if (vys < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "At least 2 items are required !");
  int xe = vxs + vnt - 1, ye = vys + vnt - 1;
  if (xe > (int)ni || ye > (int)ni) return disperr(vpname, "Item specification of x and y is incorrect !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["ce"] = vngpmap["data"] = vngpmap["ddata"] = vngpmap["line"] = vngpmap["lline"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["ce"] = vngpmap["data"] = vngpmap["ddata"] = vngpmap["line"] = vngpmap["lline"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["l"] = vresmap["r"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["l"] = vresmap["r"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["cc"] = vresmap["l"] = vresmap["r"] = vresmap["stat"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class linkcor                   // 連動係数用統計量クラス
{
private:
  int vnd;                      // 差の例数
  double sigmadx;               // Σdx
  double sigmadx2;              // Σdx2
  double sigmady;               // Σdy
  double sigmady2;              // Σdy2
  double sigmadxy;              // Σdxdy
public:
  linkcor() : vnd(0), sigmadx(0.0), sigmadx2(0.0), sigmady(0.0), sigmady2(0.0), sigmadxy(0.0)
  // デフォールトコンストラクタ
  {}
  linkcor(int nd0, double sigmadx0, double sigmadx20, double sigmady0, double sigmady20, double sigmadxy0) : vnd(nd0), sigmadx(sigmadx0), sigmadx2(sigmadx20), sigmady(sigmady0), sigmady2(sigmady20), sigmadxy(sigmadxy0)
  // コンストラクタ
  {}
  linkcor(const linkcor& linkcor0) : vnd(linkcor0.vnd), sigmadx(linkcor0.sigmadx), sigmadx2(linkcor0.sigmadx2), sigmady(linkcor0.sigmady), sigmady2(linkcor0.sigmady2), sigmadxy(linkcor0.sigmadxy)
  // コピーコンストラクタ
  {}
  virtual ~linkcor()            // デストラクタ
  {}
  const linkcor& operator=(const linkcor& linkcor0); // コピー代入演算子
  const linkcor& operator+=(const linkcor& linkcor0); // 加算代入演算子
  int sigmad(double dx, double dy); // dxとdyの累積集計
  int nd() const                    // 差の例数を返す
  { return vnd; }
  double mdx() const            // dxの平均を返す
  { return (vnd > 0) ? sigmadx / vnd : 0.0; }
  double mdy() const            // dyの平均を返す
  { return (vnd > 0) ? sigmady / vnd : 0.0; }
  double sdxx() const           // dxの単純平方和を返す
  { return sigmadx2; }
  double sdyy() const           // dyの単純平方和を返す
  { return sigmady2; }
  double sdxy() const           // dxとdyの単純積和を返す
  { return sigmadxy; }
  double sddx() const           // dxの標準変化を返す
  { return (vnd > 0) ? sqrt(sigmadx2 / vnd) : 0.0; }
  double sddy() const           // dyの標準変化を返す
  { return (vnd > 0) ? sqrt(sigmady2 / vnd) : 0.0; }
  double lamda() const;         // dxとdyの連動係数を返す
};

const linkcor& linkcor::operator=(const linkcor& linkcor0) // コピー代入演算子
{
  if (&linkcor0 == this) return *this; // 自分自身への代入を防ぐ
  vnd = linkcor0.vnd;  sigmadx = linkcor0.sigmadx;
  sigmadx2 = linkcor0.sigmadx2;  sigmady = linkcor0.sigmady;
  sigmady2 = linkcor0.sigmady2;  sigmadxy = linkcor0.sigmadxy;
  return *this;
}

const linkcor& linkcor::operator+=(const linkcor& linkcor0) // 加算代入演算子
{
  if (linkcor0.vnd < 1) return *this;
  vnd += linkcor0.vnd;  sigmadx += linkcor0.sigmadx;
  sigmadx2 += linkcor0.sigmadx2;  sigmady += linkcor0.sigmady;
  sigmady2 += linkcor0.sigmady2;  sigmadxy += linkcor0.sigmadxy;
  return *this;
}

int linkcor::sigmad(double dx, double dy) // dxとdyの累積集計
{
  if (dx == BLANKf || dy == BLANKf) return 0; // 累積せず
  ++vnd;  sigmadx += dx;  sigmadx2 += dx * dx;
  sigmady += dy;  sigmady2 += dy * dy;  sigmadxy += dx * dy;
  return 1;                     // 累積実行
}

double linkcor::lamda() const   // dxとdyの連動係数を返す
{
  double ssxy = sigmadx2 * sigmady2;
  return (ssxy > 0.0) ? sigmadxy / sqrt(ssxy) : 0.0;
}

double putlci(ostream& fout, double cc, int n, double l)
// 連動係数の信頼区間を出力する
{
  if (cc < 0.0 || 1.0 <= cc || n < 3 || fabs(l) == 1.0) return 0.0; // 計算できず
  double zl = log((1.0 + l) / (1.0 - l)) * 0.5;
  double ci = ptoz(1.0 - cc) / sqrt(double(n - 2));
  fout << "  " << cc * 100.0 << "%CI of l=tanh(" << zl << " +/- ";
  fout << ci << ")(" << tanh(zl - ci) << " - " << tanh(zl + ci) << ")\n";
  return ci;                    // 信頼幅を返す
}

class varclass                  // 変数クラス
{
private:
  int vnt;                        // 時期数
  pairdbl xrange;                 // xの最小値と最大値
  pairdbl dxrange;                // x変化量の最小値と最大値
  vector<statv<double>> xstatvec; // Xの全データ用基礎統計量ベクトル
  vector<statv<double>> ystatvec; // Yの全データ用基礎統計量ベクトル
  vector<statcor<double>> statcorvec; // 相関基礎統計量ベクトル
  vector<linkcor> linkcorvec;         // 連動係数用統計量ベクトル
public:
  varclass(int nt0) : vnt(nt0), xrange(DBL_MAX, -DBL_MAX), dxrange(DBL_MAX, -DBL_MAX), xstatvec(nt0 + 1), ystatvec(nt0 + 1), statcorvec(nt0 + 1), linkcorvec(nt0 + 1)
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int cal(const option& opt, ldata& datatlst);
  // データを読み込み変数をセットする
  double xmin() const           // xの最小値を返す
  { return xrange.first; }
  double xmax() const           // xの最大値を返す
  { return xrange.second; }
  double dxmin() const          // x変化量の最小値を返す
  { return dxrange.first; }
  double dxmax() const          // x変化量の最大値を返す
  { return dxrange.second; }
  int nt() const                // 時期数を返す
  { return vnt; }
  int xn(size_t idx) const      // xの例数を返す
  { return (idx < xstatvec.size()) ? xstatvec[idx].n() : 0; }
  int xnt() const               // xの全例数を返す
  { return xstatvec[vnt].n(); }
  double xmean(size_t idx) const // xの平均値を返す
  { return (idx < xstatvec.size()) ? xstatvec[idx].mean() : 0.0; }
  double xvr(size_t idx) const  // xの不偏分散を返す
  { return (idx < xstatvec.size()) ? xstatvec[idx].vr() : 0.0; }
  int yn(size_t idx) const      // yの例数を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].n() : 0; }
  int ynt() const               // yの全例数を返す
  { return ystatvec[vnt].n(); }
  double ymean(size_t idx) const // yの平均値を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].mean() : 0.0; }
  double yvr(size_t idx) const  // yの不偏分散を返す
  { return (idx < ystatvec.size()) ? ystatvec[idx].vr() : 0.0; }
  int cn(size_t idx) const      // xyペアの例数を返す
  { return (idx < statcorvec.size()) ? statcorvec[idx].n() : 0; }
  double cmx(size_t idx) const  // xyペアのxの平均値を返す
  { return (idx < statcorvec.size()) ? statcorvec[idx].mx() : 0.0; }
  double cmy(size_t idx) const  // xyペアのyの平均値を返す
  { return (idx < statcorvec.size()) ? statcorvec[idx].my() : 0.0; }
  double csxx(size_t idx) const // xyペアのxの平方和を返す
  { return (idx < statcorvec.size()) ? statcorvec[idx].sxx() : 0.0; }
  double csyy(size_t idx) const // xyペアのyの平方和を返す
  { return (idx < statcorvec.size()) ? statcorvec[idx].syy() : 0.0; }
  double csxy(size_t idx) const // xyペアの積和を返す
  { return (idx < statcorvec.size()) ? statcorvec[idx].sxy() : 0.0; }
  double cvxx(size_t idx) const // xyペアのxの不偏分散を返す
  { return (idx < statcorvec.size()) ? statcorvec[idx].vxx() : 0.0; }
  double cvyy(size_t idx) const // xyペアのyの不偏分散を返す
  { return (idx < statcorvec.size()) ? statcorvec[idx].vyy() : 0.0; }
  double cvxy(size_t idx) const // xyペアの共分散を返す
  { return (idx < statcorvec.size()) ? statcorvec[idx].vxy() : 0.0; }
  double crxy(size_t idx) const // xyペアの相関係数を返す
  { return (idx < statcorvec.size()) ? statcorvec[idx].rxy() : 0.0; }
  int lnd(size_t idx) const     // 変化量の例数を返す
  { return (idx < linkcorvec.size()) ? linkcorvec[idx].nd() : 0; }
  int lndt() const              // 変化量の全例数を返す
  { return linkcorvec[vnt].nd(); }
  double lmdx(size_t idx) const // x変化量の平均値を返す
  { return (idx < linkcorvec.size()) ? linkcorvec[idx].mdx() : 0.0; }
  double lmdy(size_t idx) const // y変化量の平均値を返す
  { return (idx < linkcorvec.size()) ? linkcorvec[idx].mdy() : 0.0; }
  double lsddx(size_t idx) const // x変化量の標準偏差を返す
  { return (idx < linkcorvec.size()) ? linkcorvec[idx].sddx() : 0.0; }
  double lsddy(size_t idx) const // y変化量の標準偏差を返す
  { return (idx < linkcorvec.size()) ? linkcorvec[idx].sddy() : 0.0; }
  double lsdxx(size_t idx) const // x変化量の平方和を返す
  { return (idx < linkcorvec.size()) ? linkcorvec[idx].sdxx() : 0.0; }
  double lsdyy(size_t idx) const // y変化量の平方和を返す
  { return (idx < linkcorvec.size()) ? linkcorvec[idx].sdyy() : 0.0; }
  double lsdxy(size_t idx) const // xy変化量の積和を返す
  { return (idx < linkcorvec.size()) ? linkcorvec[idx].sdxy() : 0.0; }
  double llamda(size_t idx) const // xy変化量の連動係数を返す
  { return (idx < linkcorvec.size()) ? linkcorvec[idx].lamda() : 0.0; }
};

int varclass::cal(const option& opt, ldata& datalst)
// データを読み込み変数をセットする
{
  int nc = 0;
  for (auto& cdata : datalst)
  {
    double xb = cdata.fdata(opt.xs() - 1); // Xの前値
    double yb = cdata.fdata(opt.ys() - 1); // Yの前値
    bool bok = (xb != BLANKf && yb != BLANKf) ? true : false; // 前値の有無
    for (int i = 0; i < vnt; ++i)
    {
      double x = cdata.fdata(opt.xs() + i - 1);
      double y = cdata.fdata(opt.ys() + i - 1);
      if (x != BLANKf && y != BLANKf)
      {
        xstatvec[i].sigma(x);  ystatvec[i].sigma(y);
        statcorvec[i].sigma(x, y);
        if (x < xrange.first) xrange.first = x;
        if (x > xrange.second) xrange.second = x;
        if (i > 0 && bok)
        {
          double dx = x - xb, dy = y - yb;
          linkcorvec[i].sigmad(dx, dy);
          if (dx < dxrange.first) dxrange.first = dx;
          if (dx > dxrange.second) dxrange.second = dx;
        }
        ++nc;
      }
    }
  }
  if (nc < 1) return nc;        // 有効なケースなし
  for (int i = 0; i < vnt; ++i)
  {
    if (xstatvec[i].n() > 0)
    {
      xstatvec[vnt] += xstatvec[i];
      ystatvec[vnt] += ystatvec[i];
      statcorvec[vnt] += statcorvec[i];
      linkcorvec[vnt] += linkcorvec[i];
    }
  }
  return nc;                    // 有効なケース数を返す
}

class tcoregclass               // tcoregメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout); // 分析結果のタイトルとデータ識別情報を出力する
  void put_resttid(ostream& fout, const option& opt);
  // 時期合計分析結果のタイトルとデータ識別情報を出力する
  void put_coreg(ostream& fout, option& opt, int idx, const varclass& var);
  // 相関係数と回帰直線を出力する
  void put_link(ostream& fout, option& opt, int idx, const varclass& var);
  // 連動係数と連動回帰直線を出力する
  void plot_ngraph(option& opt, const varclass& var); // Ngraph用データ出力
  void plot_ngpid(ofstream& fout);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpttid(ofstream& fout, const option& opt);
  // 時期合計Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpdata(ofstream& fout, const option& opt, int idx);
  // Ngraph用に実測値のデータを出力する
  void plot_ngpline(ofstream& fout, int idx, const varclass& var) const;
  // Ngraph用に回帰直線を出力する
  void plot_ngpce(ofstream& fout, const option& opt, int idx, const varclass& var) const;
  // Ngraph用に信頼楕円を出力する
  void plot_ngpellipse(ofstream& fout, double mx, double my, double a, double b, double h) const;
  // 楕円のデータを出力する
  void plot_ngpddata(ofstream& fout, const option& opt, int idx);
  // Ngraph用に変化量のデータを出力する
  void plot_ngplline(ofstream& fout, int idx, const varclass& var) const;
  // Ngraph用に連動回帰直線を出力する
public:
  tcoregclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~tcoregclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み集計する
  void put(option& opt);                     // 基礎統計量を出力する
};

void tcoregclass::put_resid(ostream& fout)
// 分析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
}

void tcoregclass::put_resttid(ostream& fout, const option& opt)
// 時期合計分析結果のタイトルとデータ識別情報を出力する
{
  int inos = opt.xs(), inoe = opt.xs() + opt.nt() - 1;
  for (int i = inos; i <= inoe; ++i)
  {
    fout << "EXPLANATORY VARIABLE X" << setw(3) << i - opt.xs() + 1;
    fout << ":" << id.item(i - 1) << "\n";
  }
  inos = opt.ys();  inoe = opt.ys() + opt.nt() - 1;
  for (int i = inos; i <= inoe; ++i)
  {
    fout << "CRITERION VARIABLE   Y" << setw(3) << i - opt.ys() + 1;
    fout << ":" << id.item(i - 1) << "\n";
  }
  fout << "\n";
}

void tcoregclass::put_coreg(ostream& fout, option& opt, int idx, const varclass& var)
// 相関係数と回帰直線を出力する
{
  if (opt.res("stat"))
  {
    fout << "X:";
    putmsd(fout, var.xn(idx), var.xmean(idx), var.xvr(idx));
    fout << "Y:";
    putmsd(fout, var.yn(idx), var.ymean(idx), var.yvr(idx));
    putline(fout, LENline);
  }
  int n = var.cn(idx);
  int df = n - 2;
  double sxx = var.csxx(idx), syy = var.csyy(idx);
  double cxy = var.csxy(idx);
  double r = var.crxy(idx), b = (sxx > 0.0) ? cxy / sxx : 0.0;
  double a = var.cmy(idx) - var.cmx(idx) * b;
  fout << "  Correlation Coefficient:r=" << r << "  r-squared=" << r * r << "\n";
  double zr = 0.0;
  double ci = calrci(zr, opt.cc(), n, r);
  if (opt.res("cc"))
  {
    fout << "  " << opt.cc() * 100.0 << "%CI of r:tanh(" << zr << " +/- " << ci << ")(" << tanh(zr - ci) << " - " << tanh(zr + ci) << ")\n";
  }
  double vb = (df > 0 && sxx > 0.0) ? (syy - cxy * b) / (double)df / sxx : 0.0;
  fout << "  Regression Line:y=a + bx=" << a;
  fout.setf(ios::showpos);
  fout << b << "x    SE of b=";
  fout.unsetf(ios::showpos);
  fout << sqrt(vb) << "\n";
  if (opt.res("cc"))
  {
    ci = calmcci(mcfis, opt.cc(), 1, 1, df, 1, vb);
    fout << "  " << opt.cc() * 100.0 << "%CI of b:";
    fout << b << " +/- " << ci << "(" << b - ci << " - " << b + ci << ")\n";
  }
  double vr = (df > 0 && fabs(r) < 1.0) ? (1.0 - r * r) / (double)df : 0.0;
  double t = (vr > 0.0) ? r / sqrt(vr) : 0.0;
  ldbl p = ttop(t, df);
  fout << "  Test of r and b:t=" << t << "  DF=" << df << "  p=" << p << pind(p) << "\n";
}

void tcoregclass::put_link(ostream& fout, option& opt, int idx, const varclass& var)
// 連動係数と連動回帰直線を出力する
{
  if (opt.res("stat"))
  {
    fout << "Difference of X:n=" << left << setw(6) << var.lnd(idx);
    fout << "Mean=" << setw(12) << var.lmdx(idx);
    fout << "Standard Difference=" << var.lsddx(idx) << "\n";
    fout << "Difference of Y:n=" << setw(6) << var.lnd(idx);
    fout << "Mean=" << setw(12) << var.lmdy(idx) << right;
    fout << "Standard Difference=" << var.lsddy(idx) << "\n";
    putline(fout, LENline);
  }
  double sx = var.lsdxx(idx), sy = var.lsdyy(idx), cxy = var.lsdxy(idx);
  double b = (sx > 0.0) ? cxy / sx : 0.0, l = var.llamda(idx);
  int df = var.lnd(idx) - 1;
  fout << "  Linkage Coefficient:l=" << l << "  l-squared=" << l * l << "\n";
  if (opt.res("cc")) putlci(fout, opt.cc(), df + 1, l);
  double vb = (df > 0 && sx > 0.0) ? (sy - cxy * b) / (double)df / sx : 0.0;
  fout << "  Linkage Regression Line:dy=b dx=" << b << "dx    SE of b=" << sqrt(vb) << "\n";
  double ci = calmcci(mcfis, opt.cc(), 1, 1, df, 1, vb);
  if (opt.res("cc"))
  {
    fout << "  " << opt.cc() * 100.0 << "%CI of b=";
    fout << b << " +/- " << ci << "(" << b - ci << " - " << b + ci << ")\n";
  }
  double vl = (df > 0 && fabs(l) < 1.0) ? (1.0 - l * l) / double(df) : 0.0;
  double t = (vl > 0.0) ? l / sqrt(vl) : 0.0;
  ldbl p = ttop(t, df);
  fout << "  Test of l and b:t=" << t << "  DF=" << df << "  p=" << p << pind(p) << "\n";
}

void tcoregclass::plot_ngraph(option& opt, const varclass& var)
// Ngraph用データ出力
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  plot_ngpid(fout);
  int ntt = (opt.nt() > 1) ? opt.nt() : 0;
  bool delml = false;
  for (int i = 0; i < ntt; ++i)
  {
    if (var.xn(i) > 0)
    {
      if (opt.ti() || opt.ngp("data") || opt.ngp("ddata"))
      {
        if (delml) fout << NGPdelm << "\n";
        delml = false;
        fout << "%EXPLANATORY VARIABLE X:" << id.item(opt.xs() + i - 1) << "\n";
        fout << "%CRITERION VARIABLE   Y:" << id.item(opt.ys() + i - 1) << "\n";
        if (opt.ngp("data"))    // 時期合計だけの時も出力する
        {
          plot_ngpdata(fout, opt, i);
          delml = true;
        }
        if (opt.ngp("line"))
        {
          if (delml) fout << NGPdelm << "\n";
          else delml = true;
          plot_ngpline(fout, i, var);
        }
        if (opt.ngp("ce"))
        {
          if (delml) fout << NGPdelm << "\n";
          else delml = true;
          plot_ngpce(fout, opt, i, var);
        }
        if (i > 0)
        {
          if (opt.ngp("ddata")) // 時期合計だけの時も出力する
          {
            if (delml) fout << NGPdelm << "\n";
            else delml = true;
            plot_ngpddata(fout, opt, i);
          }
          if (opt.ngp("lline"))
          {
            if (delml) fout << NGPdelm << "\n";
            else delml = true;
            plot_ngplline(fout, i, var);
          }
        }
      }
    }
  }
  if (var.xnt() < 1 || opt.tt() == false) return;
  if (opt.ngp("line") == false && opt.ngp("ce") == false && opt.ngp("lline") == false) return;
  if (delml) fout << NGPdelm << "\n";
  plot_ngpttid(fout, opt);
  delml = false;
  if (opt.ngp("line"))
  {
    plot_ngpline(fout, ntt, var);
    delml = true;
  }
  if (opt.ngp("ce"))
  {
    if (delml) fout << NGPdelm << "\n";
    else delml = true;
    plot_ngpce(fout, opt, ntt, var);
  }
  if (opt.ngp("lline"))
  {
    if (delml) fout << NGPdelm << "\n";
    plot_ngplline(fout, ntt, var);
  }
}

void tcoregclass::plot_ngpid(ofstream& fout)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
}

void tcoregclass::plot_ngpttid(ofstream& fout, const option& opt)
// 時期合計Ngraph用データのタイトルとデータ識別情報を出力する
{
  int inos = opt.xs(), inoe = opt.xs() + opt.nt() - 1;
  for (int i = inos; i <= inoe; ++i)
  {
    fout << "%EXPLANATORY VARIABLE X" << setw(3) << i - opt.xs() + 1;
    fout << ":" << id.item(i - 1) << "\n";
  }
  inos = opt.ys();  inoe = opt.ys() + opt.nt() - 1;
  for (int i = inos; i <= inoe; ++i)
  {
    fout << "%CRITERION VARIABLE   Y" << setw(3) << i - opt.ys() + 1;
    fout << ":" << id.item(i - 1) << "\n";
  }
  fout << "%TIME PERIOD TOTAL\n";
}

void tcoregclass::plot_ngpdata(ofstream& fout, const option& opt, int idx)
// Ngraph用に実測値のデータを出力する
{
  fout << "%<RAW DATA>\n";
  fout << "%-s1 -dmark\n";
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double x = ditr->fdata(opt.xs() + idx - 1);
    double y = ditr->fdata(opt.ys() + idx - 1);
    if (x != BLANKf && y != BLANKf) fout << x << " " << y << "\n";
  }
}

void tcoregclass::plot_ngpline(ofstream& fout, int idx, const varclass& var) const
// Ngraph用に回帰直線を出力する
{
  fout << "%<REGRESSION LINE>\n";
  fout << "%-s1 -dline\n";
  double mx = var.cmx(idx), my = var.cmy(idx);
  double sxx = var.csxx(idx), syy = var.csyy(idx);
  double sxy = var.csxy(idx);
  double b = (sxx > 0.0) ? sxy / sxx : 0.0;
  double a = my - b * mx;
  double xmin = var.xmin(), xmax = var.xmax();
  double xrange = xmax - xmin;
  double x = xmin - xrange * 0.5;
  double y = a + b * x;
  fout << x << " " << a + b * x << "\n";
  x = xmax + xrange * 0.5;
  fout << x << " " << a + b * x << endl;
}

void tcoregclass::plot_ngpce(ofstream& fout, const option& opt, int idx, const varclass& var) const
// Ngraph用に信頼楕円を出力する
{
  fout << "%-s1 -dpolygon\n";
  int n = var.cn(idx);
  double mx = var.cmx(idx), my = var.cmy(idx);
  double vxx = var.cvxx(idx), vyy = var.cvyy(idx), vxy = var.cvxy(idx);
  int dfr = n - 2;
  double cf = 1.0 / (double)n, rf = 1.0 + 1.0 / (double)n;
  double ccpr = opt.cc() * 100.0;
  fout << "%<" << ccpr << "% CONFIDENCE ELLIPSE>\n";
  double a = 0.0, b = 0.0, h = 0.0;
  calelp(a, b, h, opt.cc(), 2, dfr, vxx * cf, vyy * cf, vxy * cf);
  plot_ngpellipse(fout, mx, my, a, b, h);
  fout << "= =\n";
  fout << "%<" << ccpr << "% TOLERANCE ELLIPSE>\n";
  calelp(a, b, h, opt.cc(), 2, dfr, vxx, vyy, vxy);
  plot_ngpellipse(fout, mx, my, a, b, h);
  fout << "= =\n";
  fout << "%<" << ccpr << "% PREDICTION ELLIPSE>\n";
  calelp(a, b, h, opt.cc(), 2, dfr, vxx * rf, vyy * rf, vxy * rf);
  plot_ngpellipse(fout, mx, my, a, b, h);
}

void tcoregclass::plot_ngpellipse(ofstream& fout, double mx, double my, double a, double b, double h) const
// 楕円のデータを出力する
{
  constexpr int RDdef = 60;     // 楕円の分割数
  constexpr double Dt = 2.0 * M_PI / (double)RDdef;
  double x0 = 0.0, y0 = 0.0, x = 0.0, y = 0.0, t = 0.0;
  for (int i = 0; i <= RDdef; ++i, t += Dt)
  {
    x0 = a * cos(t);  y0 = b * sin(t);
    x = mx + x0 * cos(h) - y0 * sin(h);  y = my + x0 * sin(h) + y0 * cos(h);
    fout << x << " " << y << "\n";
  }
}

void tcoregclass::plot_ngpddata(ofstream& fout, const option& opt, int idx)
// Ngraph用に変化量のデータを出力する
{
  fout << "%<DIFFERENCE DATA>\n";
  fout << "%-s1 -dmark\n";
  for (auto& cdata : datalst)
  {
    double xb = cdata.fdata(opt.xs() - 1); // Xの前値
    double yb = cdata.fdata(opt.ys() - 1); // Yの前値
    if (xb != BLANKf && yb != BLANKf)
    {
      double x = cdata.fdata(opt.xs() + idx - 1);
      double y = cdata.fdata(opt.ys() + idx - 1);
      if (x != BLANKf && y != BLANKf) fout << x - xb << " " << y - yb << "\n";
    }
  }
}

void tcoregclass::plot_ngplline(ofstream& fout, int idx, const varclass& var) const
// Ngraph用に連動回帰直線を出力する
{
  fout << "%<LINKAGE REGRESSION LINE>\n";
  double sx = var.lsdxx(idx), sy = var.lsdyy(idx), cxy = var.lsdxy(idx);
  double b = (sx > 0.0) ? cxy / sx : 0.0;
  double xmin = var.dxmin(), xmax = var.dxmax();
  double xrange = xmax - xmin;
  double x = xmin - xrange * 0.5;
  double y = b * x;
  fout << x << " " << b * x << "\n";
  x = xmax + xrange * 0.5;
  fout << x << " " << b * x << "\n";
}

int tcoregclass::get(const char* c_fname, option& opt)
// データを読み込み集計する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void tcoregclass::put(option& opt) // 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var(opt.nt());           // 計算用変数オブジェクト
  if (var.cal(opt, datalst) < 1) return; // 有効なデータがない
  if (opt.res("res"))           // 計算結果出力
  {
    put_resid(cout);
    int ntt = (opt.nt() > 1) ? opt.nt() : 0;
    bool putresi = true, putrest = false;
    for (int i = 0; i < ntt; ++i)
    {
      if (var.xn(i) > 0)
      {
        if (opt.ti())
        {
          cout << "EXPLANATORY VARIABLE X:" << id.item(opt.xs() + i - 1) << "\n";
          cout << "CRITERION VARIABLE   Y:" << id.item(opt.ys() + i - 1) << "\n";
          putresi = false;
          if (opt.res("r"))
          {
            putline(cout, LENline);
            put_coreg(cout, opt, i, var);
            putresi = true;
          }
          if (opt.res("l") && var.lnd(i) > 0)
          {
            putline(cout, LENline);
            put_link(cout, opt, i, var);
            putresi = true;
          }
          if (putresi) putline(cout, LENline);
          cout << "\n";
        }
      }
    }
    if (var.xnt() > 0 && opt.tt())
    {
      if (opt.ti() == false) put_resttid(cout, opt);
      cout << "<TIME PERIOD TOTAL>\n";
      putrest = false;
      if (opt.res("r"))
      {
        putline(cout, LENline);
        put_coreg(cout, opt, ntt, var);
        putrest = true;
      }
      if (opt.res("l") && var.lndt() > 0)
      {
        putline(cout, LENline);
        put_link(cout, opt, ntt, var);
        putrest = true;
      }
    }
    if (putrest)
    {
      putline(cout, LENline);
      cout << "\n";
    }
    else if (putresi == false)
    { cout << "\n"; }
    cout << endl;
  }
  if (opt.ngpname() != "") plot_ngraph(opt, var); // Ngraph用データ出力
}

int main(int argc, char* argv[]) // メインルーチン
{
  tcoregclass tcoreg;           // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) tcoreg.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { tcoreg.get(argv[i], opt);  fok = false; }
    }
    tcoreg.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occured !");
    return 255;
  }
  return 0;
}
