// tteste.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-26 11:46:19 tac>
//
// 対応のあるt検定(1標本t検定)を多時期で行う。
// 3時期以上については、本来は二元配置分散分析を行うべきだが、
// 諸般の事情により多時期についても計算している。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W ttest ttest.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "tteste  Version 7.41e        Copyleft(Vo\\oV) 203 Sugimoto Norio",
  "USAGE:tteste [<OPTION> ...] [<DATA FILE> ...]",
  "  The paired t-test between the key item and each item.",
  "<OPTION>",
  "  -comp[={-|+}]:specify whether to analysis only complete cases(-comp:-comp=+)",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND]:show online manual(-h:-h=man)",
  "  -key=<NUMBER>:specify the key item number(default:-key=1)",
  "  -ngp={dbar|dci|ddata|dmsd|dmse|vbar|vci|vdata|vmsd|vmse}[,...]",
  "      :select the data for Ngraph",
  "    dbar:bar chart of the difference  dci:mean of the difference +/- CI",
  "    ddata:difference data  dmsd:mean of the difference +/- SD",
  "    dmse:mean of the difference +/- SE",
  "    vbar:bar chart of the raw data  vci:mean of the raw data +/- CI",
  "    vdata:raw data  vmsd:mean of the raw data +/- SD",
  "    vmse:mean of raw data +/- SE",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=tteste_ngp.dat",
  "  -res=[-][{def|all|astat|cc[<C-COEF>]|m|pstat|tt}][,...]",
  "      :select output results",
  "    -:with -, no output(-res=-:no output at all)",
  "    def:same ouput as default  all:output all",
  "    astat:basic statistcs of all data",
  "    cc[<C-COEF>]:confidence interval(0=<C-COEF<1)",
  "    m:multiple comparison  pstat:basic statistcs of paired data",
  "    tt:paired t-test  default:auto judgement according to data",
  "  -t=<TIME1>,<TIME2>[,<TIME3>,...]:specify the time of item for Ngraph"
};

const string Title = "=== Comparison of Means Between Two Periods ==="; // 表題
const string NGPNAMEdef = "tteste_ngp.dat"; // 省略時Ngraphファイル名

enum ngptype                    // Ngraph用出力データの種類
{
  ngpnon,                       // 出力なし
  ngpdbar,                      // 変化量バーチャート
  ngpdci,                       // 変化量平均値±信頼区間
  ngpddata,                     // 変化量の生データ
  ngpdmsd,                      // 変化量平均値±SD
  ngpdmse,                      // 変化量平均値±SE
  ngpvbar,                      // 実測値バーチャート
  ngpvci,                       // 実測値平均値±信頼区間
  ngpvdata,                     // 実測値の生データ
  ngpvmsd,                      // 実測値平均値±SD
  ngpvmse,                      // 実測値平均値±SE
  ngpetc                        // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  bool vcomp;                   // 欠損のないケースだけ集計フラッグ
  int vkey;                     // 基準項目番号
  ngptype vngp;                 // Ngraph用出力データの種類
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 結果出力リスト
  map<string, bool> vresmap;    // 結果出力の有無マップ
  lstr tlst;                    // 項目時間リスト
  vecdbl vtvec;                 // 項目時間ベクトル
  testtype vtt;                 // t検定出力フラッグ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  bool comp() const             // 欠損のないケースだけ集計フラッグを返す
  { return vcomp; }
  int key() const               // 基準項目番号を返す
  { return vkey; }
  ngptype ngp() const           // Ngraph用出力データの種類を返す
  { return vngp; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double t(size_t idx) const    // 項目時間を返す
  { return (idx < vtvec.size()) ? vtvec[idx] : BLANKf; }
  testtype tt() const           // t検定出力フラッグを返す
  { return vtt; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int search_g(const string& gcode) const; // 群コードを検索する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vcomp(false), vkey(1), vngp(ngpnon), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), tlst(), vtvec(), vtt(testaut)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "comp")         // 欠損のないケースだけ集計指定
  {
    if (opt[1].empty() || opt[1] == "+") vcomp = true;
    else if (opt[1] == "-") vcomp = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "key")     // 基準項目番号指定
  {
    char* ptr = nullptr;
    vkey = strtol(opt[1].c_str(), &ptr, 0);
    if (vkey < 1) err = badopt(vpname, optstr, USAGE, 1.0, (double)INT_MAX);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "t")       // 項目時間の指定
  {
    if (strsplit(tlst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (vkey > (int)ni) return disperr(vpname, "Wrong key item number !");
  vtvec.resize(ni);
  for (size_t j = 0; j < ni; ++j) vtvec[j] = (double)(j + 1);
  if (settvec(vtvec, tlst, ni) < 0) return disperr(vpname, "Wrong time of item !");
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["dbar"] = vngpmap["dci"] = vngpmap["ddata"] = vngpmap["dmsd"] = vngpmap["dmse"] = vngpmap["vbar"] = vngpmap["vci"] = vngpmap["vdata"] = vngpmap["vmsd"] = vngpmap["vmse"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["dbar"] = vngpmap["dci"] = vngpmap["ddata"] = vngpmap["dmsd"] = vngpmap["dmse"] = vngpmap["vbar"] = vngpmap["vci"] = vngpmap["vdata"] = vngpmap["vmsd"] = vngpmap["vmse"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["astat"] = vresmap["cc"] = vresmap["pstat"] = vresmap["tt"] = true;
  vresmap["m"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["astat"] = vresmap["cc"] = vresmap["pstat"] = vresmap["tt"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["astat"] = vresmap["cc"] = vresmap["pstat"] = vresmap["tt"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["astat"] = vresmap["cc"] = vresmap["m"] = vresmap["pstat"] = vresmap["tt"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  vector<statv<double>> astatvec; // 全データ用基礎統計量ベクトル
  vector<statv<double>> pstatvec; // 対データ用基礎統計量ベクトル
  vector<statv<double>> kstatvec; // 基準項目用基礎統計量ベクトル
  vector<statv<double>> dstatvec; // 変化量データ用基礎統計量ベクトル
public:
  varclass(size_t ni) : astatvec(ni), pstatvec(ni), kstatvec(ni), dstatvec(ni)
  // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int cal(const option& opt, ldata& datalst); // データを集計して変数に保存する
  int an(size_t idx) const      // 全データの例数を返す
  { return (idx < astatvec.size()) ? astatvec[idx].n() : 0; }
  double amean(size_t idx) const // 全データの平均値を返す
  { return (idx < astatvec.size()) ? astatvec[idx].mean() : 0.0; }
  double avr(size_t idx) const  // 全データの不偏分散を返す
  { return (idx < astatvec.size()) ? astatvec[idx].vr() : 0.0; }
  double asd(size_t idx) const  // 全データの標準偏差を返す
  { return (idx < astatvec.size()) ? astatvec[idx].sd() : 0.0; }
  double ase(size_t idx) const  // 全データの標準誤差を返す
  { return (idx < astatvec.size()) ? astatvec[idx].se() : 0.0; }
  int pn(size_t idx) const      // 対データの例数を返す
  { return (idx < pstatvec.size()) ? pstatvec[idx].n() : 0; }
  double pmean(size_t idx) const // 対データの平均値を返す
  { return (idx < pstatvec.size()) ? pstatvec[idx].mean() : 0.0; }
  double pvr(size_t idx) const  // 対データの不偏分散を返す
  { return (idx < pstatvec.size()) ? pstatvec[idx].vr() : 0.0; }
  double psd(size_t idx) const  // 対データの標準偏差を返す
  { return (idx < pstatvec.size()) ? pstatvec[idx].sd() : 0.0; }
  double pse(size_t idx) const  // 対データの標準誤差を返す
  { return (idx < pstatvec.size()) ? pstatvec[idx].se() : 0.0; }
  int kn(size_t idx) const      // 規準項目の例数を返す
  { return (idx < kstatvec.size()) ? kstatvec[idx].n() : 0; }
  double kmean(size_t idx) const // 規準項目の平均値を返す
  { return (idx < kstatvec.size()) ? kstatvec[idx].mean() : 0.0; }
  double kvr(size_t idx) const  // 規準項目の不偏分散を返す
  { return (idx < kstatvec.size()) ? kstatvec[idx].vr() : 0.0; }
  double ksd(size_t idx) const  // 規準項目の標準偏差を返す
  { return (idx < kstatvec.size()) ? kstatvec[idx].sd() : 0.0; }
  double kse(size_t idx) const  // 規準項目の標準誤差を返す
  { return (idx < kstatvec.size()) ? kstatvec[idx].se() : 0.0; }
  int dn(size_t idx) const      // 変化量の例数を返す
  { return (idx < dstatvec.size()) ? dstatvec[idx].n() : 0; }
  double dmean(size_t idx) const // 変化量の平均値を返す
  { return (idx < dstatvec.size()) ? dstatvec[idx].mean() : 0.0; }
  double dvr(size_t idx) const  // 変化量の不偏分散を返す
  { return (idx < dstatvec.size()) ? dstatvec[idx].vr() : 0.0; }
  double dsd(size_t idx) const  // 変化量の標準偏差を返す
  { return (idx < dstatvec.size()) ? dstatvec[idx].sd() : 0.0; }
  double dse(size_t idx) const  // 変化量の標準誤差を返す
  { return (idx < dstatvec.size()) ? dstatvec[idx].se() : 0.0; }
};

int varclass::cal(const option& opt, ldata& datalst)
// データを集計して変数に保存する
{
  int nc = 0;
  bool ok = true;
  for (auto& cdata : datalst)
  {
    if (opt.comp())             // 欠損のないケースだけ集計
    {
      ok = true;
      for (size_t i = 0; ok && i < cdata.size(); ++i)
      {
        if (cdata.fdata(i) == BLANKf) ok = false;
      }
    }
    if (ok)
    {
      double xkey = cdata.fdata(opt.key() - 1);
      for (size_t i = 0; i < cdata.size(); ++i)
      {
        double x = cdata.fdata(i);
        if (x != BLANKf)
        {
          astatvec[i].sigma(x);
          if (xkey != BLANKf)
          {
            pstatvec[i].sigma(x);  kstatvec[i].sigma(xkey);
            dstatvec[i].sigma(x - xkey);
          }
        }
      }
      ++nc;
    }
  }
  return nc;                    // 集計ケース数を返す
}

class ttestclass                // ttestメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  void put_resid(ostream& fout);
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_astat(ostream& fout, const option& opt, varclass& var) const;
  // 全データの基礎統計量を出力する
  void put_ttest(ostream& fout, option& opt, varclass& var);
  // 対応のあるデータの基礎統計量とt検定の結果を出力する
  void plot_ngraph(option& opt, varclass& var); // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, const option& opt);
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_bar(ofstream& fout, const option& opt, bool diff, varclass& var) const;
  // Ngraph用にバーチャートデータを出力する
  void plot_ci(ofstream& fout, const option& opt, bool diff, varclass& var) const;
  // Ngraph用に平均値±信頼区間を出力する
  void plot_data(ofstream& fout, const option& opt, bool diff);
  // Ngraph用に生データを出力する
  void plot_msde(ofstream& fout, const option& opt, bool sd, bool diff, varclass& var) const;
  // Ngraph用に平均値±SD/SEを出力する
public:
  ttestclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~ttestclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み集計する
  void put(option& opt);        // 対応のあるt検定結果を出力する
};

void ttestclass::put_resid(ostream& fout)
// 集計結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put(fout);
}

void ttestclass::put_astat(ostream& fout, const option& opt, varclass& var) const
// 全データの基礎統計量を出力する
{
  constexpr int LENline = 79; // 区切り線の長さ
  fout << "\n<BASIC STATISTICS OF ALL DATA>";
  if (opt.comp()) fout << "(Only Complete Cases)";
  fout << "\n";
  putline(fout, LENline);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    fout << "ITEM" << setw(2) << i + 1 << ":";
    putmsd(fout, var.an(i), var.amean(i), var.avr(i));
  }
  putline(fout, LENline);
}

void ttestclass::put_ttest(ostream& fout, option& opt, varclass& var)
// 対応のあるデータの基礎統計量とt検定の結果を出力する
{
  constexpr int LENline = 79; // 区切り線の長さ
  fout << "\n<PAIRED DATA AND PAIRED T-TEST(ONE SAMPLE T-TEST)>";
  for (size_t i = 0; i < id.ni(); ++i)
  {
    int ino = i + 1;
    if (var.pn(i) > 0 && ino != opt.key())
    {
      fout << "\nITEM" << setw(2) << ino << ":" << id.item(i) << "\n";
      if (opt.res("pstat"))
      {
        putline(fout, LENline);
        fout << "ITEM" << setw(2) << ino << ":";
        putmsd(fout, var.pn(i), var.pmean(i), var.pvr(i));
        fout << "ITEM" << setw(2) << opt.key() << ":";
        putmsd(fout, var.kn(i), var.kmean(i), var.kvr(i));
      }
      double md = var.dmean(i), sed = var.dse(i);
      double t = (sed > 0.0) ? md / sed : 0.0;
      int dfd = var.dn(i) - 1;
      ldbl pr = ttop(t, dfd);
      if (opt.res("tt") || opt.res("cc") || opt.res("m"))
      {
        putline(fout, LENline);
        fout << "Difference:";
        putmsd(fout, var.dn(i), var.dmean(i), var.dvr(i));
      }
      if (opt.res("tt"))
      {
        fout << "  t=" << t << "  DF=" << dfd << "  p=" << pr << pind(pr) << "\n";
      }
      double ci = calmcci(mcfis, opt.cc(), 2, 1, dfd, 1, sed * sed);
      if (opt.res("cc"))
      {
        fout << "  " << opt.cc() * 100.0 << "%CI for the Mean of the Dif=";
        fout << md << " +/- " << ci << "(" << md - ci << " - " << md + ci << ")\n";
      }
      if ((id.ni() > 2 && opt.res("def")) || opt.res("m"))
      {
        int dc = id.ni() - 1;
        pr = pr * (ldbl)dc;
        if (pr > 1.0L) pr = 1.0L;
        fout << "  Bonferroni Type Multiple Test:p=" << pr << pind(pr) << "\n";
        ci = calmcci(mcbon, opt.cc(), 2, 1, dfd, dc, sed * sed);
        if (opt.res("cc"))
        {
          fout << "  " << opt.cc() * 100.0 << "% Sim-CI for the Mean of the Dif=";
          fout << md << " +/- " << ci << "(" << md - ci << " - " << md + ci << ")\n";
        }
      }
      putline(fout, LENline);
    }
  }
}

void ttestclass::plot_ngraph(option& opt, varclass& var)
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("ddata"))         // 変化量生データ
  {
    fout << "-s1 -zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -dline -w40\n";
    plot_ngpid(fout, opt);
    plot_data(fout, opt, true);
    delml =true;
  }
  if (opt.ngp("dbar"))          // 変化量バーチャート
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -drectangle_solid_fill -x1 -y3\n";
      delml =true;
    }
    plot_ngpid(fout, opt);
    plot_bar(fout, opt, true, var);
  }
  if (opt.ngp("dci"))           // 変化量平均値±CI
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -dline -w80\n";
      delml =true;
    }
    plot_ngpid(fout, opt);
    plot_ci(fout, opt, true, var);
  }
  if (opt.ngp("dmsd"))          // 変化量平均値±SD
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -dline -w80\n";
      delml =true;
    }
    plot_ngpid(fout, opt);
    plot_msde(fout, opt, true, true, var);
  }
  if (opt.ngp("dmse"))          // 変化量平均値±SE
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -dline -w80\n";
      delml =true;
    }
    plot_ngpid(fout, opt);
    plot_msde(fout, opt, false, true, var);
  }
  if (opt.ngp("vdata"))         // 実測値生データ
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -dline -w40\n";
      delml =true;
    }
    plot_ngpid(fout, opt);
    plot_data(fout, opt, false);
  }
  if (opt.ngp("vbar"))          // 実測値バーチャート
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -drectangle_solid_fill -x1 -y3\n";
      delml =true;
    }
    plot_ngpid(fout, opt);
    plot_bar(fout, opt, false, var);
  }
  if (opt.ngp("vci"))           // 実測値平均値±CI
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -dline -w80\n";
      delml =true;
    }
    plot_ngpid(fout, opt);
    plot_ci(fout, opt, false, var);
  }
  if (opt.ngp("vmsd"))          // 実測値平均値±SD
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -dline -w80\n";
      delml =true;
    }
    plot_ngpid(fout, opt);
    plot_msde(fout, opt, true, false, var);
  }
  if (opt.ngp("vmse"))          // 実測値平均値±SE
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -dline -w80\n";
    plot_ngpid(fout, opt);
    plot_msde(fout, opt, false, false, var);
  }
}

void ttestclass::plot_ngpid(ofstream& fout, const option& opt)
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put(fout, formngp);
  if (opt.comp()) fout << "%Only Complete Cases\n";
}

void ttestclass::plot_bar(ofstream& fout, const option& opt, bool diff, varclass& var) const
// Ngraph用にバーチャートデータを出力する
{
  double t0 = opt.t(0);
  if (diff == false)            // 実測値プロット
  {
    fout << "%<BAR CHART FOR THE MEAN OF ALL DATA>\n";
    for (size_t i = 0; i < id.ni(); ++i)
    {
      double ti = opt.t(i);
      fout << ti - 0.2 << " 0 " << ti + 0.2 << " " << var.amean(i) << "\n";
    }
    fout << "= =\n";
  }
  fout << "%<BAR CHART FOR THE MEAN OF ";
  if (diff) fout << "PAIRED DIFFERENCE>\n";
  else fout << "PAIRED DATA>\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    double ti = opt.t(i), m0 = var.kmean(i), mi = var.pmean(i);
    if (diff) { m0 = 0.0;  mi = var.dmean(i); }
    fout << "%ITEM1-" << i + 1 << "\n";
    fout << t0 - 0.2 << " 0 " << t0 + 0.2 << " " << m0 << "\n";
    fout << ti - 0.2 << " 0 " << ti + 0.2 << " " << mi << "\n";
  }
}

void ttestclass::plot_ci(ofstream& fout, const option& opt, bool diff, varclass& var) const
// Ngraph用に平均値±信頼区間を出力する
{
  if (diff == false)            // 実測値プロット
  {
    fout << "%<MEAN OF ALL DATA +/- " << opt.cc() * 100.0 << "% CI(-zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -derrorbar_y -w0 -o500)>\n";
    for (size_t i = 0; i < id.ni(); ++i)
    {
      double t = ptot(1.0 - opt.cc(), var.an(i) - 1);
      double ci = t * var.ase(i);
      fout << opt.t(i) << " " << var.amean(i) << " " << ci << " " << -ci << "\n";
    }
    fout << "= =\n";
  }
  fout << "%<MEAN OF ";
  if (diff) fout << "PAIRED DIFFERENCE";
  else fout << "PAIRED DATA";
  fout << " +/- " << opt.cc() * 100.0 << "% CI>\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    double m0 = var.kmean(i), se0 = var.kse(i);
    double mi = var.pmean(i), sei = var.pse(i);
    int df = var.kn(i) - 1;
    if (diff)
    {
      m0 = se0 = 0.0;
      mi = var.dmean(i);  sei = var.dse(i);
    }
    if (i > 1) fout << "= =\n";
    fout << "%ITEM1-" << i + 1 << "\n";
    double t = (0.0 < opt.cc() && opt.cc() < 1.0) ? ptot(1.0 - opt.cc(), df) : 0.0;
    double ci = t * se0;
    fout << opt.t(0) << " " << m0 << " " << ci << " " << -ci << "\n";
    ci = t * sei;
    fout << opt.t(i) << " " << mi << " " << ci << " " << -ci << "\n";
  }
}

void ttestclass::plot_data(ofstream& fout, const option& opt, bool diff)
// Ngraph用に生データを出力する
{
  if (diff) fout << "%<DIFFERENCE OF DATA>\n";
  else fout << "%<RAW DATA>\n";
  bool putdelml = false;
  for (auto& cdata : datalst)
  {
    if (putdelml) fout << "= =\n";
    else putdelml = true;
    double x0 = cdata.fdata(0);
    if (x0 != BLANKf)
    {
      if (diff) fout << opt.t(0) << " " << 0.0 << "\n";
      else fout << opt.t(0) << " " << x0 << "\n";
    }
    for (size_t i = 1; i < cdata.size(); ++i)
    {
      double xi = cdata.fdata(i);
      if (xi != BLANKf)
      {
        if (diff)
        {
          if (x0 != BLANKf) fout << opt.t(i) << " " << xi - x0 << "\n";
        }
        else
        { fout << opt.t(i) << " " << xi << "\n"; }
      }
    }
  }
}

void ttestclass::plot_msde(ofstream& fout, const option& opt, bool sd, bool diff, varclass& var) const
// Ngraph用に平均値±SD/SEを出力する
{
  string strsde = sd ? "SD" : "SE";
  if (diff == false)            // 実測値プロット
  {
    fout << "%<MEAN OF ALL DATA +/- " << strsde << "(-zx0," << opt.t(id.ni() - 1) + 1.0 << ",1 -derrorbar_y -w40 -o500)>\n";
    for (size_t i = 0; i < id.ni(); ++i)
    {
      double sdei = sd ? var.asd(i) : var.ase(i);
      fout << opt.t(i) << " " << var.amean(i) << " " << sdei << " " << -sdei << "\n";
    }
    fout << "= =\n";
  }
  fout << "%<MEAN OF ";
  if (diff) fout << "PAIRED DIFFERENCE";
  else fout << "PAIRED DATA";
  fout << " +/- " << strsde << ">\n";
  for (size_t i = 1; i < id.ni(); ++i)
  {
    if (i > 1) fout << "= =\n";
    fout << "%ITEM1-" << i + 1 << "\n";
    double m0 = var.kmean(i), mi = var.pmean(i);
    double sde0 = sd ? var.ksd(i) : var.kse(i);
    double sdei = sd ? var.psd(i) : var.pse(i);
    if (diff)
    {
      m0 = sde0 = 0.0;
      mi = var.dmean(i);  sdei = sd ? var.dsd(i) : var.dse(i);
    }
    fout << opt.t(0) << " " << m0 << " " << sde0 << " " << -sde0 << "\n";
    fout << opt.t(i) << " " << mi << " " << sdei << " " << -sdei << "\n";
  }
}

int ttestclass::get(const char* c_fname, option& opt)
// データを読み込み集計する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void ttestclass::put(option& opt) // 対応のあるt検定結果を出力する
{
  if (id.nc() < 1)
  {
    if (opt.res("res"))
    {
      put_resid(cout);
      cout << "\n" << endl;
      return;
    }
  }
  varclass var(id.ni());        // 計算用変数クラス
  if (var.cal(opt, datalst) < 1) return; // 集計ケースなし
  if (opt.res("res"))           // 計算結果出力
  {
    put_resid(cout);
    if (opt.res("astat")) put_astat(cout, opt, var);
    if (opt.res("pstat") || opt.res("tt")) put_ttest(cout, opt, var);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngraph(opt, var); // グラフデータ出力
}

int main(int argc, char *argv[])
// メインルーチン
{
  ttestclass ttest;             // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) ttest.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { ttest.get(argv[i], opt);  fok = false; }
    }
    ttest.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occured !");
    return 255;
  }
  return 0;
}
